"""Configuration file for the Sphinx documentation builder."""

import os
from importlib.util import module_from_spec, spec_from_file_location

from sphinx_gallery.sorting import ExplicitOrder

from planetary_coverage import __version__


# Dynamically load the `download.py` module (https://stackoverflow.com/a/72098449)
spec = spec_from_file_location('download', 'download.py')
download = module_from_spec(spec)
spec.loader.exec_module(download)


# -- Project information -----------------------------------------------------

project = 'Planetary coverage documentation'
copyright = '2020-2025 - All right reserved'  # noqa: A001
author = 'Osuna / LPG / CNRS / Nantes Université / ESA-SOC / CNES'
version = __version__
release = __version__

# -- General configuration ---------------------------------------------------

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.napoleon',
    'sphinx.ext.viewcode',
    'sphinx.ext.intersphinx',
    'sphinx_copybutton',
    'sphinx_tabs.tabs',
    'sphinx_gallery.gen_gallery',
    'myst_nb',
    'IPython.sphinxext.ipython_console_highlighting',
    'matplotlib.sphinxext.roles',
]

master_doc = 'index'
language = 'en'
exclude_patterns = [
    '_build',
    'Thumbs.db',
    '.DS_Store',
    '.jupyter_cache',
    'jupyter_execute',
    '**.ipynb_checkpoints',
    'trim_kernels',
]

# -- Mapping -----------------------------------------------------------------

intersphinx_mapping = {
    'python': ('https://docs.python.org/3/', None),
    'numpy': ('https://numpy.org/doc/stable/', None),
    'matplotlib': ('https://matplotlib.org/stable/', None),
    'spiceypy': ('https://spiceypy.readthedocs.io/en/stable/', None),
}

# -- Options for HTML output -------------------------------------------------

html_theme = 'sphinx_book_theme'
html_title = 'Planetary coverage documentation'
html_logo = 'images/planetary-coverage-logo.svg'
html_favicon = 'images/favicon.ico'

html_theme_options = {
    'use_repository_button': False,
    'use_download_button': False,
    'use_fullscreen_button': False,
    'logo': {
        'text': html_title,
    },
}
html_context = {
    'default_mode': 'light',
}

html_static_path = ['_static']

html_css_files = [
    'css/custom.css',
]

# -- Extension configuration -------------------------------------------------
autodoc_default_options = {
    'members': True,
    'show-inheritance': True,
}

autodoc_member_order = 'bysource'
pygments_style = None

myst_enable_extensions = [
    'colon_fence',
    'dollarmath',
]

myst_heading_anchors = 3
nb_execution_mode = 'cache'

copybutton_prompt_text = r'>>>\s?|\.\.\.\s|\$\s'
copybutton_prompt_is_regexp = True
copybutton_copy_empty_lines = True

# -- LaTeX config for PDF output ---------------------------------------------

latex_engine = 'xelatex'

# -- Gallery examples --------------------------------------------------------

sphinx_gallery_conf = {
    'examples_dirs': '../examples',
    'gallery_dirs': 'gallery',
    'filename_pattern': '.py$',
    'download_all_examples': False,
    'show_signature': False,
    'remove_config_comments': True,
    'pypandoc': True,
    'notebook_images': True,
    'first_notebook_cell': (
        '# This example is part of the planetary-coverage examples gallery.\n'
        '# More examples can be found here: https://docs.planetary-coverage.org'
    ),
    'subsection_order': ExplicitOrder([
        '../examples/trajectory',
        '../examples/cross-missions',
    ]),
    'within_subsection_order': 'FileNameSortKey',
}


def setup(app):
    """Fix multiple files warning.

    See: https://github.com/executablebooks/MyST-NB/issues/363

    """
    app.registry.source_suffix.pop('.ipynb', None)


# -- Download the required kernels in advance ----------------------------------------

download.download_kernels()

# -- Read The Docs -------------------------------------------------------------------
html_baseurl = os.environ.get('READTHEDOCS_CANONICAL_URL', '')

if os.environ.get('READTHEDOCS', '') == 'True':
    html_context['READTHEDOCS'] = True
