"""Pre-build download for ReadTheDocs."""

import os
from datetime import datetime as dt
from multiprocessing import cpu_count
from pathlib import Path

from planetary_coverage import TourConfig
from planetary_coverage.debug import debug_download
from planetary_coverage.misc import logger, wget


# A custom remote can be provided at runtime with
# the KERNELS_REMOTE environnement variable.
ESA_REMOTE = os.environ.get(
    'KERNELS_REMOTE',
    'https://spiftp.esac.esa.int/data/SPICE',
)

# NASA remote
NASA_REMOTE = os.environ.get(
    'KERNELS_REMOTE',
    'https://naif.jpl.nasa.gov/pub/naif',
)

# Gallery examples kernels folder
EXAMPLES_DIR = (Path(__file__).parent / '../examples/cross-missions/data').resolve()

# Enable download debugger
if os.environ.get('DEBUG_DOWNLOAD'):
    debug_download(True)

# ESA mission metakernels requirements
ESA_EXAMPLES_MK = {
    'JUICE': ('JUICE', '5.1 150lb_23_1', 'v422_20230130_002'),
    'BEPICOLOMBO': ('MPO', 'plan', 'v200_20181212_002'),
    'ENVISION': ('ENVISION', 'study_esc_t2_south_voi_v040', 'v040_20220503_001'),
}

# NASA mission kernels requirements
NASA_KERNELS = {
    'JUNO': (
        'lsk/naif0012.tls',
        'pck/pck00010.tpc',
        'fk/juno_v12.tf',
        'ik/juno_junocam_v03.ti',
        'sclk/JNO_SCLKSCET.00114.tsc',
        'spk/de440s.bsp',
        'spk/juno_pred_orbit.bsp',
        'spk/juno_rec_orbit.bsp',
        'spk/juno_struct_v04.bsp',
        'spk/jup363.bsp',
        'spk/spk_pre_210416_210902_210608_btm34_p.bsp',
        'ck/juno_sc_rec_210607_210608_v01.bc',
    ),
}


def download(mission, kernels_dir):
    """Download mission kernels."""
    if mission in ESA_EXAMPLES_MK:
        spacecraft, mk, skd_version = ESA_EXAMPLES_MK[mission]

        _ = TourConfig(
            mk=mk,
            version=skd_version,
            spacecraft=spacecraft,
            kernels_dir=kernels_dir,
            remote_kernels=f'{ESA_REMOTE}/{mission}/kernels',
            download_kernels=True,
        )

    elif mission in NASA_KERNELS:
        for kernel in NASA_KERNELS[mission]:
            wget(
                f'{NASA_REMOTE}/{mission}/kernels/{kernel}',
                f'{kernels_dir}/{kernel}',
                skip=True,
            )

    else:
        raise ValueError(f'Unknown mission: `{mission}`.')


def download_kernels():
    """Download all ESA kernels required the gallery of examples."""
    # Add Docs logger
    log, _ = logger('Docs')
    log.warning('CPU available: %i', cpu_count())

    # Start download
    t_start = dt.now()
    log.warning('Start kernels download at: %s', t_start)

    # Juice
    if not os.environ.get('KERNELS_JUICE'):
        raise KeyError('`KERNELS_JUICE` variable must be defined.')

    download('JUICE', os.environ.get('KERNELS_JUICE'))

    t_juice = dt.now()
    log.warning('Juice download duration: %s', t_juice - t_start)

    # Bepi
    download('BEPICOLOMBO', EXAMPLES_DIR / 'BEPICOLOMBO' / 'kernels')

    t_bepi = dt.now()
    log.warning('Bepi download duration: %s', t_bepi - t_juice)

    # EnVision
    download('ENVISION', EXAMPLES_DIR / 'ENVISION' / 'kernels')

    t_envision = dt.now()
    log.warning('EnVision download duration: %s', t_envision - t_bepi)

    # JUNO
    download('JUNO', EXAMPLES_DIR / 'JUNO' / 'kernels')

    t_juno = dt.now()
    log.warning('JUNO download duration: %s', t_juno - t_envision)

    # Download complete
    t_end = dt.now()
    log.warning('Complete kernels download time: %s', t_end - t_start)


if __name__ == '__main__':
    # Download the required kernels for the docs.
    out = Path(os.sys.argv[1] if len(os.sys.argv) > 1 else '.')

    download('JUICE', out / 'JUICE' / 'kernels')
    download('BEPICOLOMBO', out / 'BEPICOLOMBO' / 'kernels')
    download('ENVISION', out / 'ENVISION' / 'kernels')
    download('JUNO', out / 'JUNO' / 'kernels')
