---
jupytext:
  formats: ipynb,md:myst
  text_representation:
    extension: .md
    format_name: myst
    format_version: 0.16
    jupytext_version: 1.14.1
kernelspec:
  display_name: Python 3
  language: python
  name: python3
---

# ESA events files

```{code-cell}
:tags: [remove-input]

from pathlib import Path

import matplotlib.pyplot as plt

from planetary_coverage import TourConfig, read_events

tour = TourConfig(mk='5.1 150lb_23_1', spacecraft='JUICE', target='Ganymede', version='v422_20230130_002')
```

Recently the ESA-SOC published a collection of event files that contains the different phases of the mission,
the dates and names of the flybys; as well as some technical information, e.g. the visibility of the DSN stations and the expected downlinks windows.

:::{Caution}
These files need to be considered as experimental and their format and location can evolve significantly in the future.
Currently they are only provided for the Juice mission (CREMA `3.0`, `5.0` and `5.0b23.1`) on the
[SOC core system toolkit](https://juicesoc.esac.esa.int/panel/#/navigation/autoindex/DATA).
:::

:::{versionadded} 1.1.0
Some of the mission are now translated into `fk` event kernels that can be loaded with the Juice metakernels.
See `TourConfig` for [more details](../trajectory/spacecraft/tour_config.md#mission-phases-and-timeline-events).
:::

## Event file format

An event file is a text file (`.csv`/`.evf`/`.itl`/`.orb`) containing either a single time event, a start or stop time event or an event with both a start and a stop time.

For example, the [mission phases file](https://juicesoc.esac.esa.int/data/DATA/crema_5_0b23_1/Mission_Phases.csv) is defined like this:

:::{Hint}
:class: margin

All the events have a `t_start` and a `t_end` time and the primary key is stored in the `Name` column.

You can add comments and blank lines in the the csv file, as long as you keep the column `header` in the first line (or explicitly mention it in `read_events()`, see below).
:::

```{code-cell}
:tags: [remove-input, output_scroll]

read_events('Mission_Phases.csv').csv
```

[In some cases](https://juicesoc.esac.esa.int/data/DATA/crema_5_0b23_1/mission_timeline_event_file_5_0b23_1.csv),
the events are only defined as single `event time`:

:::{Note}
:class: margin

The primary key (`Event name`) does not need to be unique and will be grouped together.

Here, the `XXX_START` and `XXX_END` events will be paired.

If a pair is incomplete, a _warning_ message will be send to the user at runtime.
:::

```{code-cell}
:tags: [remove-input, output_scroll]

read_events('mission_timeline_event_file_5_0b23_1.csv').csv
```

[When no explicit header is provided](https://juicesoc.esac.esa.int/data/DATA/crema_5_0b23_1/segmentation_proposal_Jupiter_Phase_all_crema_5_0b23_1_MALARGUE.csv)
in the file. By default we will assume that when it's the case, the default header should be:

:::{Tip}
:class: margin

As you can see below, any custom header can be provided.
:::

```{code-cell}
:tags: [remove-input, output_scroll]

read_events('segmentation_proposal_Jupiter_Phase_all_crema_5_0b23_1_MALARGUE.csv').csv
```

MAPPS event file (`.evf`) are also supported and should follow this pattern:

:::{Note}
:class: margin

`.evf` and `.itl` files are not currently distributed publicly by the Juice project.
:::

```{code-cell}
:tags: [remove-input, output_scroll]

print(Path('CA_EUROPA.EVF').read_text(encoding='utf-8'))
```

As well as instrument timeline file (`.itl`):

```{code-cell}
:tags: [remove-input, output_scroll]

print(Path('OTL_EUROPA.ITL').read_text(encoding='utf-8'))
```

For many space mission, the space agency can provide orbit numbers file (`.orb`) files
than can be used as events files. Usually they are stored in a `misc/orbnum/` folder as
ancillary kernels files (eg.
[Juice](https://spiftp.esac.esa.int/data/SPICE/JUICE/misc/orbnum/),
[Bepi](https://spiftp.esac.esa.int/data/SPICE/BEPICOLOMBO/misc/orbnum/),
[Cassini](https://naif.jpl.nasa.gov/pub/naif/pds/data/co-s_j_e_v-spice-6-v1.0/cosp_1000/extras/orbnum/)...)

```{code-cell}
:tags: [remove-input, output_scroll]

print(Path('juice.orb').read_text(encoding='utf-8'))
```


## Parsing an events file

We recommend to use the generic `read_events()` function that can parse all
ESA events files listed above:

```{code-cell}
from planetary_coverage import read_events

phases = read_events('Mission_Phases.csv')
timeline = read_events('mission_timeline_event_file_5_0b23_1.csv')
malargue = read_events('segmentation_proposal_Jupiter_Phase_all_crema_5_0b23_1_MALARGUE.csv')
evf = read_events('CA_EUROPA.EVF')
itl = read_events('OTL_EUROPA.ITL', evf='CA_EUROPA.EVF')
orbits = read_events('juice.orb')
```

If you already know the type of file you want to parse, you can also use
`CsvEventsFile`/`EvfEventsFile`/`ItlEventsFile`/`OrbitEventsFile` parsers:

```{code-cell}
from planetary_coverage.events import CsvEventsFile

CsvEventsFile('Mission_Phases.csv')
```

```{code-cell}
from planetary_coverage.events import EvfEventsFile

EvfEventsFile('CA_EUROPA.EVF')
```

:::{Hint}
:class: margin

The parameter `evf` is required only for relative timelines.

:::
```{code-cell}
from planetary_coverage.events import ItlEventsFile

ItlEventsFile('OTL_EUROPA.ITL', evf='CA_EUROPA.EVF')
```

```{code-cell}
from planetary_coverage.events import OrbitEventsFile

OrbitEventsFile('juice.orb')
```

In all cases, when parsed, you get a collection of `Event` and `EventWindow` objects:

```{code-cell}
:tags: [output_scroll]

phases
```


## Get an event by key

You can select a given event with its name:

:::{Tip}
:class: margin

On Jupyter and IPython environment, you can use the `<tab>`
key to auto-complete: `events['<tab>` to show the available keys.
:::

```{code-cell}
gco_500 = phases['GCO500']

gco_500
```
When multiple event with the same name are detected, they will be grouped in an `EventsList`:

:::{Tip}
:class: margin

The search key is not sensitive to the case.
:::

```{code-cell}
downlinks = timeline['downlink_malargue']

downlinks
```

:::{Hint}
:class: margin

In the case of the __flybys__, they contain a special property (`.crema_names`) that host they canonic name defined by the SOC:

```python
>>> flybys.crema_names
['1G1', '2G2', ...]
```
:::

```{code-cell}
flybys = timeline['FLYBY_GANYMEDE']

flybys
```

Then, you can select a precise flyby if you want with either its `name` or its `id`:

:::{Tip}
:class: margin

The value in the ``contextual info`` column are also parsed as additional attributes (separated with a ``;``).
:::

```{code-cell}
flyby_1G1 = flybys['1G1']  # or flybys[0]

flyby_1G1
```

Multiple inputs is also supported:

:::{Tip}
:class: margin

The events will be __sorted__ by start time and
the __duplicates__ will be __discarded__.
:::

```{code-cell}
phases['GCO500', 'GCO5000']
```

You can also get all the events starting with a given keyword:

:::{Hint}
:class: margin

If only one match is found, the events is directly returned.
If you need to the results as an `EventsDict` you can add the argument: `as_dict=True`.
:::

```{code-cell}
timeline.startswith('GANYMEDE')
```

or ending with a given keyword:

```{code-cell}
timeline.endswith('TRANSIT')
```

as well as [complex regular expression](https://docs.python.org/3/library/re.html):

:::{Note}
:class: margin

The `find()` function is the default function called by `__getitem__`
in last resort, if none of the key(s) were found.
:::

```{code-cell}
timeline.find('^VIS.*E$')  # Starting with 'VIS' and ending with a 'E'
```

### Orbit numbers

In the case of orbit numbers file (`.orb`), you can query the orbit number by `index`, `tuple` or `slice`:
```{code-cell}
orbits[1]  # Tip: negative index are accepted too
```
```{code-cell}
orbits[2, 4]
```
:::{Note}
:class: margin

Orbit events are defined as single `Event` object (usually on apo-apsis or peri-apsis point).
If you need a time window for a complete orbit, you need to select the starting orbit
and the ending orbit in the `slice`.

We assume that the 1st column of the file correspond to the event time.
If you need an other parameter, you can query it with the name of the column.
```python
>>> orbits[1]['sc lon']
340.06
```

:::
```{code-cell}
orbits[1:2]  # Note: the last element is always included
```
```{code-cell}
orbits[1:5:2]  # with step interval
```


## Events filtering

:::{Tip}
:class: margin

You can also use the comparison operators ``<``, ``<=``, ``>``, ``>=``.
But you should avoid to chain them (see `Danger` note below).
:::

The `CsvEventsFile`, `EvfEventsFile`, `ItlEventsFile`, `OrbitEventsFile`, `EventsDict` and `EventsList`
can be filtered with any date string with a `before()`, `after()` and `between()` functions:

:::{Hint}
:class: margin

If only one match is found, the events is directly returned.
If you need to the results as an `EventsDict` you can add the argument: `as_dict=True`.
:::

```{code-cell}
phases.before(2033)  # or: `phases <= 2033`
```

```{code-cell}
phases.after('2034-06-01', strict=True)  # or: `phases > '2034-06-01'`
```

:::{Caution}
- `before()` / `<` / `<=` perform a comparison with the `Event` _start_ time.
- `after()` / `>` / `>=` perform a comparison with the `Event` _stop_ time.
:::

```{code-cell}
timeline.between('2031-06', '2032-03')
```

:::{Danger}
Unfortunately, it is not possible to chain the operators without
explicit parenthesis. You should __NOT__ do:

```python
'2031-06' <= timeline <= '2032-03'
```

but you __should__ do:
```python
('2031-06' <= timeline) <= '2032-03'
```
or:
```python
'2031-06' <= (timeline <= '2032-03')
```
:::

### Trim event(s) time boundaries

If you need to trim the time boundaries of an `Event` or a collection of `Events`, you can use the `.trim()`
function. You will get a **copy** of the event(s) with the same the primary key
and properties but with a trimmed start and stop time. The original events are not edited.

:::{Note}
:class: margin

The original datetime format is not conserved and you could observed some discrepancies
in the datetime string formatting in the output trimmed event.
However, the `Event` _start_ and _stop_ times will always be returned as `numpy.datetime64` objects.
:::

```{code-cell}
downlinks.trim(before='2031-01-20T12:00:00')
```

If an element is outside the new time boundary, it will be discarded:
```{code-cell}
downlinks.trim(after='2031-01-21')
```

You can use both keywords at the same time:
```{code-cell}
downlinks.trim(before='2031-01-20T12:00:00', after='2031-01-21T12:00:00')
```

or use directly an other event object:
```{code-cell}
from planetary_coverage.events import EventWindow

event = EventWindow('DUMMY_EVENT', t_start='2031-01-20T12:00:00', t_end='2031-01-21T12:00:00')

downlinks.trim(by_event=event)
```


## Events usage

All these events objects are compatible with `TourConfig` and `Trajectory` objects, for example you can use them to select a time window:

```{code-cell}
tour[gco_500:'1h']
```

You can also use an event to defined quickly the associated flyby:

```{code-cell}
tour.flyby(flyby_1G1)
```

You can also get multiple flyby trajectories at once (here the first 3 flybys above Ganymede defined before):

```{code-cell}
tour.flyby(flybys)
```

Temporal operation is also supported on the `Event` and `EventWindow` objects, for example you can take the trajectory `12 hours` around a flyby CA event (with a regular temporal step of `10 minutes`):

:::{Hint}
:class: margin

A _subtraction_ is performed on the `start time` and an _addition_ is performed on the `stop time`.
:::

```{code-cell}
tour[flyby_1G1 - '12 h':flyby_1G1 + '12 h':'10 mins']
```

You can also filter a trajectory with any `EventWindow` or `EventsList`.
Here we choose to take all the tour (from the beginning to the end of the mission with a temporal step of `1h`) and we only show the point during the `GCO500 phase`.

:::{Caution}
:class: margin

Filtering on a large temporal windows is __very inefficient__. Here the full trajectory contains more than 4,500 points when only 130 are really relevant. You should better use the solution presented above:

```python
tour[gco_500:'1 day']
```

to get the same result in the more efficient way.
:::

```{code-cell}
tour['start':'stop':'1 day'] & gco_500
```

:::{admonition} New in version 1.1.0
If a `fk` events kernel is load into the kernel pool, you can query them directly:
:::

```{code-cell}
tour['GCO500':'1 day']
```


## Trajectory filtering with events


Similarly to the `ROI`, you can also discard an `EventWindow` or an `EventsList`. For example if you want to mask all the downlink period, you can do:

```{code-cell}
traj_no_downlinks = tour['2031-01-20' : '2031-01-22'] ^ downlinks

traj_no_downlinks
```

We can represent the filtered trajectory parameter, with the 2 masked downlink windows:

```{code-cell}
:tags: [hide-input]

_, ax = plt.subplots(figsize=(14, 3))

ax.plot(traj_no_downlinks.utc, traj_no_downlinks.inc, 'o', ms=2)

y0, y1 = ax.get_ylim()
for downlink in downlinks[:2]:
    ax.fill_between([downlink.start, downlink.stop], y0, y1, color='0.95')
    ax.text(downlink.start + (downlink.stop - downlink.start)/2, (y0 + y1) / 2,
            'Downlink\nwindow', ha='center', va='center')

ax.set_ylim(y0, y1)
ax.set_xlabel('UTC datetime in 2031')
ax.set_ylabel('Incidence angle (degree)');
```

You can also exclude an `EventWindow` or an `EventsList` directly in `TourConfig` or in the `Trajectory` definition:

:::{Tip}
:class: margin

If you know in advance the events that you want to discard, it is recommended to use this method to avoid creating trajectory points that will not be masked later on. __Only the relevant points are kept__.
:::

```{code-cell}
TourConfig(
    mk='5.1 150lb_23_1',
    spacecraft='JUICE',
    target='Ganymede',
    exclude=downlinks,
    version='v422_20230130_002',
)['2031-01-20' : '2031-01-22']
```
