---
jupytext:
  formats: ipynb,md:myst
  text_representation:
    extension: .md
    format_name: myst
    format_version: 0.16
    jupytext_version: 1.14.1
kernelspec:
  display_name: Python 3
  language: python
  name: python3
---

# ESA metakernels

:::{Note}
:class: margin

For some missions, you can find the CREMA metakernel denomination. It stands for _Consolidated Report on Mission Analysis_ and they correspond to a given planning mission scenario.
:::

ESA metakernel, like any SPICE metakernel, contains list of relevant kernels files that are needed to perform a orbital and trajectory calculations with SPICE.
These files are created and maintained by [ESA SPICE service](http://spice.esac.esa.int/) and can be found in the `kernels/mk/` folder in every [ESA missions](https://spiftp.esac.esa.int/data/SPICE/) (you can also find then on [ESA SPICE bitbucket registry](https://s2e2.cosmos.esa.int/bitbucket/projects/SPICE_KERNELS)). A plain-text summary is also available on [their ESA Cosmos website](https://www.cosmos.esa.int/web/spice/) for each mission.

Theses metakernels are considered to be the canonic setup required to compute the different orbital tour planned by ESA. It is not mandatory to use them, but they constitute an excellent starting point for the beginner user.

The `planetary-coverage` tool provides a collection of shortcuts to pull any ESA metakernels (from any given release) and all its associated kernels directly from ESA archive:

```{code-cell}
:tags: [output_scroll]
from planetary_coverage import ESA_MK

ESA_MK
```

## Juice metakernels

For Juice, we can list all the metakernels available in the latest release, very easily:

```{code-cell}
:tags: [output_scroll]
JUICE_MK = ESA_MK['JUICE']

JUICE_MK
```

As we can see, the tool automatically query the online ESA archive to get the latest tag publicly published (here `v422_20230130_002`) and pull the list of all the metakernels available in that release.

You can check if a known metakernel filename is in the latest release:

```{code-cell}
'juice_crema_5_1_150lb.tm' in JUICE_MK
```

```{code-cell}
'juice_crema_6_0.tm' in JUICE_MK
```

All the previous releases are available with the `versions` property:

```{code-cell}
:tags: [output_scroll]
JUICE_MK.versions
```

and if you need to list the metakernels available for an older version, you only have to do:

:::{Tip}
:class: margin

You don't need to provide the full release id, usually the version number is sufficient.
:::

```{code-cell}
:tags: [output_scroll]
JUICE_MK.version('v160')
```

When you know which metakernel you want to work with, you can request it by with its name:

```{code-cell}
:tags: [remove-output]
JUICE_MK['juice_crema_5_0.tm']
```
```
PosixPath('~/.planetary-coverage/esa-mk/juice/v422_20230130_002/juice_crema_5_0.tm')
```

You will get `pathlib.Path` object pointing to its location on your disk (cached in your home folder). If the file is missing locally, it will be downloaded automatically from the ESA archive.

:::{Tip}

To query a metakernel, you can use its full name or you can use these kind of shortcuts:

```python
>>> JUICE_MK[5.0]
PosixPath('.../v422_20230130_002/juice_crema_5_0.tm')

>>> JUICE_MK['4.2b22.1 (cruise only)']
PosixPath('.../v422_20230130_002/juice_crema_4_2b22_1_cruise.tm')

>>> JUICE_MK['3_2 gan N56 Pp5 Q19']
PosixPath('.../v422_20230130_002/juice_crema_3_2_gan_N56_Pp5_Q19.tm')
```

These keys follow the same pattern as the one listed on [ESA-Cosmos SPICE page](https://www.cosmos.esa.int/web/spice/spice-for-juice).

:::

:::{Hint}
The locally cached metakernel files are sorted by mission and releases. If you need to use a specific `version` of a metakernel, you can add a second argument with its version number:

```python
>>> JUICE_MK['3.2 ops', 'v160']
PosixPath('~/.planetary-coverage/esa-mk/juice/v160_20180216_001/juice_crema_3_2_ops.tm')
```

Similarly to the `version()` function, you don't need to provide the full release id, the version is usually enough.
If no explicit version is provided, the tool will search for the **latest release** available and use it by default.

When available, you can also use a `MK_IDENTIFIER` to get a specific version of a metakernel:

```python
>>> JUICE_MK['juice_crema_5_1_150lb_23_1_v422_20230130_002']
PosixPath('~/.planetary-coverage/esa-mk/juice/v422_20230130_002/juice_crema_5_1_150lb_23_1.tm')
```
:::

:::{Note}
If you need to use ESA metakernel offline or if ESA servers are not available, the tool will display a warning error and try to use the metakernels available in the cache:
```text
[ESA API] Impossible to fetch `latest` version online. Use local cache instead.
[ESA API] Impossible to fetch `juice_crema_5_0.tm` metakernel at `v422_20230130_002` version online. Use local cache instead.
```
:::

Finally, you can consume these metakernels files very easily with the [`MetaKernel`](../spice/mk.md) object:


```python
from planetary_coverage import MetaKernel

MetaKernel(JUICE_MK['5.1 150lb_23_1', 'v422'], kernels='/data/kernels')
```

## Other ESA missions

The current setup was optimized for the Juice mission but already works with all ESA missions.
```{code-cell}
:tags: [output_scroll]

ESA_MK['BEPICOLOMBO']
```

You can also request the metakernels for a spacecraft by using its SPICE name:
```{code-cell}
:tags: [output_scroll]

ESA_MK['MERCURY PLANETARY ORBITER']  # or ESA_MK['MPO']
```

You can append a metakernel name, shortcut or `MK_IDENTIFIER` and a version number.
```{code-cell}
:tags: [output_scroll]

ESA_MK['MPO', 'plan', 'v310_20221027_001']  # or ESA_MK['MPO', 'bc_ops_v310_20221027_001']
```
