# Planetary coverage documentation

The [planetary-coverage](https://docs.planetary-coverage.org)
package is a toolbox to perform surface coverage analysis based on orbital trajectory calculations.
Its main intent is to provide an easy way to compute observation
opportunities of specific region of interest above the Galilean
satellites for the ESA-Juice mission but could be extended in the
future to other space mission.

It is actively developed by the
[Observatoire des Sciences de l'Univers Nantes Atlantique](https://osuna.univ-nantes.fr)
(OSUNA, CNRS-UAR 3281) and the
[Laboratory of Planetology and Geosciences](https://lpg-umr6112.fr) (LPG, CNRS-UMR 6112)
at Nantes University (France),
under [ESA-Juice](https://sci.esa.int/web/juice) and [CNES](https://cnes.fr) founding support.

```{figure} images/logos.png
:width: 512
:alt: Logos OSUNA / LPG / Nantes Université / CNRS / CNES / ESA / Juice
:align: center
```

```{toctree}
:caption: Get started
:hidden:

get-started/presentation.md
get-started/kernels.md
```

```{toctree}
:caption: Trajectory calculations
:hidden:

trajectory/spacecraft/index.md
trajectory/instrument/index.md
trajectory/flybys/index.md
trajectory/rois/index.md
```

```{toctree}
:hidden:
:caption: SPICE tools

spice/kernel.md
spice/mk.md
spice/toolbox.md
spice/times.md
```

```{toctree}
:hidden:
:caption: ESA tools

esa/metakernel.md
esa/events.md
esa/export.md
esa/ptr.md
```

```{toctree}
:hidden:
:caption: Miscellaneous

gallery/index.rst
maps/index.md
api/index.md
CHANGELOG.md
```

```{toctree}
:hidden:
:caption: Resources
Gitlab repo <https://juigitlab.esac.esa.int/python/planetary-coverage>
Bug tracker <https://juigitlab.esac.esa.int/python/planetary-coverage/-/issues>
PyPI <https://pypi.org/project/planetary-coverage/>
SWH archive <https://archive.softwareheritage.org/browse/origin/?origin_url=https://juigitlab.esac.esa.int/python/planetary-coverage>
```


## Installation

:::{Note}
:class: margin

If you have access to a [Jupyter environnement](https://jupyter.org/),
ie. a local [JupyterLab](https://blog.jupyter.org/jupyterlab-desktop-app-now-available-b8b661b17e9a)
or a remote JupyterHub, like [ESA DataLabs](https://datalabs.esa.int/),
**we recommend to use the first method** of installation.
:::

The latest version of the `planetary-coverage` is `1.1.1` and is available on [Python Package Index](https://pypi.org/project/planetary-coverage/) (PyPI)
and [conda-forge](https://anaconda.org/conda-forge/planetary-coverage).
You can install it with `pip` or `conda`:

:::{Hint}
:class: margin

If you need to perform pointing simulations for the `ESA-Juice` mission.
You will need to install an optional dependency (`esa-ptr`). More details can be found [here](esa/ptr).
:::


:::::{tabs}

::::{tab} Jupyter

In your Jupyter environment, go to `File > New > Notebook` to create a new notebook and in the first cell, type:

```python
%pip install planetary-coverage
```

Execute the cell with <kbd>⇧ Shift</kbd> + <kbd>↵ Enter</kbd> and you should get the latest version of the `planetary-coverage`.

You can check which version was installed, restart the Jupyter kernel and run the following command:
```python
%load_ext planetary_coverage
```
::::

::::{tab} PyPI

:::{admonition} Recommended
:class: caution

If you want to use the `planetary-coverage` directly in your terminal or
in your [REPL](https://realpython.com/interacting-with-python),
we **recommend** to encapsulate it in a virtual environment (`venv`)
to avoid any conflict with your default python environment:

```bash
python -m venv venv
. venv/bin/activate
```

If you don't use a virtual environment, we recommand that you add a `--user` flag to the command below.
:::

Then install the package:

```bash
pip install planetary-coverage
```
::::

::::{tab} Conda

You need to add `conda-forge` to your channels:
```bash
conda config --add channels conda-forge
conda config --set channel_priority strict
```

:::{admonition} Optional
:class: note

If you want to isolate the `planetary-coverage`
from your `(base)` environment, we can create a dedicated
conda environment like this:

```text
conda create -n planetary-coverage python=3.10
conda activate planetary-coverage
```
:::

Then install the package:

```bash
conda install planetary-coverage
```
::::

::::{tab} Poetry

This method is recommend **only if** you need to install the **development version** from the source.

:::{admonition} Required
:class: caution

You will need to install [Poetry](https://python-poetry.org/docs/) (≥`1.6`) on your system.
On _Linux/macOS/Windows (WSL)_ you can do:
```bash
curl -sSL https://install.python-poetry.org | python3 -
```
:::

Clone the source from the [Gitlab repository](https://juigitlab.esac.esa.int/python/planetary-coverage):
```bash
git clone https://juigitlab.esac.esa.int/python/planetary-coverage.git
cd planetary-coverage
```

and install the package and its development dependencies:
```bash
poetry install
```

All the dependencies will be installed in an isolated virtual environment. Then, you can either prefix your commands
with:
```bash
poetry run <my-cmd>
```

or activate directly the virtual environment:

```bash
poetry shell
```

::::
:::::


## Changes and updates

The `planetary-coverage` adheres to [semantic versioning](https://semver.org),
ie. the version number will have the following pattern:
```
MAJOR.MINOR.PATCH
```

New releases are available on irregular basis. All the changes are reported in the [changelog](CHANGELOG).
If you want to get the latest version, you need to add a `--upgrade` flag to the installation command above.

## How to cite this package

If you use this package for your analyses, please consider using the following citation:

> Seignovert et al. 2023,
> Planetary coverage package (1.1.1),
> [planetary-coverage.org](https://docs.planetary-coverage.org/en/1.1.1/),
> [swh:1:rel:f6280865fc5144b4beadc04d4085385f7a58491d](https://archive.softwareheritage.org/browse/origin/directory/?origin_url=https://juigitlab.esac.esa.int/python/planetary-coverage&release=1.1.1)

or can use either:
- {download}`planetary-coverage.bib <../planetary-coverage.bib>`
- {download}`citation.cff <../CITATION.cff>`
- {download}`codemeta.json <../codemeta.json>`

## Issues and feedback

If you have any issue with this package, we highly recommend to take a look at:

- [this documentation](index) (the search bar on the left is there for you)
- [the gallery of examples](gallery/index.rst)
- [the example notebooks repository](https://juigitlab.esac.esa.int/notebooks/planetary-coverage)

If you did not find a solution there, don't hesitate to:

- [open an issue](https://juigitlab.esac.esa.int/python/planetary-coverage/-/issues/new), if you have an account on the [Juice Gitlab](https://juigitlab.esac.esa.int/python/planetary-coverage)
- send us an email at [&#99;&#111;&#110;&#116;&#97;&#99;&#116;&#64;&#112;&#108;&#97;&#110;&#101;&#116;&#97;&#114;&#121;&#45;&#99;&#111;&#118;&#101;&#114;&#97;&#103;&#101;&#46;&#111;&#114;&#103;](&#109;&#97;&#105;&#108;&#116;&#111;&#58;&#99;&#111;&#110;&#116;&#97;&#99;&#116;&#64;&#112;&#108;&#97;&#110;&#101;&#116;&#97;&#114;&#121;&#45;&#99;&#111;&#118;&#101;&#114;&#97;&#103;&#101;&#46;&#111;&#114;&#103;
)

The `planetary-coverage` is distributed under a open-source
[BSD-3 clauses](https://juigitlab.esac.esa.int/python/planetary-coverage/-/blob/main/LICENSE.md) license.
Bug reports and contributions are **encouraged** and **always welcome**.
