---
jupytext:
  formats: ipynb,md:myst
  text_representation:
    extension: .md
    format_name: myst
    format_version: 0.16
    jupytext_version: 1.14.1
kernelspec:
  display_name: Python 3
  language: python
  name: python3
---

# Planetary maps

To facilitate the representation of the data in their context, the `planetary-coverage` have a pre-build-in collection of planetary background maps.

```{code-cell}
:tags: [remove-input]

import matplotlib.pyplot as plt
```

:::{Warning}
By default the `planetary-coverage` computes [east planetocentric coordinates](https://naif.jpl.nasa.gov/pub/naif/toolkit_docs/Tutorials/pdf/individual_docs/17_frames_and_coordinate_systems.pdf). This means that the coordinates are provided with respect to the body reference **sphere** and the longitudes are defined __eastward__ (increasing from 0° to 360°).

The last [IAU report (2015)](https://doi.org/10.1007/s10569-017-9805-5) recommends to use [planetographic coordinates](../spice/ref.md#planetocentric-vs-planetocentric-coordinates) on celestial bodies (defined westward for prograde bodies with some exceptions). If you want to use these conventions, you will need to [convert the planetocentric coordinates into planetographic coordinates](../spice/ref.md#planetographic-to-planetocentric-coordinates) before displaying them on the map.
:::

## The Solar System

:::{Note}
All the background maps are represented here in [equirectangular](https://en.wikipedia.org/wiki/Equirectangular_projection) projection, with $\lambda_0 = 180°$ (central meridian) and $\phi_0 = 0°$ (cental parallel) and positive eastward longitudes. If you need to display west longitude ticks, see [this section](#customize-the-ticks).
:::


### Mercury

```{code-cell}
from planetary_coverage import MERCURY
```
```{code-cell}
:tags: [remove-input]

MERCURY
```

- __Source__: [USGS](https://planetarymaps.usgs.gov/mosaic/Mercury_MESSENGER_MDIS_Basemap_LOI_Mosaic_Global_166m.tif)
- __Original image size__: 46,080 × 92,160 pixels
- __Instrument__: Messenger WAC/NAC at 750 nm
- __Equatorial resolution__: 166 m / pixel
- __Mean radius__: 2,439.7 km  (from [pck00010.tpc][tpc] kernel).

[tpc]: https://spiftp.esac.esa.int/data/SPICE/JUICE/kernels/pck/pck00010.tpc

### Venus

```{code-cell}
from planetary_coverage import VENUS
```
```{code-cell}
:tags: [remove-input]

VENUS
```

- __Source__: [USGS](https://planetarymaps.usgs.gov/mosaic/Venus_Magellan_C3-MDIR_Colorized_Global_Mosaic_4641m.tif)
- __Original image size__: 8,192 × 4,096 pixels
- __Instrument__: C3-MDIR Synthetic Color Mosaic
- __Equatorial resolution__: 4.6 km / pixel
- __Mean radius__: 6,051.8 km  (from [pck00010.tpc][tpc] kernel).

### Earth

```{code-cell}
from planetary_coverage import EARTH
```
```{code-cell}
:tags: [remove-input]

EARTH
```

- __Source__: [NASA Blue Marble](https://visibleearth.nasa.gov/images/73751/july-blue-marble-next-generation-w-topography-and-bathymetry)
- __Original image size__: 86,400 × 43,200 pixels
- __Instrument__: Terra MODIS
- __Equatorial resolution__: 463 m / pixel
- __Mean radius__: 6,371.0 km  (from [pck00010.tpc][tpc] kernel).


### Moon

```{code-cell}
from planetary_coverage import MOON
```
```{code-cell}
:tags: [remove-input]

MOON
```

- __Source__: [USGS](https://planetarymaps.usgs.gov/mosaic/Lunar_LRO_LROC-WAC_Mosaic_global_100m_June2013.tif)
- __Original image size__: 109,164 × 54,582 pixels
- __Instrument__: LRO WAC
- __Equatorial resolution__: 100 m / pixel
- __Mean radius__: 1,737.4 km  (from [pck00010.tpc][tpc] kernel).

### Mars

```{code-cell}
from planetary_coverage import MARS
```
```{code-cell}
:tags: [remove-input]

MARS
```

- __Source__: [USGS](https://planetarymaps.usgs.gov/mosaic/Mars_Viking_ClrMosaic_global_925m.tif)
- __Original image size__: 11,530 × 23,059 pixels
- __Instrument__: Viking Orbiter (red and violet filters)
- __Equatorial resolution__: 925 m / pixel
- __Mean radius__: 3,389.5 km  (from [pck00010.tpc][tpc] kernel).

### Jupiter

:::{Warning}
:class: margin

For the Giant Planets (Jupiter, Saturn, Uranus and Neptune) the background maps
provided here are just for illustrative purposes but don't really represent the location
of the main feature on the top of the atmosphere (since they drift rapidly with time).

This map should be consistent with other planning tools
([Cosmographia](https://cosmoguide.org/), [MAPPS](https://www.cosmos.esa.int/web/esac-cmso/mapps),
[Juice Pointing Tool](https://juicept.esac.esa.int/) and [Gfinder](https://nmanaud.github.io/gfinder-docs/)).
:::

```{code-cell}
from planetary_coverage import JUPITER
```
```{code-cell}
:tags: [remove-input]

JUPITER
```

- __Source__: [JPL PIA07782](https://photojournal.jpl.nasa.gov/catalog/PIA07782)
- __Original image size__: 3,601 ×  1,801 pixels
- __Instrument__: Cassini ISS
- __Equatorial resolution__: 122 km / pixel
- __Mean radius__: 69,911.3 km  (from [pck00010.tpc][tpc] kernel).

### Io

```{code-cell}
from planetary_coverage import IO
```
```{code-cell}
:tags: [remove-input]

IO
```

- __Source__: [USGS](https://planetarymaps.usgs.gov/mosaic/Io_GalileoSSI-Voyager_Global_Mosaic_ClrMerge_1km.tif)
- __Original image size__: 11,445 × 5,723 pixels
- __Instrument__: Voyager ISS / Galileo SSI
- __Equatorial resolution__: 1 km / pixel
- __Mean radius__: 1,821.5 km  (from [pck00010.tpc][tpc] kernel).

### Europa

```{code-cell}
from planetary_coverage import EUROPA
```
```{code-cell}
:tags: [remove-input]

EUROPA
```

- __Source__: [NASA / JPL / Björn Jónsson](https://www.planetary.org/articles/0218-mapping-europa)
- __Original image size__: 20,000 × 10,000 pixels
- __Instrument__: Voyager ISS / Galileo SSI
- __Equatorial resolution__: 490 m / pixel
- __Mean radius__: 1,560.8 km  (from [pck00010.tpc][tpc] kernel).

### Ganymede

```{code-cell}
from planetary_coverage import GANYMEDE
```
```{code-cell}
:tags: [remove-input]

GANYMEDE
```

- __Source__: [Kersten et al. 2021 - PSS / DLR / Juno](https://janus.dlr.de/publications/ganymede/Ganymede_Mosaic_Juno.html) colorized with [Björn Jónsson](https://bjj.mmedia.is/data/ganymede/) map.
- __Original image size__: 46,080 × 23,040 pixels
- __Instrument__: Voyager ISS / Galileo SSI
- __Equatorial resolution__: 360 m / pixel
- __Mean radius__: 2,631.2 km (from [`pck00010.tpc`][tpc] kernel).

### Callisto

```{code-cell}
from planetary_coverage import CALLISTO
```
```{code-cell}
:tags: [remove-input]

CALLISTO
```

- __Source__: [USGS/ESA](https://www.cosmos.esa.int/web/juice/reference-maps-ganymede-europa-callisto-) colorized with [Björn Jónsson](https://bjj.mmedia.is/data/callisto/) map.
- __Original image size__: 15,138 × 7,569 pixels
- __Instrument__: Voyager ISS / Galileo SSI
- __Equatorial resolution__: 1 km / pixel
- __Mean radius__: 2,410.3 km (from [`pck00010.tpc`][tpc] kernel).

### Saturn

```{code-cell}
from planetary_coverage import SATURN
```
```{code-cell}
:tags: [remove-input]

SATURN
```

- __Source__: [Björn Jónsson](http://bjj.mmedia.is/data/saturn/index.html)
- __Original image size__: 2,880 ×  1,440 pixels
- __Instrument__: Cassini ISS
- __Equatorial resolution__: 122 km / pixel
- __Mean radius__: 58,232.0 km  (from [pck00010.tpc][tpc] kernel).

### Enceladus

```{code-cell}
from planetary_coverage import ENCELADUS
```
```{code-cell}
:tags: [remove-input]

ENCELADUS
```

- __Source__: [JPL PIA18435](https://photojournal.jpl.nasa.gov/catalog/PIA18435)
- __Original image size__: 15,960 ×  7,980 pixels
- __Instrument__: Cassini ISS (IR3-GRN-UV3)
- __Equatorial resolution__: 100 m / pixel
- __Mean radius__: 252.1 km  (from [pck00010.tpc][tpc] kernel).

### Titan

```{code-cell}
from planetary_coverage import TITAN
```
```{code-cell}
:tags: [remove-input]

TITAN
```

- __Source__: [JPL PIA22770](https://photojournal.jpl.nasa.gov/catalog/PIA22770)
- __Original image size__: 5,760 × 2,880 pixels
- __Instrument__: Cassini ISS (CB3 938 nm)
- __Equatorial resolution__: 2.8 km / pixel
- __Mean radius__: 2,574.8 km  (from [pck00010.tpc][tpc] kernel).

### Uranus

```{code-cell}
from planetary_coverage import URANUS
```
```{code-cell}
:tags: [remove-input]

URANUS
```

- __Source__: [Solid color (`#b6e0e3`) extracted from JPL PIA18182](https://photojournal.jpl.nasa.gov/catalog/PIA18182)
- __Original image size__: 1 × 1 pixels
- __Instrument__: Voyager ISS
- __Equatorial resolution__: 0 km / pixel
- __Mean radius__: 25,362.2 km  (from [pck00010.tpc][tpc] kernel).

### Neptune

```{code-cell}
from planetary_coverage import NEPTUNE
```
```{code-cell}
:tags: [remove-input]

NEPTUNE
```

- __Source__: [Björn Jónsson](http://bjj.mmedia.is/data/neptune/index.html)
- __Original image size__: 1,800 × 900 pixels
- __Instrument__: Voyager 2 ISS
- __Equatorial resolution__: 86 km / pixel
- __Mean radius__: 24,622.2 km  (from [pck00010.tpc][tpc] kernel).

### Pluto

```{code-cell}
from planetary_coverage import PLUTO
```
```{code-cell}
:tags: [remove-input]

PLUTO
```

- __Source__: [JPL PIA11707](https://photojournal.jpl.nasa.gov/catalog/PIA11707)
- __Original image size__: 5,926 × 2,963 pixels
- __Instrument__: New Horizon LORRI/Ralph
- __Equatorial resolution__: 1.2 km / pixel
- __Mean radius__: 1,195.0 km  (from [pck00010.tpc][tpc] kernel).

### Charon

```{code-cell}
from planetary_coverage import CHARON
```
```{code-cell}
:tags: [remove-input]

CHARON
```

- __Source__: [Antdoghalo](https://www.flickr.com/photos/60012030@N02/49983125091/)
- __Original image size__: 12,694 × 6,347 pixels
- __Instrument__: New Horizon LORRI/MVIC
- __Equatorial resolution__: 300 m / pixel
- __Mean radius__: 605.0 km  (from [pck00010.tpc][tpc] kernel).


## Get a map from the registry

All the default maps above are also available in the `MAPS` registry.
This allows you to load a `Map` programmatically with a string key:

:::{Tip}
:class: margin

The target name key is not case sensitive. You can also use a `SpiceRef` object.
:::

```{code-cell}
from planetary_coverage import MAPS

MAPS
```

## Customize the map

If none of these maps correspond to your needs, you can create your own custom map by providing a `2:1` background image centered a `180°` (preferred) or `0°`.
See [`Map`](../api/plot.md#map) object API for details.

For example, `GANYMEDE` is defined as:

```python
from planetary_coverage.maps import Map

GANYMEDE = Map('Ganymede_map_180.jpg', body='Ganymede', radius=2631.2)
```

### Plot data on the map

All these planetary maps are expected to be used in _matplotlib_ `projection` keyword:

```{code-cell}
:tags: [remove-input]

import matplotlib.pyplot as plt
from matplotlib.path import Path
```
```{code-cell}
lons_e, lats = [180, 150, 120, 150], [30, 50, 30, 10]

polygon = Path([
    (180, 30), (150, 50), (120, 30), (150, 10)
])
```
```{code-cell}
fig = plt.figure(figsize=(12, 9))
ax = fig.add_subplot(projection=GANYMEDE)

ax.plot(lons_e, lats, 'o', color='tab:red')

ax.add_path(polygon, facecolor='tab:orange', edgecolor='tab:green', alpha=.5);
```

### Customize the ticks

You can change the ticks display on the map (east longitude ticks to west longitude ticks, or/and, mirror the ticks on the secondary axis). For example:

:::{Warning}
:class: margin

As you see, the input values in `ax.plot(lons_e, lats)` must be in **east longitudes**, even if the required ticks representation is `'west'`.
:::

```{code-cell}
fig = plt.figure(figsize=(12, 9))
ax = fig.add_subplot(projection=MARS)

ax.plot([180, 150, 120, 150, 180], [30, 50, 30, 10, 30], 'o-')  # Always east longitude

ax.set_lon_ticks('west')
ax.set_lon_ticks('east', secondary=True)
ax.set_lat_ticks(secondary=True);
```
