---
jupytext:
  formats: ipynb,md:myst
  text_representation:
    extension: .md
    format_name: myst
    format_version: 0.16
    jupytext_version: 1.14.1
kernelspec:
  display_name: Python 3
  language: python
  name: python3
---

:::{admonition} Disclaimer
All the SPICE tools in this package should not be considered has a replacement of the SPICE toolkit. They follow [NAIF documentation](https://naif.jpl.nasa.gov/pub/naif/toolkit_docs/C/) as closely as possible, however, if you accounter an error with one of these functions, please [contact us](https://juigitlab.esac.esa.int/python/planetary-coverage/-/issues/) to report your issue so it can be fixed for everyone.
:::

# SPICE kernel

## Text kernel parser

```{code-cell}
from planetary_coverage.spice import kernel_parser
```

The function `kernel_parse()` allows the user to inspect quickly the content and the data in a text-kernel, without having to load it into the SPICE pool:
```{code-cell}
content, data = kernel_parser('naif0012.tls')
```

Here, the `content` variable is a full dump of the kernel content (text and data):
```{code-cell}
:tags: [output_scroll]
print(content)
```

The `data` variable on the other hand, only contains the parsed data in the kernel (between the `\begindata` and `\begintext` tags). The values are stored in a simple `dict` with the matching key:
```{code-cell}
:tags: [output_scroll]
data
```

As you can see above, the decimal value are parsed as `int` and `float`. When multiple value are present,
an parsed array is returned:
```{code-cell}
data['DELTET/M']
```

When a datetime is present (starting with a `@` symbol), it will be parsed as a `numpy.datetime64` object:
```{code-cell}
data['DELTET/DELTA_AT'][1]
```

This function also supports, line continuation (i.e. the string ending with `//`) and multi-values assignments (scattered in multiple lines or with the `+=` symbol).


## Text kernel formatter

```{code-cell}
from planetary_coverage.spice import format_data
```

The function `format_data()` is a handy shortcut to format your raw-data into a string that could be copy-pasted into your text-kernel:
```{code-cell}
format_data(single_float=1.2)
```

By default is indentation is set at 4 spaces, but you can use your own:
```{code-cell}
format_data(single_int=-1, indent=0)
```

You can define the format of your output:
```{code-cell}
format_data(single_float_d=1.657e-3, fmt='.3E')
```

The datetime are converted to SPICE time keys:
```{code-cell}
:tags: [remove-cell]
import numpy as np
```
```{code-cell}
format_data(
    single_date=np.datetime64('1987-01-31'),
    fmt=r'%d-%b-%Y',
)
```

If you provide multiple values, the `=` signs will the aligned:
```{code-cell}
print(format_data(
    vec_int_comma=(1, 3, 2),
    vec_float_comma=np.array([0., 0., 90.]),
    vec_str_comma=['FOO', 'BAR'],
    indent=2,
))
```

You can change the separation, if needed:
```{code-cell}
format_data(
    arr_int_space=(4, 5, 6),
    sep=' ',
)
```

The SPICE line constrains are enforce by default and
if the value is too long (> 132 characters), the result
will be split in multiple lines:
```{code-cell}
print(format_data(long_string=150 * 'a'))
```
