---
jupytext:
  formats: ipynb,md:myst
  text_representation:
    extension: .md
    format_name: myst
    format_version: 0.16
    jupytext_version: 1.14.1
kernelspec:
  display_name: Python 3
  language: python
  name: python3
---

```{code-cell}
:tags: ["remove-cell"]

!mk-download --kernels-dir kernels metakernel.tm
```

# SPICE metakernel

A metakernel is a configuration text-kernel with a required `KERNELS_TO_LOAD` key containing a list of kernel files to load in the SPICE pool. Most of the time, the exact location the kernels are abbreviated with a `PATH_SYMBOLS` and its `PATH_VALUES` keys. The values are usually hard coded in the file to a relative or an absolute path.

:::{admonition} Downloads
:class: margin

- {download}`metakernel.tm`
- [`naif0012.tls`](https://naif.jpl.nasa.gov/pub/naif/generic_kernels/lsk/naif0012.tls)
- [`pck00010.tpc`](https://naif.jpl.nasa.gov/pub/naif/generic_kernels/pck/pck00010.tpc)

See also [SPICE download CLI](cli).
:::

```
  PATH_VALUES       = ( '.' )

  PATH_SYMBOLS      = ( 'KERNELS' )

  KERNELS_TO_LOAD   = (

    '$KERNELS/lsk/naif0012.tls'
    '$KERNELS/pck/pck00010.tpc'
    ...

  )
```

## Read the metakernel

The `MetaKernel` object allows the user to parse these file and
if needed can change the `PATH_VALUES` at runtime:

:::{Hint}
:class: margin

You can provide any path:
```python
MetaKernel(
  'metakernel.tm',
  kernels='~/kernels/'
)
```
:::

```{code-cell}
from planetary_coverage import MetaKernel

mk = MetaKernel('metakernel.tm', kernels='kernels')

mk
```

In the example above, we have replace the initial value of `$KERNELS` (equal to `'.'`) by a custom value which corresponds to the actual location of our kernels (ie. the current directly `'kernels/'` in our case).

```{code-cell}
mk.kernels
```

If, at least one of the kernels listed is not available locally, the tool will throw a `MissingKernelError` error:

```{code-cell}
:tags: [raises-exception]
MetaKernel('metakernel.tm', kernels='/invalid/path')
```

:::{Note}
:class: margin

To speed us the kernel retrieval multiple download threads are started
at the same time. By default it is equal to the number of CPU seen by the
`multiprocessing` library but you can provide a custom value by yourself:

```python
MetaKernel(
  'metakernel.tm',
  kernels='/data/kernels/',
  download=True,
  n_threads=2,
)
```
:::
:::{Warning}
:class: margin

On Windows, due to a `RuntimeError` (see
[#80](https://juigitlab.esac.esa.int/python/planetary-coverage/-/issues/80)),
multi-thread download is disabled (`n_thread=1`) but you can
still increase the number of thread if you use it, with
``if __name__ == "__main__":`` syntax or in a Jupyter notebook.

:::

:::{Tip}
If a remote location (`https/http/ftp`) is present in the content of the metakernel (in the text or in the data fields) you can add a `download=True` argument to download the missing kernels:

```python
MetaKernel(
  'metakernel.tm',
  kernels='/data/kernels/',
  download=True,
)
```
```
[Download] https://naif.jpl.nasa.gov/pub/naif/generic_kernels/lsk/naif0012.tls
[Download] https://naif.jpl.nasa.gov/pub/naif/generic_kernels/pck/pck00010.tpc
```

You can also manually set a custom remote location (as a string) to pull your kernels from there:

```python
MetaKernel(
  'metakernel.tm',
  kernels='/data/kernels/',
  remote='https://remote.location.tld/',
  download=True,
)
```
```
[Download] https://remote.location.tld/lsk/naif0012.tls
[Download] https://remote.location.tld/pck/pck00010.tpc
```

These new kernels will stored at their expected location:
- `/data/kernels/lsk/naif0012.tls`
- `/data/kernels/pck/pck00010.tpc`

If multiple remote urls are present in the content (for example `https` and `ftp`), by default,
the tool will use the first listed (here `https`).
If you need to use an other one, you need to provide its index (as a integer) in the `remote` argument:

```python
MetaKernel(
  'metakernel.tm',
  kernels='/data/kernels/',
  remote=1,
  download=True,
)
```
```
[Download] ftp://remote.location.tld/naif0012.tls
[Download] ftp://remote.location.tld/pck00010.tpc
```

:::

To get a summary of the kernels listed in `KERNELS_TO_LOAD`, you can do:
```{code-cell}
mk.summary
```

## Load the kernels

Similarly to [`TourConfig`](../trajectory/spacecraft/tour_config.md#spice-pool-content) and [`Trajectory`](../api/trajectory.md#trajectory) objects, you can load the kernels in a `MetaKernel` at initialization (with `load_kernels=True` attribute) or later on with the `.load_kernels()` function:

```{code-cell}
MetaKernel('metakernel.tm', kernels='kernels', load_kernels=True)
# or
mk.load_kernels()
```

The `MetaKernel` object can also be used with [`spiceypy.furnsh()`](https://spiceypy.readthedocs.io/en/main/documentation.html#spiceypy.spiceypy.furnsh) to load its kernels list into the SPICE pool.

__However__, contrary to a regular metakernel, the tool will create a temporary copy of the metakernel on the fly with an edited set of `PATH_VALUES`. The original metakernel is kept untouched, the new file is stored in the system temporary folder. This allow the tool to redefined the location of the kernels at runtime (which was not possible with the default SPICE toolkit).

```{code-cell}
import spiceypy

spiceypy.kclear()    # Clear the kernel pool
spiceypy.furnsh(mk)  # Load the MetaKernel object
```

Now we can inspect the pool to see its content:

:::{Tip}
:class: margin
The same check can be performed even quicker with the `SpicePool` object (see [SPICE toolbox](pool) for details).
:::

```{code-cell}
nb_kernels_loaded = spiceypy.ktotal('ALL')
nb_kernels_loaded
```
```{code-cell}
:tags: [margin]
from planetary_coverage import SpicePool

SpicePool
```

```{code-cell}
for i in range(nb_kernels_loaded):
    print(spiceypy.kdata(i, 'ALL')[0])
```


:::{Caution}
The temporary file, `/tmp/metakernel-xxxxxxxx.tm`, is automatically removed after being loaded in the pool. If you need to keep this file open (for example to inspect it), you should use the python context manager (with `with` syntax):
:::

```{code-cell}
from pathlib import Path

with mk as f:
    print(f)                 # New tmp mk file
    print(Path(f).exists())  # The tmp mk exists only here

print(Path(f).exists())      # The tmp mk no longer exists
```


:::{Important}
Even if the name of the temporary metakernel file change each time you load it, its `hash` is always based on its kernels content and will remain constant, even if you load and purge the pool multiple times. Therefore, it is possible to perform a comparison between the SPICE pool content and the new content provided by the metakernel (to avoid to reload the pool if the list of the kernels to load is the same).
:::

```{code-cell}
SpicePool.add(mk, purge=True)

SpicePool, hash(SpicePool)
```
```{code-cell}
SpicePool.add(mk, purge=True)

SpicePool, hash(SpicePool)
```
