---
jupytext:
  formats: ipynb,md:myst
  text_representation:
    extension: .md
    format_name: myst
    format_version: 0.16
    jupytext_version: 1.14.1
kernelspec:
  display_name: Python 3
  language: python
  name: python3
---

# SPICE Pool

The SPICE calculations are performed on a loaded set of kernels, usually with the `spiceypy.furnsh()` function. These kernels are cached in the SPICE pool and accessible to all the SPICE functions.

To easily manipulate this pool, to add or remove kernels and to quickly check its content, we created a python `SpicePool` object that acts as a _singleton_.

```{code-cell}
from planetary_coverage import SpicePool
```

## Add and remove kernels to the pool

You can add one or many kernel(s) with:

```{admonition} Download
:class: margin

- [`naif0012.tls`](https://naif.jpl.nasa.gov/pub/naif/generic_kernels/lsk/naif0012.tls)
- [`pck00010.tpc`](https://naif.jpl.nasa.gov/pub/naif/generic_kernels/pck/pck00010.tpc)
```

```python
SpicePool.add('naif0012.tls')
SpicePool.add('pck00010.tpc')
```

or directly with a list of kernels:

:::{Warning}
:class: margin

If a `kernel` is already present in the pool, it will throw a `ValueError` exception to ensure that kernels are only loaded once. You need to explicitly remove it from the pool before adding it again or use `purge=True` (see below).
:::

```{code-cell}
SpicePool.add(['naif0012.tls', 'pck00010.tpc'])
```

We can inspect the content of the pool and see our two loaded kernels:
```{code-cell}
SpicePool
```

To get a summary of the pool content, grouped by kernel types, use the `.summary` property:
```{code-cell}
SpicePool.summary
```

If you need an explicit list of the loaded kernels, you can use the `.kernels` property:
```{code-cell}
SpicePool.kernels
```

You can remove a kernel from the pool with:

:::{Hint}
:class: margin

Logical operations to add/remove a kernel(s) to/from the pool are also supported:

```python
SpicePool + 'naif0012.tls'
SpicePool - 'naif0012.tls'
```
:::

```{code-cell}
SpicePool.remove('naif0012.tls')

SpicePool
```

and check if the kernel is still in the pool:

```{code-cell}
'naif0012.tls' in SpicePool
```
```{code-cell}
'pck00010.tpc' in SpicePool
```

You can also check the full content of the pool against a list of expected kernels:

```{code-cell}
SpicePool == ['pck00010.tpc']
```

## Get values from pool

Reload the kernels in the pool:

:::{Tip}
:class: margin

Here the `purge=True` ensure that only these kernels will be present
in the pool any previously loaded kernel will be purged.

You can also use `SpicePool.purge()` to explicitly purge the pool.
:::

```{code-cell}
SpicePool.add(['naif0012.tls', 'pck00010.tpc'], purge=True)

SpicePool
```

When your set of kernels is loaded, you can request any values in the pool with the python getter syntax:

```{code-cell}
SpicePool['DELTET/DELTA_T_A']
```
```{code-cell}
SpicePool['BODY503_RADII']
```

You can also search by wild string (`*`) or wild characters (`%`):
```{code-cell}
SpicePool['BODY50%_RADII']
```
```{code-cell}
SpicePool['BODY503_*']
```

If the value is not present in the pool a `ValueError` will be thrown:

```{code-cell}
:tags: [raises-exception]
SpicePool['FOO']
```

## Kernels temporal coverage

We will start by loading all the kernels from Juice 5.1 150lb metakernel:

```{code-cell}
from planetary_coverage import TourConfig

TourConfig(mk='5.1 150lb_23_1', version='v422_20230130_002', load_kernels=True)
```

Sometimes it can be difficult to get the complete range of valid times that can be use with the kernels loaded in the pool.
The `SpicePool` provide a helper that will search all `CK`, `PCK` and `SPK` kernels to find the coverage window associated with a given body:

```{code-cell}
SpicePool.windows('GANYMEDE')
```

You can select different types of output format, with the `fmt` argument:

```{code-cell}
SpicePool.windows('GANYMEDE', fmt='ET')
```

You can also provide a list of bodies to get all their coverage windows:

```{code-cell}
SpicePool.windows('JUICE', 'GANYMEDE')
```

If you need the overlapping coverage (i.e. max start time and min stop time), you can use the `.coverage()` method:

```{code-cell}
SpicePool.coverage('JUICE', 'GANYMEDE', fmt='TDB')
```

This coverage correspond to the **maximum temporal coverage** including the gaps (if some are present). If you are interested by the temporal gaps, you can retreive them with `.gaps()` method:

```{code-cell}
SpicePool.gaps('JUICE_SPACECRAFT_PLAN', fmt='TDB')
```

You can also get a list of all bodies temporal coverage with the `.brief()` method (similar to NAIF `brief -t -a` utility):
```{code-cell}
:tags: [output_scroll]

SpicePool.brief(fmt='TDB')
```

## Enable pool debugger

If you need to debug the SPICE pool, you can increase the level of the verbose with the `debug_spice_pool()` function.

```{code-cell}
from planetary_coverage.debug import debug_spice_pool

debug_spice_pool(True)
```

```{code-cell}
SpicePool.add('naif0012.tls', purge=True)
```

```{code-cell}
SpicePool.remove('naif0012.tls')
```

```{code-cell}
SpicePool.purge()
```

You can decrease the verbose level to `INFO` with an `'info'` key:

```{code-cell}
debug_spice_pool('info')
```

Now only, the purge of the pool will trigger a log output:

```{code-cell}
SpicePool.add('naif0012.tls')
```

```{code-cell}
SpicePool.remove('naif0012.tls')
```

```{code-cell}
SpicePool.purge()
```

To disable the SPICE pool verbose, pass a `False` key:

```{code-cell}
debug_spice_pool(False)
```
