---
jupytext:
  formats: ipynb,md:myst
  text_representation:
    extension: .md
    format_name: myst
    format_version: 0.16
    jupytext_version: 1.14.1
kernelspec:
  display_name: Python 3
  language: python
  name: python3
---

# SPICE References

```{code-cell}
:tags: [remove-cell]

from planetary_coverage import TourConfig

TourConfig(mk='5.1 150lb_23_1', version='v422_20230130_002', load_kernels=True)
```

Object references in the SPICE kernels are usually represented either by their `name` (which is not always unique) or by an internal `id` code. Most of the time, the user does not have to know the reference `id` and can pass directly its name to the function but it's not always the case. To provide a simpler access to the SPICE references, the `planetary-coverage` provides a `SpiceRef` object that will extract the correct information from the [`SpicePool`](pool).


```{code-cell}
from planetary_coverage import SpiceRef
```

For example, for `Jupiter barycenter` we have:

:::{Note}
:class: margin

If the reference is not known, a `spiceypy` error will be thrown with an additional `ValueError`:

```python
>>> SpiceRef('FOO')
...
NotFoundError: Spice returns not found for function: bods2c
...
ValueError: Unknown reference: `FOO`
```
:::

```{code-cell}
ref = SpiceRef('Jupiter Barycenter')
ref
```

## Generic `SpiceRef` properties

This reference can be used either as a string to get its `name` property:

```{code-cell}
str(ref)
```

or as an integer to get its `id` property:

```{code-cell}
int(ref)
```

You can compare a `SpiceRef` with its string and integer values:

```{code-cell}
ref == 'JUPITER BARYCENTER'
```

```{code-cell}
ref == 5
```

If a reference frame is available for this object you can use the `.frame` property to query it:
```{code-cell}
ref.frame
```

Similarly to the `SpicePool`, you can query the values from the pool directly on the `SpiceRef` object:

:::{Warning}
:class: margin

By default, the value is not specific to the `SpiceRef`, any value in the `SpicePool` is accessible.
:::
```{code-cell}
ref['BODY599_RADII']
```

## `SpiceBody` properties

If the `SpiceRef` is the Sun, a [Planet][naif-ids-planets], a [Moon][naif-ids-planets]
or [Satellite][naif-ids-planets], a [Comet][naif-ids-commets] or an [Asteroid][naif-ids-asteroids],
the `SpiceRef` is promoted to a `SpiceBody`:

[naif-ids-planets]: https://naif.jpl.nasa.gov/pub/naif/toolkit_docs/FORTRAN/req/naif_ids.html#Planets%20and%20Satellites
[naif-ids-commets]: https://naif.jpl.nasa.gov/pub/naif/toolkit_docs/FORTRAN/req/naif_ids.html#Comets
[naif-ids-asteroids]: https://naif.jpl.nasa.gov/pub/naif/toolkit_docs/FORTRAN/req/naif_ids.html#Asteroids

```{code-cell}
body = SpiceRef('Ganymede')
body
```

```{code-cell}
SpiceRef('Sun'), SpiceRef('Mars'), SpiceRef('CHURYUMOV-GERASIMENKO'), SpiceRef('RYUGU')
```
:::{versionadded} 1.2.0
Comet and Asteroid support.
:::

You can distinguish the different types of body with `.is_planet`, `.is_satellite`, `.is_comet` and
`.is_asteroid` properties:

```{code-cell}
body.is_planet, body.is_satellite, body.is_comet, body.is_asteroid
```

For a `SpiceBody`, you have access to additional attributes (mainly from the `pck` kernels).
For example, the body `radii` values:

```{code-cell}
body.radii  # [km]
```

its _geometric-mean_ radius:

```{code-cell}
body.radius  # [km]
```

or its centered body-fixed frame:

```{code-cell}
body.frame
```

You can also derived the name of the parent body:

```{code-cell}
body.parent
```

and its barycenter:

```{code-cell}
body.barycenter
```

For a `SpiceBody`, the item getter is prefix with `BODYXXX_` with `XXX` being the body code. Therefore you can do:

```{code-cell}
body['GM']  # Alias to `BODY503_GM`
```

## `SpiceSpacecraft` properties

If you select a spacecraft reference, it will be promoted to a `SpiceSpacecraft`:

```{code-cell}
sc = SpiceRef('JUICE')
sc
```

`SpiceSpacecraft` includes a `frame` and `boresight` attributes:

```{code-cell}
sc.frame
```
```{code-cell}
sc.boresight
```

You can also access the list of the spacecraft instruments available in the SPICE pool:

```{code-cell}
:tags: [output_scroll]

sc.instruments
```

or select a single one:

```{code-cell}
inst = sc.instr('JANUS')
inst
```

## `SpiceInstrument` properties

When an instrument is selected (as a `SpiceInstrument`), you can get its field of view parameters:

```{code-cell}
inst.boresight
```
```{code-cell}
inst.bounds
```
```{code-cell}
inst.shape
```
```{code-cell}
inst.frame
```

You can also retrieve the detector size:
```{code-cell}
inst.ns, inst.nl
```

or the full field-of-view along the track and cross track:
```{code-cell}
inst.fov_along_track, inst.fov_cross_track  # in radians
```

:::{Caution}
:class: margin
We assume that the sample direction is
aligned with the cross-track direction (ie. 1-line acquisition
in push-broom mode should be in the direction of flight).

This convention is not strictly enforced in the instrument kernel
and **the user must be careful** and adjust theses values according
to the detector orientation relative to the spacecraft motion.

:::

:::{Warning}
:class: margin
`JUICE_JANUS` instrument in `v06` does not follow this convention.
We manually manage this exception for the moment (see
[MR !27](https://juigitlab.esac.esa.int/python/planetary-coverage/-/merge_requests/27)
for more details).
:::

as well as the pixel instantaneous field-of-view:
```{code-cell}
inst.ifov_along_track, inst.ifov_cross_track  # in radians
```

For a `SpiceInstrument`, the item getter is prefix with `INSXXX_` with `XXX` being the instrument code. Therefore you can do:

```{code-cell}
inst['PIXEL_SAMPLES']  # Alias to `INS-28200_PIXEL_SAMPLES`
```

You can display the `boresight`, the `bounds` vectors and `rays` of the field of view contour in the instrument frame:

```{code-cell}
inst.display();
```

Or you can represent it in the plane orthogonal to the field of view:

```{code-cell}
inst.view();
```

## `SpiceFrame` properties

If the `SpiceRef` is a SPICE frame, the `SpiceRef` is promoted to a `SpiceFrame`:

```{code-cell}
frame = SpiceRef('JUICE_SPACECRAFT_PLAN')
frame
```

You can retrieve the type of frame:
```{code-cell}
frame.class_type
```

The frame body center:
```{code-cell}
frame.center
```

and when available, the associated `sclk` and `spk`:
```{code-cell}
frame.sclk
```
```{code-cell}
frame.spk
```


:::{Danger}
SPICE codes can overlap SPICE frame ids. For example here, both `MERCURY_BARYCENTER` and `J2000` frame have an `id` of `1`. It is recommend to be very carefull with theses cases and use explicit names:
:::

```{code-cell}
SpiceRef('MERCURY_BARYCENTER')
```
```{code-cell}
SpiceRef('J2000')
```



## Planetocentric vs. planetocentric coordinates

Recommended reading: [NAIF reference frames and coordinates system](https://naif.jpl.nasa.gov/pub/naif/toolkit_docs/Tutorials/pdf/individual_docs/17_frames_and_coordinate_systems.pdf)

### Planetocentric coordinates

**NAIF definition:**

For planets and their satellites the `+Z` axis (`+90` latitude) always points
to the north side of the invariable plane (the plane whose normal vector is
the angular momentum vector of the solar system):

- Planetocentric longitude increases positively eastward (-180° to +180°)
- Planetocentric latitude increases positively northward (-90° to +90°)

```{image} ../images/NAIF_planetocentric.png
:width: 300
:alt: NAIF planetocentric coordinates
:align: center
```

:::{Note}

In the `planetary-coverage`, the planetocentric longitudes are defined as strictly positive eastward (`[0°, 360°[`).
:::


### Planetographic coordinates

**NAIF definition:**

For planet and satellite planetographic coordinate systems:

- Planetographic longitude is usually defined such that the
  **sub-observer longitude increases** with time as seen by a
  distant, fixed observer (0° to 360°)
- The Earth, Moon and Sun are exceptions; planetographic
  longitude is positive east by default (0° to 360°)
- Planetographic latitude is planetodetic latitude (-90° to +90°), ie.
  is **tied to a reference ellipsoid** and for a point, P, on a reference
  ellipsoid, the angle measured from the X-Y plane to the surface normal
  at the point of interest. For points not on the ellipsoid, equals latitude
  at the nearest point on the reference ellipsoid and increases positively northward.

```{image} ../images/NAIF_planetographic.png
:width: 300
:alt: NAIF planetographic coordinates
:align: center
```

### Planetographic to Planetocentric coordinates.

:::{Warning}
:class: margin

Depending on the body, the output will be either in **east** or in **west** longitude.
:::

It is possible to convert planetographic coordinates to planetographic coordinates with the `ocentric2ographic()` function:

:::{Note}
:class: margin

This calculation requires that a `tpc` kernel is loaded into the SPICE pool.
:::
```{code-cell}
from planetary_coverage.spice import ocentric2ographic

lon_w, lat = ocentric2ographic('MARS', 90, 45)  # east longitude, latitude

lon_w, lat  # west longitudes for MARS
```
