"""CK and SPK kernels trimmer."""

from argparse import ArgumentParser
from datetime import timedelta
from pathlib import Path
from shutil import copytree
from subprocess import run
from sys import stdout

from toml import load

from planetary_coverage.misc import wget
from planetary_coverage.spice.kernel import format_data, get_data


def cli():
    """Command line interface."""
    parser = ArgumentParser(description='CK and SPK kernel trimmer.')

    parser.add_argument(
        '--kernels-dir', default='originals', help='Kernels dataset original location.'
    )

    parser.add_argument(
        '--output', default='trimmed', help='Trimmed kernels dataset output location.'
    )

    parser.add_argument(
        '-q', '--quiet', action='store_true', help='Disable verbose output.'
    )

    parser.add_argument(
        '--keep-config', action='store_true', help='Keep the temporary conf files.'
    )

    parser.add_argument(
        'toml',
        default='config.toml',
        nargs='?',
        metavar='config.toml',
        help='Input TOML configuration file.',
    )

    args, _ = parser.parse_known_args()

    # Copy all the originals kernels in a trimmed folder
    if not args.quiet:
        stdout.write(
            f'>>> Duplicate `{args.kernels_dir}` kernels dataset into `{args.output}`.\n'
        )

    copytree(args.kernels_dir, args.output, dirs_exist_ok=True)

    # Trim CK kernels
    if not args.quiet:
        stdout.write(f'>>> Trim CK files from {args.toml}.\n')

    ckslicer(args.kernels_dir, args.output, args.toml)

    # Trim SPK kernels
    if not args.quiet:
        stdout.write(f'>>> Trim SPK files from {args.toml}.\n')

    spkmerge(args.kernels_dir, args.output, args.toml, keep_config=args.keep_config)

    # Trim DSK kernels
    if not args.quiet:
        stdout.write(f'>>> Trim DSK files from {args.toml}.\n')

    mkdsk(args.kernels_dir, args.output, args.toml, keep_config=args.keep_config)

    # Trim PCK kernels
    if not args.quiet:
        stdout.write(f'>>> Trim PCK files from {args.toml}.\n')

    pck_bpc(args.output, args.toml)


def ckslicer(kernels_dir, kernels_output, config_toml):
    """Trim CK kernels."""
    data = load(config_toml)

    lsk = f'{kernels_dir}/{data["lsk"]}'

    for key in data:
        if key != 'lsk':
            for fname, segments in data[key].items():
                if fname.endswith('.bc'):
                    inputck = f'{kernels_dir}/{key}/{fname}'
                    outputck = f'{kernels_output}/{key}/{fname}'
                    kernels_input = f'{kernels_dir}/{key.rsplit("/", 1)[0]}'

                    if len(segments) > 1:
                        raise ValueError('Only 1 segment can be provided in CKSLICER.')

                    for seg in segments:
                        # Remove existing output CK
                        Path(outputck).unlink(missing_ok=True)

                        # Execute NAIF CKSLICER function
                        # (https://naif.jpl.nasa.gov/naif/utilities.html)
                        run(
                            [
                                'ckslicer',
                                '-lsk',
                                lsk,
                                '-sclk',
                                f'{kernels_input}/{seg["sclk"]}',
                                '-inputck',
                                inputck,
                                '-outputck',
                                outputck,
                                '-start',
                                str(seg['begin'] - timedelta(days=1)),
                                '-stop',
                                str(seg['end'] + timedelta(days=1)),
                            ],
                            check=True,
                        )


def spkmerge(kernels_dir, kernels_output, config_toml, keep_config=False):
    """Trim SPK kernels."""
    data = load(config_toml)

    conf = f'LEAPSECONDS_KERNEL    = {kernels_dir}/{data["lsk"]}\n'

    kernels = []
    for key in data:
        if key != 'lsk':
            for fname, segments in data[key].items():
                if fname.endswith('.bsp'):
                    kernels.append(f'{key}/{fname}')
                    conf += f'\nSPK_KERNEL            = {kernels_output}/{key}/{fname}\n'

                    for seg in segments:
                        # Increase time window by ±1 day
                        begin = seg['begin'] - timedelta(days=1)
                        end = seg['end'] + timedelta(days=1)

                        conf += f'    SOURCE_SPK_KERNEL = {kernels_dir}/{key}/{fname}\n'
                        conf += f'        BODIES        = {seg["bodies"]}\n'
                        conf += f'        BEGIN_TIME    = {begin.isoformat()}\n'
                        conf += f'        END_TIME      = {end.isoformat()}\n'

    if kernels:
        # Save the configuration file
        cfg = Path('spkmerge.cfg')
        cfg.write_text(conf, encoding='utf-8')

        # Remove existing output SPK
        for kernel in kernels:
            Path(f'{kernels_output}/{kernel}').unlink(missing_ok=True)

        # Execute NAIF SPKMERGE function (https://naif.jpl.nasa.gov/naif/utilities.html)
        run(['spkmerge', cfg], check=True)

        if not keep_config:
            cfg.unlink()


def mkdsk(kernels_dir, kernels_output, config_toml, keep_config=False):
    """Trim DSK kernels."""
    data = load(config_toml)

    for key in data:
        if key != 'lsk':
            for fname, patches in data[key].items():
                if fname.endswith('.bds'):
                    output_dsk_file = Path(f'{kernels_output}/{key}/{fname}')

                    # Remove if exists
                    if output_dsk_file.exists():
                        output_dsk_file.unlink()

                    # Get previous config
                    params = dsk_commnt(kernels_dir, key, fname, patches)

                    # Create config file
                    dsk_cfg = '\n'.join([
                        '*****************************************************',
                        'Planetary-coverage trimmed DSK file',
                        '',
                        'DO NOT USE this kernel in your production environment',
                        '',
                        '*****************************************************',
                        '',
                        r'\begindata',
                        '',
                        format_data(
                            input_shape_file='dummy_cube.obj',
                            output_dsk_file=output_dsk_file,
                            comment_file=' ',
                            kernels_to_load=params['KERNELS_TO_LOAD'],
                            surface_name=params['SURFACE_NAME'],
                            center_name=params['CENTER_NAME'],
                            ref_frame_name=params['REF_FRAME_NAME'],
                            start_time='1950-JAN-1/00:00:00',
                            stop_time='2050-JAN-1/00:00:00',
                            data_class=1,
                            input_data_units=(
                                'ANGLES    = DEGREES',
                                'DISTANCES = KILOMETERS',
                            ),
                            coordinate_system='LATITUDINAL',
                            minimum_latitude=-90.0,
                            maximum_latitude=90.0,
                            minimum_longitude=-180.0,
                            maximum_longitude=180.0,
                            data_type=2,
                            plate_type=3,
                        ),
                        format_data(
                            naif_surface_name=params['NAIF_SURFACE_NAME'][0],
                            naif_surface_code=params['NAIF_SURFACE_CODE'][0],
                            naif_surface_body=params['NAIF_SURFACE_BODY'][0],
                            eq='+=',
                        ),
                        '',
                    ])

                    cfg = Path(fname.replace('.bds', '.cfg'))
                    cfg.write_text(dsk_cfg, encoding='utf-8')

                    # Execute NAIF MKDSK function
                    # (https://naif.jpl.nasa.gov/naif/utilities.html)
                    run(['mkdsk', '-setup', cfg], check=True)

                    if not keep_config:
                        cfg.unlink()


def dsk_commnt(kernels_dir, key, fname, patches):
    """Extract dsk comments values."""
    # Load the DSK parameters from the commnt function
    out = run(
        [
            'commnt',
            '-r',
            f'{kernels_dir}/{key}/{fname}',
        ],
        check=True,
        capture_output=True,
        text=True,
    )

    params = get_data(out.stdout.replace('*', ''))  # Remove ***** lines

    # Fix relative paths
    for i, kernel in enumerate(params['KERNELS_TO_LOAD']):
        if '..' in kernel:
            _, folder, k = kernel.rsplit('/', 2)
            params['KERNELS_TO_LOAD'][i] = f'{folder}/{k}'

    # Fix old fk frames
    for old, new in patches.items():
        if old in params['KERNELS_TO_LOAD']:
            i = params['KERNELS_TO_LOAD'].index(old)
            params['KERNELS_TO_LOAD'][i] = new

    # Update list kernels to load
    kernels_input = f'{kernels_dir}/{key.rsplit("/", 1)[0]}'

    # Prepend the kernels paths
    params['KERNELS_TO_LOAD'] = [
        f'{kernels_input}/{k}' for k in params['KERNELS_TO_LOAD']
    ]

    return params


def pck_bpc(kernels_output, config_toml):
    """Replace PCK binary kernels."""
    data = load(config_toml)

    for key in data:
        if key != 'lsk':
            for fname, segments in data[key].items():
                if fname.endswith('.bpc'):
                    kernel = f'{kernels_output}/{key}/{fname}'
                    url = segments[0]['url']
                    wget(url, kernel, force=True)


if __name__ == '__main__':
    cli()
