"""Test CLI kernel module."""

import sys
from urllib.error import HTTPError

from pytest import raises

import planetary_coverage.cli.kernel as mk_module
from planetary_coverage.cli import cli_kernel_download


def test_cli_kernel(monkeypatch, capsys):
    """Test CLI kernel."""

    def mock_wget_stdout(url, fname, **kwargs):
        """Mock wget function arguments to stdout."""
        sys.stdout.write(f'{url} {fname}')

        for key, value in kwargs.items():
            if value:
                sys.stdout.write(f' {key}={value}')

        sys.stdout.write('\n')

    monkeypatch.setattr(mk_module, 'wget', mock_wget_stdout)

    # Single kernel provided (use NAIF generic kernels)
    cli_kernel_download(['lsk/naif0012.tls'])
    assert capsys.readouterr().out == (
        'https://naif.jpl.nasa.gov/pub/naif/generic_kernels/lsk/naif0012.tls '
        'lsk/naif0012.tls\n'
    )

    # Multiple kernels
    cli_kernel_download(['lsk/naif0012.tls', 'lsk/naif0012.tls'])
    assert capsys.readouterr().out == (
        'https://naif.jpl.nasa.gov/pub/naif/generic_kernels/lsk/naif0012.tls '
        'lsk/naif0012.tls\n'
        'https://naif.jpl.nasa.gov/pub/naif/generic_kernels/lsk/naif0012.tls '
        'lsk/naif0012.tls\n'
    )

    # Custom remote location
    cli_kernel_download(['--remote', 'https://example.org', 'foo.ck'])
    assert capsys.readouterr().out == 'https://example.org/foo.ck foo.ck\n'

    # Custom output kernel directory
    cli_kernel_download(['--kernels-dir', 'kernels', 'foo.ck'])
    assert capsys.readouterr().out == (
        'https://naif.jpl.nasa.gov/pub/naif/generic_kernels/foo.ck kernels/foo.ck\n'
    )

    # NASA remote
    cli_kernel_download(['--nasa', 'JUNO', 'foo.ck'])
    assert capsys.readouterr().out == (
        'https://naif.jpl.nasa.gov/pub/naif/JUNO/kernels/foo.ck foo.ck\n'
    )

    # ESA remote
    cli_kernel_download(['--esa', 'JUICE', 'foo.ck'])
    assert capsys.readouterr().out == (
        'https://spiftp.esac.esa.int/data/SPICE/JUICE/kernels/foo.ck foo.ck\n'
    )

    # JAXA remote
    cli_kernel_download(['--jaxa', 'HAYABUSA', 'foo.ck'])
    assert capsys.readouterr().out == (
        'https://data.darts.isas.jaxa.jp/pub/spice/HAYABUSA/kernels/foo.ck foo.ck\n'
    )

    # Force flag
    cli_kernel_download(['--force', '--esa', 'JUICE', 'foo.ck'])
    assert capsys.readouterr().out == (
        'https://spiftp.esac.esa.int/data/SPICE/JUICE/kernels/foo.ck foo.ck force=True\n'
    )

    # Skip flag
    cli_kernel_download(['--skip', '--esa', 'JUICE', 'foo.ck'])
    assert capsys.readouterr().out == (
        'https://spiftp.esac.esa.int/data/SPICE/JUICE/kernels/foo.ck foo.ck skip=True\n'
    )

    # Remote location (no kernels requested)
    cli_kernel_download(['--esa', 'JUICE'])
    assert capsys.readouterr().out == (
        'https://spiftp.esac.esa.int/data/SPICE/JUICE/kernels/\n'
    )


def test_cli_kernel_errors(monkeypatch, capsys):
    """Test CLI kernel errors."""
    # Multiple remote error
    with raises(SystemExit):
        cli_kernel_download(['--nasa', 'JUNO', '--esa', 'JUICE', 'foo.ck'])

    assert capsys.readouterr().err == (
        '>>> ERROR: 2 remotes provided (nasa, esa). Only one is accepted.\n'
    )

    # wget with FileExistsError
    def mock_wget_file_exits_error(*args, **kwargs):
        """Mock wget with FileExistsError."""
        raise FileExistsError

    monkeypatch.setattr(mk_module, 'wget', mock_wget_file_exits_error)

    with raises(SystemExit):
        cli_kernel_download(['--esa', 'JUICE', 'foo.ck'])

    assert capsys.readouterr().err == (
        '>>> ERROR: foo.ck is already present locally. '
        'Use `--force` to overwrite it or `--skip` to continue.\n'
    )

    # wget with HTTPError
    def mock_wget_http_error(*args, **kwargs):
        """Mock wget with HTTP error."""
        raise HTTPError('example.org', 404, 'Not found', None, None)

    monkeypatch.setattr(mk_module, 'wget', mock_wget_http_error)

    with raises(SystemExit):
        cli_kernel_download(['--esa', 'JUICE', 'foo.ck'])

    assert capsys.readouterr().err == ('>>> ERROR: kernel not found.\n')
