"""Test ESA API module."""

import re
from io import BytesIO
from json import JSONDecodeError
from pathlib import Path
from urllib.error import HTTPError, URLError

from pytest import fixture, raises

import planetary_coverage.esa.api as api_module
from planetary_coverage.esa import get_mk, get_tag
from planetary_coverage.esa.api import (
    EsaApiNotAvailableError,
    EsaApiNotFoundError,
    EsaMissionMetakernels,
    crema_ref,
    esa_api,
    esa_mk_name,
    get_mk_offline,
    get_tag_offline,
)


SKD_VERSION = re.compile(r'v\d{3}_\d{8}_\d{3}')

MOCK_ESA_MK_CACHE = Path(__file__).parent / '..' / 'data' / 'esa-mk'


@fixture
def mock_url_resp(monkeypatch):
    """Mock URL response."""

    def url_response(data, err=None):
        """Dummy url response data/err."""
        resp = iter(data if isinstance(data, list) else [data])

        class DummyResponse:
            """Dummy data response."""

            def __init__(self, url):
                self.url = url

            def __enter__(self):
                """Dummy reader function."""
                return BytesIO(next(resp).encode())

            def __exit__(self, *args):
                pass

        def dummy_error(_):
            """Dummy error response."""
            raise err

        monkeypatch.setattr(api_module, 'urlopen', dummy_error if err else DummyResponse)

    return url_response


def test_esa_api_cache(mock_url_resp):
    """Test ESA API cache."""
    # The first response to cache
    mock_url_resp('{"values":true}')
    assert esa_api('foo')

    # Change response on a new key
    mock_url_resp('{"children":{"values":false}}')
    assert not esa_api('bar')

    # Reload from cache (even if the response evolved)
    assert esa_api('foo')

    # Response splitted in 3 payloads
    mock_url_resp([
        '{"values":[0, 1],"nextPageStart":2,"start":0}',
        '{"values":[2, 3],"nextPageStart":4,"start":2}',
        '{"values":[4],"start":4}',
    ])
    assert esa_api('baz') == [0, 1, 2, 3, 4]

    # Response splitted in 2 payload but with a limit
    mock_url_resp(['{"values":[0, 1],"nextPageStart":2}', '{"values":[2]}'])
    assert esa_api('qux', limit=2) == [0, 1]


def test_esa_api_err(mock_url_resp):
    """Test ESA API error."""
    # Bad request (404 error)
    mock_url_resp(None, err=HTTPError('example.org', 404, 'Bad request', {}, BytesIO()))
    with raises(EsaApiNotFoundError, match='HTTP 404 error: `example.org/404`.'):
        _ = esa_api('example.org/404')

    # Service not available (503 error)
    mock_url_resp(None, err=URLError('503 Service Unavailable'))
    with raises(EsaApiNotAvailableError, match='ESA server not available'):
        _ = esa_api('example.org/503-service-unavailable')

    # The response is not a JSON
    mock_url_resp('<html><body><p>Not a JSON</p></body></html>.')
    with raises(JSONDecodeError):
        _ = esa_api('html')

    # JSON without values keyword
    mock_url_resp('{"size":1,"limit":25,"isLastPage":true}')
    with raises(KeyError):
        _ = esa_api('missing-values')

    # Valid JSON without values
    mock_url_resp('{"size":1,"limit":25,"isLastPage":true,"values":[]}')
    with raises(ValueError):
        _ = get_tag('empty-values')


def test_esa_api_get_tag():
    """Test get tags from ESA API."""
    # Latest tag
    latest_tag = get_tag('juice')

    assert SKD_VERSION.match(latest_tag)
    assert get_tag('juice', 'latest') == latest_tag

    # Get an older version from short version key
    assert get_tag('juice', 'v422') == 'v422_20230130_002'

    # Get an older version from long version key
    assert get_tag('juice', 'v422_20230130_001') == 'v422_20230130_001'

    # Try to get an invalid version
    with raises(ValueError, match='Version `v999` is not available.'):
        _ = get_tag('juice', 'v999')

    # Get all tags
    tags = get_tag('juice', 'all', limit=2)

    assert len(tags) == 2
    assert tags[0] == latest_tag

    # Invalid mission name
    with raises(AttributeError, match='The mission name must be defined.'):
        _ = get_tag(None)


def test_esa_api_get_mk():
    """Test get mk form ESA API."""
    # Latest metakernel for the latest version
    mk = get_mk('JUICE')

    assert mk.name.endswith('.tm')

    latest_tag = mk.parent.name  # parent folder

    assert SKD_VERSION.match(latest_tag)
    assert latest_tag == get_tag('juice', 'latest')

    assert mk.parent.parent.name == 'juice'

    # Get metakernel for the latest version (with shortcut)
    mk = get_mk('juice', 'crema_5_0')
    assert mk.name == 'juice_crema_5_0.tm'

    # Get a metakernel for the latest version (with crema shortcut)
    mk = get_mk('juice', 5.0)
    assert mk.name == 'juice_crema_5_0.tm'

    # Get an old metakernel from an old version (with full tag)
    mk = get_mk('juice', 'juice_crema_5_0', 'v320_20220725_001')
    assert mk.name == 'juice_crema_5_0.tm'
    assert mk.parent.name == 'v320_20220725_001'

    # Get an old metakernel from an old version (with short version)
    mk = get_mk('juice', '5.0', 'v320')
    assert mk.name == 'juice_crema_5_0.tm'
    assert mk.parent.name == 'v320_20220725_001'

    # Get all the metakernels for a given tag
    mks = get_mk('juice', mk='all', version='v320_20220725_001')

    assert len(mks) == 13
    assert mks[0] == 'juice_crema_3_0.tm'
    assert mks[-1] == 'juice_crema_5_0b23_1.tm'

    # Get metakernels in uppercase
    mks = get_mk('ROSETTA', mk='all')

    assert len(mks) == 2
    assert 'ROS_OPS.TM' in mks

    # Get all the metakernels for all tag
    mks = get_mk('juice', mk='all', version='all')

    assert isinstance(mks, EsaMissionMetakernels)

    assert latest_tag in mks  # latest
    assert 'v160_20180216_001' in mks  # oldest

    assert len(mks['v160_20180216_001']) == 5
    assert mks['v160_20180216_001'][0] == 'juice_crema_2_0_ops.tm'
    assert mks['v160_20180216_001'][-1] == 'juice_crema_3_2_ops.tm'

    # Invalid version in MkDict
    assert 'foo' not in mks

    with raises(KeyError):
        _ = mks['foo']

    # Remove and re-downloaded the latest Juice crema
    mk = get_mk('juice')
    mk.unlink()

    mk = get_mk('juice')
    assert mk.exists()

    # Invalid mission name
    with raises(AttributeError):
        _ = get_mk(None)

    # Invalid metakernel name
    with raises(FileNotFoundError):
        _ = get_mk('juice', '1.0')


def test_esa_api_offline(monkeypatch, caplog):
    """Test ESA API offline getter."""
    # Setup dummy local metakernel cache folder
    monkeypatch.setattr(api_module, 'ESA_MK_CACHE', MOCK_ESA_MK_CACHE)

    juice_mk_cache = MOCK_ESA_MK_CACHE / 'juice'
    juice_mk_cache.mkdir(parents=True, exist_ok=True)

    for version in ['v123', 'v456']:
        tag = juice_mk_cache / f'{version}_yyyymmdd_000'
        tag.mkdir(exist_ok=True)
        (tag / 'juice_foo.tm').write_text(f'Metakernel FOO {version}')

    (tag / 'juice_bar.tm').write_text(f'Metakernel BAR {version}')

    # Get latest local tag
    assert get_tag_offline('juice') == 'v456_yyyymmdd_000'

    # Get short tag version
    assert get_tag_offline('juice', 'v123') == 'v123_yyyymmdd_000'

    # Get long tag version
    assert get_tag_offline('juice', 'v456_yyyymmdd_000') == 'v456_yyyymmdd_000'

    # Get all tags
    assert get_tag_offline('juice', 'all') == [
        'v456_yyyymmdd_000',
        'v123_yyyymmdd_000',
    ]

    # Missing local tag
    err = 'Version `v789_yyyymmdd_000` not found in the cache.'
    with raises(ValueError, match=err):
        _ = get_tag_offline('juice', 'v789_yyyymmdd_000')

    # Invalid mission name
    with raises(AttributeError, match='No version found in the cache for mission `foo`.'):
        _ = get_tag_offline('foo')

    # Latest metakernel for the latest version in the cache
    assert (
        get_mk_offline('juice') == juice_mk_cache / 'v456_yyyymmdd_000' / 'juice_foo.tm'
    )

    # All metakernels for the all versions in the cache
    assert get_mk_offline('juice', mk='all', version='all') == {
        'v456_yyyymmdd_000': [
            juice_mk_cache / 'v456_yyyymmdd_000' / 'juice_foo.tm',
            juice_mk_cache / 'v456_yyyymmdd_000' / 'juice_bar.tm',
        ],
        'v123_yyyymmdd_000': [
            juice_mk_cache / 'v123_yyyymmdd_000' / 'juice_foo.tm',
        ],
    }

    # Latest metakernels for the all versions in the cache
    assert get_mk_offline('juice', mk='latest', version='all') == {
        'v456_yyyymmdd_000': juice_mk_cache / 'v456_yyyymmdd_000' / 'juice_foo.tm',
        'v123_yyyymmdd_000': juice_mk_cache / 'v123_yyyymmdd_000' / 'juice_foo.tm',
    }

    # All metakernels for the latest version in the cache
    assert get_mk_offline('juice', mk='all', version='latest') == [
        juice_mk_cache / 'v456_yyyymmdd_000' / 'juice_foo.tm',
        juice_mk_cache / 'v456_yyyymmdd_000' / 'juice_bar.tm',
    ]

    # Single mk match in the cache (for the latest version)
    assert (
        get_mk_offline('juice', mk='bar')
        == juice_mk_cache / 'v456_yyyymmdd_000' / 'juice_bar.tm'
    )

    # Multiple mk match in the cache (for the latest version)
    assert get_mk_offline('juice', mk='juice_') == [
        juice_mk_cache / 'v456_yyyymmdd_000' / 'juice_foo.tm',
        juice_mk_cache / 'v456_yyyymmdd_000' / 'juice_bar.tm',
    ]

    # No mk match in the cache (for the latest version)
    err = '`baz` metakernel at `latest` is not in the cache.'
    with raises(FileNotFoundError, match=err):
        _ = get_mk_offline('juice', mk='baz')

    # Remove a metakernel from the cache (corrupted)
    (juice_mk_cache / 'v123_yyyymmdd_000' / 'juice_foo.tm').unlink()

    # Empty mission tag folder error (usually corrupted cache)
    err = 'No metakernel found locally for mission `juice` at version `v123`.'
    with raises(AttributeError, match=err):
        _ = get_mk_offline('juice', mk='foo', version='v123')

    # No warning message so far
    assert not caplog.messages

    # Disable ESA API
    def esa_api_not_available(uri, **params):
        """ESA service not available."""
        raise EsaApiNotAvailableError('ESA server not available')

    monkeypatch.setattr(api_module, 'esa_api', esa_api_not_available)

    def urlretrieve_available(url):
        """Disable URL retrieve."""
        raise URLError('URL retrieve disabled')

    monkeypatch.setattr(api_module, 'urlretrieve', urlretrieve_available)

    # Try offline to get the latest tag (warning msg)
    assert get_tag('juice') == 'v456_yyyymmdd_000'

    assert caplog.messages == [
        'Impossible to fetch `latest` version online. Use local cache instead.',
    ]
    caplog.clear()

    # Try offline to get a given tag (warning msg)
    assert get_tag('juice', version='v123') == 'v123_yyyymmdd_000'

    assert caplog.messages == [
        'Impossible to fetch `v123` version online. Use local cache instead.',
    ]
    caplog.clear()

    # Try offline to get the latest mk for the latest version (warning msg)
    assert get_mk('juice') == juice_mk_cache / 'v456_yyyymmdd_000' / 'juice_foo.tm'

    assert caplog.messages == [
        'Impossible to fetch `latest` version online. Use local cache instead.',
    ]
    caplog.clear()

    # Try offline to get mk for a short tag version (warning msg)
    assert (
        get_mk('juice', version='v456')
        == juice_mk_cache / 'v456_yyyymmdd_000' / 'juice_foo.tm'
    )

    assert caplog.messages == [
        'Impossible to fetch `v456` version online. Use local cache instead.',
    ]
    caplog.clear()

    # Try offline to get mk for a long tag version (no warning msg, by-pass)
    assert (
        get_mk('juice', mk='bar', version='v456_yyyymmdd_000')
        == juice_mk_cache / 'v456_yyyymmdd_000' / 'juice_bar.tm'
    )

    assert not caplog.messages

    # Try offline to get missing mk for a long tag version (by-pass but warning msg)
    err = '`juice_baz.tm` metakernel at `v456_yyyymmdd_000` is not in the cache.'
    with raises(FileNotFoundError, match=err):
        _ = get_mk('juice', mk='baz', version='v456_yyyymmdd_000')

    assert caplog.messages == [
        'Impossible to fetch `juice_baz.tm` metakernel '
        'at `v456_yyyymmdd_000` version online. Use local cache instead.'
    ]
    caplog.clear()

    # Try offline to get mk for an invalid long tag version (no warning msg, by-pass)
    err = 'Version `v789_yyyymmdd_000` not found in the cache.'
    with raises(ValueError, match=err):
        _ = get_mk('juice', mk='foo', version='v789_yyyymmdd_000')

    assert caplog.messages == [
        'Impossible to fetch `juice_foo.tm` metakernel '
        'at `v789_yyyymmdd_000` version online. Use local cache instead.'
    ]


def test_esa_api_mk_name():
    """Test ESA metakernel names shortcuts."""
    assert esa_mk_name('BEPICOLOMBO', 'ops') == 'bc_ops.tm'
    assert esa_mk_name('COMET-INTERCEPTOR', 'study_v02') == 'interceptor_study_v02.tm'
    assert (
        esa_mk_name('ENVISION', 'study_esc_t2_south_voi_v040')
        == 'envision_study_esc_t2_south_voi_v040.tm'
    )
    assert esa_mk_name('ExoMars2016', 'plan') == 'em16_plan.tm'
    assert (
        esa_mk_name('ExoMarsRSP', 'test_tlm_0003_v009') == 'emrsp_test_tlm_0003_v009.tm'
    )
    assert esa_mk_name('Gaia', 'plan') == 'gaia_plan.tm'
    assert esa_mk_name('hera', 'study_PO_EMA_2024') == 'hera_study_PO_EMA_2024.tm'
    assert esa_mk_name('HUYGENS', 'V040') == 'HUYGENS_V040.TM'
    assert esa_mk_name('INTEGRAL', 'ops') == 'integral_ops.tm'
    assert esa_mk_name('JUICE', 'plan') == 'juice_plan.tm'
    assert esa_mk_name('JUICE', 'ops') == 'juice_ops.tm'
    assert esa_mk_name('JUICE', 5.0) == 'juice_crema_5_0.tm'
    assert esa_mk_name('JUICE', '5.1 150lb_23_1') == 'juice_crema_5_1_150lb_23_1.tm'
    assert esa_mk_name('JUICE', 'crema_5_0') == 'juice_crema_5_0.tm'
    assert esa_mk_name('JWST', 'plan') == 'jwst_plan.tm'
    assert esa_mk_name('LUNAR-GATEWAY', 'study') == 'lg_study.tm'
    assert esa_mk_name('MARS-EXPRESS', 'ops') == 'MEX_OPS.TM'
    assert esa_mk_name('ROSETTA', 'ops') == 'ROS_OPS.TM'
    assert esa_mk_name('SMART-1', 'ops') == 'SMART1_OPS.TM'
    assert esa_mk_name('SOLAR-ORBITER', 'ANC_soc-flown-mk') == 'solo_ANC_soc-flown-mk.tm'
    assert esa_mk_name('VENUS-EXPRESS', 'OBS') == 'VEX_OBS.TM'

    # Prefix with the mission name (with a full key or MK_IDENTIFIER)
    assert (
        esa_mk_name('JUICE', 'juice_crema_5_1_150lb_v320_20220725_001')
        == 'juice_crema_5_1_150lb_v320_20220725_001.tm'
    )

    assert (
        esa_mk_name('ENVISION', 'envision_study_esc_t2_south_voi_v040')
        == 'envision_study_esc_t2_south_voi_v040.tm'
    )

    # Unknown ESA mission name
    err = 'SPICE metakernel not found for mission FOO on ESA SPICE repository.'
    with raises(ValueError, match=err):
        _ = esa_mk_name('FOO', 'obs')


def test_esa_api_crema_ref():
    """Test ESA CReMA reference formatter."""
    assert crema_ref(3.0) == '3_0'
    assert crema_ref('3.0') == '3_0'
    assert crema_ref('4.0b') == '4_0b'
    assert crema_ref('3_2 gan N56 Pp5 Q19') == '3_2_gan_N56_Pp5_Q19'
    assert crema_ref('3.2 gan N56 Pp5 Q19') == '3_2_gan_N56_Pp5_Q19'
    assert crema_ref('3.2-gan-N56-Pp5-Q19') == '3_2_gan_N56_Pp5_Q19'
    assert crema_ref('4.2 N56 Pp5 Q19') == '4_2_N56_Pp5_Q19'
    assert crema_ref('4.2 N56 Pn5 Q19') == '4_2_N56_Pn5_Q19'
    assert crema_ref('4.2 N54 Pp5 Q19') == '4_2_N54_Pp5_Q19'
    assert crema_ref('4.2b22.1') == '4_2b22_1'
    assert crema_ref('4_2b22_1') == '4_2b22_1'
    assert crema_ref('4.2b22.1 (cruise only)') == '4_2b22_1_cruise'
    assert crema_ref(5.0) == '5_0'
    assert crema_ref('5.0b23.1') == '5_0b23_1'
    assert crema_ref('5.1 150la') == '5_1_150la'
    assert crema_ref('5.1 150lb_23_1') == '5_1_150lb_23_1'
