"""Test ESA metakernel module."""

from pytest import raises

import planetary_coverage.esa.metakernel as mk_module
from planetary_coverage import ESA_MK
from planetary_coverage.esa.metakernel import EsaMetakernels


def test_esa_mk(monkeypatch):
    """Test ESA metakernels collection."""
    assert len(ESA_MK) == 16

    # Missions
    assert 'JUICE' in ESA_MK
    assert 'BEPICOLOMBO' in ESA_MK
    assert 'FOO' not in ESA_MK

    # SPICE spacecraft alias
    assert len(ESA_MK.aliases) == 43

    assert 'MPO' in ESA_MK
    assert ESA_MK.aliases['MPO'] == 'BEPICOLOMBO'

    # MK getter
    assert isinstance(ESA_MK['JUICE'], EsaMetakernels)

    # Get item for the latest version
    monkeypatch.setattr(
        mk_module,
        'get_mk',
        lambda mission, mk='latest', version='latest': f'{version}/{mission}_{mk}.tm',
    )

    # Explicit pinned version
    mk = ESA_MK['JUICE', '5_0_v422_20230130_002']
    assert mk == 'v422_20230130_002/JUICE_5_0.tm'

    # With an explicit tag
    mk = ESA_MK['JUICE', '5_0', 'v320_20220725_001']
    assert mk == 'v320_20220725_001/JUICE_5_0.tm'

    # Without an explicit version (use 'latest')
    monkeypatch.setattr(
        mk_module, 'get_tag', lambda mission, version='latest': 'v123_yyyymmdd_000'
    )

    mk = ESA_MK['JUICE', '5_0']
    assert mk == 'v123_yyyymmdd_000/JUICE_5_0.tm'

    # Too many arguments
    with raises(KeyError):
        _ = ESA_MK['JUICE', '5_0', 'v320_20220725_001', 'foo']

    # Invalid key
    with raises(KeyError):
        _ = ESA_MK['FOO']


def test_esa_mk_juice(monkeypatch):
    """Test ESA metakernel collection for Juice."""
    juice_mk = ESA_MK['JUICE']

    assert str(juice_mk) == 'JUICE'
    assert juice_mk.mission == 'JUICE'
    assert juice_mk.doi == '10.5270/esa-ybmj68p'

    assert juice_mk.alias == []

    # Latest tag
    del juice_mk.latest  # Clear cache property

    monkeypatch.setattr(mk_module, 'get_tag', lambda *args, **kwargs: 'latest_version')
    assert juice_mk.latest == 'latest_version'

    # Check cache
    monkeypatch.setattr(mk_module, 'get_tag', lambda *args, **kwargs: 'new_version')
    assert juice_mk.latest == 'latest_version'

    # Metakernel list
    monkeypatch.setattr(
        mk_module, 'get_mk', lambda *args, **kwargs: ['juice_mk_1.tm', 'juice_mk_2.tm']
    )

    assert juice_mk.mks == ['juice_mk_2.tm', 'juice_mk_1.tm']  # The list is inverted
    assert len(juice_mk) == 2

    # __contains__
    assert 'juice_mk_1.tm' in juice_mk
    assert 'juice_mk_3.tm' not in juice_mk

    # __iter__
    for mk in juice_mk:
        assert mk == 'juice_mk_2.tm'
        break

    # Get item for the latest version
    monkeypatch.setattr(
        mk_module,
        'get_mk',
        lambda mission, mk='latest', version='latest': f'{version}/{mission}_{mk}.tm',
    )

    assert juice_mk[3] == 'latest_version/JUICE_3.tm'
    assert juice_mk['3_0'] == 'latest_version/JUICE_3_0.tm'

    # Get item for an SKD version suffix
    assert juice_mk['3_0_v300_20220101_001'] == 'v300_20220101_001/JUICE_3_0.tm'

    # Get item for an explicit version
    assert juice_mk['3_0', 'v300_20220101_001'] == 'v300_20220101_001/JUICE_3_0.tm'

    # Too many arguments
    with raises(KeyError):
        _ = juice_mk[3.0, 'latest', 'foo']

    # Get an older list of metakernels based on an explicit version
    monkeypatch.setattr(
        mk_module, 'get_mk', lambda *args, **kwargs: ['juice_mk_3.tm', 'juice_mk_4.tm']
    )

    assert juice_mk.version('previous') == ['juice_mk_4.tm', 'juice_mk_3.tm']
    assert juice_mk.mks == ['juice_mk_2.tm', 'juice_mk_1.tm']  # From the cache

    # Get all the previous versions
    monkeypatch.setattr(mk_module, 'get_tag', lambda *args, **kwargs: ['v1', 'v2'])
    assert juice_mk.versions == ['v1', 'v2']


def test_esa_mk_bepi():
    """Test ESA metakernel collection for Bepi."""
    bepi_mk = ESA_MK['MPO']  # Use alias

    assert bepi_mk.alias == (
        'MPO',
        'BEPICOLOMBO MPO',
        'MERCURY PLANETARY ORBITER',
        'MTM',
        'BEPICOLOMBO MTM',
        'MERCURY TRANSFER MODULE',
        'MMO',
        'BEPICOLOMBO MMO',
        'MERCURY MAGNETOSPHERIC ORBITER',
    )

    assert str(bepi_mk) == (
        'BEPICOLOMBO / '
        'MPO / '
        'BEPICOLOMBO MPO / '
        'MERCURY PLANETARY ORBITER / '
        'MTM / '
        'BEPICOLOMBO MTM / '
        'MERCURY TRANSFER MODULE / '
        'MMO / '
        'BEPICOLOMBO MMO / '
        'MERCURY MAGNETOSPHERIC ORBITER'
    )
