"""Test events module."""

from numpy import datetime64 as dt
from numpy import timedelta64 as td
from numpy.testing import assert_array_almost_equal as assert_array

from pytest import fixture, raises

from planetary_coverage.events import Event, EventsDict, EventsList, EventWindow
from planetary_coverage.events.event import AbstractEvent, AbstractEventsCollection
from planetary_coverage.html import display_html


@fixture
def event_dict():
    """Single event attributes."""
    return {
        'event time [utc]': '2031-07-21T07:14:23Z',
        'contextual info': 'Crema name= 1G1;Altitude above ganymede [km]= 400.00697',
    }


@fixture
def event(event_dict):
    """Single event."""
    return Event('FLYBY_GANYMEDE', **event_dict)


@fixture
def event_window():
    """Windows event."""
    return EventWindow(
        'GCO500', t_start='2035-05-22T06:00:00', t_end='2035-10-05T16:01:00'
    )


@fixture
def events_dict(event, event_window):
    """Events dict."""
    return EventsDict([event_window, event])


@fixture
def events_list(event):
    """Events list."""
    return EventsList([event, event])


def test_event_single(event, event_dict):
    """Test single event attributes."""
    assert isinstance(event, Event)
    assert str(event) == 'FLYBY_GANYMEDE (2031-07-21)'

    assert event.start == event.stop == dt('2031-07-21T07:14:23')
    assert event.start_date == event.stop_date == '2031-07-21'

    assert isinstance(hash(event), int)

    # Contextual info are extracted with the data (but not parsed)
    assert tuple(event.keys()) == (
        'event time [utc]',
        'Crema name',
        'Altitude above ganymede [km]',
    )

    assert tuple(event.values()) == ('2031-07-21T07:14:23Z', '1G1', '400.00697')

    assert 'Crema name' in event
    assert event['Crema name'] == '1G1'
    assert event.get('Crema name') == '1G1'

    assert 'FOO' not in event
    assert event.get('FOO') is None

    # Single event init with an EventWindow
    event = EventWindow('FLYBY_GANYMEDE', **event_dict)
    assert isinstance(event, Event)
    assert isinstance(hash(event), int)

    # Custom time input
    assert Event('FOO', **{'event time [utc]': '2031-07-21T07:14:23Z'}).start == dt(
        '2031-07-21T07:14:23'
    )
    assert Event('FOO', **{'event utc apo': '2031 JUL 21 07:14:23'}).start == dt(
        '2031-07-21T07:14:23'
    )
    assert Event('FOO', **{'event utc peri': '2031 JUL 21 07:14:23'}).start == dt(
        '2031-07-21T07:14:23'
    )
    assert Event('FOO', time='2031 JUL 21 07:14:23').start == dt('2031-07-21T07:14:23')

    event = Event('FOO', **{'event utc peri': '2031 JUL 21 07:14:23'})
    del event['event utc peri']

    with raises(KeyError):
        _ = event.start


def test_event_window(event_window):
    """Test event window attributes."""
    assert isinstance(event_window, EventWindow)
    assert str(event_window) == 'GCO500 (2035-05-22 -> 2035-10-05)'

    assert event_window.start == dt('2035-05-22T06:00:00')
    assert event_window.stop == dt('2035-10-05T16:01:00')
    assert event_window.start_date == '2035-05-22'
    assert event_window.stop_date == '2035-10-05'

    assert isinstance(hash(event_window), int)


def test_events_dict(event, event_window, events_dict):
    """Test events dictionary."""
    events = events_dict

    assert isinstance(events, EventsDict)
    assert len(events) == 2

    # Sorted by increasing time
    assert str(events) == '\n'.join([
        '(2031-07-21 -> 2035-10-05 | 2 keys):',
        ' - FLYBY_GANYMEDE (2031-07-21)',
        ' - GCO500 (2035-05-22 -> 2035-10-05)',
    ])

    # Getter
    assert events['FLYBY_GANYMEDE'] == event
    assert events['GCO500'] == event_window

    assert events[0] == event
    assert events[1] == event_window
    assert events[-1] == event_window
    assert events[-2] == event

    with raises(IndexError):  # out of range index
        _ = events[2]

    with raises(IndexError):  # out of range index
        _ = events[-3]

    items = events[:]
    assert isinstance(items, EventsDict)
    assert len(items) == 2

    items = events['GCO', 'FLYBY_', 'FLYBY_']  # Unsorted tuple with duplicates
    assert isinstance(items, EventsDict)
    assert len(items) == 2

    assert items[0] == event  # Sorted without duplicates
    assert items[1] == event_window

    item = events[r'GCO\d+']  # Single regex search (1 output only)
    assert isinstance(item, EventWindow)
    assert item == event_window

    with raises(KeyError, match='`FOO` not found'):
        _ = events['FOO']

    # Function filters
    items = events.find('G')
    assert isinstance(items, EventsDict)
    assert len(items) == 2

    item = events.startswith('g')  # Ignore case
    assert isinstance(item, EventWindow)

    item = events.endswith('E', 'DE')  # multi-letters and duplicates
    assert isinstance(item, Event)

    with raises(KeyError, match='`FOO`, `BAR` not found'):
        _ = events.find('FOO', 'BAR')

    # Keep EventsDict results
    events_dict = events.find('GANYMEDE', as_dict=True)  # Only 1 match by as EventsDict
    assert isinstance(events_dict, EventsDict)
    assert len(events_dict) == 1

    for ev in events_dict:
        assert ev.key == 'FLYBY_GANYMEDE'

    with raises(KeyError):
        _ = events.find('FOO', as_dict=True)  # No match always raise KeyError

    # Iter
    for i, ev in enumerate(events):
        assert ev == event if i == 0 else event_window

    # Hash
    assert isinstance(hash(events), int)

    # Contains
    assert 'FLYBY_GANYMEDE' in events
    assert 'FOO' not in events

    assert '2035-05-23' in events
    assert '2035-10-05' in events
    assert '2035-05-22' not in events
    assert '2035-10-06' not in events

    # Keys
    assert tuple(events.keys()) == ('FLYBY_GANYMEDE', 'GCO500')

    # Event start and Stop times
    assert events.starts == [
        dt('2031-07-21T07:14:23'),
        dt('2035-05-22T06:00:00'),
    ]
    assert events.stops == [
        dt('2031-07-21T07:14:23'),
        dt('2035-10-05T16:01:00'),
    ]
    assert events.windows == [
        (dt('2031-07-21T07:14:23'), dt('2031-07-21T07:14:23')),
        (dt('2035-05-22T06:00:00'), dt('2035-10-05T16:01:00')),
    ]

    assert events.start == dt('2031-07-21T07:14:23')
    assert events.stop == dt('2035-10-05T16:01:00')
    assert events.start_date == '2031-07-21'
    assert events.stop_date == '2035-10-05'


def test_events_dict_with_single_input():
    """Test events dict with a single input."""
    events = EventsDict(
        EventWindow('FOO', t_start='2035-01-01', t_end='2035-01-02'),
    )

    assert isinstance(events, EventsDict)
    assert len(events) == 1

    assert isinstance(events['FOO'], EventWindow)

    # Empty array
    events = EventsDict([])

    assert isinstance(events, EventsDict)
    assert len(events) == 0


def test_events_dict_with_duplicated_keys():
    """Test events dict with duplicated keys."""
    events = EventsDict([
        EventWindow('FOO', t_start='2035-01-01', t_end='2035-01-02'),
        EventWindow('FOO', t_start='2035-01-03', t_end='2035-01-04'),
        EventWindow('FOO', t_start='2035-01-05', t_end='2035-01-06'),
    ])

    assert isinstance(events, EventsDict)
    assert len(events) == 1

    items = events['FOO']

    assert isinstance(items, EventsList)
    assert len(items) == 3

    assert items.start == dt('2035-01-01')
    assert items.stop == dt('2035-01-06')

    # EventsList input
    events = EventsDict(
        EventsList([
            EventWindow('BAR', t_start='2035-01-01', t_end='2035-01-02'),
            EventWindow('BAR', t_start='2035-01-03', t_end='2035-01-04'),
        ])
    )

    assert isinstance(events, EventsDict)
    assert len(events) == 1

    items = events['BAR']

    assert isinstance(items, EventsList)
    assert len(items) == 2

    assert items.start == dt('2035-01-01')
    assert items.stop == dt('2035-01-04')


def test_events_list(event, events_list):
    """Test events dictionary."""
    events = events_list

    assert isinstance(events, EventsList)
    assert str(events) == 'FLYBY_GANYMEDE (2031-07-21 -> 2031-07-21 | 2 events)'
    assert len(events) == 2

    # Getter
    assert events[0] == event
    assert events[1] == event

    items = events[:1]  # with a slice
    assert isinstance(items, EventsList)
    assert len(items) == 1
    assert items[0] == event

    # Iter
    for ev in events:
        assert ev == event

    # Get crema names if present
    assert events.crema_names == ['1G1', '1G1']

    # Get item by flyby crema name
    assert events['1G1'] == event

    # Get item with multiple flyby names (tuple)
    items = events['1G1', '1G1']
    assert isinstance(items, EventsList)
    assert len(items) == 2
    assert items[0]['Crema name'] == '1G1'

    # Event start and Stop times
    assert events.starts == [
        dt('2031-07-21T07:14:23'),
        dt('2031-07-21T07:14:23'),
    ]
    assert events.stops == [
        dt('2031-07-21T07:14:23'),
        dt('2031-07-21T07:14:23'),
    ]
    assert events.windows == [
        (dt('2031-07-21T07:14:23'), dt('2031-07-21T07:14:23')),
        (dt('2031-07-21T07:14:23'), dt('2031-07-21T07:14:23')),
    ]

    assert events.start == events.stop == dt('2031-07-21T07:14:23')
    assert events.start_date == events.stop_date == '2031-07-21'

    # Hash
    assert isinstance(hash(events), int)

    # Contains
    assert '1G1' in events  # Crema names
    assert 'FLYBY_GANYMEDE' not in events  # no keys

    assert '2031-07-21T07:14:23' in events
    assert '2031-07-21' not in events
    assert '2031-07-22' not in events


def test_events_operations(event, event_window):
    """Test events operations."""
    # Add/subtract temporal step as a string
    assert event - '12 h' == dt('2031-07-20T19:14:23')
    assert event + '12 h' == dt('2031-07-21T19:14:23')

    # Add/subtract temporal step as timedelta64
    assert event - td(12, 'h') == dt('2031-07-20T19:14:23')
    assert event + td(12, 'h') == dt('2031-07-21T19:14:23')

    # With EventWindow
    assert event_window - '12 h' == dt('2035-05-21T18:00:00')
    assert event_window + '12 h' == dt('2035-10-06T04:01:00')


def test_events_intersection(event_window):
    """Test events intersection on an event window and events list."""
    # With a single time string
    assert '2035-05-23' in event_window
    assert '2035-10-05' in event_window
    assert '2035-05-22' not in event_window
    assert '2035-10-06' not in event_window

    # With a numpy.datetime64
    assert dt('2035-05-23') in event_window

    # With a list / tuple
    assert ['2035-05-22', '2035-05-23', '2035-10-05', '2035-10-06'] in event_window
    assert_array(
        event_window.contains(['2035-05-22', '2035-05-23', '2035-10-05', '2035-10-06']),
        [False, True, True, False],
    )
    assert [dt('2035-05-23')] in event_window

    # With a custom instance
    class A:
        """Dummy class."""

        @property
        def utc(self):
            """UTC times"""
            return [
                '2035-05-22',
                '2035-05-23',
                '2035-10-05',
                '2035-10-06',
            ]

    a = A()

    assert a in event_window
    assert_array(event_window.contains(a), [False, True, True, False])

    # Same with an EventsList
    events = EventsList([event_window])

    # With a single time string
    assert '2035-05-23' in events
    assert '2035-10-05' in events
    assert '2035-05-22' not in events
    assert '2035-10-06' not in events

    # With a numpy.datetime64
    assert dt('2035-05-23') in events

    # With a list / tuple
    assert ['2035-05-22', '2035-05-23', '2035-10-05', '2035-10-06'] in events
    assert_array(
        events.contains(['2035-05-22', '2035-05-23', '2035-10-05', '2035-10-06']),
        [False, True, True, False],
    )
    assert [dt('2035-05-23')] in events

    # With a custom instance
    assert a in events
    assert_array(events.contains(a), [False, True, True, False])


def test_events_window_comparisons(event_window):
    """Test event window comparison operations."""
    assert not event_window > '2035-05-22T06:00:00'
    assert event_window >= '2035-05-22T06:00:00'
    assert not event_window > '2035-10-05T16:01:00'
    assert not event_window >= '2035-10-05T16:01:00'

    assert event_window > '2035-05-22'
    assert not event_window > '2035-05-23'
    assert not event_window < '2035-10-05'
    assert event_window < '2035-10-06'

    assert not event_window < '2035-05-22T06:00:00'
    assert not event_window <= '2035-05-22T06:00:00'
    assert not event_window < '2035-10-05T16:01:00'
    assert event_window <= '2035-10-05T16:01:00'

    # Check input types
    assert event_window > '2035'
    assert event_window > 2035
    assert event_window > dt('2035')


def test_events_dict_comparisons():
    """Test events dict comparisons."""
    events = EventsDict([
        EventWindow('EVENT_1', t_start='2035-01-01', t_end='2035-01-02'),
        EventWindow('EVENT_2', t_start='2035-01-03', t_end='2035-01-04'),
        EventWindow('EVENT_3', t_start='2035-01-05', t_end='2035-01-06'),
    ])

    # Before
    res = events.before('2035-01-02')
    assert isinstance(res, EventWindow)
    assert res.key == 'EVENT_1'

    res = events.before('2035-01-06', strict=True)
    assert isinstance(res, EventsDict)
    assert len(res) == 2
    assert res[0].key == 'EVENT_1'
    assert res[1].key == 'EVENT_2'

    res = events.before('2035-01-02', as_dict=True)
    assert isinstance(res, EventsDict)
    assert len(res) == 1
    assert isinstance(res['EVENT_1'], EventWindow)

    with raises(LookupError):
        _ = events.before('2035-01-01')

    # After
    res = events.after('2035-01-04')
    assert isinstance(res, EventWindow)
    assert res.key == 'EVENT_3'

    res = events.after('2035-01-01', strict=True)
    assert isinstance(res, EventsDict)
    assert len(res) == 2
    assert res[0].key == 'EVENT_2'
    assert res[1].key == 'EVENT_3'

    res = events.after('2035-01-04', as_dict=True)
    assert isinstance(res, EventsDict)
    assert len(res) == 1
    assert isinstance(res['EVENT_3'], EventWindow)

    with raises(LookupError):
        _ = events.after('2035-01-06')

    # Between
    res = events.between('2035-01-01', '2035-01-06', strict=True)
    assert isinstance(res, EventWindow)
    assert res.key == 'EVENT_2'

    res = events.between('2035-01-01', '2035-01-06')
    assert isinstance(res, EventsDict)
    assert len(res) == 3
    assert res[0].key == 'EVENT_1'
    assert res[1].key == 'EVENT_2'
    assert res[2].key == 'EVENT_3'

    res = events.between('2035-01-01', '2035-01-06', strict=True, as_dict=True)
    assert isinstance(res, EventsDict)
    assert len(res) == 1
    assert isinstance(res['EVENT_2'], EventWindow)

    with raises(LookupError):
        _ = events.between('2035-01-02', '2035-01-03')


def test_events_list_comparisons():
    """Test events list comparisons."""
    events = EventsList([
        EventWindow('EVENT', t_start='2035-01-01', t_end='2035-01-02'),
        EventWindow('EVENT', t_start='2035-01-03', t_end='2035-01-04'),
        EventWindow('EVENT', t_start='2035-01-05', t_end='2035-01-06'),
    ])

    # Before
    res = events.before('2035-01-02')
    assert isinstance(res, EventWindow)
    assert res.start == dt('2035-01-01')
    assert res.stop == dt('2035-01-02')

    res = events.before('2035-01-06', strict=True)
    assert isinstance(res, EventsList)
    assert len(res) == 2
    assert res.start == dt('2035-01-01')
    assert res.stop == dt('2035-01-04')

    res = events.before('2035-01-02', as_dict=True)
    assert isinstance(res, EventsDict)
    assert len(res) == 1

    event = res['EVENT']
    assert isinstance(event, EventWindow)
    assert event.start == dt('2035-01-01')
    assert event.stop == dt('2035-01-02')

    with raises(LookupError):
        _ = events.before('2035-01-01')

    # After
    res = events.after('2035-01-04')
    assert isinstance(res, EventWindow)
    assert res.start == dt('2035-01-05')
    assert res.stop == dt('2035-01-06')

    res = events.after('2035-01-01', strict=True)
    assert isinstance(res, EventsList)
    assert len(res) == 2
    assert res.start == dt('2035-01-03')
    assert res.stop == dt('2035-01-06')

    res = events.after('2035-01-04', as_dict=True)
    assert isinstance(res, EventsDict)
    assert len(res) == 1

    event = res['EVENT']
    assert isinstance(event, EventWindow)
    assert event.start == dt('2035-01-05')
    assert event.stop == dt('2035-01-06')

    with raises(LookupError):
        _ = events.after('2035-01-06')

    # Between
    res = events.between('2035-01-01', '2035-01-06', strict=True)
    assert isinstance(res, EventWindow)
    assert res.start == dt('2035-01-03')
    assert res.stop == dt('2035-01-04')

    res = events.between('2035-01-01', '2035-01-06')
    assert isinstance(res, EventsList)
    assert len(res) == 3
    assert res.start == dt('2035-01-01')
    assert res.stop == dt('2035-01-06')

    res = events.between('2035-01-01', '2035-01-06', as_dict=True)
    assert isinstance(res, EventsDict)
    assert len(res) == 1

    event = res['EVENT']
    assert isinstance(event, EventsList)
    assert len(event) == 3
    assert event.start == dt('2035-01-01')
    assert event.stop == dt('2035-01-06')

    with raises(LookupError):
        _ = events.between('2035-01-02', '2035-01-03')


def test_events_cut_event(event, event_window):
    """Test event trim on event."""
    # No change
    e = event.trim()

    assert isinstance(e, Event)
    assert e.key == 'FLYBY_GANYMEDE'
    assert e.start == dt('2031-07-21T07:14:23')
    assert e.stop == dt('2031-07-21T07:14:23')

    # Inside the time boundaries
    e = event.trim(before='2031-07-21', after='2031-07-22')

    assert isinstance(e, Event)
    assert e.key == 'FLYBY_GANYMEDE'
    assert e.start == dt('2031-07-21T07:14:23')
    assert e.stop == dt('2031-07-21T07:14:23')

    # Before start
    assert event.trim(before='2031-07-22') is None

    # After stop
    assert event.trim(after='2031-07-21') is None

    # By event (outside the time boundaries)
    assert event.trim(by_event=event_window) is None

    # By event (self)
    e = event.trim(by_event=event)

    assert isinstance(e, Event)
    assert e.key == 'FLYBY_GANYMEDE'
    assert e.start == dt('2031-07-21T07:14:23')
    assert e.stop == dt('2031-07-21T07:14:23')

    # Errors (explicit `before`, `after` or `by_event` required)
    with raises(TypeError):
        _ = event.trim('foo')


def test_events_cut_event_window(event_window, event):
    """Test event trim on event window."""
    # No change
    e = event_window.trim()

    assert isinstance(e, EventWindow)
    assert e.key == 'GCO500'
    assert e.start == dt('2035-05-22T06')
    assert e.stop == dt('2035-10-05T16:01')

    # Inside the time boundaries
    e = event_window.trim(before='2035-05-01', after='2035-11-01')

    assert isinstance(e, EventWindow)
    assert e.key == 'GCO500'
    assert e.start == dt('2035-05-22T06')
    assert e.stop == dt('2035-10-05T16:01')

    # Trim boundaries
    e = event_window.trim(before='2035-06-01', after='2035-10-01')

    assert isinstance(e, EventWindow)
    assert e.key == 'GCO500'
    assert e.start == dt('2035-06-01')
    assert e.stop == dt('2035-10-01')

    # Before start
    assert event_window.trim(before='2035-11-01') is None

    # After stop
    assert event_window.trim(after='2035-05-01') is None

    # By event (outside the time boundaries)
    assert event_window.trim(by_event=event) is None

    # By event (self)
    e = event_window.trim(by_event=event_window)

    assert isinstance(e, EventWindow)
    assert e.key == 'GCO500'
    assert e.start == dt('2035-05-22T06')
    assert e.stop == dt('2035-10-05T16:01')

    # Errors (explicit `before`, `after` or `by_event` required)
    with raises(TypeError):
        _ = event_window.trim('foo')


def test_events_cut_events_list(event_window, event):
    """Test event trim on event list."""
    events_list = EventsList([event_window, event_window])

    # No change
    events = events_list.trim()

    assert isinstance(events, EventsList)
    assert len(events) == 2
    assert events.start == dt('2035-05-22T06')
    assert events.stop == dt('2035-10-05T16:01')

    # Inside the time boundaries
    events = events_list.trim(before='2035-05-01', after='2035-11-01')

    assert isinstance(events, EventsList)
    assert len(events) == 2
    assert events.start == dt('2035-05-22T06')
    assert events.stop == dt('2035-10-05T16:01')

    # Trim boundaries
    events = events_list.trim(before='2035-06-01', after='2035-10-01')

    assert isinstance(events, EventsList)
    assert len(events) == 2
    assert events.start == dt('2035-06-01')
    assert events.stop == dt('2035-10-01')

    for e in events:
        assert e.key == 'GCO500'
        assert e.start == dt('2035-06-01')
        assert e.stop == dt('2035-10-01')

    # Before start
    assert events_list.trim(before='2035-11-01') is None

    # After stop
    assert events_list.trim(after='2035-05-01') is None

    # By event (outside the time boundaries)
    assert events_list.trim(by_event=event) is None

    # By event (self)
    events = events_list.trim(by_event=events_list)

    assert isinstance(events, EventsList)
    assert len(events) == 2
    assert events.start == dt('2035-05-22T06')
    assert events.stop == dt('2035-10-05T16:01')

    # Errors (explicit `before`, `after` or `by_event` required)
    with raises(TypeError):
        _ = events_list.trim('foo')


def test_events_cut_events_dict(events_dict, event_window, event):
    """Test event trim on event dict."""
    # No change
    events = events_dict.trim()

    assert isinstance(events, EventsDict)
    assert len(events) == 2
    assert events.start == dt('2031-07-21T07:14:23')
    assert events.stop == dt('2035-10-05T16:01')

    # Inside the time boundaries
    events = events_dict.trim(before='2031-05-01', after='2035-11-01')

    assert isinstance(events, EventsDict)
    assert len(events) == 2
    assert events.start == dt('2031-07-21T07:14:23')
    assert events.stop == dt('2035-10-05T16:01')

    # Trim boundaries (one element is discarded)
    events = events_dict.trim(before='2035-06-01', after='2035-10-01')

    assert isinstance(events, EventsDict)
    assert len(events) == 1
    assert events.start == dt('2035-06-01')
    assert events.stop == dt('2035-10-01')

    for e in events:
        assert e.key == 'GCO500'
        assert e.start == dt('2035-06-01')
        assert e.stop == dt('2035-10-01')

    # Before start
    assert events_dict.trim(before='2035-11-01') is None

    # After stop
    assert events_dict.trim(after='2031-05-01') is None

    # By event
    events = events_dict.trim(by_event=event)

    assert isinstance(events, EventsDict)
    assert len(events) == 1
    assert events.start == dt('2031-07-21T07:14:23')
    assert events.stop == dt('2031-07-21T07:14:23')

    events = events_dict.trim(by_event=event_window)

    assert isinstance(events, EventsDict)
    assert len(events) == 1
    assert events.start == dt('2035-05-22T06')
    assert events.stop == dt('2035-10-05T16:01')

    # Errors (explicit `before`, `after` or `by_event` required)
    with raises(TypeError):
        _ = events_dict.trim('foo')


def test_events_errors():
    """Test events errors."""
    with raises(ValueError):
        _ = AbstractEvent('missing-time', a=1)

    class DummyEvent(AbstractEvent):
        """Dummy event object."""

        def __init__(self, *args, **kwargs):
            pass

    event = DummyEvent()
    with raises(NotImplementedError):
        _ = event.start

    with raises(NotImplementedError):
        _ = event.stop

    class DummyEvents(AbstractEventsCollection):
        """Dummy event object."""

        def __init__(self, *args, **kwargs):
            pass

    events = DummyEvents()
    with raises(NotImplementedError):
        _ = hash(events)


def test_events_ipython_autocompletion(event, event_window, events_dict, events_list):
    """Test Events IPython autocompletion."""

    def tab(obj):
        """Test IPython <tab> autocompletion."""
        return tuple(obj._ipython_key_completions_())

    assert tab(event) == (
        'event time [utc]',
        'Crema name',
        'Altitude above ganymede [km]',
    )
    assert tab(event_window) == ('t_start', 't_end')
    assert tab(events_dict) == ('FLYBY_GANYMEDE', 'GCO500')
    assert tab(events_list) == ('1G1', '1G1')


def test_events_html_repr(event, event_window, events_dict, events_list):
    """Test Events HTML representation."""
    assert display_html(event) == (
        '<table>'
        '<tr>'
        '<td>event time [utc]</td>'
        '<td>2031-07-21T07:14:23Z</td>'
        '</tr>'
        '<tr>'
        '<td>Crema name</td>'
        '<td>1G1</td>'
        '</tr>'
        '<tr>'
        '<td>Altitude above ganymede [km]</td>'
        '<td>400.00697</td>'
        '</tr>'
        '</table>'
    )

    assert display_html(event_window) == (
        '<table>'
        '<tr>'
        '<td>t_start</td>'
        '<td>2035-05-22T06:00:00</td>'
        '</tr>'
        '<tr>'
        '<td>t_end</td>'
        '<td>2035-10-05T16:01:00</td>'
        '</tr>'
        '</table>'
    )

    assert display_html(events_dict) == (
        '<table>'
        '<thead>'
        '<tr>'
        '<th>'
        '</th>'
        '<th>event</th>'
        '<th>#</th>'
        '<th>t_start</th>'
        '<th>t_stop</th>'
        '</tr>'
        '</thead>'
        '<tbody>'
        '<tr>'
        '<td>'
        '<em>0</em>'
        '</td>'
        '<td>FLYBY_GANYMEDE</td>'
        '<td>-</td>'
        '<td>2031-07-21</td>'
        '<td>2031-07-21</td>'
        '</tr>'
        '<tr>'
        '<td>'
        '<em>1</em>'
        '</td>'
        '<td>GCO500</td>'
        '<td>-</td>'
        '<td>2035-05-22</td>'
        '<td>2035-10-05</td>'
        '</tr>'
        '</tbody>'
        '</table>'
    )

    assert display_html(events_list) == (
        '<table>'
        '<thead>'
        '<tr>'
        '<th>'
        '</th>'
        '<th>event time [utc]</th>'
        '<th>Crema name</th>'
        '<th>Altitude above ganymede [km]</th>'
        '</tr>'
        '</thead>'
        '<tbody>'
        '<tr>'
        '<td>'
        '<em>0</em>'
        '</td>'
        '<td>2031-07-21T07:14:23Z</td>'
        '<td>1G1</td>'
        '<td>400.00697</td>'
        '</tr>'
        '<tr>'
        '<td>'
        '<em>1</em>'
        '</td>'
        '<td>2031-07-21T07:14:23Z</td>'
        '<td>1G1</td>'
        '<td>400.00697</td>'
        '</tr>'
        '</tbody>'
        '</table>'
    )
