"""Test events file csv parser."""

from pathlib import Path

import numpy as np

from pytest import fixture, raises

from planetary_coverage.events import CsvEventsFile, Event, EventsDict, EventWindow
from planetary_coverage.events.event import EventsList
from planetary_coverage.html import Html, display_html


DATA = Path(__file__).parent / '..' / 'data'
PHASES_CSV = DATA / 'mission_phases.csv'
TIMELINE_CSV = DATA / 'mission_timeline.csv'
DOWNLINK_CSV = DATA / 'downlink_visibility.csv'


@fixture
def phases():
    """Mission phases."""
    return CsvEventsFile(PHASES_CSV)


@fixture
def timeline():
    """Mission timeline."""
    return CsvEventsFile(TIMELINE_CSV, primary_key='Event Name')


def test_events_file_parser_phases(phases):
    """Test events csv parser for mission phases."""
    assert str(phases) == 'mission_phases.csv'
    assert len(phases) == 4  # number of events

    assert phases.fields == ['column_0', 'name', 't_start', 't_end', 'column_4']

    assert phases.keys() == {
        'Approach and first ellipse',
        'GCO500',
        'All Ganymede phases',
        'All mission phases',
    }

    # Phases start and stop
    assert phases.start == np.datetime64('2031-01-19T18:50:41')
    assert phases.stop == np.datetime64('2035-10-05T16:01:00')

    # __repr__
    assert repr(phases) == '\n'.join([
        '<CsvEventsFile> mission_phases.csv (2031-01-19 -> 2035-10-05 | 4 keys):',
        ' - Approach and first ellipse (2031-01-19 -> 2032-02-08)',
        ' - GCO500 (2035-05-22 -> 2035-10-05)',
        ' - All Ganymede phases (2034-12-19 -> 2035-10-05)',
        ' - All mission phases (2031-01-19 -> 2035-10-05)',
    ])

    # __getitem__ with key
    assert phases['column_0'] == [
        'Jupiter_Phase_1',
        'Ganymede_Phase_6_5',
        'Ganymede_Phase_all',
        'Mission_phase_all',
    ]

    # __getitem__ with slice
    for event in phases[1:2]:  # [event('GCO500')] only
        assert isinstance(event, EventWindow)
        assert str(event) == 'GCO500 (2035-05-22 -> 2035-10-05)'

    # __getitem__ with int
    event = phases[1]  # event('GCO500')
    assert isinstance(event, EventWindow)
    assert str(event) == 'GCO500 (2035-05-22 -> 2035-10-05)'

    # IPython auto-completion
    assert tuple(phases._ipython_key_completions_()) == (
        'Approach and first ellipse',  # keys
        'GCO500',
        'All Ganymede phases',
        'All mission phases',
        'column_0',
        'name',
        't_start',
        't_end',
        'column_4',  # Append csv fields
    )

    # Check EventWindow
    event = phases['GCO500']

    assert isinstance(event, EventWindow)
    assert event.key == event['name'] == 'GCO500'
    assert event.start == np.datetime64('2035-05-22T06:00:00')
    assert event.stop == np.datetime64('2035-10-05T16:01:00')
    assert str(event) == 'GCO500 (2035-05-22 -> 2035-10-05)'
    assert repr(event) == '\n'.join([
        '<EventWindow> GCO500 (2035-05-22 -> 2035-10-05):',
        ' - column_0: Ganymede_Phase_6_5',
        ' - name: GCO500',
        ' - t_start: 2035-05-22T06:00:00',
        ' - t_end: 2035-10-05T16:01:00',
        ' - column_4: ',
    ])

    # Getitem error
    with raises(KeyError):
        _ = phases['foo']


def test_events_file_parser_timeline(timeline):
    """Test events csv parser for mission timeline."""
    assert str(timeline) == 'mission_timeline.csv'
    assert len(timeline) == 11  # number of events != number of rows

    assert timeline.fields == ['event name', 'event time [utc]', 'contextual info']
    assert not timeline.comments

    assert timeline.start == np.datetime64('2031-01-20T06:31:08')
    assert timeline.stop == np.datetime64('2032-02-15T06:07:11')

    # __repr__
    assert repr(timeline) == '\n'.join([
        '<CsvEventsFile> mission_timeline.csv (2031-01-20 -> 2032-02-15 | 11 keys):',
        ' - VISIBILITY_MADRID (2031-01-20 -> 2031-01-20)',
        ' - VISIBILITY_CEBREROS (2031-01-20 -> 2031-01-20)',
        ' - DOWNLINK_MALARGUE (2031-01-20 -> 2032-02-14 | 2 events)',
        ' - VISIBILITY_MALARGUE (2031-01-20 -> 2032-02-14 | 2 events)',
        ' - VISIBILITY_GOLDSTONE (2031-01-20 -> 2032-02-14 | 2 events)',
        ' - VISIBILITY_CANBERRA (2031-01-20 -> 2032-02-15 | 2 events)',
        ' - PHASE_010 (2031-07-20)',
        ' - FLYBY_GANYMEDE (2031-07-21 -> 2032-02-13 | 2 events)',
        ' - PERIJOVE_1PJ (2031-07-21)',
        ' - PHASE_090 (2031-07-22 -> 2032-02-14 | 2 events)',
        ' - PERIJOVE_2PJ (2032-02-14)',
    ])

    # Get a single Event without contextual info
    assert 'PHASE_010' in timeline

    event = timeline['PHASE_010']
    assert isinstance(event, Event)
    assert event.key == 'PHASE_010'
    assert event.start == event.stop == np.datetime64('2031-07-20T14:09:31')
    assert str(event) == 'PHASE_010 (2031-07-20)'
    assert repr(event) == '\n'.join([
        '<Event> PHASE_010 (2031-07-20):',
        ' - event name: PHASE_010',
        ' - event time [utc]: 2031-07-20T14:09:31Z',
    ])

    # Get a single Event with contextual info
    assert 'PERIJOVE_2PJ' in timeline

    event = timeline['PERIJOVE_2PJ']
    assert isinstance(event, Event)
    assert event.key == 'PERIJOVE_2PJ'
    assert event.start == event.stop == np.datetime64('2032-02-14T14:20:18')
    assert str(event) == 'PERIJOVE_2PJ (2032-02-14)'
    assert repr(event) == '\n'.join([
        '<Event> PERIJOVE_2PJ (2032-02-14):',
        ' - event name: PERIJOVE_2PJ',
        ' - event time [utc]: 2032-02-14T14:20:18Z',
        ' - Distance to Jupiter [RJ]: 12.77',
        ' - Sub-SC lon [deg]: 231.1',
        ' - Sub-SC lat [deg]: 1.7',
        ' - Sub-SC phase [deg]: 64.6',
        ' - Sub-SC loctime [hours]: 16.3',
        ' - Jupiter angular size [deg]: 9.0',
        ' - Energy generated over +-50h around CA [Wh]: 99684.2',
        ' - Energy for science over +-50h around CA [Wh]: 39684.2',
        ' - Relative generated energy amount for science compared to PJ2: 1.000',
        ' - Malargue visibility: YES',
        ' - Delta-time to closest moon Flyby: 15.25 hours after the flyby  2G2',
    ])
    assert event['Distance to Jupiter [RJ]'] == '12.77'

    # Get list of Event
    assert 'FLYBY_GANYMEDE' in timeline

    flybys = timeline['FLYBY_GANYMEDE']
    assert isinstance(flybys, EventsList)
    assert len(flybys) == 2

    assert flybys.windows == [
        (np.datetime64('2031-07-21T07:14:23'), np.datetime64('2031-07-21T07:14:23')),
        (np.datetime64('2032-02-13T23:04:51'), np.datetime64('2032-02-13T23:04:51')),
    ]

    assert flybys.start == np.datetime64('2031-07-21T07:14:23')
    assert flybys.stop == np.datetime64('2032-02-13T23:04:51')

    # Get flyby by name
    flyby = flybys['1G1']
    assert isinstance(flyby, Event)
    assert flyby.start == flyby.stop == np.datetime64('2031-07-21T07:14:23')
    assert flyby['Crema name'] == '1G1'

    # Get flyby by index
    assert flyby == flybys[0]

    # Get flybys crema names
    assert flybys.crema_names == ['1G1', '2G2']

    # Invalid flyby name
    with raises(KeyError):
        _ = flybys['FOO']

    # Get a single EventWindow
    assert 'VISIBILITY_MADRID' in timeline

    event = timeline['VISIBILITY_MADRID']
    assert isinstance(event, EventWindow)
    assert event.key == 'VISIBILITY_MADRID'
    assert event.start == np.datetime64('2031-01-20T06:31:08')
    assert event.stop == np.datetime64('2031-01-20T13:29:11')
    assert repr(event) == '\n'.join([
        '<EventWindow> VISIBILITY_MADRID (2031-01-20 -> 2031-01-20):',
        ' - event name: VISIBILITY_MADRID',
        ' - t_start: 2031-01-20T06:31:08Z',
        ' - t_end: 2031-01-20T13:29:11Z',
    ])

    # Get list of EventWindow
    assert 'DOWNLINK_MALARGUE' in timeline

    events = timeline['DOWNLINK_MALARGUE']
    assert isinstance(events, EventsList)
    assert len(events) == 2

    assert events.starts == [
        np.datetime64('2031-01-20T08:03:38'),
        np.datetime64('2032-02-14T08:16:38'),
    ]
    assert events.stops == [
        np.datetime64('2031-01-20T17:03:41'),
        np.datetime64('2032-02-14T17:17:11'),
    ]
    assert events.windows == [
        (np.datetime64('2031-01-20T08:03:38'), np.datetime64('2031-01-20T17:03:41')),
        (np.datetime64('2032-02-14T08:16:38'), np.datetime64('2032-02-14T17:17:11')),
    ]

    assert events.start == np.datetime64('2031-01-20T08:03:38')
    assert events.stop == np.datetime64('2032-02-14T17:17:11')

    event_0, event_1 = events[0], events[1]
    assert isinstance(event_0, EventWindow)
    assert isinstance(event_1, EventWindow)

    assert event_0.start == np.datetime64('2031-01-20T08:03:38')
    assert event_0.stop == np.datetime64('2031-01-20T17:03:41')
    assert event_1.start == np.datetime64('2032-02-14T08:16:38')
    assert event_1.stop == np.datetime64('2032-02-14T17:17:11')

    # Get event by name (only valid for the flybys with `Crema name` attribute)
    with raises(KeyError):
        _ = events['FOO']


def test_events_file_parser_without_header():
    """Test events csv parser without an explicit header."""
    events = CsvEventsFile(
        DOWNLINK_CSV,
        header='# name, t_start, t_end, subgroup, working_group',
    )

    # Only 1 (repeated) element is present
    assert len(events) == 1
    assert 'DL_' in events

    downlinks = events['DL_']

    assert isinstance(downlinks, EventsList)
    assert len(downlinks) == 4

    assert downlinks.start == np.datetime64('2034-12-19T17:55:17')
    assert downlinks.stop == np.datetime64('2035-09-28T11:38:51')

    downlink = downlinks[0]

    assert isinstance(downlink, EventWindow)
    assert downlink.start == np.datetime64('2034-12-19T17:55:17')
    assert downlink.stop == np.datetime64('2034-12-20T04:10:31')


def test_events_file_parser_errors(caplog):
    """Test events csv parser errors."""
    # Invalid events file format
    with raises((KeyError, BufferError)):
        _ = CsvEventsFile(DATA / 'dummy.evf')

    # Missing primary key
    err = 'Primary key `name` not found'
    with raises(KeyError, match=err):
        _ = CsvEventsFile(TIMELINE_CSV)

    # Init dummy timeline
    csv = Path('dummy_timeline.csv')
    if csv.exists():
        csv.unlink()

    # Missing input file
    with raises(FileNotFoundError):
        _ = CsvEventsFile(csv)

    # Create dummy timeline with missing `VISIBILITY_CEBREROS_STOP`
    # and `VISIBILITY_MADRID_START`
    csv.write_text(
        '\n'.join([
            '#name, event time [utc], contextual info',
            'VISIBILITY_CEBREROS_START,2031-01-20T06:31:38Z,',
            'VISIBILITY_MADRID_END,2031-01-20T13:29:11Z,',
        ]),
        encoding='utf-8',
    )

    timeline = CsvEventsFile(csv)

    assert len(timeline) == 1

    assert 'VISIBILITY_CEBREROS' in timeline
    assert 'VISIBILITY_MADRID' not in timeline

    assert caplog.messages == [
        'Found `VISIBILITY_MADRID_END` (at 2031-01-20T13:29:11Z) '
        'without `VISIBILITY_MADRID_START`.'
    ]

    event = timeline['VISIBILITY_CEBREROS']

    assert event.start == np.datetime64('2031-01-20T06:31:38')
    assert np.isnat(event.stop)

    assert str(event) == 'VISIBILITY_CEBREROS (2031-01-20 -> NaT)'

    csv.unlink()


def test_event_file_getter(timeline):
    """Test event file getter."""
    events = timeline
    assert len(events) == 11

    # File header column
    items = events['event name']
    assert len(items) == 27
    assert items[0] == 'VISIBILITY_MADRID_START'
    assert items[-1] == 'VISIBILITY_CANBERRA_END'

    # With a slice
    items = events[1:3]
    assert isinstance(items, EventsDict)
    assert len(items) == 2
    assert str(items[0]) == 'VISIBILITY_CEBREROS (2031-01-20 -> 2031-01-20)'
    assert str(items[1]) == 'DOWNLINK_MALARGUE (2031-01-20 -> 2032-02-14 | 2 events)'

    # With an int
    assert str(events[0]) == 'VISIBILITY_MADRID (2031-01-20 -> 2031-01-20)'
    assert str(events[1]) == 'VISIBILITY_CEBREROS (2031-01-20 -> 2031-01-20)'
    assert str(events[-1]) == 'PERIJOVE_2PJ (2032-02-14)'
    assert str(events[-11]) == 'VISIBILITY_MADRID (2031-01-20 -> 2031-01-20)'

    with raises(IndexError, match='Event index out of range'):
        _ = events[11]

    with raises(IndexError, match='Event index out of range'):
        _ = events[-12]

    # Full keyword with a single Event/EventWindow/EventsList
    assert str(events['PHASE_010']) == 'PHASE_010 (2031-07-20)'

    assert (
        str(events['VISIBILITY_MADRID']) == 'VISIBILITY_MADRID (2031-01-20 -> 2031-01-20)'
    )

    assert (
        str(events['DOWNLINK_MALARGUE'])
        == 'DOWNLINK_MALARGUE (2031-01-20 -> 2032-02-14 | 2 events)'
    )

    # Partial keyword with multiple Event/EventWindow/EventsList
    items = events['PHASE_']
    assert len(items) == 2
    assert str(items[0]) == 'PHASE_010 (2031-07-20)'
    assert str(items[1]) == 'PHASE_090 (2031-07-22 -> 2032-02-14 | 2 events)'

    # With a multiple keywords (sorted by dates)
    items = events['PHASE_090', 'PHASE_010']
    assert len(items) == 2
    assert str(items[0]) == 'PHASE_010 (2031-07-20)'
    assert str(items[1]) == 'PHASE_090 (2031-07-22 -> 2032-02-14 | 2 events)'

    # With a multiple keywords and dummy values (single output here)
    items = events['PHASE_010', 'FOO']
    assert str(items) == 'PHASE_010 (2031-07-20)'

    # With a multiple keywords with duplicated (discarded)
    items = events['PHASE_01', 'PHASE_010']
    assert str(items) == 'PHASE_010 (2031-07-20)'

    # Error (no output event find)
    with raises(KeyError, match='`FOO` not found'):
        _ = events['FOO']

    # Global regex finder
    items = events[r'^V.*E$']
    assert isinstance(items, EventsDict)
    assert len(items) == 2
    assert str(items[0]) == 'VISIBILITY_MALARGUE (2031-01-20 -> 2032-02-14 | 2 events)'
    assert str(items[1]) == 'VISIBILITY_GOLDSTONE (2031-01-20 -> 2032-02-14 | 2 events)'

    # Not found
    with raises(KeyError, match='`FOO` not found'):
        _ = events.find('FOO')

    # Startswith
    items = events.startswith('FLYBY')
    assert isinstance(items, EventsList)
    assert len(items) == 2
    assert str(items) == 'FLYBY_GANYMEDE (2031-07-21 -> 2032-02-13 | 2 events)'

    # Endswith
    items = events.endswith('PJ')
    assert isinstance(items, EventsDict)
    assert len(items) == 2
    assert str(items[0]) == 'PERIJOVE_1PJ (2031-07-21)'
    assert str(items[1]) == 'PERIJOVE_2PJ (2032-02-14)'


def test_events_file_comparisons(timeline):
    """Test multi-events comparison operations."""
    assert len(timeline) == 11

    events = timeline.after(2032)

    assert isinstance(events, EventsDict)
    assert len(events) == 7
    assert 'FLYBY_GANYMEDE' in events
    assert 'DOWNLINK_MALARGUE' in events
    assert 'VISIBILITY_MADRID' not in events

    event = events['FLYBY_GANYMEDE']  # Only 1 of the 2 flybys is kept
    assert isinstance(event, Event)
    assert event['Crema name'] == '2G2'

    events = timeline.before('2032')

    assert isinstance(events, EventsDict)
    assert len(events) == 10
    assert 'FLYBY_GANYMEDE' in events
    assert 'VISIBILITY_MADRID' in events
    assert 'DOWNLINK_MALARGUE' in events
    assert 'PERIJOVE_2PJ' not in events

    event = events['FLYBY_GANYMEDE']  # Only 1 of the 2 flybys is kept
    assert isinstance(event, Event)
    assert event['Crema name'] == '1G1'

    event = events['DOWNLINK_MALARGUE']  # Only 1 of the 2 events is kept
    assert isinstance(event, EventWindow)
    assert event >= '2031-01-20T08:03:38'
    assert event <= '2031-01-20T17:03:41'

    events = timeline.between('2031-06', '2032-03', strict=True)
    assert isinstance(events, EventsDict)
    assert len(events) == 9

    # The events are reordered
    assert isinstance(events[0], Event)
    assert str(events[0]) == 'PHASE_010 (2031-07-20)'

    assert isinstance(events[1], EventsList)
    assert str(events[1]) == 'FLYBY_GANYMEDE (2031-07-21 -> 2032-02-13 | 2 events)'

    assert isinstance(events[-1], EventWindow)
    assert str(events[-1]) == 'VISIBILITY_CANBERRA (2032-02-14 -> 2032-02-15)'

    # Errors
    with raises(LookupError, match='mission_timeline.csv > 2031'):
        _ = timeline.before('2031')

    with raises(LookupError, match='mission_timeline.csv < 2033'):
        _ = timeline.after('2033')


def test_events_file_html_repr(phases, timeline):
    """Test CsvEventsFile HTML representation."""
    assert display_html(phases) == (
        '<table>'
        '<thead>'
        '<tr><th></th><th>event</th><th>#</th><th>t_start</th><th>t_stop</th></tr>'
        '</thead>'
        '<tbody>'
        '<tr><td><em>0</em></td><td>Approach and first ellipse</td><td>-</td>'
        '<td>2031-01-19</td><td>2032-02-08</td></tr>'
        '<tr><td><em>1</em></td><td>GCO500</td><td>-</td>'
        '<td>2035-05-22</td><td>2035-10-05</td></tr>'
        '<tr><td><em>2</em></td><td>All Ganymede phases</td><td>-</td>'
        '<td>2034-12-19</td><td>2035-10-05</td></tr>'
        '<tr><td><em>3</em></td><td>All mission phases</td><td>-</td>'
        '<td>2031-01-19</td><td>2035-10-05</td></tr>'
        '</tbody>'
        '</table>'
    )

    assert display_html(timeline) == (
        '<table>'
        '<thead>'
        '<tr><th></th><th>event</th><th>#</th><th>t_start</th><th>t_stop</th></tr>'
        '</thead>'
        '<tbody>'
        '<tr><td><em>0</em></td><td>VISIBILITY_MADRID</td><td>-</td>'
        '<td>2031-01-20</td><td>2031-01-20</td></tr>'
        '<tr><td><em>1</em></td><td>VISIBILITY_CEBREROS</td><td>-</td>'
        '<td>2031-01-20</td><td>2031-01-20</td></tr>'
        '<tr><td><em>2</em></td><td>DOWNLINK_MALARGUE</td><td>2</td>'
        '<td>2031-01-20</td><td>2032-02-14</td></tr>'
        '<tr><td><em>3</em></td><td>VISIBILITY_MALARGUE</td><td>2</td>'
        '<td>2031-01-20</td><td>2032-02-14</td></tr>'
        '<tr><td><em>4</em></td><td>VISIBILITY_GOLDSTONE</td><td>2</td>'
        '<td>2031-01-20</td><td>2032-02-14</td></tr>'
        '<tr><td><em>5</em></td><td>VISIBILITY_CANBERRA</td><td>2</td>'
        '<td>2031-01-20</td><td>2032-02-15</td></tr>'
        '<tr><td><em>6</em></td><td>PHASE_010</td><td>-</td>'
        '<td>2031-07-20</td><td>2031-07-20</td></tr>'
        '<tr><td><em>7</em></td><td>FLYBY_GANYMEDE</td><td>2</td>'
        '<td>2031-07-21</td><td>2032-02-13</td></tr>'
        '<tr><td><em>8</em></td><td>PERIJOVE_1PJ</td><td>-</td>'
        '<td>2031-07-21</td><td>2031-07-21</td></tr>'
        '<tr><td><em>9</em></td><td>PHASE_090</td><td>2</td>'
        '<td>2031-07-22</td><td>2032-02-14</td></tr>'
        '<tr><td><em>10</em></td><td>PERIJOVE_2PJ</td><td>-</td>'
        '<td>2032-02-14</td><td>2032-02-14</td></tr>'
        '</tbody>'
        '</table>'
    )


def test_events_file_csv_repr(phases):
    """Test CsvEventsFile CSV representation as a HTML table."""
    assert isinstance(phases.csv, Html)

    assert display_html(phases.csv) == (
        '<table>'
        '<thead>'
        '<tr>'
        '<th>column_0</th><th>name</th><th>t_start</th><th>t_end</th><th>column_4</th>'
        '</tr>'
        '</thead>'
        '<tbody>'
        '<tr><td>Jupiter_Phase_1</td><td>Approach and first ellipse</td>'
        '<td>2031-01-19T18:50:41</td><td>2032-02-08T23:05:31</td><td></td></tr>'
        '<tr><td>Ganymede_Phase_6_5</td><td>GCO500</td>'
        '<td>2035-05-22T06:00:00</td><td>2035-10-05T16:01:00</td><td></td></tr>'
        '<tr><td>Ganymede_Phase_all</td><td>All Ganymede phases</td>'
        '<td>2034-12-19T12:00:00</td><td>2035-10-05T16:01:00</td><td></td></tr>'
        '<tr><td>Mission_phase_all</td><td>All mission phases</td>'
        '<td>2031-01-19T18:50:41</td><td>2035-10-05T16:01:00</td><td></td></tr>'
        '</tbody>'
        '</table>'
    )
