"""Test EVF Event File module."""

from pathlib import Path

from numpy import datetime64 as dt

from pytest import raises

from planetary_coverage.events import Event, EventsList, EventWindow, EvfEventsFile
from planetary_coverage.events.evf import evf_key, evf_rows
from planetary_coverage.html import display_html


DATA = Path(__file__).parent / '..' / 'data'


def test_events_evf_key_parser():
    """Test EVF key parser."""
    assert evf_key('CA_EUROPA') == ('CA_EUROPA', None)  # No count
    assert evf_key('CA_Europa') == ('CA_EUROPA', None)  # Uppercase
    assert evf_key('CA_EUROPA (COUNT = 1)') == ('CA_EUROPA', 1)  # With COUNT
    assert evf_key('CA_EUROPA      (COUNT =  1)') == ('CA_EUROPA', 1)  # Strip spaces
    assert evf_key('CA_EUROPA(COUNT=1)') == ('CA_EUROPA', 1)  # No spaces

    with raises(KeyError, match='Invalid EVF key: ?'):
        _ = evf_key('?')


def test_events_evf_rows_parser():
    """Test EVF rows parser"""
    comments, data = evf_rows(
        '\n'.join([
            '# foo',
            '#   bar',
            '# baz  ',
            '22-JAN-2031_18:54:03    CA_EUROPA    (COUNT =  1)',
            '29-JAN-2031_21:21:56    CA_EUROPA    (COUNT =  2)',
            '26-JAN-2031_08:08:00    CA_GANYMEDE  (COUNT =  1)',
        ])
    )

    assert comments == ['foo', 'bar', 'baz']  # Stripped comments
    assert data == {
        'CA_EUROPA': [
            '2031-01-22T18:54:03',
            '2031-01-29T21:21:56',
        ],
        'CA_GANYMEDE': ['2031-01-26T08:08:00'],
    }

    with raises(KeyError, match=r'`CA_EUROPA` missing \(COUNT = 2\)'):
        _ = evf_rows(
            '\n'.join([
                '22-JAN-2031_18:54:03    CA_EUROPA    (COUNT =  1)',
                '29-JAN-2031_21:21:56    CA_EUROPA    (COUNT =  3)',
            ])
        )


def test_events_evf_file_parser():
    """Test EVF file parser."""
    evf = EvfEventsFile(DATA / 'dummy.evf')

    assert str(evf) == 'dummy.evf'
    assert evf.fname == DATA / 'dummy.evf'

    # __len__
    assert len(evf) == 3

    # Keys
    assert evf.keys() == {'CA_EUROPA', 'CA_GANYMEDE', 'OCC'}

    # Properties
    assert evf.comments == ['Dummy examples']
    assert evf.fields == ['event time [utc]', 'name']
    assert evf.rows == [
        # Sorted by time
        ('2030-01-01T00:00:00', 'OCC_START'),
        ('2030-01-02T00:00:00', 'OCC_END'),
        ('2031-01-01T00:00:00', 'OCC_START'),
        ('2031-01-02T00:00:00', 'OCC_END'),
        ('2031-01-22T18:54:03', 'CA_EUROPA'),
        ('2031-01-26T08:08:00', 'CA_GANYMEDE'),
        ('2031-01-29T21:21:56', 'CA_EUROPA'),
        ('2031-02-02T10:35:52', 'CA_EUROPA'),
        ('2031-02-05T23:49:50', 'CA_EUROPA'),
    ]

    # __repr__
    assert repr(evf) == '\n'.join([
        '<EvfEventsFile> dummy.evf (2030-01-01 -> 2031-02-05 | 3 keys):',
        ' - OCC (2030-01-01 -> 2031-01-02 | 2 events)',
        ' - CA_EUROPA (2031-01-22 -> 2031-02-05 | 4 events)',
        ' - CA_GANYMEDE (2031-01-26)',
    ])

    # __contains__
    assert 'CA_EUROPA' in evf
    assert 'CA_EUROPA (COUNT = 1)' in evf
    assert 'OCC_START' not in evf
    assert 'OCC' in evf

    # Get EventsList
    events = evf['CA_EUROPA']

    assert isinstance(events, EventsList)
    assert len(events) == 4

    assert events.windows == [
        (dt('2031-01-22T18:54:03'), dt('2031-01-22T18:54:03')),
        (dt('2031-01-29T21:21:56'), dt('2031-01-29T21:21:56')),
        (dt('2031-02-02T10:35:52'), dt('2031-02-02T10:35:52')),
        (dt('2031-02-05T23:49:50'), dt('2031-02-05T23:49:50')),
    ]

    assert events.start == dt('2031-01-22T18:54:03')
    assert events.stop == dt('2031-02-05T23:49:50')

    # Get a single Event
    event = evf['CA_EUROPA (COUNT = 1)']

    assert isinstance(event, Event)
    assert str(event) == 'CA_EUROPA (2031-01-22)'
    assert event.start == dt('2031-01-22T18:54:03')
    assert event.stop == dt('2031-01-22T18:54:03')

    # Get a single EventWindow
    event = evf['OCC (COUNT = 1)']

    assert isinstance(event, EventWindow)
    assert str(event) == 'OCC (2030-01-01 -> 2030-01-02)'
    assert event.start == dt('2030-01-01')
    assert event.stop == dt('2030-01-02')

    # HTML repr
    assert display_html(evf) == (
        '<table>'
        '<thead>'
        '<tr><th>event</th>'
        '<th>#</th><th>t_start</th><th>t_stop</th></tr>'
        '</thead>'
        '<tbody>'
        '<tr><td>OCC</td>'
        '<td>2</td><td>2030-01-01</td><td>2031-01-02</td></tr>'
        '<tr><td>CA_EUROPA</td>'
        '<td>4</td><td>2031-01-22</td><td>2031-02-05</td></tr>'
        '<tr><td>CA_GANYMEDE</td>'
        '<td>-</td><td>2031-01-26</td><td>2031-01-26</td></tr>'
        '</tbody>'
        '</table>'
    )

    # Errors
    with raises(IndexError):
        _ = evf['CA_EUROPA (COUNT = 5)']

    with raises(KeyError):
        _ = evf['OCC_START']


def test_events_evf_errors():
    """Test EVF errors."""
    # Invalid events file format
    with raises((ValueError, BufferError)):
        _ = EvfEventsFile(DATA / 'dummy.itl')
