"""Test generic Events file reader module"""

from pathlib import Path

from pytest import raises

from planetary_coverage import read_events
from planetary_coverage.events import (
    CsvEventsFile,
    EvfEventsFile,
    ItlEventsFile,
    OrbitEventsFile,
)


DATA = Path(__file__).parent / '..' / 'data'


def test_events_reader_evf():
    """Test events EVF reader."""
    evf = read_events(DATA / 'dummy.evf')

    assert isinstance(evf, EvfEventsFile)
    assert len(evf) == 3
    assert evf.keys() == {'CA_EUROPA', 'CA_GANYMEDE', 'OCC'}


def test_events_reader_itl():
    """Test events ITL reader."""
    itl = read_events(DATA / 'dummy.itl', evf=DATA / 'dummy.evf')

    assert isinstance(itl, ItlEventsFile)
    assert len(itl) == 2
    assert itl.keys() == {'OBS', 'SWITCH_MODE'}
    assert 'UVS_SAT_STELL_OCC_A' in itl
    assert 'MAJIS_FLYBY_HR_01_001' in itl

    # Error is EVF is missing
    with raises(FileNotFoundError):
        _ = read_events(DATA / 'dummy.itl')


def test_events_reader_csv_header():
    """Test events CSV reader with explicit header."""
    downlinks = read_events(
        DATA / 'downlink_visibility.csv',
        header='# name, t_start, t_end, subgroup, working_group',
    )

    assert isinstance(downlinks, CsvEventsFile)
    assert len(downlinks) == 1
    assert downlinks.keys() == {'DL_'}


def test_events_reader_csv_phases():
    """Test events CSV reader for mission phase file."""
    phases = read_events(DATA / 'mission_phases.csv')

    assert isinstance(phases, CsvEventsFile)
    assert len(phases) == 4
    assert phases.keys() == {
        'Approach and first ellipse',
        'GCO500',
        'All Ganymede phases',
        'All mission phases',
    }


def test_events_reader_csv_timeline():
    """Test events CSV reader for mission timeline file."""
    timeline = read_events(DATA / 'mission_timeline.csv')

    assert isinstance(timeline, CsvEventsFile)
    assert len(timeline) == 11
    assert 'VISIBILITY_MADRID' in timeline


def test_events_reader_csv_events():
    """Test events CSV reader for generic mission events."""
    downlinks = read_events(DATA / 'downlink_visibility.csv')

    assert isinstance(downlinks, CsvEventsFile)
    assert len(downlinks) == 1
    assert downlinks.keys() == {'DL_'}


def test_events_reader_orb():
    """Test events orbit number reader."""
    orb = read_events(DATA / 'orbnum.orb')

    assert isinstance(orb, OrbitEventsFile)
    assert len(orb) == 5
    assert list(orb.keys()) == ['1', '2', '3', '4', '5']


def test_events_reader_errors():
    """Test events reader errors."""
    # File does not exist
    with raises(FileNotFoundError):
        _ = read_events('foo')

    # File can not be parsed
    with raises(BufferError):
        _ = read_events(DATA / 'dummy.tf')
