"""Test ITL Event File module."""

from pathlib import Path

from numpy import datetime64 as dt

from pytest import raises

import planetary_coverage.events.itl as itl_submodule
from planetary_coverage.events import EventsList, EventWindow, ItlEventsFile
from planetary_coverage.html import display_html


DATA = Path(__file__).parent / '..' / 'data'


def test_events_itl_parser():
    """Test ITL file parser"""
    itl = ItlEventsFile(DATA / 'dummy.itl', evf=DATA / 'dummy.evf')

    assert str(itl) == 'dummy.itl'
    assert itl.fname == DATA / 'dummy.itl'

    # __len__
    assert len(itl) == 2

    # Keys
    assert itl.keys() == {'OBS', 'SWITCH_MODE'}

    # Properties
    assert not itl.comments
    assert itl.fields == ['event', 'event time [utc]', 'contextual info']
    assert itl.rows == [
        (
            'OBS_START',
            '2032-07-02T04:18:40',
            'instrument=UVS;observation name=UVS_SAT_STELL_OCC_A;prime=true;'
            'comments=Block comment with absolute timing',
        ),
        (
            'OBS_END',
            '2032-07-02T04:32:00',
            'instrument=UVS;observation name=UVS_SAT_STELL_OCC_A;comments=--',
        ),
        (
            'OBS_START',
            '2031-02-05T23:22:50',
            'instrument=MAJIS;observation name=MAJIS_FLYBY_HR_01_001;prime=true;'
            'comments=Multi-lines comment block with EVF relative timing '
            '(related to `dummy.evf`):\n'
            '  05-FEB-2031_23:49:50    CA_EUROPA    (COUNT ＝  4)\n'
            'with a tab, a semi-column , and an equal ＝ sign.',
        ),
        (
            'OBS_END',
            '2031-02-05T23:24:44',
            'instrument=MAJIS;observation name=MAJIS_FLYBY_HR_01_001;comments=--',
        ),
        (
            'SWITCH_MODE',
            '2031-02-06T00:16:25',
            'instrument=MAJIS;current_mode=service3 [eng];'
            'comments=Switch mode absolute block',
        ),
        (
            'SWITCH_MODE',
            '2031-02-06T04:26:45',
            'instrument=MAJIS;current_mode=off [eng];'
            'comments=Switch mode relative block',
        ),
    ]

    # __repr__
    assert repr(itl) == '\n'.join([
        '<ItlEventsFile> dummy.itl (2031-02-05 -> 2032-07-02 | 2 keys):',
        ' - OBS (2031-02-05 -> 2032-07-02 | 2 events)',
        ' - SWITCH_MODE (2031-02-06 -> 2031-02-06 | 2 events)',
    ])

    # __contains__
    assert 'OBS' in itl
    assert 'SWITCH_MODE' in itl
    assert 'FOO' not in itl
    assert 'UVS_SAT_STELL_OCC_A' in itl  # check observation names as well

    # Get EventsList
    events = itl['SWITCH_MODE']

    assert isinstance(events, EventsList)
    assert len(events) == 2

    assert events.windows == [
        (dt('2031-02-06T00:16:25'), dt('2031-02-06T00:16:25')),
        (dt('2031-02-06T04:26:45'), dt('2031-02-06T04:26:45')),
    ]

    assert events.start == dt('2031-02-06T00:16:25')
    assert events.stop == dt('2031-02-06T04:26:45')

    # Get all observations
    obs = itl.observations

    assert isinstance(events, EventsList)
    assert len(events) == 2

    assert 'UVS_SAT_STELL_OCC_A' in obs
    assert 'MAJIS_FLYBY_HR_01_001' in obs
    assert 'OBS' not in obs

    # Get a single EventWindow from observations
    event = obs['UVS_SAT_STELL_OCC_A']

    assert isinstance(event, EventWindow)
    assert event['observation name'] == 'UVS_SAT_STELL_OCC_A'
    assert event.start == dt('2032-07-02T04:18:40')
    assert event.stop == dt('2032-07-02T04:32:00')

    # Get a single EventWindow from itl itself
    event = itl['MAJIS_FLYBY_HR_01_001']

    assert isinstance(event, EventWindow)
    assert event['observation name'] == 'MAJIS_FLYBY_HR_01_001'
    assert event.start == dt('2031-02-05T23:22:50')
    assert event.stop == dt('2031-02-05T23:24:44')

    # HTML repr
    assert display_html(itl) == (
        '<table>'
        '<thead>'
        '<tr><th>event</th><th>#</th><th>t_start</th><th>t_stop</th></tr>'
        '</thead>'
        '<tbody>'
        '<tr><td>OBS</td><td>2</td><td>2031-02-05</td><td>2032-07-02</td></tr>'
        '<tr><td>SWITCH_MODE</td><td>2</td><td>2031-02-06</td><td>2031-02-06</td></tr>'
        '</tbody>'
        '</table>'
    )

    # IPython auto-completion
    assert tuple(itl._ipython_key_completions_()) == (
        # Primary keys
        'OBS',
        'SWITCH_MODE',
        # Observations names
        'UVS_SAT_STELL_OCC_A',
        'MAJIS_FLYBY_HR_01_001',
    )


def test_events_itl_errors(monkeypatch):
    """Test ITL errors."""
    # Disable EvfEventsFile object
    monkeypatch.setattr(itl_submodule, 'EvfEventsFile', lambda x: {'foo': x})

    # Missing EVF file for relative times
    with raises(FileNotFoundError):
        _ = ItlEventsFile(DATA / 'dummy.itl')

    # Missing EVF key for relative times
    with raises(KeyError):
        _ = ItlEventsFile(DATA / 'dummy.itl', evf='foo.evf')

    # Invalid events file format
    with raises(BufferError):
        _ = ItlEventsFile(DATA / 'dummy.evf')
