"""Test orbit number event file module."""

from pathlib import Path

from numpy import datetime64 as dt

from pytest import raises

from planetary_coverage.events import Event, EventsList, OrbitEventsFile
from planetary_coverage.html import display_html


DATA = Path(__file__).parent / '..' / 'data'


def test_events_orb():
    """Test orbit number parser."""
    orbits = OrbitEventsFile(DATA / 'orbnum.orb')

    assert len(orbits) == 5
    assert list(orbits.keys()) == ['1', '2', '3', '4', '5']

    assert orbits.fields == [
        'no.',
        'event utc peri',
        'event sclk peri',
        'orbit start',
        'orbit end',
        'lonnode',
        'arg per',
        'sol dist',
        'semi axis',
    ]

    # Get orbit by positive index (No. value)
    orbit = orbits[1]

    assert isinstance(orbit, Event)
    assert orbit.start == dt('2031-07-21T20:46:34')
    assert orbit['event sclk peri'] == '1/0288391662.57647'
    assert orbit['semi axis'] == '-8734123.78'  # not parsed as float

    # Get orbit by negative index
    orbit = orbits[-1]

    assert isinstance(orbit, Event)
    assert orbit.start == dt('2032-06-02T01:35:53')
    assert orbit['no.'] == '5'

    # Get orbits by tuple
    orbits_135 = orbits[1, 3, 5]

    assert isinstance(orbits_135, EventsList)
    assert len(orbits_135) == 3
    assert orbits_135.start == dt('2031-07-21T20:46:34')
    assert orbits_135.stop == dt('2032-06-02T01:35:53')
    assert orbits_135[0]['no.'] == '1'
    assert orbits_135[1]['no.'] == '3'
    assert orbits_135[2]['no.'] == '5'

    # Get orbits by list
    orbits_24 = orbits[[2, 4]]

    assert isinstance(orbits_24, EventsList)
    assert len(orbits_24) == 2
    assert orbits_24.start == dt('2032-02-14T14:19:45')
    assert orbits_24.stop == dt('2032-05-10T14:44:53')
    assert orbits_24[0]['no.'] == '2'
    assert orbits_24[1]['no.'] == '4'

    # Get orbits by slice
    orbits_123 = orbits[1:3]  # the last element is included!

    assert isinstance(orbits_123, EventsList)
    assert len(orbits_123) == 3
    assert orbits_123.start == dt('2031-07-21T20:46:34')
    assert orbits_123.stop == dt('2032-04-11T22:47:53')
    assert orbits_123[0]['no.'] == '1'
    assert orbits_123[1]['no.'] == '2'
    assert orbits_123[2]['no.'] == '3'

    # Get orbits by slice with negative start/stop
    orbits_345 = orbits[-3:-1]

    assert isinstance(orbits_345, EventsList)
    assert len(orbits_345) == 3
    assert orbits_345[0]['no.'] == '3'
    assert orbits_345[1]['no.'] == '4'
    assert orbits_345[2]['no.'] == '5'

    # Get orbits by slice with negative start/stop (no edge)
    orbits_34 = orbits[-3:-2]

    assert isinstance(orbits_34, EventsList)
    assert len(orbits_34) == 2
    assert orbits_34[0]['no.'] == '3'
    assert orbits_34[1]['no.'] == '4'

    # Get orbits by slice with step
    orbits_135 = orbits[1:5:2]

    assert isinstance(orbits_135, EventsList)
    assert len(orbits_135) == 3
    assert orbits_135[0]['no.'] == '1'
    assert orbits_135[1]['no.'] == '3'
    assert orbits_135[2]['no.'] == '5'

    # Get orbits by slice with None start
    orbits_135 = orbits[:5:2]

    assert isinstance(orbits_135, EventsList)
    assert len(orbits_135) == 3
    assert orbits_135[0]['no.'] == '1'
    assert orbits_135[1]['no.'] == '3'
    assert orbits_135[2]['no.'] == '5'

    # Get orbits by slice with None stop
    orbits_135 = orbits[1::2]

    assert isinstance(orbits_135, EventsList)
    assert len(orbits_135) == 3
    assert orbits_135[0]['no.'] == '1'
    assert orbits_135[1]['no.'] == '3'
    assert orbits_135[2]['no.'] == '5'

    # Errors (out of range)
    msg = '0 out of range. Orbit available: 1 to 5'
    with raises(IndexError, match=msg):
        _ = orbits[0]

    msg = '6 out of range. Orbit available: 1 to 5'
    with raises(IndexError, match=msg):
        _ = orbits[6]

    msg = 'Only `int`, `tuple`, `list` and `slice` are accepted. `float` provided.'
    with raises(TypeError, match=msg):
        _ = orbits[1.2]

    # HTML repr
    assert display_html(orbits) == (
        '<table>'
        '<thead>'
        '<tr>'
        '<th>no.</th><th>event utc peri</th><th>event sclk peri</th>'
        '<th>orbit start</th><th>orbit end</th>'
        '<th>lonnode</th><th>arg per</th><th>sol dist</th><th>semi axis</th>'
        '</tr>'
        '</thead>'
        '<tbody>'
        '<tr>'
        '<td>1</td><td>2031 JUL 21 20:46:34</td><td>1/0288391662.57647</td>'
        '<td>N/A</td><td>2031 NOV 02 02:33:58</td>'
        '<td>6.09</td><td>153.67</td><td>788073210.9</td><td>-8734123.78</td>'
        '</tr><tr>'
        '<td>2</td><td>2032 FEB 14 14:19:45</td><td>1/0306339653.54137</td>'
        '<td>2031 NOV 02 02:33:58</td><td>2032 MAR 14 05:22:01</td>'
        '<td>3.52</td><td>164.99</td><td>776960863.8</td><td>4313415.24</td>'
        '</tr><tr>'
        '<td>3</td><td>2032 APR 11 22:47:53</td><td>1/0311294942.33451</td>'
        '<td>2032 MAR 14 05:22:01</td><td>2032 APR 26 06:02:38</td>'
        '<td>358.81</td><td>180.15</td><td>773845024.9</td><td>2712784.67</td>'
        '</tr><tr>'
        '<td>4</td><td>2032 MAY 10 14:44:53</td><td>1/0313771562.20434</td>'
        '<td>2032 APR 26 06:02:38</td><td>2032 MAY 21 20:13:05</td>'
        '<td>358.07</td><td>186.26</td><td>772293776.2</td><td>2298334.65</td>'
        '</tr><tr>'
        '<td>5</td><td>2032 JUN 02 01:35:53</td><td>1/0315711422.21023</td>'
        '<td>2032 MAY 21 20:13:05</td><td>2032 JUN 10 04:29:47</td>'
        '<td>358.07</td><td>186.24</td><td>771083464.1</td><td>2298210.38</td>'
        '</tr>'
        '</tbody>'
        '</table>'
    )

    # IPython auto-completion
    assert orbits._ipython_key_completions_() == [
        1,
        2,
        3,
        4,
        5,
    ]
