"""Test math auxiliary spherical formula."""

import numpy as np
from numpy.testing import assert_array_almost_equal as assert_array

from pytest import approx, raises

from planetary_coverage.math import hav_dist
from planetary_coverage.math.sphere import hav, sph_pixel_scale


def test_hav():
    """Test half versine function."""
    assert hav(0) == approx(0, abs=1e-9)
    assert hav(np.pi / 2) == approx(0.5, abs=1e-9)
    assert hav(np.pi) == approx(1, abs=1e-9)
    assert hav(2 * np.pi) == approx(0, abs=1e-9)

    # 1D array
    assert_array(hav([0, np.pi / 2, np.pi, 2 * np.pi]), [0, 0.5, 1, 0], decimal=9)

    # 2D array
    assert_array(hav([[0, np.pi / 2], [np.pi, 2 * np.pi]]), [[0, 0.5], [1, 0]], decimal=9)


def test_hav_dist():
    """Test the haversine distance on a sphere."""
    assert hav_dist(0, 0, 90, 0) == approx(np.pi / 2, abs=1e-9)
    assert hav_dist(0, 0, 0, 45) == approx(np.pi / 4, abs=1e-9)

    assert hav_dist(90, 0, 180, 0) == approx(np.pi / 2, abs=1e-9)
    assert hav_dist(0, -90, 0, 0) == approx(np.pi / 2, abs=1e-9)

    assert hav_dist(90, 0, 180, 0) == approx(np.pi / 2, abs=1e-9)
    assert hav_dist(0, -90, 0, 0) == approx(np.pi / 2, abs=1e-9)

    # Antipodes
    assert hav_dist(10, 10, 190, -10) == approx(np.pi, abs=1e-6)

    # Same point
    assert hav_dist(10, 10, 10, 10) == approx(0, abs=1e-6)

    # 1D array
    assert_array(hav_dist([90, 0], [0, 45], 0, 0), [np.pi / 2, np.pi / 4], decimal=6)
    assert_array(hav_dist(0, 0, [90, 0], [0, 45]), [np.pi / 2, np.pi / 4], decimal=6)

    assert_array(
        hav_dist([0, 10], [0, 10], [90, 190], [0, -10]), [np.pi / 2, np.pi], decimal=6
    )

    # 2D array
    assert_array(
        hav_dist([[0], [10]], [[0], [10]], [[90], [190]], [[0], [-10]]),
        [[np.pi / 2], [np.pi]],
        decimal=6,
    )

    assert_array(
        hav_dist([[0, 90]], [[0, 45]], [[90]], [[0]]), [[np.pi / 2, np.pi / 4]], decimal=6
    )

    assert_array(
        hav_dist([[90]], [[0]], [[0, 90]], [[0, 45]]), [[np.pi / 2, np.pi / 4]], decimal=6
    )

    assert_array(
        hav_dist([[0, 90]], [[0, 45]], [[90]], [[0]]), [[np.pi / 2, np.pi / 4]], decimal=6
    )

    assert_array(
        hav_dist([[0, 90]], [[0, 45]], [[90], [180], [0]], [[0], [0], [0]]),
        [[np.pi / 2, np.pi / 4], [np.pi, np.pi / 2], [0, np.pi / 2]],
        decimal=6,
    )


def test_sph_pixel_scale():
    """Test pixel scale in spherical geometry."""
    # iFOV = 0 -> 0
    assert sph_pixel_scale(30, 0, 3, 1) == 0

    # e = 0 -> iFOV * alt
    assert sph_pixel_scale(0, 0.1, 3, 1) == approx(0.2, abs=1e-2)
    assert sph_pixel_scale(0, 0.1, 4, 1) == approx(0.3, abs=1e-2)
    assert sph_pixel_scale(0, 0.1, 6, 2) == approx(0.4, abs=1e-2)
    assert sph_pixel_scale(30, 0.1, 3, 2) == approx(0.13, abs=1e-2)

    # Test with valid values
    assert sph_pixel_scale(30, np.radians(5), 3, 1) == approx(np.radians(12.19), abs=1e-2)
    assert sph_pixel_scale(30, np.radians(10), 3, 1) == approx(np.radians(25.2), abs=1e-2)
    assert sph_pixel_scale(60, np.radians(5), 3, 1) == approx(np.radians(29.53), abs=1e-2)

    # alpha_2 > alpha_max = 19.47° -> alpha = 16.97° -> e = 61°
    assert np.isnan(sph_pixel_scale(62, np.radians(5), 3, 1))
    assert sph_pixel_scale(62, np.radians(4), 3, 1) == approx(np.radians(23.8), abs=1e-2)

    # alpha_1 = 0° -> +ifov/2 = 2.5° -> e = 7.52° -> ~ iFOV * alt
    assert sph_pixel_scale(7.5, np.radians(5), 3, 1) == approx(np.radians(10.1), abs=1e-2)

    # alpha_1 < 0
    assert sph_pixel_scale(20, np.radians(20), 3, 1) == approx(np.radians(49.1), abs=1e-2)

    # Invalid distance < target radius
    with raises(ValueError):
        _ = sph_pixel_scale(0, 0, 1, 3)

    # List input
    assert_array(
        sph_pixel_scale([30, 60, 62], np.radians(5), 3, 1),
        np.radians([12.19, 29.53, np.nan]),
        decimal=2,
    )

    assert_array(sph_pixel_scale(0, 0.1, [3, 4], 1), [0.2, 0.3], decimal=2)

    assert_array(sph_pixel_scale([0, 30], 0.1, [6, 3], 2), [0.4, 0.13], decimal=2)

    # Array size mismatch
    with raises(ValueError):
        _ = sph_pixel_scale([0, 30, 60], 0.1, [6, 3], 2)
