"""Test miscellaneous depreciation module."""

from planetary_coverage.misc import DepreciationHelper, depreciated


def test_depreciation_renamed(caplog):
    """Test depreciation renamed class decorator."""

    class New:
        """Dummy new object."""

        def __init__(self, a, b=0):
            self.a = a
            self.b = b

    @depreciated('1.2.3')
    class Old(New):
        """Dummy old object.

        Multi-line docstring.

        """

    old = Old(1, b=10)

    assert isinstance(old, Old)

    # Use parent behavior
    assert old.a == 1
    assert old.b == 10

    # Update object docstring with depreciation warning
    assert old.__doc__ == (
        '[Depreciated] Dummy old object.\n\n'
        '        Multi-line docstring.\n\n'
        '.. deprecated:: 1.2.3\n'
        '    `Old` is depreciated in favor of `New`.'
    )

    # Update class docstring with depreciation warning
    assert Old.__doc__ == (
        '[Depreciated] Dummy old object.\n\n'
        '        Multi-line docstring.\n\n'
        '.. deprecated:: 1.2.3\n'
        '    `Old` is depreciated in favor of `New`.'
    )

    # Warning message
    assert caplog.messages == ['`Old` is depreciated. Use `New` instead. ']


def test_depreciation_new_name(caplog):
    """Test depreciation new name, class decorator."""

    @depreciated('1.2.3', 'Bar')
    class Foo:
        """Dummy foo object."""

        def __str__(self):
            return 'foo'

    foo = Foo()

    assert isinstance(foo, Foo)

    # Old behavior is kept
    assert str(foo) == 'foo'

    # Update object docstring with depreciation warning
    assert foo.__doc__ == (
        '[Depreciated] Dummy foo object.\n\n'
        '.. deprecated:: 1.2.3\n'
        '    `Foo` is depreciated in favor of `Bar`.'
    )

    # Update class docstring with depreciation warning
    assert Foo.__doc__ == (
        '[Depreciated] Dummy foo object.\n\n'
        '.. deprecated:: 1.2.3\n'
        '    `Foo` is depreciated in favor of `Bar`.'
    )

    # Warning message
    assert caplog.messages == ['`Foo` is depreciated. Use `Bar` instead. ']


def test_depreciation_helper_dict(caplog):
    """Test depreciation replaced helper on a dict."""
    new = {'key': 'value'}

    old = DepreciationHelper('OLD', 'NEW', new)

    assert repr(old) == repr(new)

    assert 'key' in old
    assert 'missing' not in old

    assert old['key'] == 'value'

    assert len(old) == 1

    for el in old:
        assert el == 'key'

    assert len(caplog.messages) == 6

    assert set(caplog.messages) == {'`OLD` has been replaced by `NEW`. '}


def test_depreciation_helper_func(caplog):
    """Test depreciation replaced helper on a function."""

    def new(a, b=0):
        """Dummy new function."""
        return a + b

    old = DepreciationHelper('old', 'new', new)

    assert old(1) == 1
    assert old(1, 2) == 3
    assert old(1, b=3) == 4

    assert len(caplog.messages) == 3

    assert set(caplog.messages) == {'`old` has been replaced by `new`. '}


def test_depreciation_helper_class(caplog):
    """Test depreciation replaced helper on a class."""

    class New:
        """Dummy new class."""

        A = 2

        def __init__(self, a):
            self.a = a

        def __repr__(self):
            return 'new-function'

        def add(self, b):
            """Dummy add function."""
            return self.a + b

        @classmethod
        def attr(cls):
            """Dummy class method."""
            return cls.A

    old_depreciation = DepreciationHelper('Old', 'New', New)

    old = old_depreciation(1)  # log __call__

    assert repr(old) == 'new-function'

    assert old.a == 1
    assert old.add(2) == 3

    assert old.A == 2

    assert old_depreciation.attr() == 2  # log __getattr__

    assert len(caplog.messages) == 2

    assert set(caplog.messages) == {'`Old` has been replaced by `New`. '}
