"""Test dot environment files helper module."""

from os import environ
from pathlib import Path

from planetary_coverage.debug import debug_env
from planetary_coverage.misc.dotenv import (
    find_dotenv,
    getenv,
    parse_dotenv,
    print_kernels_dir,
)


DATA = (Path(__file__).parent / '..' / 'data').resolve()


def test_dotenv_getenv(caplog):
    """Test dotenv get environment variable."""
    # Remove .env file if exists
    env = Path().resolve() / '.env.test'
    env.unlink(missing_ok=True)

    # Enable debugger
    debug_env('info')
    caplog.clear()

    # Get variable globally
    assert getenv('HOME') == environ['HOME']

    # Get variable from default
    assert getenv('FOO') is None
    assert getenv('FOO', default='BAR') == 'BAR'

    # Create .env file
    env.write_text('FOO = BAZ', encoding='utf-8')
    assert getenv('FOO', dotenv='.env.test') == 'BAZ'

    # Disable .env file search
    assert getenv('FOO', dotenv=None) is None

    # Check debugger info messages
    assert caplog.messages == [
        f'Search for `HOME`. Found value: `{environ["HOME"]}` (globally)',
        'Search for `FOO`. Not found, use default: `None`',
        'Search for `FOO`. Not found, use default: `BAR`',
        f'Search for `FOO`. Found value: `BAZ` (in {env})',
        'Search for `FOO`. Not found, use default: `None`',
    ]

    # Remove file
    env.unlink()


def test_dotenv_finder(caplog):
    """Test dotenv finder."""
    # Remove .env file if exists
    env = Path().resolve() / '.env.test'
    env.unlink(missing_ok=True)

    # Enable debugger
    debug_env(True)
    caplog.clear()

    # Not present
    assert find_dotenv(fname='.env.test') is None

    # Create .env file
    env.write_text('', encoding='utf-8')
    assert str(find_dotenv(fname='.env.test')) == str(env)

    # Check debugger
    assert caplog.messages == [f'{env} found']

    # Remove file
    env.unlink()


def test_dotenv_parser():
    """Test dotenv parser."""
    data = parse_dotenv(DATA / '.env.sample')

    assert data['SIMPLE'] == 'xyz123'
    assert data['USER'] == 'admin'
    assert data['DOMAIN'] == 'example.org'
    assert data['MULTILINE'] == '\nlong text here,\ne.g. a private SSH key\n'

    # Comments
    assert data['SECRET_KEY'] == 'YOURSECRETKEYGOESHERE'  # noqa: S105
    assert data['SECRET_HASH'] == 'something-with-a-hash-#-this-is-not-a-comment'  # noqa: S105

    # Interpolation
    assert data['EMAIL'] == 'admin@example.org'
    assert data['DATABASE_URL'] == 'postgres://admin@localhost/my_database'
    assert data['CACHE_DIR'] == f'{environ["PWD"]}/cache'

    # Non-Interpolated
    assert data['PASSWORD'] == '!@G0${k}k'  # noqa: S105
    assert data['MESSAGE_TEMPLATE'] == '\n    Hello ${PERSON},\n\n    Nice to meet you!\n'

    assert len(data) == 11


def test_dotenv_print_kernels_dir(monkeypatch, capsys):
    """Test dotenv print env kernels dir list."""
    # Create `.env` file
    env = Path().resolve() / '.env.test'
    env.unlink(missing_ok=True)

    # Discard ENV variables KERNELS_*
    for key in environ:
        if key.startswith('KERNELS_'):
            monkeypatch.delenv(key)

    print_kernels_dir(dotenv='.env.test')

    assert capsys.readouterr().out == 'Kernels ENV variables:\n- None\n'

    # Set env variable globally
    monkeypatch.setenv('KERNELS_JUICE', '/data/kernels/JUICE')

    print_kernels_dir(dotenv='.env.test')

    assert capsys.readouterr().out == (
        'Kernels ENV variables:\n- KERNELS_JUICE: /data/kernels/JUICE\n'
    )

    # Add .env file content (override global variable)
    env.write_text('KERNELS_JUICE = ../kernels/JUICE', encoding='utf-8')

    print_kernels_dir(dotenv='.env.test')

    assert capsys.readouterr().out == (
        'Dotenv file found:\n'
        f'- {env}\n'
        '\n'
        'Kernels ENV variables:\n'
        '- KERNELS_JUICE: ../kernels/JUICE (.env)\n'
    )

    # Remove `.env` file
    env.unlink()
