"""Test ground projection axes module."""

import numpy as np
from numpy.testing import assert_array_almost_equal as assert_array

import matplotlib.pyplot as plt
from matplotlib.collections import LineCollection, PatchCollection
from matplotlib.colorbar import Colorbar
from matplotlib.lines import Line2D

from pytest import fixture, raises

from planetary_coverage.maps import GANYMEDE
from planetary_coverage.projections.axes import ProjectionMapTargetError
from planetary_coverage.rois import CallistoROIs, GanymedeROIs
from planetary_coverage.ticks import km_ticks, m_pix_ticks


@fixture
def traj():
    """Default trajectory."""

    class Trajectory:
        """Dummy trajectory."""

        lonlat = [(0, 10, 20), (0, 5, 10)]
        alt = [0, 100, 200]
        inc = [10, 30, 45]
        pixel_scale = [0.1, 0.3, 0.45]

    return Trajectory()


def test_axes_attr():
    """Test projection axes attributes."""
    fig = plt.figure()
    ax = fig.add_subplot(projection=GANYMEDE)

    assert ax.proj.PROJ4 == 'eqc'
    assert ax.bg == GANYMEDE.fname
    assert ax.bg_extent == [0, 360, -90, 90]
    assert ax.target == 'Ganymede'

    plt.close()


def test_axes_plot(traj):
    """Test projection axes plot."""
    fig = plt.figure()
    ax = fig.add_subplot(projection=GANYMEDE)

    # Trajectory-like (with `lonlat` attribute)
    ax.plot(traj)

    # Lon and Lat array
    lonlat = [(0, 10, 20), (0, 5, 10)]
    ax.plot(lonlat)

    # Lon and Lat individual variables
    lon, lat = (0, 10, 20), (0, 5, 10)
    ax.plot(lon, lat)

    # With dots
    ax.plot(lon, lat, 'ro')
    ax.plot(lon, lat, '.', color='red')

    # With lines
    ax.plot(lon, lat, 'r--')
    ax.plot(lon, lat, '--', color='red')

    # With 1 single point
    ax.plot(lon[0], lat[0])
    ax.plot(lon[0], lat[0], '.', color='red')

    plt.close()


def test_axes_scatter(traj):
    """Test projection axes scatter."""
    fig = plt.figure()
    ax = fig.add_subplot(projection=GANYMEDE)

    # Trajectory-like (with `lonlat` attribute)
    ax.scatter(traj)
    ax.scatter(traj, c='tab:red')
    ax.scatter(traj, c=traj.inc, vmin=0, vmax=90, cmap='Reds')
    ax.scatter(traj, s=10, edgecolor='tab:red', facecolor='none')

    # Trajectory-like with attribute property
    ax.scatter(traj, 'alt')
    ax.scatter(traj, 'inc', vmin=0, vmax=90, cmap='Reds', cbar=True)

    # Lon and Lat individual variables
    lon, lat = (0, 10, 20), (0, 5, 10)
    ax.scatter(lon, lat)
    ax.scatter(lon, lat, c='tab:blue')
    ax.scatter(lon, lat, c=['C0', 'C1', 'C2'])
    ax.scatter(lon, lat, s=1)
    ax.scatter(lon, lat, s=[1, 2, 3])
    ax.scatter(lon, lat, c='tab:blue', s=[1, 2, 3])
    ax.scatter(lon, lat, c=['C0', 'C1', 'C2'], s=1)

    # With 1 single point
    ax.scatter(lon[0], lat[0])
    ax.scatter(lon[1], lat[1], c='red')
    ax.scatter(lon[1], lat[1], s=1)

    plt.close()


def test_axes_color_lines(traj):
    """Test projection axes color lines."""
    fig = plt.figure()
    ax = fig.add_subplot(projection=GANYMEDE)

    # Color lines with colorbar
    cbar = ax.plot(traj, 'alt')

    assert isinstance(cbar, Colorbar)
    assert cbar.ax.get_xlabel() == 'Altitude'
    assert cbar.cmap.name == 'turbo_r'
    assert cbar.vmin == 0
    assert cbar.vmax == 200
    assert_array(cbar.get_ticks(), [0, 50, 100, 150, 200])

    assert cbar.formatter(0) == '0 km'
    assert cbar.formatter(200) == '200 km'

    # Override vmin/vmax and label
    cbar = ax.plot(traj, 'inc', cmap='Reds', vmin=0, vmax=90, label='foo')

    assert isinstance(cbar, Colorbar)
    assert cbar.ax.get_xlabel() == 'foo'
    assert cbar.cmap.name == 'Reds'
    assert cbar.vmin == 0
    assert cbar.vmax == 90
    assert_array(cbar.get_ticks(), [0, 20, 40, 60, 80, 100])

    assert cbar.formatter(0) == '0°'
    assert cbar.formatter(90) == '90°'

    # Override format
    cbar = ax.plot(traj, 'pixel_scale', vmin=0, vmax=1, fmt=m_pix_ticks)

    assert isinstance(cbar, Colorbar)
    assert cbar.ax.get_xlabel() == 'Pixel scale'
    assert cbar.vmin == 0
    assert cbar.vmax == 1
    assert_array(cbar.get_ticks(), [0, 0.2, 0.4, 0.6, 0.8, 1])

    assert cbar.formatter(0) == '0 m/pix'
    assert cbar.formatter(1) == '1,000 m/pix'

    # Color lines without colorbar
    lines = ax.plot(traj, 'inc', cbar=False)
    assert isinstance(lines, LineCollection)

    # Valid matplotlib property
    pts = ax.plot(traj, 'ro')

    assert isinstance(pts, list)
    assert len(pts) == 1

    pt = pts[0]

    assert isinstance(pt, Line2D)
    assert pt.get_color() == 'r'
    assert pt.get_marker() == 'o'
    assert pt.get_linestyle() == 'None'
    assert pt.get_linewidth() == 1.5

    plt.close()

    # Invalid traj property and matplotlib property
    fig = plt.figure()
    ax = fig.add_subplot(projection=GANYMEDE)

    err_msg = (
        'The second argument `foo` must be a `Trajectory` property '
        r'or a valid matplotlib format string \(eg. `ro`\).'
    )

    with raises(ValueError, match=err_msg):
        ax.plot(traj, 'foo')


def test_axes_text():
    """Test projection text on the map."""
    fig = plt.figure()
    ax = fig.add_subplot(projection=GANYMEDE)

    # Enable clip by default
    txt = ax.text(-10, 10, 'foo')

    assert_array(txt.get_position(), (350, 10))
    assert txt.get_text() == 'foo'
    assert txt.get_clip_on()

    # Disable text clipping
    txt = ax.text(0, 0, 'bar', clip_on=False)

    assert_array(txt.get_position(), (0, 0))
    assert not txt.get_clip_on()


def test_axes_colorbar():
    """Test projection standalone colorbar."""
    fig = plt.figure()
    ax = fig.add_subplot(projection=GANYMEDE)

    # No initial colorbar
    with raises(ValueError, match='No parent colorbar found.'):
        ax.twin_colorbar()

    # Add a standalone horizontal colorbar
    cbar = ax.colorbar(vmin=0, vmax=90, cmap='Reds', label='inc')

    assert isinstance(cbar, Colorbar)
    assert cbar.ax.get_xlabel() == 'Incidence angle'
    assert cbar.cmap.name == 'Reds'
    assert cbar.vmin == 0
    assert cbar.vmax == 90
    assert_array(cbar.get_ticks(), [0, 20, 40, 60, 80, 100])

    assert cbar.formatter(0) == '0°'
    assert cbar.formatter(90) == '90°'

    # Check twin colorbar
    cax = ax.twin_colorbar(
        label='Twin label', format=2 * km_ticks, ticks=[0, 20, 40, 80, 100]
    )

    assert cax.get_xlabel() == 'Twin label'
    assert cax.get_xlim() == (0, 90)
    assert_array(cax.get_xticks(), [0, 10, 20, 40, 50])

    fmt = cax.xaxis.get_major_formatter()
    assert fmt(0) == '0 km'
    assert fmt(100) == '200 km'

    ax.colorbar(vmin=0, vmax=1, orientation='vertical')
    cax = ax.twin_colorbar(
        label='Vertical twin', format=3 * km_ticks, ticks=[0, 0.3, 0.9]
    )

    assert cax.get_ylabel() == 'Vertical twin'
    assert cax.get_ylim() == (0, 1)

    assert_array(cax.get_yticks(), [0, 0.1, 0.3])

    fmt = cax.yaxis.get_major_formatter()
    assert fmt(0) == '0 km'
    assert fmt(100) == '300 km'

    plt.close()


def test_axes_patch_coll():
    """Test projection axes patch collection plot."""
    fig = plt.figure()
    ax = fig.add_subplot(projection=GANYMEDE)

    ax.add_path(GanymedeROIs[2, 0, 4].path)

    ax.add_patch(GanymedeROIs[2, 0, 4])
    ax.add_patch(GanymedeROIs[2, 0, 4](facecolor='blue'))

    ax.add_collection(GanymedeROIs)
    ax.add_collection(GanymedeROIs(facecolors='blue'))

    plt.close()


def test_axes_patch_coll_legend():
    """Test projection axes patch collection legend."""
    fig = plt.figure()
    ax = fig.add_subplot(projection=GANYMEDE)

    ax.plot([], [], label='Foo')

    handles, labels = ax.get_legend_handles_labels()
    assert len(handles) == len(labels) == 1
    assert labels[0] == 'Foo'

    leg = ax.legend()
    handler_map = leg.get_legend_handler_map()

    assert PatchCollection in handler_map

    plt.close()


def test_axes_proj_target_error(traj):
    """Test projection target errors."""
    fig = plt.figure()
    ax = fig.add_subplot(projection=GANYMEDE)

    traj.target = 'Callisto'
    with raises(ProjectionMapTargetError):
        _ = ax.plot(traj)

    with raises(ProjectionMapTargetError):
        _ = ax.add_patch(CallistoROIs[1, 0, 4])

    with raises(ProjectionMapTargetError):
        _ = ax.add_patch(CallistoROIs[1, 0, 4](facecolor='blue'))

    with raises(ProjectionMapTargetError):
        _ = ax.add_collection(CallistoROIs)

    with raises(ProjectionMapTargetError):
        _ = ax.add_collection(CallistoROIs(facecolors='blue'))

    plt.close()


def test_axes_proj_ticks():
    """Test projection axes ticks."""
    fig = plt.figure()
    ax = fig.add_subplot(projection=GANYMEDE)

    ax.set_lon_ticks('west')
    ax.set_lon_ticks('east', secondary=True)
    ax.set_lat_ticks(secondary=True)

    def assert_ticks(tick_labels, values):
        """Extract position and text from tick labels."""
        for label, (x, y, t) in zip(tick_labels, values):
            assert label.get_position() == (x, y)
            assert label.get_text() == t

    # Bottom primary -> West longitudes ticks
    assert_ticks(
        ax.get_xticklabels(),
        [
            (0, 0, '0°'),
            (30, 0, '330°W'),
            (60, 0, '300°W'),
            (90, 0, '270°W'),
            (120, 0, '240°W'),
            (150, 0, '210°W'),
            (180, 0, '180°'),
            (210, 0, '150°W'),
            (240, 0, '120°W'),
            (270, 0, '90°W'),
            (300, 0, '60°W'),
            (330, 0, '30°W'),
            (360, 0, '0°'),
        ],
    )

    # Left primary -> Latitudes ticks
    assert_ticks(
        ax.get_yticklabels(),
        [
            (0, -90, 'S.P.'),
            (0, -60, '60°S'),
            (0, -30, '30°S'),
            (0, 0, 'Eq.'),
            (0, 30, '30°N'),
            (0, 60, '60°N'),
            (0, 90, 'N.P.'),
        ],
    )

    # Top secondary -> East longitudes ticks
    assert_ticks(
        ax.child_axes[0].get_xticklabels(),
        [
            (0, 1, '0°'),
            (30, 1, '30°E'),
            (60, 1, '60°E'),
            (90, 1, '90°E'),
            (120, 1, '120°E'),
            (150, 1, '150°E'),
            (180, 1, '180°'),
            (210, 1, '210°E'),
            (240, 1, '240°E'),
            (270, 1, '270°E'),
            (300, 1, '300°E'),
            (330, 1, '330°E'),
            (360, 1, '0°'),
        ],
    )

    # Right secondary -> Latitudes ticks
    assert_ticks(
        ax.child_axes[1].get_yticklabels(),
        [
            (1, -90, 'S.P.'),
            (1, -60, '60°S'),
            (1, -30, '30°S'),
            (1, 0, 'Eq.'),
            (1, 30, '30°N'),
            (1, 60, '60°N'),
            (1, 90, 'N.P.'),
        ],
    )

    with raises(KeyError):
        ax.set_lon_ticks('foo')

    plt.close()


def test_axes_proj_limits():
    """Test projection axes limits."""
    fig = plt.figure()
    ax = fig.add_subplot(projection=GANYMEDE)

    # Check grid xticks
    ax.set_xlim(0, 360)
    assert ax.get_xlim() == (0, 360)
    assert_array(ax.get_xticks(), np.arange(0, 361, 30))

    ax.set_xlim(0, 180)
    assert ax.get_xlim() == (0, 180)
    assert_array(ax.get_xticks(), np.arange(0, 361, 15))

    ax.set_xlim(10, 100)
    assert ax.get_xlim() == (10, 100)
    assert_array(ax.get_xticks(), np.arange(0, 361, 10))

    ax.set_xlim(0, 45)
    assert ax.get_xlim() == (0, 45)
    assert_array(ax.get_xticks(), np.arange(0, 361, 5))

    ax.set_xlim(0, 15)
    assert ax.get_xlim() == (0, 15)
    assert_array(ax.get_xticks(), np.arange(0, 361, 2))

    ax.set_xlim(0, 5)
    assert ax.get_xlim() == (0, 5)
    assert_array(ax.get_xticks(), np.arange(0, 361, 1))

    # If None provided, the ticks grid is unchanged
    ax.set_xlim(None, 360)
    assert ax.get_xlim() == (0, 360)
    assert_array(ax.get_xticks(), np.arange(0, 361, 1))

    ax.set_xlim(180, None)
    assert ax.get_xlim() == (180, 360)
    assert_array(ax.get_xticks(), np.arange(0, 361, 1))

    # Check grid yticks
    ax.set_ylim(-90, 90)
    assert ax.get_ylim() == (-90, 90)
    assert_array(ax.get_yticks(), np.arange(-90, 91, 30))

    ax.set_ylim(-45, 45)
    assert ax.get_ylim() == (-45, 45)
    assert_array(ax.get_yticks(), np.arange(-90, 91, 10))

    ax.set_ylim(0, 45)
    assert ax.get_ylim() == (0, 45)
    assert_array(ax.get_yticks(), np.arange(-90, 91, 5))

    ax.set_ylim(0, 15)
    assert ax.get_ylim() == (0, 15)
    assert_array(ax.get_yticks(), np.arange(-90, 91, 2))

    ax.set_ylim(0, 5)
    assert ax.get_ylim() == (0, 5)
    assert_array(ax.get_yticks(), np.arange(-90, 91, 1))

    # If None provided, the ticks grid is unchanged
    ax.set_ylim(None, 90)
    assert ax.get_ylim() == (0, 90)
    assert_array(ax.get_yticks(), np.arange(-90, 91, 1))

    ax.set_ylim(-90, None)
    assert ax.get_ylim() == (-90, 90)
    assert_array(ax.get_yticks(), np.arange(-90, 91, 1))

    plt.close()


def test_axes_proj_view_setter(traj):
    """Test projection axes view setter."""
    fig = plt.figure()
    ax = fig.add_subplot(projection=GANYMEDE)

    # Zoom on trajectory
    ax.set_view(traj)
    assert_array(ax.get_xlim(), (0, 21))  # Clipped at 0 (instead of -1)
    assert_array(ax.get_ylim(), (-1, 11))

    ax.set_view(traj, margin=0)
    assert_array(ax.get_xlim(), (0, 20))
    assert_array(ax.get_ylim(), (0, 10))

    # Zoom on ROI
    ax.set_view(GanymedeROIs[2, 0, 4], margin=0)
    assert_array(ax.get_xlim(), (167, 173))
    assert_array(ax.get_ylim(), (-34, -28))

    # Wrong projection target
    with raises(ProjectionMapTargetError):
        ax.set_view(CallistoROIs[1, 0, 4], margin=0)

    # Explicit coordinates
    ax.set_view(10, 20, 30, 40, margin=0)
    assert_array(ax.get_xlim(), (10, 20))
    assert_array(ax.get_ylim(), (30, 40))

    ax.set_view([10, 20], [30, 40])  # 5 % margin (default)
    assert_array(ax.get_xlim(), (9.5, 20.5))
    assert_array(ax.get_ylim(), (29.5, 40.5))

    ax.set_view([10, 20, 30, 40], margin=10)
    assert_array(ax.get_xlim(), (9, 21))
    assert_array(ax.get_ylim(), (29, 41))

    # Clip the limits on projection extent
    ax.set_view(0, 360, -100, 100, margin=100)
    assert_array(ax.get_xlim(), (0, 360))
    assert_array(ax.get_ylim(), (-90, 90))

    # Use negative values
    ax.set_view(370, -10, -100, 100, margin=0)
    assert_array(ax.get_xlim(), (10, 350))
    assert_array(ax.get_ylim(), (-90, 90))

    # Invalid input
    with raises(ValueError):
        ax.set_view(1, 2, 3)

    plt.close()
