"""Test stereographic projection module."""

import numpy as np
from numpy.testing import assert_array_almost_equal as assert_array

from matplotlib.collections import PatchCollection, PathCollection
from matplotlib.colors import Normalize
from matplotlib.patches import PathPatch
from matplotlib.path import Path

from pytest import approx, fixture, raises

from planetary_coverage.projections.proj import GroundProjection, Projection


@fixture
def proj():
    """Default abstract projection."""
    return Projection()


@fixture
def xy_proj(monkeypatch, proj):
    """Default projection with `xy` invariant conversion."""

    def xy(_, lon_e, lat, value=None):
        if np.ndim(lon_e) > 0 and np.ndim(lat) == 0:
            lat = np.broadcast_to(lat, np.shape(lon_e))

        if np.ndim(lon_e) == 0 and np.ndim(lat) > 0:
            lon_e = np.broadcast_to(lon_e, np.shape(lat))

        return (lon_e, lat) if value is None else (lon_e, lat, value)

    monkeypatch.setattr(Projection, 'xy', xy)
    return proj


@fixture
def ground_proj():
    """Default ground projection."""
    return GroundProjection(lon_e_0=30)


def test_projection_attr(proj):
    """Test default projection attributes."""
    assert str(proj) == 'Projection'

    assert proj.EPSILON == 1e-10
    assert proj.PROJ4 is None


def test_projection_err(proj):
    """Test default projection errors."""
    # Path/Patch/Collection
    path = Path([(None, None)])
    patch = PathPatch(path)
    coll = PatchCollection([patch])

    with raises(NotImplementedError):
        _ = proj(path)

    with raises(NotImplementedError):
        _ = proj(patch)

    with raises(NotImplementedError):
        _ = proj(coll)

    # (Lon_e, Lat)
    with raises(NotImplementedError):
        _ = proj(0, 0)

    # (X, Y)
    with raises(NotImplementedError):
        _ = proj(0, 0, invert=True)

    # (X, Y, Z) not ortho
    with raises(ValueError):
        _ = proj(0, 0, 0, invert=True)

    # Extent
    with raises(NotImplementedError):
        _ = proj.extent


def test_projection_xy_mock(xy_proj):
    """Test mock invariant conversion."""
    lon_e, lat = xy_proj(1, 2)
    assert lon_e == 1
    assert lat == 2

    # 1D array
    lons_e, lats = xy_proj(1, [2, 3])
    assert_array(lons_e, [1, 1])
    assert_array(lats, [2, 3])

    lons_e, lats = xy_proj([1, 2], 3)
    assert_array(lons_e, [1, 2])
    assert_array(lats, [3, 3])

    lons_e, lats = xy_proj([1, 2], [3, 4])
    assert_array(lons_e, [1, 2])
    assert_array(lats, [3, 4])

    # 2D array
    lons_e, lats = xy_proj([[1], [2]], [[3], [4]])
    assert_array(lons_e, [[1], [2]])
    assert_array(lats, [[3], [4]])

    # With values
    lon_e, lat, value = xy_proj(1, 2, value=3)
    assert lon_e == 1
    assert lat == 2
    assert value == 3

    # 1D array
    lons_e, lats, values = xy_proj([1, 2], [3, 4], value=[5, 6])
    assert_array(lons_e, [1, 2])
    assert_array(lats, [3, 4])
    assert_array(values, [5, 6])

    # 2D array
    lons_e, lats, values = xy_proj([[1], [2]], [[3], [4]], value=[[5], [6]])
    assert_array(lons_e, [[1], [2]])
    assert_array(lats, [[3], [4]])
    assert_array(values, [[5], [6]])


def test_projection_xy_plot(xy_proj):
    """Test default projection vertices and codes."""
    x, y = xy_proj.xy_plot([1, 2], [3, 4])

    assert x == [1, 2]
    assert y == [3, 4]

    x, y, v = xy_proj.xy_plot([1, 2], [3, 4], values=[5, 6])

    assert x == [1, 2]
    assert y == [3, 4]
    assert v == [5, 6]


def test_projection_vertices(xy_proj):
    """Test default projection vertices and codes."""
    # Closed polygon with codes (invariant)
    v, c = xy_proj.vc(
        Path([(1, 2), (3, 4), (1, 2)], [Path.MOVETO, Path.LINETO, Path.CLOSEPOLY])
    )
    assert len(v) == len(c) == 3
    assert_array(v, [(1, 2), (3, 4), (1, 2)])
    assert_array(c, [Path.MOVETO, Path.LINETO, Path.CLOSEPOLY])

    # Closed polygon without codes (add codes when missing)
    v, c = xy_proj.vc(Path([(1, 2), (3, 4), (1, 2)]))
    assert len(v) == len(c) == 3
    assert_array(v, [(1, 2), (3, 4), (1, 2)])
    assert_array(c, [Path.MOVETO, Path.LINETO, Path.CLOSEPOLY])

    # Open polygon without codes (close the polygon + add codes when missing)
    v, c = xy_proj.vc(Path([(1, 2), (3, 4)]))
    assert len(v) == len(c) == 3
    assert_array(v, [(1, 2), (3, 4), (1, 2)])
    assert_array(c, [Path.MOVETO, Path.LINETO, Path.CLOSEPOLY])

    # Open polygon with wrong closed code codes (close polygon and fix codes when missing)
    v, c = xy_proj.vc(Path([(1, 2), (3, 4)], [Path.MOVETO, Path.LINETO]))
    assert len(v) == len(c) == 3
    assert_array(v, [(1, 2), (3, 4), (1, 2)])
    assert_array(c, [Path.MOVETO, Path.LINETO, Path.CLOSEPOLY])


def test_projection_xy_path_and_patch(xy_proj):
    """Test default projection path and patch."""
    # Valid Path (not closed)
    _path = Path([(1, 2), (3, 4)])

    path = xy_proj(_path)
    assert isinstance(path, Path)
    assert_array(path.vertices, [(1, 2), (3, 4), (1, 2)])
    assert_array(path.codes, [Path.MOVETO, Path.LINETO, Path.CLOSEPOLY])

    # Null Path
    assert xy_proj.xy_path(None) is None

    # Patch
    _patch = PathPatch(_path)

    patch = xy_proj.xy_patch(_patch)
    assert isinstance(patch, PathPatch)
    assert_array(patch.get_path().vertices, [(1, 2), (3, 4), (1, 2)])
    assert_array(patch.get_path().codes, [Path.MOVETO, Path.LINETO, Path.CLOSEPOLY])

    # Null patch
    _patch_null = PathPatch(
        None,
        facecolor='r',
        edgecolor='b',
        alpha=0.5,
        lw=3,
        ls='--',
        zorder=10,
        label='Foo',
    )
    patch = xy_proj.xy_patch(_patch_null)

    assert patch.get_path() is None
    assert patch.get_alpha() == 0.5
    assert patch.get_edgecolor() == (0.0, 0.0, 1.0, 0.5)
    assert patch.get_facecolor() == (1.0, 0.0, 0.0, 0.5)
    assert patch.get_linewidth() == 3
    assert patch.get_linestyle() == '--'
    assert patch.get_zorder() == 10
    assert patch.get_label() == 'Foo'


def test_projection_path_collection(xy_proj):
    """Test default projection paths collection."""
    _path = Path([(1, 2), (3, 4)])

    # Simple path collection (no size)
    _coll = PathCollection((_path,))

    coll = xy_proj.xy_collection(_coll)

    assert isinstance(coll, PathCollection)

    paths = coll.get_paths()

    assert len(paths) == 1

    assert_array(paths[0].vertices, [(1, 2), (3, 4), (1, 2)])
    assert_array(paths[0].codes, [Path.MOVETO, Path.LINETO, Path.CLOSEPOLY])

    # Paths with single size
    _coll = PathCollection((_path, _path), sizes=[1])

    coll = xy_proj.xy_collection(_coll)
    paths = coll.get_paths()

    assert len(paths) == 2

    for path in paths:
        assert_array(path.vertices, [(0.5, 1), (1.5, 2), (0.5, 1)])  # 1/2 values
        assert_array(path.codes, [Path.MOVETO, Path.LINETO, Path.CLOSEPOLY])

    # Paths with multiple sizes (same size as paths)
    _coll = PathCollection((_path, _path, _path), sizes=[1, 4, 16])

    coll = xy_proj.xy_collection(_coll)
    paths = coll.get_paths()

    assert len(paths) == 3

    assert_array(paths[0].vertices, [(0.5, 1), (1.5, 2), (0.5, 1)])  # size 1: 1/2x values
    assert_array(paths[1].vertices, [(1, 2), (3, 4), (1, 2)])  # size 4: 1x values
    assert_array(paths[2].vertices, [(2, 4), (6, 8), (2, 4)])  # size 16: 2x values

    for path in paths:
        assert_array(path.codes, [Path.MOVETO, Path.LINETO, Path.CLOSEPOLY])

    # Paths and sizes with different shapes
    with raises(ValueError):
        _coll = PathCollection((_path, _path, _path), sizes=[1, 4])
        _ = xy_proj.xy_collection(_coll)

    # Paths with single offset
    _coll = PathCollection((_path, _path))
    _coll.set_offsets([10, 20])

    coll = xy_proj.xy_collection(_coll)
    paths = coll.get_paths()

    assert len(paths) == 2

    for path in paths:
        assert_array(path.vertices, [(11, 22), (13, 24), (11, 22)])
        assert_array(path.codes, [Path.MOVETO, Path.LINETO, Path.CLOSEPOLY])

    # Single paths with multiple offsets
    _coll = PathCollection((_path,))
    _coll.set_offsets([(10, 20), (30, 40), (50, 60)])

    coll = xy_proj.xy_collection(_coll)
    paths = coll.get_paths()

    assert len(paths) == 3

    assert_array(paths[0].vertices, [(11, 22), (13, 24), (11, 22)])
    assert_array(paths[1].vertices, [(31, 42), (33, 44), (31, 42)])
    assert_array(paths[2].vertices, [(51, 62), (53, 64), (51, 62)])

    for path in paths:
        assert_array(path.codes, [Path.MOVETO, Path.LINETO, Path.CLOSEPOLY])

    # Paths and sizes with different shapes
    with raises(ValueError):
        _coll = PathCollection((_path, _path))
        _coll.set_offsets([(10, 20), (30, 40), (50, 60)])
        _ = xy_proj.xy_collection(_coll)

    # Paths with data array (discard color edges)
    _coll = PathCollection((_path, _path, _path))
    _coll.set_edgecolor('red')
    _coll.set_cmap('turbo')
    _coll.set_norm(Normalize(0, 3))
    _coll.set_array([1, 2, 4])

    coll = xy_proj.xy_collection(_coll)

    assert coll.get_cmap().name == 'turbo'
    assert coll.norm.vmin == 0
    assert coll.norm.vmax == 3
    assert_array(coll.get_array(), [1, 2, 4])
    assert_array(coll.get_facecolors(), [[0.121, 0.466, 0.705, 1]], decimal=3)
    assert_array(coll.get_edgecolors(), np.empty((0, 4)))

    # With None Path
    _coll = PathCollection((None,))
    coll = xy_proj.xy_collection(_coll)

    assert coll.get_paths() == [None]


def test_projection_patch_collection(xy_proj):
    """Test default projection patches collection."""
    # Valid Path (not closed)
    _patch = PathPatch(Path([(1, 2), (3, 4)]))
    _patch_null = PathPatch(
        None,
        facecolor='r',
        edgecolor='b',
        alpha=0.5,
        lw=3,
        ls='--',
        zorder=10,
        label='Foo',
    )

    # Patch Collection
    _coll = PatchCollection([_patch])

    coll = xy_proj.xy_collection(_coll)
    assert isinstance(coll, PatchCollection)
    assert_array(coll.get_paths()[0].vertices, [(1, 2), (3, 4), (1, 2)])
    assert_array(coll.get_paths()[0].codes, [Path.MOVETO, Path.LINETO, Path.CLOSEPOLY])

    # Null Collection
    coll = xy_proj.xy_collection(
        PatchCollection(
            [_patch_null, _patch],
            facecolor='r',
            edgecolor='b',
            alpha=0.5,
            lw=3,
            ls='--',
            zorder=10,
            label='Foo',
        )
    )

    assert coll.get_paths()[0] is None
    assert coll.get_paths()[1] is not None

    assert_array(coll.get_paths()[1].vertices, [(1, 2), (3, 4), (1, 2)])
    assert_array(coll.get_paths()[1].codes, [Path.MOVETO, Path.LINETO, Path.CLOSEPOLY])

    assert coll.get_alpha() == 0.5
    assert_array(coll.get_edgecolor(), [(0.0, 0.0, 1.0, 0.5)])
    assert_array(coll.get_facecolor(), [(1.0, 0.0, 0.0, 0.5)])

    assert coll.get_linewidths() == 3
    assert_array(coll.get_linestyles()[0][1], [33.3, 14.4])
    assert coll.get_zorder() == 10
    assert coll.get_label() == 'Foo'


def test_projection_meridians(xy_proj):
    """Test projection meridians."""
    meridians = xy_proj.meridians()

    assert meridians.shape == (2, 65, 13)

    assert_array(
        meridians[0, 0], [0, 30, 60, 90, 120, 150, 180, 210, 240, 270, 300, 330, 360]
    )
    assert_array(meridians[1, 1], 13 * [-77.5])

    assert meridians[1, 0, 0] == -80
    assert meridians[1, -1, 0] == 80

    # Set lons list with exclusion and extended latitude range.
    meridians = xy_proj.meridians(
        lons_e=[0, 150, 180, 360], exclude=150, lat_min=-90, lat_max=90, nlat=7
    )

    assert meridians.shape == (2, 7, 3)
    assert_array(meridians[0, 0], [0, 180, 360])
    assert_array(meridians[1, :, 0], [-90, -60, -30, 0, 30, 60, 90])

    # Set a single east longitude.
    meridians = xy_proj.meridians(lons_e=150, nlat=3)

    assert meridians.shape == (2, 3, 1)
    assert_array(meridians[0, 0], [150])
    assert_array(meridians[1, :, 0], [-80, 0, 80])


def test_projection_parallels(xy_proj):
    """Test projection parallels."""
    parallels = xy_proj.parallels()

    assert parallels.shape == (2, 73, 17)

    assert_array(parallels[0, 1], 17 * [5])
    assert_array(
        parallels[1, 0],
        [-80, -70, -60, -50, -40, -30, -20, -10, 0, 10, 20, 30, 40, 50, 60, 70, 80],
    )

    assert parallels[0, 0, 0] == 0
    assert parallels[0, -1, 0] == 360

    # Set lons list with exclusion and extended latitude range.
    parallels = xy_proj.parallels(
        lats=[-90, 0, 30, 90], exclude=30, lon_e_min=-180, lon_e_max=180, nlon_e=7
    )

    assert parallels.shape == (2, 7, 3)
    assert_array(parallels[0, :, 0], [-180, -120, -60, 0, 60, 120, 180])
    assert_array(parallels[1, 0], [-90, 0, 90])

    # Set a single east longitude.
    parallels = xy_proj.parallels(lats=30, nlon_e=3)

    assert parallels.shape == (2, 3, 1)
    assert_array(parallels[0, :, 0], [0, 180, 360])
    assert_array(parallels[1, 0], [30])


def test_ground_projection_attr(ground_proj):
    """Test ground projection attributes."""
    assert str(ground_proj) == 'GroundProjection'

    assert ground_proj.DEFAULT_RADIUS_KM == 1e-3

    assert ground_proj.lat_0 == 0
    assert ground_proj.lon_e_0 == 30

    assert ground_proj.clat0 == 1
    assert ground_proj.slat0 == 0
    assert ground_proj.clon0_e == approx(np.sqrt(3) / 2)
    assert ground_proj.slon0_e == approx(0.5)

    assert ground_proj.target == 'Undefined'
    assert ground_proj.radius == 1e-3  # Default radius in km
    assert ground_proj.r == 1  # Default radius in km

    # Attributes on a known body
    proj = GroundProjection(target='Ganymede', radius=2631.2)
    assert proj.target == 'Ganymede'
    assert proj.radius == approx(2631.2)  # [km]
    assert proj.r == 2_631_200  # [m]

    assert proj.proj4 == (
        '+proj=None +lat_0=0 +lon_0=0 +k=1 +x_0=0 +y_0=0 '
        '+a=2631200.0 +b=2631200.0 +units=m +no_defs'
    )

    assert proj.wkt == (
        'PROJCS["PROJCS_Ganymede_GroundProjection",'
        'GEOGCS["GCS_Ganymede",'
        'DATUM["D_Ganymede",'
        'SPHEROID["Ganymede_Mean_Sphere", 2631200, 0]],'
        'PRIMEM["Greenwich",0],'
        'UNIT["Degree",0.017453292519943295]],'
        'PROJECTION["GroundProjection"],'
        'PARAMETER["false_easting", 0],'
        'PARAMETER["false_northing", 0],'
        'PARAMETER["scale_factor", 1],'
        'PARAMETER["central_meridian", 0],'
        'PARAMETER["latitude_of_origin", 0],'
        'UNIT["Meter", 1]]'
    )
