"""Test GeoJson ROI module."""

import pathlib

from matplotlib.path import Path

from pytest import fixture, raises

from planetary_coverage.rois import GeoJsonROI


DATA = pathlib.Path(__file__).parent / '..' / 'data'


@fixture
def roi():
    """Default test GeoJson ROI."""
    return GeoJsonROI('roi-key', DATA / 'test.geojson')


def test_roi_geojson(roi):
    """Test GeoJson ROI."""
    assert roi.key == 'roi-key'

    assert not hasattr(roi, 'lons_e')
    assert not hasattr(roi, 'lats')

    assert roi.color == 'blue'

    assert roi.vertices == [
        (100.0, 0.0),
        (101.0, 0.0),
        (101.0, 1.0),
        (100.0, 1.0),
        (100.0, 0.0),
    ]
    assert roi.codes == [Path.MOVETO] + 3 * [Path.LINETO] + [Path.CLOSEPOLY]


def test_toi_geojson_loader(roi):
    """Test GeoJson ROI loader."""
    # Load single feature
    roi.load_geojson({
        'type': 'Feature',
        'properties': {'color': 'blue'},
        'geometry': {
            'type': 'Polygon',
            'coordinates': [
                [[100.1, 0.0], [101.1, 0.0], [101.1, 1.0], [100.1, 1.0], [100.1, 0.0]]
            ],
        },
    })

    assert roi.vertices == [
        (100.1, 0.0),
        (101.1, 0.0),
        (101.1, 1.0),
        (100.1, 1.0),
        (100.1, 0.0),
    ]

    assert roi.codes == [Path.MOVETO] + 3 * [Path.LINETO] + [Path.CLOSEPOLY]

    # Load single polygon
    roi.load_geojson({
        'type': 'Polygon',
        'coordinates': [
            [[100.2, 0.0], [101.2, 0.0], [101.2, 1.0], [100.2, 1.0], [100.2, 0.0]]
        ],
    })

    assert roi.vertices == [
        (100.2, 0.0),
        (101.2, 0.0),
        (101.2, 1.0),
        (100.2, 1.0),
        (100.2, 0.0),
    ]

    assert roi.codes == [Path.MOVETO] + 3 * [Path.LINETO] + [Path.CLOSEPOLY]

    # Load single polygon with a hole
    roi.load_geojson({
        'type': 'Polygon',
        'coordinates': [
            [[0, 0], [4, 0], [4, 4], [0, 4], [0, 0]],
            # Note: Holes must be counter clockwise
            [[1, 1], [1, 2], [2, 2], [2, 1], [1, 1]],
        ],
    })

    assert roi.vertices == [
        (0, 0),
        (4, 0),
        (4, 4),
        (0, 4),
        (0, 0),
        (1, 1),
        (1, 2),
        (2, 2),
        (2, 1),
        (1, 1),
    ]

    assert roi.codes == 2 * ([Path.MOVETO] + 3 * [Path.LINETO] + [Path.CLOSEPOLY])


def test_toi_geojson_loader_err(roi):
    """Test GeoJson ROI loader errors."""
    # Invalid type
    with raises(NotImplementedError):
        _ = roi.load_geojson({'type': 'FOO'})

    # Missing features key
    with raises(ValueError):
        _ = roi.load_geojson({'type': 'FeatureCollection'})

    # Empty features
    with raises(ValueError):
        _ = roi.load_geojson({'type': 'FeatureCollection', 'features': []})

    # Too many features
    with raises(NotImplementedError):
        _ = roi.load_geojson({'type': 'FeatureCollection', 'features': [1, 2]})

    # Missing geometry key
    with raises(ValueError):
        _ = roi.load_geojson({
            'type': 'FeatureCollection',
            'features': [
                {
                    'type': 'Feature',
                }
            ],
        })

    # Invalid geometry type
    with raises(NotImplementedError):
        _ = roi.load_geojson({
            'type': 'FeatureCollection',
            'features': [
                {
                    'type': 'Feature',
                    'geometry': {
                        'type': 'Point',
                    },
                }
            ],
        })

    # Missing coordinates
    with raises(ValueError):
        _ = roi.load_geojson({
            'type': 'FeatureCollection',
            'features': [
                {
                    'type': 'Feature',
                    'geometry': {
                        'type': 'Polygon',
                    },
                }
            ],
        })

    # Invalid coordinates
    with raises(ValueError):
        _ = roi.load_geojson({
            'type': 'FeatureCollection',
            'features': [
                {
                    'type': 'Feature',
                    'geometry': {'type': 'Polygon', 'coordinates': 'foo'},
                }
            ],
        })
