"""Test KML ROI module."""

import pathlib

from numpy.testing import assert_array_almost_equal as assert_array

from pytest import raises

from planetary_coverage import KmlROIsCollection
from planetary_coverage.rois.kml import KmlParser, read_kml, read_kmz


DATA = pathlib.Path(__file__).parent / '..' / 'data'


def test_kml_parser():
    """Test KML parser."""
    kml = KmlParser()

    # Coordinates parser
    assert_array(
        kml.coordinates('110,-35,0 -240,-30,0 130,-25,0'),
        [
            (110, 120, 130),
            (-35, -30, -25),
            (0, 0, 0),
        ],
    )

    assert_array(
        kml.coordinates('110,-35,0 -240,-30,0 130,-25,0', line=True),
        [
            (110, 120, 130, 130, 120, 110),
            (-35, -30, -25, -25, -30, -35),
            (0, 0, 0, 0, 0, 0),
        ],
    )

    # Color parser
    assert kml.rgb('12345678') == '#78563412'


def test_kml_reader():
    """Test KML reader."""
    kml = read_kml(DATA / 'with-folder.kml')

    assert isinstance(kml, list)
    assert len(kml) == 1

    # Check content values
    data = kml[0]

    assert data['name'] == 'Collection of ROIs'
    assert data['description'] == 'Collection description'

    placemarks = data['placemarks']

    assert len(placemarks) == 2

    # LineString
    line = placemarks[0]

    assert line['name'] == 'Linear feature'
    assert line['facecolor'] == '#ff202180'
    assert line['edgecolor'] == '#7c71ffff'
    assert line['linewidth'] == 3
    assert line['geometry'] == 'LineString'
    assert line['extrude'] == '1'
    assert line['tessellate'] == '1'
    assert line['altitudeMode'] == 'absolute'

    assert 'id' not in line

    assert_array(line['lons_e'], [110, 120, 130, 130, 120, 110])
    assert_array(line['lats'], [-35, -30, -25, -25, -30, -35])
    assert_array(line['alts'], [0, 0, 0, 0, 0, 0])

    # Polygon
    polygon = placemarks[1]

    assert polygon['name'] == 'Polygon feature'
    assert polygon['description'] == 'Polygon feature description'
    assert polygon['facecolor'] == '#ff202180'
    assert polygon['edgecolor'] == '#7c71ffff'
    assert polygon['linewidth'] == 3
    assert polygon['geometry'] == 'LinearRing'
    assert polygon['extrude'] == '1'
    assert polygon['tessellate'] == '1'
    assert polygon['altitudeMode'] == 'absolute'

    assert 'id' not in polygon

    assert_array(polygon['lons_e'], [270, 330, 330, 270, 270])
    assert_array(polygon['lats'], [30, 30, 0, 0, 30])
    assert_array(polygon['alts'], [0, 0, 0, 0, 0])

    # Load kml without a folder root directory
    kml = read_kml(DATA / 'without-folder.kml')

    assert isinstance(kml, list)
    assert len(kml) == 1

    data = kml[0]

    assert data['name'] == 'Burr crater'
    assert data['description'] is None

    placemarks = data['placemarks']

    assert len(placemarks) == 1

    polygon = placemarks[0]

    assert polygon['name'] == 'Burr crater'

    assert_array(
        polygon['lons_e'],
        [
            227.9,
            228.7,
            229.1,
            227.5,
            226.4,
            225.0,
            224.0,
            223.9,
            223.3,
            221.7,
            221.7,
            222.9,
            222.6,
            224.7,
            226.5,
            227.5,
        ],
    )
    assert_array(
        polygon['lats'],
        [
            42.4,
            42.9,
            44.0,
            45.2,
            45.8,
            45.0,
            46.3,
            44.9,
            44.1,
            44.2,
            43.3,
            42.6,
            41.9,
            40.8,
            41.3,
            41.4,
        ],
    )
    assert_array(
        polygon['alts'],
        [
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
        ],
    )

    assert 'description' not in polygon
    assert 'facecolor' not in polygon
    assert 'edgecolor' not in polygon
    assert 'linewidth' not in polygon

    # File missing
    with raises(FileNotFoundError):
        _ = read_kml('dummy.kml')

    # Invalid file extension
    with raises(ValueError):
        _ = read_kml(DATA / 'test.geojson')


def test_kmz_reader():
    """Test KMZ reader."""
    kmz = read_kmz(DATA / 'test.kmz')

    assert len(kmz) == 2
    assert kmz[0]['name'] == 'Collection of ROIs'
    assert kmz[1]['name'] == 'Burr crater'

    # File missing
    with raises(FileNotFoundError):
        _ = read_kmz('dummy.kmz')

    # Invalid file extension
    with raises(ValueError):
        _ = read_kmz(DATA / 'test.geojson')


def test_kml_rois():
    """Test KML ROIs collection."""
    kml = KmlROIsCollection(
        DATA / 'with-folder.kml',
        DATA / 'without-folder.kml',
        target='my-target',
    )

    assert kml.filenames == [DATA / 'with-folder.kml', DATA / 'without-folder.kml']
    assert kml.target == 'my-target'

    # Check categories
    categories = kml.categories

    assert len(categories) == 2
    assert 'Collection of ROIs' in categories
    assert 'Burr crater' in categories

    cat = categories['Collection of ROIs']

    assert cat.name == 'Collection Of Rois'
    assert cat.description == 'Collection description'
    assert cat.filename == DATA / 'with-folder.kml'
    assert cat.target == 'my-target'

    # Check ROIs content
    assert len(kml) == len(kml.rois) == 3

    rois = kml.rois

    assert len(rois) == 3
    assert '(1) Linear feature' in rois
    assert '(2) Polygon feature' in rois
    assert '(3) Burr crater' in rois
    assert 'foo' not in rois

    # Check ROI properties from `with-folder.kml`
    roi = kml['(2) Polygon feature']

    assert roi.name == 'Polygon Feature'
    assert roi.description == 'Polygon feature description'
    assert_array(roi.lons_e, [270, 330, 330, 270, 270])
    assert_array(roi.lats, [30, 30, 0, 0, 30])
    assert roi.facecolor == '#ff202180'
    assert roi.edgecolor == '#7c71ffff'
    assert roi.linewidth == 3
    assert roi.zorder == 2
    assert roi.category == cat

    # ROI patch properties
    patch = roi.patch()

    assert patch.get_alpha() is None
    assert_array(patch.get_fc(), [1, 0.12, 0.13, 0.5], decimal=2)  # FF202180
    assert_array(patch.get_ec(), [0.49, 0.44, 1, 1], decimal=2)  # 7c71FFFF
    assert roi.get_linewidth() == 3
    assert roi.get_linestyle() == 'solid'  # ROI default
    assert roi.get_zorder() == 2

    # Check ROI intersection
    assert (300, 15) in roi
    assert (300, -15) not in roi

    # Check ROIsCollection intersection
    assert (300, 15) in rois
    assert (300, -15) not in rois

    # Check ROI properties from `without-folder.kml`
    roi = kml['(3) Burr crater']

    assert roi.name == 'Burr Crater'

    assert_array(
        roi.lons_e,
        [
            227.9,
            228.7,
            229.1,
            227.5,
            226.4,
            225.0,
            224.0,
            223.9,
            223.3,
            221.7,
            221.7,
            222.9,
            222.6,
            224.7,
            226.5,
            227.5,
        ],
    )
    assert_array(
        roi.lats,
        [
            42.4,
            42.9,
            44.0,
            45.2,
            45.8,
            45.0,
            46.3,
            44.9,
            44.1,
            44.2,
            43.3,
            42.6,
            41.9,
            40.8,
            41.3,
            41.4,
        ],
    )

    assert not hasattr(roi, 'description')
    assert not hasattr(roi, 'facecolor')
    assert not hasattr(roi, 'edgecolor')
    assert not hasattr(roi, 'linewidth')

    assert roi.category.filename == DATA / 'without-folder.kml'

    # KML without target
    kml = KmlROIsCollection(DATA / 'with-folder.kml')
    cat = kml.categories['Collection of ROIs']
    roi = kml['(2) Polygon feature']

    assert kml.target is None
    assert not hasattr(cat, 'target')
    assert not hasattr(roi, 'target')

    # Duplicated categories
    with raises(KeyError):
        _ = KmlROIsCollection(DATA / 'with-folder.kml', DATA / 'with-folder.kml')


def test_kmz_rois():
    """Test KMZ ROIs collection."""
    kmz = KmlROIsCollection(DATA / 'test.kmz')

    assert kmz.filenames == [DATA / 'test.kmz']

    # Check categories
    categories = kmz.categories

    assert len(categories) == 2
    assert 'Collection of ROIs' in categories
    assert 'Burr crater' in categories

    # Check ROIs content
    assert len(kmz) == len(kmz.rois) == 3

    rois = kmz.rois

    assert len(rois) == 3
    assert '(1) Linear feature' in rois
    assert '(2) Polygon feature' in rois
    assert '(3) Burr crater' in rois
    assert 'foo' not in rois

    # Check ROI properties from `test.kmz`
    roi = kmz['(2) Polygon feature']

    assert roi.name == 'Polygon Feature'
