"""Test ROIs collections."""

import numpy as np
from numpy.testing import assert_array_almost_equal as assert_array

from pytest import fixture, raises

from planetary_coverage.rois import ROI, Category, ROIsCollection, SubCategory
from planetary_coverage.rois.rois import (
    AbstractEmptyROI,
    EmptyROI,
    ROIsCollectionWithCategories,
)


@fixture
def roi_1_0_01():
    """Ganymede ROI."""
    return {
        'name': 'Hammamat Patera',
        'lons_e': [38, 45, 45, 38],
        'lats': [-28, -28, -20, -20],
        'desc': 'Possible caldera in Nicholson Regio (Galileo SSI obs. 28GSCALDERA01/02)',
        'science_rationale': 'Study composition (content of volatiles etc.) and '
        'geo context and subsurface of possible calderas and their '
        'association with very smooth terrain or large impacts',
        'special_requirements': 'Best spatial resolution and SNR possible required to '
        'identify source of activity',
        'color': '#FECCCC',
        'target': 'Ganymede',
    }


@fixture
def category_1():
    """Ganymede Category 1."""
    return {
        'name': 'Potential Cryovolcanic Regions',
        'desc': 'Geologically young surface features of possible cryovolcanic origin',
        'science_rationale': 'Understand the formation of surface features and '
        'search for past and present activity (GB/GD/GE)',
        'special_requirements': 'Best spatial resolution and SNR possible required '
        'to identify source of activity',
        'color': '#FECCCC',
    }


@fixture
def subcategories_4_3():
    """Ganymede Subcategory 4.3."""
    return {
        'name': 'Dome / Anomaleous Dome Craters',
        'desc': 'Fresh impact craters with a central dome',
        'science_rationale': 'Study subsurface properties related to dome formation',
        'color': '#CCCDFD',
        'category': '4',
    }


def test_rois_collection(roi_1_0_01):
    """Test ROIs collection."""
    rois = ROIsCollection(target='Callisto')

    # Add ROI
    rois['JUICE_ROI_GAN_1_0_01'] = roi_1_0_01

    assert len(rois) == 1
    assert rois.count == '1 roi'
    assert rois.target == 'Callisto'

    roi = rois['JUICE_ROI_GAN_1_0_01']

    assert isinstance(roi, ROI)
    assert str(roi) == 'JUICE_ROI_GAN_1_0_01'
    assert roi == 'JUICE_ROI_GAN_1_0_01'
    assert roi.name == 'Hammamat Patera'
    assert roi.lons_e == [38, 45, 45, 38]
    assert roi.lats == [-28, -28, -20, -20]
    assert roi.color == '#FECCCC'
    assert roi.target == 'Ganymede'

    # Path
    path = roi.get_path()

    assert_array(
        path.vertices,
        [
            [38, -28],
            [45, -28],
            [45, -20],
            [38, -20],
            [38, -28],
        ],
    )

    # Default patch attributes
    assert roi.get_fc() == 'none'
    assert roi.get_ec() == '#FECCCC'
    assert roi.get_alpha() is None
    assert roi.get_linewidth() == 1.5
    assert roi.get_linestyle() == 'solid'
    assert roi.get_zorder() == 1
    assert not roi.get_label()

    # Patch (defaults)
    patch = roi()

    assert_array(patch.get_facecolor(), (0.0, 0.0, 0.0, 0.0), decimal=3)
    assert_array(patch.get_edgecolor(), (0.996, 0.8, 0.8, 1.0), decimal=3)  # #FECCCC

    assert patch.get_alpha() is None
    assert patch.get_linewidth() == 1.5
    assert patch.get_linestyle() == 'solid'
    assert patch.get_zorder() == 1
    assert not patch.get_label()

    # Change ROI attribute at runtime
    roi.alpha = 0.5
    roi.linewidth = 2
    roi.linestyle = '--'
    roi.zorder = 3
    roi.label = 'Foo'

    patch = roi(color='C0')  # __call__

    assert_array(patch.get_facecolor(), (0.121, 0.466, 0.705, 0.5), decimal=3)  # C0 + 0.5
    assert_array(patch.get_edgecolor(), (0.121, 0.466, 0.705, 0.5), decimal=3)

    assert patch.get_alpha() == 0.5
    assert patch.get_linewidth() == 2
    assert patch.get_linestyle() == '--'
    assert patch.get_zorder() == 3
    assert patch.get_label() == 'Foo'

    # Add ROI with a single point coordinate
    rois['JUICE_ROI_GAN_1_0_02'] = {
        'name': 'Natrun/Rum Patera',
        'lons_e': 177,
        'lats': -30.8,
    }

    assert len(rois) == 2
    assert rois.count == '2 rois'

    roi = rois['JUICE_ROI_GAN_1_0_02']

    assert roi == 'JUICE_ROI_GAN_1_0_02'
    assert roi.name == 'Natrun/Rum Patera'
    assert roi.lons_e == 177
    assert roi.lats == -30.8

    # ROIs Patch collection
    coll = rois()
    paths = coll.get_paths()

    assert len(paths) == 2
    assert_array(paths[1].vertices, [(177, -30.8)])

    assert_array(coll.get_facecolor(), [(0, 0, 0, 0), (0, 0, 0, 0)])  # None x2
    assert_array(coll.get_edgecolor()[0], (0.996, 0.8, 0.8, 1), decimal=3)  # FECCCC
    assert_array(coll.get_edgecolor()[1], (0.84, 0.16, 0.16, 1), decimal=2)  # tab:red

    assert coll.get_alpha() is None  # JUICE_ROI_GAN_1_0_02 alpha is None
    assert all(coll.get_linewidth() == [2, 1.5])  # [Custom, Default]

    assert coll.get_linestyle()[0][0] == 0  # Custom ('--' -> (0.0, [7.4, 3.2]))
    assert_array(coll.get_linestyle()[0][1], [7.4, 3.2])
    assert coll.get_linestyle()[1][0] == 0.0  # Default ('solid' -> (0.0, None))
    assert coll.get_linestyle()[1][1] is None

    assert coll.get_zorder() == 3
    assert coll.get_label() == 'Foo'  # Merge ['Foo', '']

    # Custom colors
    coll = rois(facecolors='C0', edgecolors='C1')

    assert len(coll.get_facecolor()) == 1
    assert len(coll.get_edgecolor()) == 1

    assert_array(coll.get_facecolor()[0], (0.121, 0.466, 0.705, 1), decimal=3)  # C0
    assert_array(coll.get_edgecolor()[0], (1, 0.498, 0.055, 1), decimal=3)  # C1

    # Default paths
    paths = rois.get_paths()

    assert len(paths) == 2
    assert_array(paths[1].vertices, [(177, -30.8)])

    # Default colors
    assert rois.get_facecolor() == ['none', 'none']
    assert rois.get_edgecolor() == ['#FECCCC', 'tab:red']

    # Remove ROI
    del rois['JUICE_ROI_GAN_1_0_02']

    assert len(rois) == 1
    assert 'JUICE_ROI_GAN_1_0_01' in rois
    assert 'JUICE_ROI_GAN_1_0_02' not in rois

    assert rois.get_alpha() == [0.5]  # JUICE_ROI_GAN_1_0_01 alpha is not None

    # Invalid ROI coordinates
    with raises(ValueError):
        _ = rois['JUICE_ROI_GAN_1_0_03'] = {
            'lons_e': [176.7, 37.5],
            'lats': -30.9,
        }


def test_rois_with_categories(category_1, subcategories_4_3, roi_1_0_01):  # noqa: PLR0915
    """Test ROIs collection with categories."""
    rois = ROIsCollectionWithCategories()

    assert len(rois) == 0
    assert len(rois.rois) == 0
    assert len(rois.categories) == 0
    assert len(rois.subcategories) == 0

    # Output rois collection
    assert isinstance(rois.rois, ROIsCollection)

    # Add categories
    rois.add_category(1, **category_1)

    assert len(rois) == 0
    assert len(rois.categories) == 1

    category = rois.categories['1']

    assert isinstance(category, Category)
    assert category == '1'
    assert category.key == 1
    assert category.name == 'Potential Cryovolcanic Regions'

    # Add subcategories without category attribute
    with raises(AttributeError):
        rois.add_subcategory('4.3')

    # Add subcategories with an unknown category
    with raises(ValueError):
        rois.add_subcategory('4.3', **subcategories_4_3)

    # Add missing category
    rois.add_category('4', name='Impact Crater Morphologies', color='#CB9CFC')

    category = rois.categories['4']

    assert isinstance(category, Category)
    assert category == '4'
    assert category.key == '4'
    assert category.name == 'Impact Crater Morphologies'

    # Add a valid subcategory
    rois.add_subcategory('4.3', **subcategories_4_3)

    assert len(rois) == 0
    assert len(rois.categories) == 2
    assert len(rois.subcategories) == 1

    subcategory = rois.subcategories['4.3']

    assert isinstance(subcategory, SubCategory)
    assert subcategory == '4.3'
    assert subcategory.key == '4.3'
    assert subcategory.name == 'Dome / Anomaleous Dome Craters'

    assert subcategory.category == '4'
    assert subcategory.category.name == 'Impact Crater Morphologies'

    # Add ROI without category
    with raises(AttributeError):
        rois['JUICE_ROI_GAN_0_0_01'] = roi_1_0_01

    # Add ROI with an unknown category
    with raises(ValueError):
        rois['JUICE_ROI_GAN_2_0_01'] = {**roi_1_0_01, 'category': 2}

    # Add ROI with a known category
    rois['JUICE_ROI_GAN_1_0_01'] = {**roi_1_0_01, 'category': 1}

    assert len(rois) == 1
    assert len(rois.rois) == 1
    assert len(rois.categories) == 2
    assert len(rois.subcategories) == 1

    roi = rois['JUICE_ROI_GAN_1_0_01']

    assert roi == 'JUICE_ROI_GAN_1_0_01'
    assert roi.name == 'Hammamat Patera'
    assert roi.lons_e == [38, 45, 45, 38]
    assert roi.lats == [-28, -28, -20, -20]
    assert roi.color == '#FECCCC'

    # Add ROI with an unknown subcategory
    with raises(ValueError):
        rois['JUICE_ROI_GAN_1_1_01'] = {**roi_1_0_01, 'subcategory': '1.1'}

    # Add ROI with a known subcategory
    rois['JUICE_ROI_GAN_4_3_01'] = {
        'name': 'Melkart',
        'lons_e': 0,
        'lats': 0,
        'subcategory': '4.3',
    }

    assert len(rois) == 2
    assert len(rois.rois) == 2
    assert len(rois.categories) == 2
    assert len(rois.subcategories) == 1

    roi = rois['JUICE_ROI_GAN_4_3_01']

    assert roi == 'JUICE_ROI_GAN_4_3_01'
    assert roi.name == 'Melkart'
    assert roi.subcategory == '4.3'
    assert roi.category == '4'

    # Add ROI item without category
    with raises(AttributeError):
        rois['JUICE_ROI_GAN_1_0_02'] = ROI(
            'JUICE_ROI_GAN_1_0_00', name='Natrun/Rum Patera', lons_e=177, lats=-30.8
        )

    # Add ROI item directly with a new key in a category
    rois['JUICE_ROI_GAN_1_0_02'] = ROI(
        'JUICE_ROI_GAN_1_0_00',
        name='Natrun/Rum Patera',
        lons_e=177,
        lats=-30.8,
        category=1,
    )

    assert len(rois) == 3

    roi = rois['JUICE_ROI_GAN_1_0_02']

    assert isinstance(roi, ROI)
    assert roi == 'JUICE_ROI_GAN_1_0_02'
    assert roi.name == 'Natrun/Rum Patera'

    assert isinstance(roi.category, Category)
    assert roi.category == '1'

    assert not hasattr(roi, 'subcategory')

    # Add ROI item directly with a new key in a subcategory
    rois['JUICE_ROI_GAN_4_3_02'] = ROI(
        'JUICE_ROI_GAN_4_3_00', name='Enkidu', lons_e=34.9, lats=-26.6, subcategory='4.3'
    )

    assert len(rois) == 4

    assert 'JUICE_ROI_GAN_4_3_02' in rois

    roi = rois['JUICE_ROI_GAN_4_3_02']

    assert isinstance(roi, ROI)
    assert roi == 'JUICE_ROI_GAN_4_3_02'
    assert roi.name == 'Enkidu'

    assert isinstance(roi.category, Category)
    assert roi.category == '4'

    assert isinstance(roi.subcategory, SubCategory)
    assert roi.subcategory == '4.3'

    # Iter (on values)
    for roi in rois:
        assert isinstance(roi, ROI)
        assert roi == 'JUICE_ROI_GAN_1_0_01'
        break

    # Get items by categories
    assert '1' in rois

    rois_cat_1 = rois['1']

    assert isinstance(rois_cat_1, ROIsCollection)
    assert len(rois_cat_1) == 2

    assert 'JUICE_ROI_GAN_1_0_01' in rois_cat_1
    assert 'JUICE_ROI_GAN_1_0_02' in rois_cat_1
    assert 'JUICE_ROI_GAN_4_3_01' not in rois_cat_1

    assert rois_cat_1['JUICE_ROI_GAN_1_0_01'].name == 'Hammamat Patera'
    assert rois_cat_1['JUICE_ROI_GAN_1_0_02'].name == 'Natrun/Rum Patera'

    # Get items by subcategories
    assert '4.3' in rois

    rois_subcat_4_3 = rois['4.3']

    assert isinstance(rois_subcat_4_3, ROIsCollection)
    assert len(rois_subcat_4_3) == 2

    assert 'JUICE_ROI_GAN_4_3_01' in rois_subcat_4_3
    assert 'JUICE_ROI_GAN_4_3_02' in rois_subcat_4_3
    assert 'JUICE_ROI_GAN_1_0_01' not in rois_subcat_4_3

    assert rois_subcat_4_3['JUICE_ROI_GAN_4_3_01'].name == 'Melkart'
    assert rois_subcat_4_3['JUICE_ROI_GAN_4_3_02'].name == 'Enkidu'

    # Get item without a valid key
    assert 'Unknown key' not in rois

    with raises(KeyError):
        _ = rois['Unknown key']


def test_roi_contains_pt():
    """Test ROI point intersection."""
    roi = ROI(
        key='TEST_ROI_ANTI_MERIDIAN',
        lons_e=[330, 30, 30, 330, 330],
        lats=[30, 30, -30, -30, 30],
    )

    # Single point
    assert (350, 0) in roi
    # assert (0, 0) not in roi  # FIXME: Point on the edge of the contour
    assert (10, 0) in roi
    assert (180, 0) not in roi

    # Multi-points
    traj = [(0, 0), (180, 50), (350, 0)]
    assert traj in roi
    assert traj[:2] not in roi

    lons_e, lats = (0, 180, 350), (0, 50, 0)
    assert (lons_e, lats) in roi


def test_roi_trajectory_intersection():
    """Test ROI intersection with a trajectory."""

    class Trajectory:
        """Dummy trajectory object."""

        lonlat = [[307.82], [-71.79]]

    traj = Trajectory()

    # ROI surrounding the sub-observer point
    roi = ROI(
        key='ROI_IN',
        lons_e=[300, 310, 310, 300],
        lats=[-70, -70, -75, -75],
    )

    assert roi.contains(traj) == [True]
    assert traj in roi

    # Compte the non-null intersection
    intersected_roi = roi & traj
    assert isinstance(intersected_roi, ROI)
    assert intersected_roi is roi

    # Compte the null exclusion
    not_intersected_roi = roi ^ traj
    assert isinstance(not_intersected_roi, AbstractEmptyROI)
    assert not_intersected_roi is EmptyROI

    # ROI not surrounding the sub-observer point
    roi = ROI(
        key='ROI_OUT',
        lons_e=[300, 310, 310, 300],
        lats=[-70, -70, -65, -65],
    )

    assert roi.contains(traj) == [False]
    assert traj not in roi

    # Compte the null intersection
    intersected_roi = roi & traj
    assert isinstance(intersected_roi, AbstractEmptyROI)
    assert intersected_roi is EmptyROI

    # Compte the non-null exclusion
    not_intersected_roi = roi ^ traj
    assert isinstance(not_intersected_roi, ROI)
    assert not_intersected_roi is roi


def test_roi_trajectory_intersection_target():
    """Test ROI intersection with a targeted trajectory."""

    class Trajectory:
        """Dummy trajectory object."""

        lonlat = [[307.82], [-71.79]]
        target = 'Ganymede'

    traj = Trajectory()

    # ROI surrounding the sub-observer point with matching target
    roi = ROI(
        key='GAN_ROI',
        lons_e=[300, 310, 310, 300],
        lats=[-70, -70, -75, -75],
        target='ganymede',
    )

    assert roi.contains(traj) == [True]
    assert traj in roi

    # Mismatch target name
    roi.target = 'Callisto'

    assert roi.contains(traj) == [False]
    assert traj not in roi


def test_rois_collection_intersection():
    """Test ROIs Collection intersection with a trajectory."""

    class Trajectory:
        """Dummy trajectory object."""

        lonlat = [[306, 307, 308], [-71, -68, -63]]

    traj = Trajectory()

    rois = ROIsCollection()

    rois['ROI_IN_OUT_OUT'] = {
        'lons_e': [300, 310, 310, 300],
        'lats': [-70, -70, -75, -75],
    }

    rois['ROI_OUT_IN_OUT'] = {
        'lons_e': [300, 310, 310, 300],
        'lats': [-70, -70, -65, -65],
    }

    assert len(rois) == 2
    assert traj in rois
    assert_array(rois.contains(traj), [True, True, False])

    # Compute intersection
    intersected_rois = rois & traj
    assert isinstance(intersected_rois, ROIsCollection)
    assert len(intersected_rois) == 2

    keys = intersected_rois.keys()
    assert 'ROI_IN_OUT_OUT' in keys
    assert 'ROI_OUT_IN_OUT' in keys

    # Compute exclusion
    not_intersected_rois = rois ^ traj
    assert isinstance(not_intersected_rois, ROIsCollection)
    assert len(not_intersected_rois) == 0


def test_rois_collection_intersection_target():
    """Test ROIs Collection intersection with a targeted trajectory."""

    class Trajectory:
        """Dummy trajectory object."""

        lonlat = [[306, 307, 308], [-71, -68, -63]]
        target = 'Ganymede'

    traj = Trajectory()

    rois = ROIsCollection()

    rois['GAN_ROI_IN_OUT_OUT'] = {
        'lons_e': [300, 310, 310, 300],
        'lats': [-70, -70, -75, -75],
        'target': 'Ganymede',
    }

    rois['GAN_ROI_OUT_IN_OUT'] = {
        'lons_e': [300, 310, 310, 300],
        'lats': [-70, -70, -65, -65],
        'target': 'Ganymede',
    }

    rois['GAN_ROI_OUT_OUT_OUT'] = {
        'lons_e': [300, 310, 310, 300],
        'lats': [70, 70, 75, 75],
        'target': 'Ganymede',
    }

    rois['CAL_ROI_IN_OUT_OUT'] = {
        'lons_e': [300, 310, 310, 300],
        'lats': [-70, -70, -75, -75],
        'target': 'Callisto',
    }

    rois['CAL_ROI_OUT_OUT_IN'] = {
        'lons_e': [300, 310, 310, 300],
        'lats': [-65, -65, -70, -70],
        'target': 'Callisto',
    }

    assert len(rois) == 5
    # Intersection with a point
    assert (307, -68) in rois
    assert [308, -63] not in rois

    # Intersection with a trajectory
    assert traj in rois
    assert_array(rois.contains(traj), [True, True, False])

    # Compute intersection
    intersected_rois = rois & traj
    assert isinstance(intersected_rois, ROIsCollection)
    assert len(intersected_rois) == 2

    keys = intersected_rois.keys()
    assert 'GAN_ROI_IN_OUT_OUT' in keys
    assert 'GAN_ROI_OUT_IN_OUT' in keys
    assert 'GAN_ROI_OUT_OUT_OUT' not in keys
    assert 'CAL_ROI_IN_OUT_OUT' not in keys
    assert 'CAL_ROI_OUT_OUT_IN' not in keys

    # Compute exclusion
    not_intersected_rois = rois ^ traj
    assert isinstance(not_intersected_rois, ROIsCollection)
    assert len(not_intersected_rois) == 3

    keys = not_intersected_rois.keys()
    assert 'GAN_ROI_IN_OUT_OUT' not in keys
    assert 'GAN_ROI_OUT_IN_OUT' not in keys
    assert 'GAN_ROI_OUT_OUT_OUT' in keys
    assert 'CAL_ROI_IN_OUT_OUT' in keys
    assert 'CAL_ROI_OUT_OUT_IN' in keys


def test_empty_roi():
    """Test Null ROI object."""
    assert not str(EmptyROI)

    assert_array(EmptyROI(foo='bar').get_path().vertices, [(np.nan, np.nan)])

    assert 'foo' not in EmptyROI
    assert (EmptyROI & 'foo') is EmptyROI
    assert (EmptyROI ^ 'foo') is EmptyROI

    assert_array(EmptyROI.contains([1, 2, 3]), [False, False, False])


def test_roi_style():
    """Test ROI style."""
    # No style (test defaults)
    roi = ROI(key='NO_STYLE', lons_e=0, lats=0)

    assert_array(roi.vertices, [(0, 0)])

    assert roi.get_alpha() is None
    assert roi.get_fc() == 'none'
    assert roi.get_ec() == 'tab:red'
    assert roi.get_linewidth() == 1.5
    assert roi.get_linestyle() == 'solid'
    assert roi.get_zorder() == 1
    assert not roi.get_label()

    patch = roi.patch()

    assert patch.get_alpha() is None
    assert_array(patch.get_facecolor(), [0, 0, 0, 0], decimal=2)
    assert_array(patch.get_edgecolor(), [0.84, 0.15, 0.16, 1], decimal=2)
    assert patch.get_linewidth() == 1.5
    assert patch.get_linestyle() == 'solid'
    assert patch.get_zorder() == 1
    assert not patch.get_label()

    # Overwrite defaults when calling patch
    patch = roi.patch(facecolor='green', edgecolor='blue', linewidth=10, alpha=0.2)

    assert patch.get_alpha() == 0.2
    assert_array(patch.get_facecolor(), [0, 0.5, 0, 0.2], decimal=2)
    assert_array(patch.get_edgecolor(), [0, 0, 1, 0.2], decimal=2)
    assert patch.get_linewidth() == 10

    # Explicit color (overwrite edgecolor and set facecolor to none)
    roi = ROI(key='COLOR', lons_e=0, lats=0, color='red')

    assert roi.get_fc() == 'none'  # No effect on facecolor
    assert roi.get_ec() == 'red'

    patch = roi.patch()

    assert_array(patch.get_facecolor(), [0, 0, 0, 0], decimal=2)  # No effect
    assert_array(patch.get_edgecolor(), [1, 0, 0, 1], decimal=2)

    patch = roi.patch(color='blue')

    assert_array(patch.get_facecolor(), [0, 0, 1, 1], decimal=2)  # Overwrite facecolor
    assert_array(patch.get_edgecolor(), [0, 0, 1, 1], decimal=2)

    # Fully color coded ROI
    roi = ROI(
        key='STYLED',
        lons_e=0,
        lats=0,
        facecolor='#ff202180',
        edgecolor='#ff0000ff',
        linewidth=10,
    )

    assert roi.get_alpha() is None
    assert roi.get_fc() == '#ff202180'
    assert roi.get_ec() == '#ff0000ff'
    assert roi.get_linewidth() == 10

    patch = roi.patch()

    assert_array(patch.get_facecolor(), [1, 0.13, 0.13, 0.5], decimal=2)
    assert_array(patch.get_edgecolor(), [1, 0, 0, 1], decimal=2)
    assert patch.get_linewidth() == 10
