"""Test SPICE light time aberration correction module."""

from numpy.testing import assert_array_almost_equal as assert_array

from pytest import approx, raises

from planetary_coverage.spice import SpiceAbCorr


def test_abcorr_type():
    """Test SPICE aberration correction type."""
    abcorr = SpiceAbCorr('NONE')
    assert isinstance(abcorr, SpiceAbCorr)
    assert isinstance(abcorr, str)


def test_abcorr_inputs():
    """Test SPICE aberration correction inputs."""
    # Empty/Null values
    assert SpiceAbCorr() == 'NONE'
    assert SpiceAbCorr('') == 'NONE'
    assert SpiceAbCorr('None') == 'NONE'
    assert SpiceAbCorr('NONE') == 'NONE'
    assert SpiceAbCorr(None) == 'NONE'

    # Valid inputs
    assert SpiceAbCorr('LT') == 'LT'
    assert SpiceAbCorr('LT+S') == 'LT+S'
    assert SpiceAbCorr('CN') == 'CN'
    assert SpiceAbCorr('CN+S') == 'CN+S'
    assert SpiceAbCorr('XLT') == 'XLT'
    assert SpiceAbCorr('XLT+S') == 'XLT+S'
    assert SpiceAbCorr('XCN') == 'XCN'
    assert SpiceAbCorr('XCN+S') == 'XCN+S'

    # Check lower case
    assert SpiceAbCorr('lt') == 'LT'
    assert SpiceAbCorr('xCn+s') == 'XCN+S'

    # Invalid values
    with raises(KeyError):
        _ = SpiceAbCorr('foo')

    with raises(KeyError):
        _ = SpiceAbCorr('foo_LT')

    with raises(KeyError):
        _ = SpiceAbCorr('LT_foo')

    # Chain AbCorr objects
    assert SpiceAbCorr(SpiceAbCorr('LT')) == 'LT'


def test_abcorr_restrictions():
    """Test SPICE aberration correction restrictions."""
    restrict = ('LT', 'LT+S')

    # Valid
    assert SpiceAbCorr('LT', restrict=restrict) == 'LT'
    assert SpiceAbCorr('LT+S', restrict=restrict) == 'LT+S'

    # Invalid
    with raises(KeyError):
        _ = SpiceAbCorr(restrict=restrict)

    with raises(KeyError):
        _ = SpiceAbCorr('CN', restrict=restrict)


def test_abcorr_properties():
    """Test SPICE aberration correction properties."""
    abcorr = SpiceAbCorr('NONE')

    assert abcorr == 'NONE'
    assert not abcorr.reception
    assert not abcorr.transmission
    assert not abcorr.stellar
    assert not abcorr.oneway
    assert not abcorr.converged

    abcorr = SpiceAbCorr('LT')

    assert abcorr.reception
    assert not abcorr.transmission
    assert not abcorr.stellar
    assert abcorr.oneway
    assert not abcorr.converged

    abcorr = SpiceAbCorr('LT+S')

    assert abcorr.reception
    assert not abcorr.transmission
    assert abcorr.stellar
    assert abcorr.oneway
    assert not abcorr.converged

    abcorr = SpiceAbCorr('XCN+S')

    assert not abcorr.reception
    assert abcorr.transmission
    assert abcorr.stellar
    assert not abcorr.oneway
    assert abcorr.converged


def test_abcorr_corrections():
    """Test SPICE aberration correction corrections."""
    abcorr = SpiceAbCorr('NONE')
    assert abcorr.lt_corr(10, 2) == 10

    # Reception case
    abcorr = SpiceAbCorr('LT')
    assert abcorr.lt_corr(10, 2) == 8

    abcorr = SpiceAbCorr('LT+S')
    assert abcorr.lt_corr(10, 2) == 8

    abcorr = SpiceAbCorr('CN')
    assert abcorr.lt_corr(10, 2) == 8

    # Transmission case
    abcorr = SpiceAbCorr('XCN+S')
    assert abcorr.lt_corr(10, 2) == 12

    # With arrays
    abcorr = SpiceAbCorr('LT')
    assert_array(abcorr.lt_corr([10, 20], 2), [8, 18])
    assert_array(abcorr.lt_corr(10, [2, 4]), [8, 6])
    assert_array(abcorr.lt_corr([10, 20], [2, 4]), [8, 16])

    # Invalid arrays sizes
    with raises(ValueError):
        _ = abcorr.lt_corr([10, 20, 30], [2, 4])

    # Distance correction
    abcorr = SpiceAbCorr('NONE')
    assert abcorr.dist_corr(10, 3e5) == 10

    abcorr = SpiceAbCorr('LT')
    assert abcorr.dist_corr(10, 3e5) == approx(9, abs=1e-3)

    abcorr = SpiceAbCorr('XLT')
    assert abcorr.dist_corr(10, 3e5) == approx(11, abs=1e-3)

    # __call__ -> dist_corr
    assert abcorr(10, 3e5) == approx(11, abs=1e-3)
