"""Test SPICE datetime module."""

from datetime import datetime as dt_native

import numpy as np
from numpy import datetime64 as dt
from numpy import timedelta64 as td

import spiceypy as sp

from pytest import approx, raises

from planetary_coverage import datetime, sorted_datetimes
from planetary_coverage.spice import iso, jd, jdn_hms, mapps_datetime, timedelta, ymd
from planetary_coverage.spice.datetime import clean_time, np_date_str, np_datetime_str


def test_spice_datetime_jdn_hms():
    """Test SPICE julian day and time parser."""
    assert jdn_hms('March-7-1987-3:10:39.221') == (2446862, 3, 10, 39, 221)
    assert jdn_hms('1999-12-02T19:09:15.84') == (2451515, 19, 9, 15, 840)
    assert jdn_hms('2021-08-12T10:48') == (2459439, 10, 48, 0, 0)

    def jdn(string):
        """Extract only the Julian day part."""
        return jdn_hms(string)[0]

    assert jdn('January 1, 2000') == 2451545
    assert jdn('1 A.D. Jan 1') == 1721426
    assert jdn('November 24, 4714 B.C.') == 0

    # JDN is always in the Gregorian Calendar
    assert jdn('15 October, 1582') == 2299161  # First Gregorian day
    assert jdn('14 October, 1582') == 2299160  # This date should not exist
    assert jdn('5 October, 1582') == 2299151  # This date should not exist
    assert jdn('4 October, 1582') == 2299150  # Last Julian day


def test_spice_datetime_jd():
    """Test SPICE decimal julian date."""
    assert jd('2000-01-01T12:00:00') == 2_451_545.000
    assert jd('2023-08-10 00:05:09') == approx(2_460_166.503_576, abs=1e-6)
    assert jd('December 31, 1 BC') == 1_721_424.5
    assert jd('1 A.D. Jan 1') == 1_721_425.5


def test_spice_datetime_ymd():
    """Test SPICE datetime JDN to YMD convertor."""
    assert ymd(2446862) == (1987, 3, 7)
    assert ymd(2451515) == (1999, 12, 2)
    assert ymd(2459439) == (2021, 8, 12)

    assert ymd(2451545) == (2000, 1, 1)
    assert ymd(1721426) == (1, 1, 1)  # SPICE expected value
    assert ymd(0) == (-4713, 11, 24)  # SPICE expected value

    assert ymd(2299161) == (1582, 10, 15)  # First Gregorian day
    assert ymd(2299160) == (1582, 10, 14)  # This date should not exist
    assert ymd(2299151) == (1582, 10, 5)  # This date should not exist
    assert ymd(2299150) == (1582, 10, 4)  # Last Julian day (SPICE expected)


def test_spice_datetime_cleaner():
    """Test SPICE datetime string cleaner."""
    assert clean_time('2000-01-01T00:00:00.000001') == '2000-01-01T00:00:00.000001'
    assert clean_time('2000-01-01T00:00:00.001000') == '2000-01-01T00:00:00.001'
    assert clean_time('2000-01-01T00:00:00.100000') == '2000-01-01T00:00:00.1'
    assert clean_time('2000-01-01T00:00:01.000000') == '2000-01-01T00:00:01'
    assert clean_time('2000-01-01T00:00:10.000000') == '2000-01-01T00:00:10'
    assert clean_time('2000-01-01T00:01:00.000000') == '2000-01-01T00:01'
    assert clean_time('2000-01-01T01:00:00.000000') == '2000-01-01T01'
    assert clean_time('2000-01-02T00:00:00.000000') == '2000-01-02'
    assert clean_time('2000-02-01T00:00:00.000000') == '2000-02'
    assert clean_time('2000-01-01T00:00:00.000000') == '2000'
    assert clean_time('0001-01-01T00:00:00.000000') == '0001'
    assert clean_time('-001-01-01T00:00:00.000000') == '-001'


def test_spice_datetime_parser():
    """Test SPICE datetime parser."""
    sp.kclear()  # It does not require any kernel to be in the pool

    # ISO (T) formats
    assert datetime('1996-12-18T12:28:28') == dt('1996-12-18T12:28:28')
    assert datetime('1986-01-18T12') == dt('1986-01-18T12')
    assert datetime('1986-01-18T12:19') == dt('1986-01-18T12:19')
    assert datetime('1986-01-18T12:19:52.18') == dt('1986-01-18T12:19:52.18')
    assert datetime('1995-08T18:28:12') == dt('1995-01-08T18:28:12')  # DOY
    assert datetime('1995-18T') == dt('1995-01-18')  # DOY
    assert datetime('1996-12-18T12:28:28Z') == dt('1996-12-18T12:28:28')  # ISO + Z

    # Calendar formats
    assert datetime('Tue Aug  6 11:10:57  1996') == dt('1996-08-06T11:10:57')
    assert datetime('1 DEC 1997 12:28:29.192') == dt('1997-12-01T12:28:29.192')
    assert datetime('2/3/1996 17:18:12.002') == dt('1996-02-03T17:18:12.002')
    assert datetime('Mar 2 12:18:17.287 1993') == dt('1993-03-02T12:18:17.287')
    assert datetime('1992 11:18:28  3 Jul') == dt('1992-07-03T11:18:28')
    assert datetime('June 12, 1989 01:21') == dt('1989-06-12T01:21:00')
    assert datetime('1978/3/12 23:28:59.29') == dt('1978-03-12T23:28:59.29')
    assert datetime('17JUN1982 18:28:28') == dt('1982-06-17T18:28:28')
    assert datetime('13:28:28.128 1992 27 Jun') == dt('1992-06-27T13:28:28.128')
    assert datetime('1972 27 jun 12:29') == dt('1972-06-27T12:29:00')
    assert datetime("'93 Jan 23 12:29:47.289") == dt('1993-01-23T12:29:47.289')  # * 19xy
    assert datetime('27 Jan 3, 19:12:28.182') == dt('2027-01-03T19:12:28.182')  # * 19xy
    assert datetime('23 A.D. APR 4, 18:28:29.29') == dt('0023-04-04T18:28:29.29')
    assert datetime('18 B.C. Jun 3, 12:29:28.291') == dt('-017-06-03T12:29:28.291')
    assert datetime('29 Jun  30 12:29:29.298') == dt('2029-06-30T12:29:29.298')  # + case
    assert datetime("29 Jun '30 12:29:29.298") == dt('2030-06-29T12:29:29.298')  # * 20xy

    # Abbreviated years: 1969/2068 switch
    assert datetime("29 Jun '68 12:29:29.298") == dt('2068-06-29T12:29:29.298')  # * 20xy
    assert datetime("29 Jun '69 12:29:29.298") == dt('1969-06-29T12:29:29.298')  # * 19xy

    # Day of Year formats
    assert datetime('1997-162::12:18:28.827') == dt('1997-06-11T12:18:28.827')
    assert datetime('162-1996/12:28:28.287') == dt('1996-06-10T12:28:28.287')  # Leapyear
    assert datetime('1993-321/12:28:28.287') == dt('1993-11-17T12:28:28.287')
    assert datetime('1992 183// 12:18:19') == dt('1992-07-01T12:18:19')  # [Fixed]
    assert datetime('17:28:01.287 1992-272//') == dt('1992-09-28T17:28:01.287')
    assert datetime('17:28:01.282 272-1994//') == dt('1994-09-29T17:28:01.282')
    assert datetime("'92-271/ 12:28:30.291") == dt('1992-09-27T12:28:30.291')  # * 19xy
    assert datetime('92-182/ 18:28:28.281') == dt('1992-06-30T18:28:28.281')  # + case
    assert datetime('182-92/ 12:29:29.192') == dt('0182-04-02T12:29:29.192')  # + case
    assert datetime("182-'92/ 12:28:29.182") == dt('1992-06-30T12:28:29.182')  # * 19xy

    # # Julian Date strings
    assert datetime('2451515.2981 (JD)') == dt('1999-12-02T19:09:15.84')
    assert datetime('2451515.2981 JD ') == dt('1999-12-02T19:09:15.84')
    assert datetime('jd 28272.291') == dt('-4635-04-21T18:59:02.4')

    # Other examples
    assert datetime('31-JAN-1987') == dt('1987-01-31')
    assert datetime('feb/4/1987') == dt('1987-02-04')
    assert datetime('March-7-1987-3:10:39.221') == dt('1987-03-07T03:10:39.221')
    assert datetime('1999-12-02T19:09:15.84') == dt('1999-12-02T19:09:15.84')
    assert datetime('2021-08-12T10:48') == dt('2021-08-12T10:48')

    # MAPPS format (with `_`)
    assert datetime('2001-JAN-01_12:34:56.789') == dt('2001-01-01T12:34:56.789')

    # SPICE and Numpy are always expressed in Gregorian Calendar
    assert datetime('15 October, 1582') == dt('1582-10-15')  # First Gregorian day
    assert datetime('14 October, 1582') == dt('1582-10-14')  # This date should not exist
    assert datetime('5 October, 1582') == dt('1582-10-05')  # This date should not exist
    assert datetime('4 October, 1582') == dt('1582-10-04')  # Last Julian day

    # Calendar node values
    assert datetime('January 1, 2000') == dt('2000')
    assert datetime('1 A.D. Jan 1') == dt('0001-01-01')
    assert datetime('1 B.C. Jan 1') == dt('0000-01-01')
    assert datetime('2 B.C. Jan 1') == dt('-0001-01-01')
    assert datetime('November 24, 4714 B.C.') == dt('-4713-11-24')

    # With leading `@`
    assert datetime('@31-JAN-1987') == dt('1987-01-31')
    assert datetime('@feb/4/1987') == dt('1987-02-04')
    assert datetime('@March-7-1987-3:10:39.221') == dt('1987-03-07T03:10:39.221')

    # With numpy.datetime64 or datetime native
    assert datetime(dt('2010')) == dt('2010')
    assert datetime(dt_native(2022, 1, 2, 3)) == dt('2022-01-02T03')

    # NaT values
    assert np.isnat(datetime('NaT'))
    assert np.isnat(datetime('N/A'))
    assert np.isnat(datetime('Unable to determine'))
    assert np.isnat(datetime(dt('NaT')))

    # Multiple inputs
    times = datetime('31-JAN-1987', '@feb/4/1987', 'March-7-1987-3:10:39.221')
    assert len(times) == 3
    assert times[0] == dt('1987-01-31')
    assert times[1] == dt('1987-02-04')
    assert times[2] == dt('1987-03-07T03:10:39.221')

    # Input an iterable
    times = datetime(['31-JAN-1987', '@feb/4/1987', 'March-7-1987-3:10:39.221'])
    assert len(times) == 3
    assert times[0] == dt('1987-01-31')
    assert times[1] == dt('1987-02-04')
    assert times[2] == dt('1987-03-07T03:10:39.221')

    # Input as list of list
    times = datetime([['31-JAN-1987', '@feb/4/1987', 'March-7-1987-3:10:39.221']])
    assert len(times) == 1
    assert len(times[0]) == 3
    assert times[0][0] == dt('1987-01-31')
    assert times[0][1] == dt('1987-02-04')
    assert times[0][2] == dt('1987-03-07T03:10:39.221')

    # Input as list of tuples
    times = datetime([('31-JAN-1987', '@feb/4/1987'), ('March-7-1987-3:10:39.221',)])
    assert len(times) == 2
    assert len(times[0]) == 2
    assert len(times[1]) == 1
    assert times[0][0] == dt('1987-01-31')
    assert times[0][1] == dt('1987-02-04')
    assert times[1][0] == dt('1987-03-07T03:10:39.221')

    # Invalid input
    with raises(TypeError):
        _ = datetime(123)

    # Parsing error
    with raises(ValueError, match='^TPARSE does not support the specification'):
        _ = datetime('2022 SEP 04 TDB')


def test_spice_datetime_sorting():
    """Test datetime sorting."""
    # Simple list
    dates = [
        '2020-01-01T12:34:56.789',
        '2020-01-01T12:34:56Z',
        '2020-01-01T12',
        '2020-01-01',
        'January 1, 2000',
        dt('2010'),
        dt_native(2022, 1, 2, 3),
    ]

    assert sorted_datetimes(dates) == [
        'January 1, 2000',
        dt('2010'),
        '2020-01-01',
        '2020-01-01T12',
        '2020-01-01T12:34:56Z',  # tailing `Z` should be before the dot `.`
        '2020-01-01T12:34:56.789',
        dt_native(2022, 1, 2, 3),
    ]

    # Reverse the list
    assert sorted_datetimes(dates, reverse=True) == [
        dt_native(2022, 1, 2, 3),
        '2020-01-01T12:34:56.789',
        '2020-01-01T12:34:56Z',
        '2020-01-01T12',
        '2020-01-01',
        dt('2010'),
        'January 1, 2000',
    ]

    # List of list, with a known datetime index
    dates = [
        ['A', '2031-01-01'],
        ['B', '2032-01-01T01:23:45'],
        ['C', '2031-01-02'],
        ['D', '2032-01-01T12:34:56.7890'],
        ['E', '2032-01-01T12:34:56Z'],
        ['F', '2031-01-02T12'],
        ['G', '2031-01-01T12'],
        ['H', '2031-01-02'],
    ]

    assert sorted_datetimes(dates, index=1) == [
        ['A', '2031-01-01'],
        ['G', '2031-01-01T12'],
        ['C', '2031-01-02'],
        ['H', '2031-01-02'],  # Duplicate, keep original order
        ['F', '2031-01-02T12'],
        ['B', '2032-01-01T01:23:45'],
        ['E', '2032-01-01T12:34:56Z'],
        ['D', '2032-01-01T12:34:56.7890'],
    ]

    # List of list with multiple indexes
    dates = [
        ['A', '2031-01-01', '2031-01-01T12'],
        ['B', '2032-01-01T01:23:45', '2032-01-01T12:34:56.7890'],
        ['C', '2031-01-02', '2031-01-02T12'],
        ['D', '2032-01-01T01:23:45', '2032-01-01T12:34:56Z'],
    ]

    assert sorted_datetimes(dates, index=(1, 2)) == [
        ['A', '2031-01-01', '2031-01-01T12'],
        ['C', '2031-01-02', '2031-01-02T12'],
        ['D', '2032-01-01T01:23:45', '2032-01-01T12:34:56Z'],
        ['B', '2032-01-01T01:23:45', '2032-01-01T12:34:56.7890'],
    ]

    # Invalid index input
    with raises(TypeError):
        _ = sorted_datetimes(dates, index='foo')


def test_spice_datetime_iso():
    """Test datetime to ISO formatting."""
    assert iso('2020-01-01') == '2020-01-01T00:00:00Z'
    assert iso('2020-01-01T12') == '2020-01-01T12:00:00Z'
    assert iso('2020-01-01T12:34:56') == '2020-01-01T12:34:56Z'
    assert iso('2020-01-01T12:34:56.789') == '2020-01-01T12:34:56.789Z'
    assert iso('2020-01-01T12:34:56.789Z') == '2020-01-01T12:34:56.789Z'

    assert iso('January 1, 2000') == '2000-01-01T00:00:00Z'
    assert iso(dt('2010')) == '2010-01-01T00:00:00Z'
    assert iso(dt_native(2022, 1, 2, 3)) == '2022-01-02T03:00:00Z'

    with raises(ValueError):
        _ = iso('foo')


def test_spice_datetime_mapps():
    """Test datetime to MAPPS formatting."""
    assert mapps_datetime('2020-01-01T12:34:56.789Z') == '2020-JAN-01_12:34:56.789'
    assert mapps_datetime('2020-MAR-01_12:34:56') == '2020-MAR-01_12:34:56'


def test_spice_np_datetime_str():
    """Test numpy datetime to ISO str."""
    assert np_datetime_str(dt('2020-01-01')) == '2020-01-01T00:00:00'
    assert np_datetime_str(dt('2020-01-01T00:00:00')) == '2020-01-01T00:00:00'
    assert np_datetime_str(dt('2020-01-01T12:34:56')) == '2020-01-01T12:34:56'

    # Strip trailing zeros
    assert np_datetime_str(dt('2020-01-01T12:34:56.700')) == '2020-01-01T12:34:56.7'
    # Strip trailing dot
    assert np_datetime_str(dt('2020-01-01T12:34:56.000')) == '2020-01-01T12:34:56'

    assert np_datetime_str(dt('NaT')) == 'NaT'


def test_spice_np_date_str():
    """Test numpy datetime to date string."""
    assert np_date_str(dt('2020-01-01')) == '2020-01-01'
    assert np_date_str(dt('2020-01-01T00:00:00')) == '2020-01-01'
    assert np_date_str(dt('2020-01-01T12:34:56')) == '2020-01-01'
    assert np_date_str(dt('NaT')) == 'NaT'


def test_spice_timedelta():
    """Test timedelta inputs."""
    assert timedelta('1 s') == td(1, 's')
    assert timedelta('1 sec') == td(1, 's')
    assert timedelta('2 seconds') == td(2, 's')
    assert timedelta('1s') == td(1, 's')

    assert timedelta('1 m') == td(1, 'm')
    assert timedelta('1 min') == td(1, 'm')
    assert timedelta('1 minute') == td(1, 'm')

    assert timedelta('12 h') == td(12, 'h')
    assert timedelta('12 hour') == td(12, 'h')
    assert timedelta('12 H') == td(12, 'h')

    assert timedelta('1 D') == td(1, 'D')
    assert timedelta('1 d') == td(1, 'D')
    assert timedelta('2 days') == td(2, 'D')

    assert timedelta('1 M') == td(1, 'M')
    assert timedelta('1 month') == td(1, 'M')

    assert timedelta('1 Y') == td(1, 'Y')
    assert timedelta('1 y') == td(1, 'Y')
    assert timedelta('2 years') == td(2, 'Y')

    # Explicit ±HH:MM:SS timedelta string
    assert timedelta('00:00:00') == td(0, 's')
    assert timedelta('00:00:01') == td(1, 's')
    assert timedelta('+00:00:02') == td(2, 's')
    assert timedelta('-00:00:03') == td(-3, 's')
    assert timedelta('00:01:00') == td(1, 'm')
    assert timedelta('+00:02:00') == td(2, 'm')
    assert timedelta('-00:03:00') == td(-3, 'm')
    assert timedelta('01:00:00') == td(1, 'h')
    assert timedelta('+02:00:00') == td(2, 'h')
    assert timedelta('-03:00:00') == td(-3, 'h')
    assert timedelta('01:02:03') == td(1, 'h') + td(2, 'm') + td(3, 's')
    assert timedelta('+02:04:06') == td(2, 'h') + td(4, 'm') + td(6, 's')
    assert timedelta('-03:06:09') == -(td(3, 'h') + td(6, 'm') + td(9, 's'))

    assert timedelta('00:00:00.1') == td(100, 'ms')
    assert timedelta('00:00:00.001') == td(1, 'ms')

    # With timedelta input (no change)
    assert timedelta(td(1, 'Y')) == td(1, 'Y')

    # Invalid input (only int are accepted)
    with raises(ValueError):
        _ = timedelta('1.5 s')
