"""Test instrument field of view module."""

from pathlib import Path

import numpy as np
from numpy.testing import assert_array_almost_equal as assert_array

import matplotlib.pyplot as plt
from matplotlib.patches import Circle, Ellipse, Polygon

import spiceypy as sp

from pytest import raises

from planetary_coverage.spice.fov import SpiceFieldOfView


KERNELS = str((Path(__file__).parent / '..' / 'data').resolve())
ti = KERNELS + '/dummy.ti'


def test_fov_attr():
    """Test SPICE instrument field of view attributes."""
    sp.kclear()
    sp.furnsh(ti)

    # Circular FOV with explicit boundary vectors
    fov = SpiceFieldOfView(-11_111)

    assert fov.shape == 'CIRCLE'
    assert fov.frame == 'FRAME_FOR_INS-11111'

    assert tuple(fov.boresight) == (0, 0, 1)

    assert len(fov.bounds) == 1
    assert_array(fov.bounds, [[0, 1, 4]])
    assert_array(fov.corners, [[0], [0.25]])
    assert_array(fov.extent, [-0.25, 0.25, -0.25, 0.25])

    # Limits margins
    assert fov.MARGIN == 1.1
    assert fov.xlim == (-0.275, 0.275)
    assert fov.ylim == (-0.275, 0.275)

    # Circular FOV with half angular extents
    fov = SpiceFieldOfView(-11_112)

    assert fov.shape == 'CIRCLE'
    assert fov.frame == 'FRAME_FOR_INS-11112'

    assert tuple(fov.boresight) == (0, 0, 1)

    assert len(fov.bounds) == 1
    assert_array(fov.bounds, [[0, 0.24, 0.97]], decimal=2)
    assert_array(fov.corners, [[0], [0.25]])
    assert_array(fov.extent, [-0.25, 0.25, -0.25, 0.25])

    # Elliptical FOV with explicit boundary vectors
    fov = SpiceFieldOfView(-22_222)

    assert fov.shape == 'ELLIPSE'
    assert fov.frame == 'FRAME_FOR_INS-22222'

    assert tuple(fov.boresight) == (0, 0, 1)

    assert len(fov.bounds) == 2
    assert_array(fov.bounds, [[0, 1, 4], [-2, 0, 4]])
    assert_array(fov.corners, [[0, -0.5], [0.25, 0]])
    assert_array(fov.extent, [-0.5, 0.5, -0.25, 0.25])

    # Elliptical FOV with half angular extents
    fov = SpiceFieldOfView(-22_223)

    assert fov.shape == 'ELLIPSE'
    assert fov.frame == 'FRAME_FOR_INS-22223'

    assert tuple(fov.boresight) == (0, 0, 1)

    assert len(fov.bounds) == 2
    assert_array(fov.bounds, [[0, 0.24, 0.97], [-0.45, 0, 0.89]], decimal=2)
    assert_array(fov.corners, [[0, -0.5], [0.25, 0]])
    assert_array(fov.extent, [-0.5, 0.5, -0.25, 0.25])

    # Rectangular FOV with explicit boundary vectors
    fov = SpiceFieldOfView(-33_333)

    assert fov.shape == 'RECTANGLE'
    assert fov.frame == 'FRAME_FOR_INS-33333'

    assert tuple(fov.boresight) == (0, 0, 1)

    assert len(fov.bounds) == 4
    assert_array(fov.bounds, [[2, 1, 4], [-2, 1, 4], [-2, -1, 4], [2, -1, 4]])
    assert_array(fov.corners, [[0.5, -0.5, -0.5, 0.5], [0.25, 0.25, -0.25, -0.25]])
    assert_array(fov.extent, [-0.5, 0.5, -0.25, 0.25])

    # Rectangular FOV with half angular extents
    fov = SpiceFieldOfView(-33_334)

    assert fov.shape == 'RECTANGLE'
    assert fov.frame == 'FRAME_FOR_INS-33334'

    assert tuple(fov.boresight) == (0, 0, 1)

    assert len(fov.bounds) == 4
    assert_array(
        fov.bounds,
        [
            [-0.44, 0.22, 0.87],
            [-0.44, -0.22, 0.87],
            [0.44, -0.22, 0.87],
            [0.44, 0.22, 0.87],
        ],
        decimal=2,
    )
    assert_array(fov.corners, [[-0.5, -0.5, 0.5, 0.5], [0.25, -0.25, -0.25, 0.25]])
    assert_array(fov.extent, [-0.5, 0.5, -0.25, 0.25])

    # Polygonal FOV with explicit boundary vectors
    fov = SpiceFieldOfView(-44_444)

    assert fov.shape == 'POLYGON'
    assert fov.frame == 'FRAME_FOR_INS-44444'

    assert tuple(fov.boresight) == (0, 0, 1)

    assert len(fov.bounds) == 4
    assert_array(fov.bounds, [[1, 1, 4], [-1, 1, 4], [-2, -1, 4], [2, -1, 4]])
    assert_array(fov.corners, [[0.25, -0.25, -0.5, 0.5], [0.25, 0.25, -0.25, -0.25]])
    assert_array(fov.extent, [-0.5, 0.5, -0.25, 0.25])

    # Elliptical FOV with a non-z axis boresight
    fov = SpiceFieldOfView(-55_555)

    assert fov.shape == 'ELLIPSE'
    assert fov.frame == 'FRAME_FOR_INS-55555'

    assert tuple(fov.boresight) == (1, 0, 0)

    assert len(fov.bounds) == 2
    assert_array(fov.bounds, [[4, 0, 1], [4, -2, 0]])
    assert_array(fov.corners, [[-0.25, 0], [0, -0.5]])
    assert_array(fov.extent, [-0.25, 0.25, -0.5, 0.5])

    # Rectangle FOV with a non-z axis boresight
    fov = SpiceFieldOfView(-55_556)

    assert fov.shape == 'RECTANGLE'
    assert fov.frame == 'FRAME_FOR_INS-55556'

    assert tuple(fov.boresight) == (0, 1, 0)

    assert len(fov.bounds) == 4
    assert_array(fov.bounds, [[1, 4, 1], [-1, 4, 1], [-2, 4, -1], [2, 4, -1]])
    assert_array(fov.corners, [[-0.25, -0.25, 0.25, 0.25], [-0.25, 0.25, 0.5, -0.5]])
    assert_array(fov.extent, [-0.25, 0.25, -0.5, 0.5])

    # Polygon FOV with a non-axis boresight
    fov = SpiceFieldOfView(-66_666)

    assert fov.shape == 'POLYGON'
    assert fov.frame == 'FRAME_FOR_INS-66666'

    assert tuple(fov.boresight) == (1, 1, 1)

    assert len(fov.bounds) == 3
    assert_array(fov.bounds, [[1, 0, 0], [0, 1, 0], [0, 0, 1]])

    a, r = np.sqrt(3 / 2), np.sqrt(2)
    assert_array(fov.corners, [[r / 2, r / 2, -r], [-a, a, 0]], decimal=3)
    assert_array(fov.extent, [-r, r / 2, -a, a])

    # Invalid FOV
    with raises(ValueError):
        _ = SpiceFieldOfView(-1)


def test_fov_rotation_from_boresight():
    """Test FOV rotation from the boresight plane to the instrument frame."""
    sp.kclear()
    sp.furnsh(ti)

    # Invariant rotation with 1 corner
    fov = SpiceFieldOfView(-11_111)

    assert tuple(fov.boresight) == (0, 0, 1)

    assert_array(fov.from_boresight([0, 0]), [[0, 0, 1]])
    assert_array(fov.from_boresight([1, 0]), [[1, 0, 1]])
    assert_array(fov.from_boresight([0, 1]), [[0, 1, 1]])
    assert_array(fov.from_boresight([1, 1]), [[1, 1, 1]])
    assert_array(
        fov.from_boresight([0, 0], [1, 0], [0, 1], [1, 1]),
        [
            [0, 0, 1],
            [1, 0, 1],
            [0, 1, 1],
            [1, 1, 1],
        ],
    )

    assert_array(fov.from_boresight(*fov.corners.T), [[0, 0.25, 1]])

    # Invariant rotation with 2 corners
    fov = SpiceFieldOfView(-22_222)

    assert tuple(fov.boresight) == (0, 0, 1)

    assert_array(fov.from_boresight(*fov.corners.T), [[0, 0.25, 1], [-0.5, 0, 1]])

    # Invariant rotation with 4 corners
    fov = SpiceFieldOfView(-44_444)

    assert_array(
        fov.from_boresight(*fov.corners.T),
        [[0.25, 0.25, 1], [-0.25, 0.25, 1], [-0.5, -0.25, 1], [0.5, -0.25, 1]],
    )

    # With a x axis boresight
    fov = SpiceFieldOfView(-55_555)

    assert tuple(fov.boresight) == (1, 0, 0)

    assert_array(fov.from_boresight(*fov.corners.T), [[1, 0, 0.25], [1, -0.5, 0]])

    # With a y axis boresight
    fov = SpiceFieldOfView(-55_556)

    assert tuple(fov.boresight) == (0, 1, 0)

    assert_array(
        fov.from_boresight(*fov.corners.T),
        [[0.25, 1, 0.25], [-0.25, 1, 0.25], [-0.5, 1, -0.25], [0.5, 1, -0.25]],
    )

    # With a non-axis boresight
    fov = SpiceFieldOfView(-66_666)

    assert tuple(fov.boresight) == (1, 1, 1)

    a = np.sqrt(3)
    assert_array(
        fov.from_boresight(*fov.corners.T), [[a, 0, 0], [0, a, 0], [0, 0, a]], decimal=6
    )


def test_fov_rays_shapes():
    """Test SPICE fov rays shapes."""
    # Circle
    assert_array(
        SpiceFieldOfView.rays_circle(2, 4),  # (r, npt)
        [[2, 0], [0, 2], [-2, 0], [0, -2]],
    )

    # Ellipse
    assert_array(
        SpiceFieldOfView.rays_ellipse(1, 1, 0),  # (a, b, npt)
        [[1, 0], [0, 1]],  # pts on primary axes only (npt=0)
    )
    assert_array(
        SpiceFieldOfView.rays_ellipse(2, 3, 2),
        [[2, 0], [0, 3]],  # pts on primary axes only (npt=2)
    )
    assert_array(
        SpiceFieldOfView.rays_ellipse(1, 5, 4),  # (a, b, npt)
        [[1, 0], [0, 5], [-1, 0], [0, -5]],
    )

    # Rectangle/Polygon corners
    corners = [[-1, 1], [1, 1], [1, -1], [-1, -1]]

    # Corners only (no interpolation) -> 4 points
    assert_array(
        SpiceFieldOfView.rays_edges(corners, 0),
        [[-1, 1], [1, 1], [1, -1], [-1, -1]],
    )
    assert_array(
        SpiceFieldOfView.rays_edges(corners, 4),
        [[-1, 1], [1, 1], [1, -1], [-1, -1]],
    )

    # Corners + 1 point in the middle of each edge -> 8 points
    assert_array(
        SpiceFieldOfView.rays_edges(corners, 8),
        [[-1, 1], [0, 1], [1, 1], [1, 0], [1, -1], [0, -1], [-1, -1], [-1, 0]],
    )

    # Corners + 3 points per edge -> 16 points (16 to 19 should give the same result)
    assert_array(
        SpiceFieldOfView.rays_edges(corners, 19),
        [
            # Corner 1
            [-1, 1],
            # +3 pts
            [-0.5, 1],
            [0, 1],
            [0.5, 1],
            # Corner 2
            [1, 1],
            # +3 pts
            [1, 0.5],
            [1, 0],
            [1, -0.5],
            # Corner 3
            [1, -1],
            # +3 pts
            [0.5, -1],
            [0, -1],
            [-0.5, -1],
            # Corner 4
            [-1, -1],
            # +3 pts
            [-1, -0.5],
            [-1, 0],
            [-1, 0.5],
        ],
    )

    # Tuple => interpolation number of points **per edge** (with cycle)
    assert_array(
        SpiceFieldOfView.rays_edges(corners, (0,)),  # (no interpolation)
        [[-1, 1], [1, 1], [1, -1], [-1, -1]],
    )
    assert_array(
        SpiceFieldOfView.rays_edges(corners, (1,)),  # 1 pt per edge
        [[-1, 1], [0, 1], [1, 1], [1, 0], [1, -1], [0, -1], [-1, -1], [-1, 0]],
    )
    assert_array(
        SpiceFieldOfView.rays_edges(corners, (1, 0)),  # +1 pt every 2 edges
        [[-1, 1], [0, 1], [1, 1], [1, -1], [0, -1], [-1, -1]],
    )
    assert_array(
        SpiceFieldOfView.rays_edges(corners, (0, 1)),  # +1 pt ever 2 edges with 1 offset
        [[-1, 1], [1, 1], [1, 0], [1, -1], [-1, -1], [-1, 0]],
    )

    # Uneven number of points en each edge (without cycle)
    assert_array(
        SpiceFieldOfView.rays_edges(corners, (-1, 0, 1, 3)),
        [
            # Corner 1 (no interpolation)
            [-1, 1],
            # Corner 2 (no interpolation)
            [1, 1],
            # Corner 3
            [1, -1],
            # +1 pt
            [0, -1],
            # Corner 4
            [-1, -1],
            # +3 points
            [-1, -0.5],
            [-1, 0],
            [-1, 0.5],
        ],
    )


def test_fov_rays():
    """Test SPICE fov rays."""
    sp.kclear()
    sp.furnsh(ti)

    # Circular FOV with explicit boundary rays
    fov = SpiceFieldOfView(-11_111)

    assert len(fov.rays()) == 24

    assert_array(fov.rays(4), [[0.25, 0, 1], [0, 0.25, 1], [-0.25, 0, 1], [0, -0.25, 1]])
    assert_array(fov.rays(0), [[0.25, 0, 1]])  # 1 point minimum

    # Circular FOV with half angular extents
    fov = SpiceFieldOfView(-11_112)

    assert_array(fov.rays(4), [[0.25, 0, 1], [0, 0.25, 1], [-0.25, 0, 1], [0, -0.25, 1]])
    assert_array(fov.rays(0), [[0.25, 0, 1]])  # 1 point minimum

    with raises(TypeError):
        fov.rays((0,))  # tuple are not supported on CIRCLE shapes

    # Elliptical FOV
    fov = SpiceFieldOfView(-22_222)

    assert len(fov.rays()) == 24

    assert_array(fov.rays(4), [[0.5, 0, 1], [0, 0.25, 1], [-0.5, 0, 1], [0, -0.25, 1]])
    assert_array(fov.rays(0), [[0.5, 0, 1], [0, 0.25, 1]])  # 2 points minimum

    with raises(TypeError):
        fov.rays((0,))  # tuple are not supported on ELLIPSE shapes

    # Rectangular FOV
    fov = SpiceFieldOfView(-33_333)

    assert len(fov.rays()) == 24  # default

    assert_array(
        fov.rays(8),
        [
            [0.5, 0.25, 1],
            [0, 0.25, 1],
            [-0.5, 0.25, 1],
            [-0.5, 0, 1],
            [-0.5, -0.25, 1],
            [0, -0.25, 1],
            [0.5, -0.25, 1],
            [0.5, 0, 1],
        ],
    )

    assert_array(
        fov.rays(0), [[0.5, 0.25, 1], [-0.5, 0.25, 1], [-0.5, -0.25, 1], [0.5, -0.25, 1]]
    )  # 4 points minimum (corners only)

    assert_array(
        fov.rays((1, 0)),  # tuple
        [
            # Corner 1
            [0.5, 0.25, 1],
            # +1 pt
            [0, 0.25, 1],
            # Corner 2
            [-0.5, 0.25, 1],
            # +0 pt
            # Corner 3
            [-0.5, -0.25, 1],
            # +1 pt
            [0, -0.25, 1],
            # Corner 4
            [0.5, -0.25, 1],
            # +0 pt
        ],
    )

    # Polygonal FOV
    fov = SpiceFieldOfView(-44_444)

    assert len(fov.rays()) == 24

    assert_array(
        fov.rays(8),
        [
            [0.25, 0.25, 1],
            [0, 0.25, 1],
            [-0.25, 0.25, 1],
            [-0.375, 0, 1],
            [-0.5, -0.25, 1],
            [0, -0.25, 1],
            [0.5, -0.25, 1],
            [0.375, 0, 1],
        ],
    )

    assert_array(
        fov.rays(0),
        [[0.25, 0.25, 1], [-0.25, 0.25, 1], [-0.5, -0.25, 1], [0.5, -0.25, 1]],
    )

    assert_array(
        fov.rays((1, 0, 3)),  # tuple
        [
            # Corner 1
            [0.25, 0.25, 1],
            # +1 pt
            [0, 0.25, 1],
            # Corner 2
            [-0.25, 0.25, 1],
            # +0 pt
            # Corner 3
            [-0.5, -0.25, 1],
            # +3 pt
            [-0.25, -0.25, 1],
            [0, -0.25, 1],
            [0.25, -0.25, 1],
            # Corner 4
            [0.5, -0.25, 1],
            # +1 pt (cycled)
            [0.375, 0, 1],
        ],
    )

    # Elliptical FOV with a non-z axis boresight
    fov = SpiceFieldOfView(-55_555)

    assert_array(
        fov.rays(4),
        [
            [1, 0, -0.25],
            [1, 0.5, 0],
            [1, 0, 0.25],
            [1, -0.5, 0],
        ],
    )

    # Rectangle FOV with a non-z axis boresight
    fov = SpiceFieldOfView(-55_556)

    assert_array(
        fov.rays(0),
        [[0.25, 1, 0.25], [-0.25, 1, 0.25], [-0.5, 1, -0.25], [0.5, 1, -0.25]],
    )

    # Polygon FOV with a non-axis boresight
    fov = SpiceFieldOfView(-66_666)

    a = np.sqrt(3)
    assert_array(fov.rays(0), [[a, 0, 0], [0, a, 0], [0, 0, a]])

    assert_array(
        fov.rays(6),
        [
            [a, 0, 0],
            [a / 2, a / 2, 0],
            [0, a, 0],
            [0, a / 2, a / 2],
            [0, 0, a],
            [a / 2, 0, a / 2],
        ],
    )


def test_fov_contour():
    """Test SPICE fov contour."""
    sp.kclear()
    sp.furnsh(ti)

    # Circular FOV
    fov = SpiceFieldOfView(-11_111)

    assert np.shape(fov.contour()) == (3, 25)

    assert_array(
        fov.contour(0),
        [
            [0.25, 0.177, 0, -0.177, -0.25, -0.177, 0, 0.177, 0.25],
            [0, 0.177, 0.25, 0.177, 0, -0.177, -0.25, -0.177, 0],
            [1, 1, 1, 1, 1, 1, 1, 1, 1],
        ],
        decimal=3,
    )  # 9 points minimum

    # Elliptical FOV
    fov = SpiceFieldOfView(-22_222)

    assert np.shape(fov.contour()) == (3, 25)

    assert_array(
        fov.contour(0),
        [
            [0.5, 0.224, 0, -0.224, -0.5, -0.224, 0, 0.224, 0.5],
            [0, 0.224, 0.25, 0.224, 0, -0.224, -0.25, -0.224, 0],
            [1, 1, 1, 1, 1, 1, 1, 1, 1],
        ],
        decimal=3,
    )  # 9 points minimum

    # Rectangular FOV
    fov = SpiceFieldOfView(-33_333)

    assert np.shape(fov.contour()) == (3, 25)

    assert_array(
        fov.contour(0),
        [
            [0.5, 0, -0.5, -0.5, -0.5, 0, 0.5, 0.5, 0.5],
            [0.25, 0.25, 0.25, 0, -0.25, -0.25, -0.25, 0, 0.25],
            [1, 1, 1, 1, 1, 1, 1, 1, 1],
        ],
    )

    assert_array(
        fov.contour((0, 1)),  # with a tuple
        [
            # [C, C, P, C, C, P, C]
            [0.5, -0.5, -0.5, -0.5, 0.5, 0.5, 0.5],
            [0.25, 0.25, 0, -0.25, -0.25, 0, 0.25],
            [1, 1, 1, 1, 1, 1, 1],
        ],
    )

    # Polygonal FOV
    fov = SpiceFieldOfView(-44_444)

    assert np.shape(fov.contour()) == (3, 25)

    assert_array(
        fov.contour(9),
        [
            [0.25, 0, -0.25, -0.375, -0.5, 0, 0.5, 0.375, 0.25],
            [0.25, 0.25, 0.25, 0, -0.25, -0.25, -0.25, 0, 0.25],
            [1, 1, 1, 1, 1, 1, 1, 1, 1],
        ],
    )


def test_fov_patch():
    """Test SPICE fov patch."""
    sp.kclear()
    sp.furnsh(ti)

    # Circle
    patch = SpiceFieldOfView(-11_111).patch(facecolor='r', edgecolor='b')

    assert isinstance(patch, Circle)

    vertices = patch.get_path().vertices

    assert len(vertices) == 26
    assert_array(
        vertices[[0, 6, 12, 18, 25]], [[0, -1], [1, 0], [0, 1], [-1, 0], [0, -1]]
    )

    # Ellipse
    patch = SpiceFieldOfView(-22_222).patch(facecolor='r', edgecolor='b')

    assert isinstance(patch, Ellipse)

    vertices = patch.get_path().vertices

    assert len(vertices) == 26
    assert_array(
        vertices[[0, 6, 12, 18, 25]],
        [
            # Path before the Affine2D transform
            [0, -1],
            [1, 0],
            [0, 1],
            [-1, 0],
            [0, -1],
        ],
    )

    m = patch.get_patch_transform().get_matrix()

    assert_array(m, [[0.5, 0, 0], [0, 0.25, 0], [0, 0, 1]])

    # Rectangle
    patch = SpiceFieldOfView(-33_333).patch(facecolor='r', edgecolor='b')

    assert isinstance(patch, Polygon)

    vertices = patch.get_path().vertices

    assert len(vertices) == 4
    assert_array(vertices, [[0.5, 0.25], [-0.5, 0.25], [-0.5, -0.25], [0.5, -0.25]])

    # Polygon
    patch = SpiceFieldOfView(-44_444).patch(facecolor='r', edgecolor='b')

    assert isinstance(patch, Polygon)

    vertices = patch.get_path().vertices

    assert len(vertices) == 4
    assert_array(vertices, [[0.25, 0.25], [-0.25, 0.25], [-0.5, -0.25], [0.5, -0.25]])


def test_fov_display():
    """Test SPICE fov 2D and 3D display."""
    sp.kclear()
    sp.furnsh(ti)
    plt.close()

    # 2D
    assert SpiceFieldOfView(-11_111).view()
    assert SpiceFieldOfView(-22_222).view()
    assert SpiceFieldOfView(-33_333).view()
    assert SpiceFieldOfView(-44_444).view()
    assert SpiceFieldOfView(-55_555).view()
    assert SpiceFieldOfView(-55_556).view()
    assert SpiceFieldOfView(-66_666).view()

    # 3D
    assert SpiceFieldOfView(-11_111).display(show_bounds=True)
    assert SpiceFieldOfView(-22_222).display()
    assert SpiceFieldOfView(-33_333).display()
    assert SpiceFieldOfView(-44_444).display()

    _, axes = plt.subplots(1, 3, subplot_kw={'projection': '3d'})
    assert SpiceFieldOfView(-55_555).display(ax=axes[0])
    assert SpiceFieldOfView(-55_556).display(ax=axes[1])
    assert SpiceFieldOfView(-66_666).display(ax=axes[2])

    with raises(KeyError):
        _, ax = plt.subplots()
        _ = SpiceFieldOfView(-11_111).display(ax=ax)

    plt.close()
