"""Test kernel data."""

import datetime as dt
from pathlib import Path

import numpy as np

from pytest import raises

from planetary_coverage.spice import format_data
from planetary_coverage.spice.kernel import chunk_string, format_value


def test_chunk_string():
    """Test chunk method."""
    assert list(chunk_string('1234567890', 10)) == ["'1234567890'"]
    assert list(chunk_string('1234567890', 5)) == ["'123//'", "'456//'", "'7890'"]
    assert list(chunk_string('1234567890', length=6, continuation='+')) == [
        "'12345+'",
        "'67890'",
    ]


def test_format_value():
    """Test kernel format value."""
    assert format_value(-1) == '-1'
    assert format_value(1.2) == '1.2'
    assert format_value(1.657e-3) == '0.001657'
    assert format_value(1.657e-3, fmt='.3E') == '1.657E-03'
    assert format_value('FOO') == "'FOO'"
    assert format_value('FOO', fmt='^5') == "' FOO '"

    # Pathlib Path
    assert format_value(Path('foo.txt')) == "'foo.txt'"
    assert format_value('~') == f"'{Path.home()}'"

    # Datetime
    assert (
        format_value(np.datetime64('1987-03-07T03:10:39.221'))
        == '@1987-03-07T03:10:39.221'
    )
    assert (
        format_value(np.datetime64('1987-03-07T03:10:39.221'), fmt=r'%d-%b-%Y')
        == '@07-Mar-1987'
    )
    assert format_value(dt.date(1987, 3, 7)) == '@1987-03-07'
    assert (
        format_value(dt.datetime(1987, 3, 7, 3, 10, 39, 221000), fmt=r'%d-%b-%Y')
        == '@07-Mar-1987'
    )

    # Single value vector
    assert format_value((-1,)) == ['-1']
    assert format_value([1.2]) == ['1.2']
    assert format_value([1.657e-3], fmt='.3E') == ['1.657E-03']
    assert format_value(('FOO',), fmt='^5') == ["' FOO '"]
    assert format_value([np.datetime64('1987-03-07T03:10:39.221')]) == [
        '@1987-03-07T03:10:39.221'
    ]

    # Doubling quotes
    assert format_value(("You can't always get what you want.",)) == [
        "'You can''t always get what you want.'"
    ]

    # Multi value vector
    assert format_value((1, 3, 2)) == ['1', '3', '2']
    assert format_value(np.array([0.0, 0.0, 90.0])) == ['0.0', '0.0', '90.0']
    assert format_value(np.array([0.0, 1.657e-3]), fmt='.1E') == ['0.0E+00', '1.7E-03']
    assert format_value(['FOO', 'BAR']) == ["'FOO'", "'BAR'"]
    assert format_value(['FOO', 'BAR'], fmt='^5') == ["' FOO '", "' BAR '"]

    # Very long string (with continuation) - Must be a flat list
    assert format_value(80 * 'a') == f"'{80 * 'a'}'"
    assert format_value(78 * 'a' + 3 * 'b') == [f"'{78 * 'a'}//'", f"'{3 * 'b'}'"]

    assert format_value([80 * 'a']) == [f"'{80 * 'a'}'"]
    assert format_value([78 * 'a' + 3 * 'b']) == [f"'{78 * 'a'}//'", f"'{3 * 'b'}'"]
    assert format_value([80 * 'a', 3 * 'b']) == [f"'{80 * 'a'}'", f"'{3 * 'b'}'"]

    assert format_value([78 * 'a' + 3 * 'b', 'c']) == [
        f"'{78 * 'a'}//'",
        f"'{3 * 'b'}'",
        "'c'",
    ]

    assert format_value(['a', 78 * 'b' + 3 * 'c']) == [
        "'a'",
        f"'{78 * 'b'}//'",
        f"'{3 * 'c'}'",
    ]

    # Escaping values
    assert format_value(['foo/', '/bar', r'\baz', '\\qux\\']) == [
        "'foo/'",
        "'/bar'",
        "'\\baz'",
        "'\\qux\\'",
    ]


def test_format_data():
    """Test kernel format data."""
    assert format_data(single_int=-1, indent=0) == 'SINGLE_INT = -1'
    assert format_data(single_float=1.2) == '    SINGLE_FLOAT = 1.2'
    assert (
        format_data(single_float_d=1.657e-3, fmt='.3E')
        == '    SINGLE_FLOAT_D = 1.657E-03'
    )
    assert format_data(single_string='FOO', fmt='^5') == "    SINGLE_STRING = ' FOO '"
    assert (
        format_data(single_date=np.datetime64('1987-01-31'), fmt=r'%d-%b-%Y')
        == '    SINGLE_DATE = @31-Jan-1987'
    )

    # Vector with a single value
    assert format_data(
        vec_int_single=[1],
        vec_float_single=(0.0,),
        vec_str_single=["You can't always get what you want."],
    ) == (
        '    VEC_INT_SINGLE   = ( 1 )\n'
        '    VEC_FLOAT_SINGLE = ( 0.0 )\n'
        "    VEC_STR_SINGLE   = ( 'You can''t always get what you want.' )"
    )

    # Vector with multiple values
    assert format_data(
        vec_int_comma=(1, 3, 2),
        vec_float_comma=np.array([0.0, 0.0, 90.0]),
        vec_str_comma=['FOO', 'BAR'],
        indent=2,
    ) == (
        '  VEC_INT_COMMA   = ( 1, 3, 2 )\n'
        '  VEC_FLOAT_COMMA = ( 0.0, 0.0, 90.0 )\n'
        "  VEC_STR_COMMA   = ( 'FOO', 'BAR' )"
    )

    # Change separator in vector
    assert format_data(
        arr_int_space=(4, 5, 6),
        arr_float_space=np.array([1.0, 10.0, -25.0]),
        arr_str_space=['BAZ', 'QUX'],
        indent=0,
        sep=' ',
    ) == (
        'ARR_INT_SPACE   = ( 4 5 6 )\n'
        'ARR_FLOAT_SPACE = ( 1.0 10.0 -25.0 )\n'
        "ARR_STR_SPACE   = ( 'BAZ' 'QUX' )"
    )

    # Multi lines
    assert format_data(
        VEC_FLOAT_MULTI_LINES=np.arange(0, 100, 2),
        indent=10,
    ) == (
        f'{10 * " "}VEC_FLOAT_MULTI_LINES = ( '
        '0, 2, 4, 6, 8, 10, 12, 14, 16, 18, 20, 22, 24, '
        '26, 28, 30, 32, 34, 36, 38, 40, 42, 44, 46, 48,\n'
        f'{36 * " "}50, 52, 54, 56, 58, 60, 62, 64, 66, 68, '
        '70, 72, 74, 76, 78, 80, 82, 84, 86, 88, 90, 92, 94,\n'
        f'{36 * " "}96, 98{89 * " "})'
    )

    # Continued string
    long_string = 80 * 'a'
    very_long_string = 78 * 'a' + 3 * 'b'
    very_very_long_string = 78 * 'a' + 50 * 'b'

    s1, s2, s3 = f'{78 * "a"}//', 'bbb', 50 * 'b'

    assert format_data(
        long_string=long_string,
        very_long_string=very_long_string,
        very_very_long_string=very_very_long_string,
        vec_long_string=[long_string],
        vec_very_long_string=[very_long_string],
        vec_very_very_long_string=[very_very_long_string],
    ) == (
        f"    LONG_STRING               = '{long_string}'\n"
        f"    VERY_LONG_STRING          = ( '{s1}', '{s2}' )\n"
        f"    VERY_VERY_LONG_STRING     = ( '{s1}',\n"
        f"                                  '{s3}'{32 * ' '})\n"
        f"    VEC_LONG_STRING           = ( '{long_string}' )\n"
        f"    VEC_VERY_LONG_STRING      = ( '{s1}', '{s2}' )\n"
        f"    VEC_VERY_VERY_LONG_STRING = ( '{s1}',\n"
        f"                                  '{s3}'{32 * ' '})"
    )

    # Continued string fro kernels to load
    long_path = 120 * 'a'
    long_kernel = 120 * 'b'

    assert format_data(
        long_path=[long_path, long_kernel],
        path_values=[long_path],
        kernels_to_load=[long_kernel],
    ) == (
        f"    LONG_PATH       = ( '{78 * 'a'}//',\n"
        f"                        '{42 * 'a'}',\n"
        f"                        '{78 * 'b'}//',\n"
        f"                        '{42 * 'b'}'{40 * ' '})\n"
        f"    PATH_VALUES     = ( '{79 * 'a'}+',\n"
        f"                        '{41 * 'a'}'{41 * ' '})\n"
        f"    KERNELS_TO_LOAD = ( '{79 * 'b'}+',\n"
        f"                        '{41 * 'b'}'{41 * ' '})"
    )

    # Assignation formatting
    assert format_data(
        assign_value=1,
        assign_list=[2, 3, 4],
        eq='+=',
    ) == ('    ASSIGN_VALUE += 1\n    ASSIGN_LIST  += ( 2, 3, 4 )')

    # Escaping values
    assert format_data(escaped=['foo/', '/bar', r'\baz', '\\qux\\']) == (
        "    ESCAPED = ( 'foo/', '/bar', '\\baz', '\\qux\\' )"
    )

    # Tilde character
    assert format_data(home='~/foo') == f"    HOME = '{Path.home()}/foo'"

    # Test errors - SPICE constrains violations
    with raises(KeyError):
        _ = format_data(**{33 * 'a': 'foo'})

    assert format_data(foo='bar', indent=40) == 40 * ' ' + "FOO = 'bar'"

    with raises(ValueError):
        _ = format_data(foo='bar', indent=41)

    assert format_data(foo='bar', indent=0, sep=42 * ',') == "FOO = 'bar'"

    with raises(ValueError):
        _ = format_data(foo='bar', indent=0, sep=43 * ',')
