"""Test kernel parser."""

from pathlib import Path

import numpy as np
from numpy.testing import assert_array_equal

from pytest import raises

from planetary_coverage.spice import kernel_parser
from planetary_coverage.spice.kernel import (
    CONTINUATION,
    concatenate,
    get_data,
    get_item,
    parse,
    read,
)
from planetary_coverage.spice.pool import SpicePool


KERNELS = str((Path(__file__).parent / '..' / 'data').resolve())


def test_continuation_character():
    """Test value continuation character based on its key."""
    assert CONTINUATION['PATH_VALUES'] == '+'
    assert CONTINUATION['KERNELS_TO_LOAD'] == '+'
    assert CONTINUATION['ANY_KEY'] == '//'


def test_read_single():
    """Test single value parsing."""
    # int
    assert read('-1') == -1
    assert isinstance(read('-1'), int)

    # float
    assert read('+1.2') == 1.2
    assert isinstance(read('+1.2'), float)

    # scientific/engineering notation
    assert read('1.657e-3') == 1.657e-3
    assert read('1.657E-3') == 1.657e-3
    assert read('1.657D-3') == 1.657e-3

    # date
    assert read('@31-JAN-1987') == np.datetime64('1987-01-31')
    assert read('@feb/4/1987') == np.datetime64('1987-02-04')
    assert read('@March-7-1987-3:10:39.221') == np.datetime64('1987-03-07T03:10:39.221')

    # string
    assert read("'FOO'") == 'FOO'
    assert read("'Sentence with space and dot.'") == 'Sentence with space and dot.'
    assert read(" 'leading space'") == 'leading space'
    assert read("'trailing space' ") == 'trailing space'
    assert read("'trailing comma',") == 'trailing comma'

    # string with special characters
    assert read("'Here''s a single quote'") == "Here's a single quote"
    assert read("'Continued sting //'") == 'Continued sting //'
    assert read("'Continued sting with space // '") == 'Continued sting with space //'
    assert read("'Filename_continuation+'") == 'Filename_continuation+'
    assert read("'equal = sign'") == 'equal = sign'


def test_read_vec():
    """Test multi value parsing."""
    # Single value
    assert read('(  -1 )') == [-1]
    assert read("( 'FOO' )") == 'FOO'
    assert read('(') == []
    assert read("( 'FOO = BAR' )") == 'FOO = BAR'

    # Commas separator
    assert read('(   1,      3,      2 )') == [1, 3, 2]
    assert read("( 'FOO', 'BAR' )") == ['FOO', 'BAR']

    # Space separator
    assert read('(   4       5       6 )') == [4, 5, 6]
    assert read("( 'BAZ' 'QUX' )") == ['BAZ', 'QUX']

    # Open/close with value
    assert read(" ( 'This //  ',") == 'This //'
    assert read(" 'is //  ',    ") == 'is //'
    assert read(" 'string.'    )") == 'string.'

    assert read(" ( 'KILOMETERS','SECONDS'") == ['KILOMETERS', 'SECONDS']
    assert read("   'KILOMETERS/SECOND'  )") == 'KILOMETERS/SECOND'

    assert read(' 6 7 ') == [6, 7]
    assert read(' 8, @31-JAN-1987 ') == [8, np.datetime64('1987-01-31')]

    # Open/close without value
    assert read('') is None
    assert read(')') is None


def test_parse_single():
    """Test single value parser."""
    assert parse(' SINGLE_INT=-1        ') == ('SINGLE_INT', -1)
    assert parse('SINGLE_FLOAT   = +1.2 ') == ('SINGLE_FLOAT', 1.2)
    assert parse('SINGLE_FLOAT_D = 1.657D-3') == ('SINGLE_FLOAT_D', 1.657e-3)
    assert parse("SINGLE_STRING  = 'BAR'") == ('SINGLE_STRING', 'BAR')
    assert parse('SINGLE_DATE  = @31-JAN-1987') == (
        'SINGLE_DATE',
        np.datetime64('1987-01-31'),
    )

    # Special characters
    assert parse('SINGLE/SLASH = 32.184') == ('SINGLE/SLASH', 32.184)

    # Assignment
    assert parse("SINGLE_APPEND += 'FOO'") == ('SINGLE_APPEND +', 'FOO')
    assert parse('SINGLE_APPEND += -1') == ('SINGLE_APPEND +', -1)


def test_parse_vector():
    """Test vector values parser."""
    # Single value
    assert parse('VEC_INT_SINGLE  = ( 1 )') == ('VEC_INT_SINGLE', [1])
    assert parse('VEC_FLOAT_SINGLE  = ( 0.0 )') == ('VEC_FLOAT_SINGLE', [0.0])
    assert parse("VEC_STR_SINGLE  = ('You can''t always get what you want.')") == (
        'VEC_STR_SINGLE',
        ["You can't always get what you want."],
    )

    # Comma separator
    assert parse('VEC_INT_COMMA   = ( 1, 3, 2 )') == ('VEC_INT_COMMA', [1, 3, 2])
    assert parse('VEC_FLOAT_COMMA = (0.0, 0.0, 90.0)') == (
        'VEC_FLOAT_COMMA',
        [0.0, 0.0, 90.0],
    )

    assert parse("VEC_STR_COMMA   = ('FOO', 'BAR')") == ('VEC_STR_COMMA', ['FOO', 'BAR'])

    # Space separator
    assert parse('ARR_INT_SPACE   = (   4       5       6 )') == (
        'ARR_INT_SPACE',
        [4, 5, 6],
    )

    assert parse('ARR_FLOAT_SPACE = ( 1.0   +10.0   -25.0 )') == (
        'ARR_FLOAT_SPACE',
        [1.0, 10.0, -25.0],
    )

    assert parse("ARR_STR_SPACE   = ( 'BAZ'  'QUX' )") == (
        'ARR_STR_SPACE',
        ['BAZ', 'QUX'],
    )


def test_parse_multi():
    """Test multiline parser."""
    assert parse('VEC_FLOAT_MULTI_LINES = (') == ('VEC_FLOAT_MULTI_LINES', [])
    assert parse('    -1.2') == (None, -1.2)
    assert parse('    ') == (None, None)
    assert parse("    6 'FOO'") == (None, [6, 'FOO'])
    assert parse('    7, @31-JAN-1987') == (None, [7, np.datetime64('1987-01-31')])
    assert parse(' ) ') == (None, None)

    assert parse("    VEC_STR_MULTI_LINES = ( 'KILOMETERS','SECONDS'") == (
        'VEC_STR_MULTI_LINES',
        ['KILOMETERS', 'SECONDS'],
    )

    assert parse("                            'KILOMETERS/SECOND' )") == (
        None,
        'KILOMETERS/SECOND',
    )

    assert parse("CONTINUED_STRINGS = ( 'This // ',") == (
        'CONTINUED_STRINGS',
        ['This //'],
    )

    assert parse("                      'is //  ', ") == (None, 'is //')
    assert parse("                      'string.' )") == (None, 'string.')

    assert parse("FOO = ( 'BAR = BAZ'   ") == ('FOO', ['BAR = BAZ'])
    assert parse("        'QUX = QUUX'  ") == (None, 'QUX = QUUX')
    assert parse("        'CORGE = GRAULT' )") == (None, 'CORGE = GRAULT')


def test_concatenate_str():
    """Test string concatenation."""
    assert concatenate('SINGLE_INT', 1) == 1
    assert concatenate('VEC_INT_SINGLE', [1]) == [1]
    assert concatenate('VEC_INT', [1, 3, 2]) == [1, 3, 2]

    assert concatenate('SINGLE_STRING', 'FOO') == 'FOO'
    assert concatenate('VEC_STR_SINGLE', ["You can't always get what you want."]) == [
        "You can't always get what you want."
    ]

    # Concatenate stings
    assert concatenate(
        'CONTINUED_STRINGS',
        [
            'This //',
            'is //',
            'just //',
            'one long //',
            'string.',
            "Here's a second //",
            'continued //',
            'string.',
        ],
    ) == [
        'This is just one long string.',
        "Here's a second continued string.",
    ]

    assert concatenate(
        'CONTINUED_STRINGS',
        [
            'This +',
            'is +',
            'just +',
            'one long +',
            'string.',
            "Here's a second +",
            'continued +',
            'string.',
        ],
    ) == [
        'This +',
        'is +',
        'just +',
        'one long +',
        'string.',
        "Here's a second +",
        'continued +',
        'string.',
    ]

    # Special case for KERNELS_TO_LOAD and PATH_VALUES (continuation with ``+``).
    assert concatenate(
        'KERNELS_TO_LOAD',
        [
            '$KERNELS/ck/example.bc',
            '$KERNELS/fk/+',
            'example.tf',
        ],
    ) == [
        '$KERNELS/ck/example.bc',
        '$KERNELS/fk/example.tf',
    ]

    assert concatenate(
        'PATH_VALUES',
        [
            '$KERNELS/ck/example.bc',
            '$KERNELS/fk//',
            'example.tf',
        ],
    ) == [
        '$KERNELS/ck/example.bc',
        '$KERNELS/fk//',
        'example.tf',
    ]


def test_get_data():
    """Test kernel data getter."""
    # Empty value
    empty = r'\begindata'
    assert get_data(empty) == {}

    # Single value
    mk = r"""\begindata
SINGLE_INT     = -1
SINGLE_FLOAT   = +1.2
SINGLE_FLOAT_D = 1.657D-3
SINGLE_STRING  = 'FOO'
SINGLE_DATE    = @31-JAN-1987
SINGLE/SLASH   = 32.184
"""
    assert get_data(mk) == {
        'SINGLE_INT': -1,
        'SINGLE_FLOAT': 1.2,
        'SINGLE_FLOAT_D': 1.657e-3,
        'SINGLE_STRING': 'FOO',
        'SINGLE_DATE': np.datetime64('1987-01-31'),
        'SINGLE/SLASH': 32.184,
    }

    # Vector with single value
    mk = r"""\begindata
VEC_INT_SINGLE   = (   1 )
VEC_FLOAT_SINGLE = ( 0.0 )
VEC_STR_SINGLE   = ( 'You can''t always get what you want.' )
"""
    assert get_data(mk) == {
        'VEC_INT_SINGLE': [1],
        'VEC_FLOAT_SINGLE': [0.0],
        'VEC_STR_SINGLE': ["You can't always get what you want."],
    }

    # Comma separator
    mk = r"""\begindata
VEC_INT_COMMA   = (   1,      3,      2 )
VEC_FLOAT_COMMA = ( 0.0,    0.0,   90.0 )
VEC_STR_COMMA   = ( 'FOO', 'BAR' )
"""
    assert get_data(mk) == {
        'VEC_INT_COMMA': [1, 3, 2],
        'VEC_FLOAT_COMMA': [0.0, 0.0, 90.0],
        'VEC_STR_COMMA': ['FOO', 'BAR'],
    }

    # Space separator
    mk = r"""\begindata
ARR_INT_SPACE   = (   4       5       6 )
ARR_FLOAT_SPACE = ( 1.0   +10.0   -25.0 )
ARR_STR_SPACE   = ( 'BAZ'  'QUX' )
"""
    assert get_data(mk) == {
        'ARR_INT_SPACE': [4, 5, 6],
        'ARR_FLOAT_SPACE': [1.0, 10.0, -25.0],
        'ARR_STR_SPACE': ['BAZ', 'QUX'],
    }

    # Multi-lines
    mk = r"""\begindata
VEC_FLOAT_MULTI_LINES = (

    -1.2

    3.4
    +4.5
)

VEC_STR_MULTI_LINES = ( 'KILOMETERS','SECONDS'
                        'KILOMETERS/SECOND' )

VEC_STR_MULTI_LINES_2 = ( 'KILOMETERS'
                          'SECONDS', 'KILOMETERS/SECOND' )

VEC_STR_MULTI_LINES_3 = ( 'ANGLES    = DEGREES'
                          'DISTANCES = KILOMETERS' )

"""
    assert get_data(mk) == {
        'VEC_FLOAT_MULTI_LINES': [-1.2, 3.4, 4.5],
        'VEC_STR_MULTI_LINES': ['KILOMETERS', 'SECONDS', 'KILOMETERS/SECOND'],
        'VEC_STR_MULTI_LINES_2': ['KILOMETERS', 'SECONDS', 'KILOMETERS/SECOND'],
        'VEC_STR_MULTI_LINES_3': ['ANGLES    = DEGREES', 'DISTANCES = KILOMETERS'],
    }

    # Continued string in a generic key
    mk = r"""\begindata
CONTINUED_STRINGS = ( 'This //  ',
                      'is //  ',
                      'just //',
                      'one long //',
                      'string.',
                      'Here''s a second //',
                      'continued //'
                      'string.'              )
"""
    assert get_data(mk) == {
        'CONTINUED_STRINGS': [
            'This is just one long string.',
            "Here's a second continued string.",
        ]
    }

    # Assignment value
    mk = r"""\begindata
ASSIGN_VALUE  = -3.878
ASSIGN_VALUE += -0.120
ASSIGN_VALUE += +0.070
ASSIGN_VALUE += -0.017
ASSIGN_VALUE += 0.

ASSIGN_EMPTY += 1
"""
    assert get_data(mk) == {
        'ASSIGN_VALUE': [-3.878, -0.120, 0.070, -0.017, 0],
        'ASSIGN_EMPTY': [1],
    }

    # Concatenate mk value
    mk = r"""\begindata
PATH_VALUES = ( '/a/very/long/+'
                'path' )
"""
    assert get_data(mk) == {
        'PATH_VALUES': ['/a/very/long/path'],
    }

    # Concatenate filenames
    mk = r"""\begindata
KERNELS_TO_LOAD   = (

    '$KERNELS/ck/example.bc'

    '$KERNELS/fk/+'
    'example.tf'
)
"""
    assert get_data(mk) == {
        'KERNELS_TO_LOAD': ['$KERNELS/ck/example.bc', '$KERNELS/fk/example.tf'],
    }


def test_kernel_parser():
    """Test kernel parser."""
    fname = Path(KERNELS + '/metakernel.tm')

    content, data = kernel_parser(fname)

    assert content == fname.read_text(encoding='utf-8')

    assert data == {
        'SINGLE_INT': -1,
        'SINGLE_FLOAT': 1.2,
        'SINGLE_FLOAT_D': 1.657e-3,
        'SINGLE_STRING': 'FOO',
        'SINGLE_DATE': np.datetime64('1987-01-31'),
        'SINGLE/SLASH': 32.184,
        'VEC_INT_SINGLE': [1],
        'VEC_FLOAT_SINGLE': [0.0],
        'VEC_STR_SINGLE': ["You can't always get what you want."],
        'VEC_INT_COMMA': [1, 3, 2],
        'VEC_FLOAT_COMMA': [0.0, 0.0, 90.0],
        'VEC_STR_COMMA': ['FOO', 'BAR'],
        'ARR_INT_SPACE': [4, 5, 6],
        'ARR_FLOAT_SPACE': [1.0, 10.0, -25.0],
        'ARR_STR_SPACE': ['BAZ', 'QUX'],
        'VEC_FLOAT_MULTI_LINES': [-1.2, 3.4, 4.5, 6, 7, 8, np.datetime64('1987-01-31')],
        'VEC_STR_MULTI_LINES': ['KILOMETERS', 'SECONDS', 'KILOMETERS/SECOND'],
        'VEC_STR_MULTI_LINES_2': ['KILOMETERS', 'SECONDS', 'KILOMETERS/SECOND'],
        'VEC_STR_MULTI_LINES_3': ['ANGLES    = DEGREES', 'DISTANCES = KILOMETERS'],
        'CONTINUED_STRINGS': [
            'This is just one long string.',
            "Here's a second continued string.",
        ],
        'ASSIGN_VALUE': [-3.878, -0.120, 0.070, -0.017, 0.0],
        'ASSIGN_EMPTY': [1],
        'PATH_VALUES': ['.'],
        'PATH_SYMBOLS': ['KERNELS'],
        'KERNELS_TO_LOAD': ['$KERNELS/naif0012.tls', '$KERNELS/dummy.tf'],
        'SKD_VERSION': 'v999_20210531_001',
        'MK_IDENTIFIER': 'metakernel_v999_20210531_001',
    }


def test_kernel_item_getter():
    """Test SPICE Pool item getter."""
    SpicePool.add(KERNELS + '/dummy.ti', purge=True)

    assert get_item('INS-28200_PIXEL_SAMPLES') == 1504
    assert get_item('INS-28200_FOV_FRAME') == 'JUICE_JANUS'

    assert_array_equal(get_item('INS-28200_CCD_CENTER'), [1000.5, 752.5])

    # Change chunk size
    SpicePool.add(KERNELS + '/events.tf', purge=True)

    assert get_item('JUICE_MISSION_PHASE_NAME', chunk_size=2) == [
        'Jupiter_Phase_1',  # 1st chunk
        'Jupiter_Phase_2',  # 1st chunk
        'Jupiter_Phase_3',  # 2nd chunk / 2 elements -> extend
        'Jupiter_Phase_4',  # 2nd chunk
        'Jupiter_Phase_5',  # 3rd chunk / 1 element -> append
    ]

    # Wildcard
    assert get_item('*', chunk_size=3) == [  # sorted
        'JUICE_MISSION_PHASE_DESC',
        'JUICE_MISSION_PHASE_NAME',
        'JUICE_MISSION_PHASE_STOP',
        'JUICE_MISSION_PHASE_STRT',
        'JUICE_TIMELINE_EVENT_NAME',
        'JUICE_TIMELINE_EVENT_TIME',
        'JUICE_TIMELINE_EVENT_TYPE',
    ]

    assert get_item('JUICE_MISSION_PHASE_*') == [  # sorted
        'JUICE_MISSION_PHASE_DESC',
        'JUICE_MISSION_PHASE_NAME',
        'JUICE_MISSION_PHASE_STOP',
        'JUICE_MISSION_PHASE_STRT',
    ]

    assert get_item('JUICE_MISSION_PHASE_ST%%') == [  # sorted
        'JUICE_MISSION_PHASE_STOP',
        'JUICE_MISSION_PHASE_STRT',
    ]

    # Invalid key
    with raises(KeyError, match=r'No item matching `DUMMY_\*` found in the kernel pool.'):
        _ = get_item('DUMMY_*')

    # Invalid key
    with raises(KeyError, match='`DUMMY` was not found in the kernel pool.'):
        _ = get_item('DUMMY')

    # Purge the pool - No key available
    SpicePool.purge()

    with raises(KeyError):
        _ = get_item('INS-28200_PIXEL_SAMPLES')
