"""Test MetaKernel object."""

import re
from pathlib import Path

from pytest import fixture, raises

import planetary_coverage.spice.metakernel as _metakernel
from planetary_coverage import MetaKernel
from planetary_coverage.html import display_html
from planetary_coverage.spice import SpicePool, debug_spice_pool
from planetary_coverage.spice.metakernel import (
    MissingKernelError,
    MissingKernelsRemoteError,
)


KERNELS = str((Path(__file__).parent / '..' / 'data').resolve())


@fixture
def mk():
    """Default metakernel."""
    return MetaKernel(KERNELS + '/metakernel.tm', kernels=KERNELS)


@fixture(name='mock_mk_content')
def fixture_mk_content(monkeypatch):
    """Metakernel content parser fixture."""

    def mk_content(content, data):
        """Mock kernel parser content and data."""
        monkeypatch.setattr(_metakernel, 'kernel_parser', lambda _: (content, data))

    return mk_content


def test_metakernel_attr(mk):
    """Test MetaKernel attributes."""
    # __eq__
    assert mk == MetaKernel(KERNELS + '/////metakernel.tm', kernels=KERNELS)
    assert mk != MetaKernel(KERNELS + '/metakernel.tm', kernels=KERNELS + '/')
    assert mk != 'metakernel.tm'

    assert str(mk) == 'metakernel.tm'

    assert mk.remote == 'ftp://example.com/kernels'
    assert isinstance(mk.n_threads, int)

    # Context manager: __enter__/__exit__/create a temporary file, exists only in the loop
    with mk as tmp_mk:
        # Check if a temporary file was created
        _mk = Path(tmp_mk)

        assert _mk.exists()
        assert _mk.name.startswith('metakernel-')
        assert _mk.name.endswith('.tm')

        # Check if the file have the edited values
        content = _mk.read_text(encoding='utf-8')

        assert KERNELS[:79] in content  # Check only the beginning of kernels path
        assert '$KERNELS/naif0012.tls' in content
        assert 'Dummy text' not in content

    assert not _mk.exists()

    # __iter__ (create a temporary file, exists only in the loop)
    for tmp_mk in mk:
        _mk = Path(tmp_mk)
        assert _mk.exists()

    assert not _mk.exists()

    # Missing file
    err = f'{Path.home()}/foo.tm'  # expanded user path
    with raises(FileNotFoundError, match=err):
        _ = MetaKernel('~/foo.tm')


def test_metakernel_content(mk):
    """Test MetaKernel content."""
    assert 'KPL/MK' in mk.content
    assert 'Dummy kernel' in mk.content
    assert r'\begintext' in mk.content
    assert r'\begindata' in mk.content
    assert 'End of the File' in mk.content


def test_metakernel_data(mk):
    """Test MetaKernel data."""
    assert mk.data['PATH_VALUES'] == [str(KERNELS)]
    assert mk.data['PATH_SYMBOLS'] == ['KERNELS']
    assert mk.data['KERNELS_TO_LOAD'] == ['$KERNELS/naif0012.tls', '$KERNELS/dummy.tf']


def test_metakernel_remote(mk):
    """Test MetaKernel remote."""
    assert mk.remote == 'ftp://example.com/kernels'
    assert mk.urls == [
        'ftp://example.com/kernels/naif0012.tls',
        'ftp://example.com/kernels/dummy.tf',
    ]

    # Change remote
    mk.remote = 'https://example.com/kernels'
    assert mk.remote == 'https://example.com/kernels'
    assert mk.urls == [
        'https://example.com/kernels/naif0012.tls',
        'https://example.com/kernels/dummy.tf',
    ]

    # Try to load the remote from the content
    mk.remote = 0
    assert mk.remote == 'ftp://example.com/kernels'
    assert mk.urls == [
        'ftp://example.com/kernels/naif0012.tls',
        'ftp://example.com/kernels/dummy.tf',
    ]


def test_metakernel_path_values(monkeypatch, mock_mk_content):
    """Test MetaKernel kernels."""

    def mock_mk_check(*args, **kwargs):
        """Mock metakernel check function."""

    monkeypatch.setattr(MetaKernel, 'check', mock_mk_check)

    # Default content
    mk = MetaKernel(KERNELS + '/metakernel.tm')
    assert "PATH_VALUES = ( '.' )" in mk.content

    # Posix path
    mk = MetaKernel(KERNELS + '/metakernel.tm', kernels='foo/')  # trailing /
    assert "PATH_VALUES = ( 'foo/' )" in mk.content

    mk = MetaKernel(KERNELS + '/metakernel.tm', kernels='./bar')
    assert "PATH_VALUES = ( './bar' )" in mk.content

    mk = MetaKernel(KERNELS + '/metakernel.tm', kernels='../baz')  # parent path
    assert "PATH_VALUES = ( '../baz' )" in mk.content

    mk = MetaKernel(KERNELS + '/metakernel.tm', kernels='/qux')  # absolute path
    assert "PATH_VALUES = ( '/qux' )" in mk.content

    # Windows path
    mk = MetaKernel(KERNELS + '/metakernel.tm', kernels=r'.\foo')  # relative r-string
    assert "PATH_VALUES = ( '.\\foo' )" in mk.content

    mk = MetaKernel(KERNELS + '/metakernel.tm', kernels=r'C:\bar')  # absolute r-string
    assert "PATH_VALUES = ( 'C:\\bar' )" in mk.content

    mk = MetaKernel(KERNELS + '/metakernel.tm', kernels='C:\\baz')  # \\ escape
    assert "PATH_VALUES = ( 'C:\\baz' )" in mk.content

    mk = MetaKernel(KERNELS + '/metakernel.tm', kernels='qux\\')  # trailing \\
    assert "PATH_VALUES = ( 'qux\\' )" in mk.content

    mk = MetaKernel(KERNELS + '/metakernel.tm', kernels=r'\Users\quux')  # \U escape
    assert "PATH_VALUES = ( '\\Users\\quux' )" in mk.content

    # With Path object
    mk = MetaKernel(KERNELS + '/metakernel.tm', kernels=Path('../foo'))
    assert "PATH_VALUES = ( '../foo' )" in mk.content

    mk = MetaKernel(KERNELS + '/metakernel.tm', kernels=Path('C:/Users/bar'))
    assert "PATH_VALUES = ( 'C:/Users/bar' )" in mk.content

    mk = MetaKernel(KERNELS + '/metakernel.tm', kernels=Path(r'C:\Users\baz'))
    assert "PATH_VALUES = ( 'C:\\Users\\baz' )" in mk.content

    mk = MetaKernel(KERNELS + '/metakernel.tm', kernels=Path('C:\\Users\\qux\\'))
    assert "PATH_VALUES = ( 'C:\\Users\\qux\\' )" in mk.content

    # Replace ~ for home value
    mk = MetaKernel(KERNELS + '/metakernel.tm', kernels='~/foo')
    assert f"PATH_VALUES = ( '{Path.home()}/foo' )" in mk.content

    # Path value with relative posix path
    mock_mk_content(
        'empty',
        {
            'PATH_VALUES': ['../foo'],
            'PATH_SYMBOLS': ['FOO'],
            'KERNELS_TO_LOAD': ['$FOO/dummy.ti'],
        },
    )
    mk = MetaKernel(KERNELS + '/metakernel.tm')

    assert len(mk.kernels) == 1
    assert mk.kernels[0] == '../foo/dummy.ti'

    mk = MetaKernel(KERNELS + '/metakernel.tm', foo='/bar')

    assert len(mk.kernels) == 1
    assert mk.kernels[0] == '/bar/dummy.ti'

    # Path value with absolute posix path
    mock_mk_content(
        'empty',
        {
            'PATH_VALUES': ['/baz'],
            'PATH_SYMBOLS': ['BAZ'],
            'KERNELS_TO_LOAD': ['$BAZ/dummy.ti'],
        },
    )

    mk = MetaKernel(KERNELS + '/metakernel.tm')
    assert mk.kernels[0] == '/baz/dummy.ti'

    mk = MetaKernel(KERNELS + '/metakernel.tm', baz=r'\qux')
    assert mk.kernels[0] == r'\qux/dummy.ti'

    # Path value with relative windows path
    mock_mk_content(
        'empty',
        {
            'PATH_VALUES': [r'.\foo'],
            'PATH_SYMBOLS': ['FOO'],
            'KERNELS_TO_LOAD': ['$FOO/dummy.ti'],
        },
    )

    mk = MetaKernel(KERNELS + '/metakernel.tm')
    assert mk.kernels[0] == '.\\foo/dummy.ti'

    mk = MetaKernel(KERNELS + '/metakernel.tm', foo=r'/bar')
    assert mk.kernels[0] == '/bar/dummy.ti'

    # Path value with absolute windows path
    mock_mk_content(
        'empty',
        {
            'PATH_VALUES': ['C:\\Users\\with space\\bar'],
            'PATH_SYMBOLS': ['BAZ'],
            'KERNELS_TO_LOAD': [r'$BAZ\dummy.ti'],
        },
    )

    mk = MetaKernel(KERNELS + '/metakernel.tm')
    assert mk.kernels[0] == 'C:\\Users\\with space\\bar\\dummy.ti'

    mk = MetaKernel(KERNELS + '/metakernel.tm', baz='\\Users\\qux')
    assert mk.kernels[0] == '\\Users\\qux\\dummy.ti'


def test_metakernel_kernels(mk):
    """Test MetaKernel kernels."""
    assert len(mk.kernels) == 2

    # Contains
    assert KERNELS + '/naif0012.tls' in mk.kernels
    assert KERNELS + '/dummy.tf' in mk.kernels


def test_metakernel_content_data(mock_mk_content):
    """Test MetaKernel content and data parser."""
    # Empty metakernel (without KERNELS_TO_LOAD key)
    mock_mk_content('missing_kernels_to_load', {'MISSING_KERNELS_TO_LOAD': None})

    err_msg = re.escape('KERNELS_TO_LOAD key is missing')

    with raises(KeyError, match=err_msg):
        _ = MetaKernel(KERNELS + '/metakernel.tm', kernels=KERNELS)

    # Update value without PATH_SYMBOLS and PATH_VALUES keys
    mock_mk_content('empty', {'KERNELS_TO_LOAD': ['foo.bar']})

    err_msg = re.escape('PATH_VALUES key is missing')

    with raises(KeyError, match=err_msg):
        _ = MetaKernel(KERNELS + '/metakernel.tm', foo='bar')

    # Update value with symbol not in PATH_SYMBOLS
    mock_mk_content(
        'empty',
        {
            'PATH_VALUES': ['bar'],
            'PATH_SYMBOLS': ['FOO'],
            'KERNELS_TO_LOAD': ['$FOO/bar'],
        },
    )

    err_msg = re.escape('Symbol: `BAZ` is not part of PATH_SYMBOLS (`FOO`)')

    with raises(ValueError, match=err_msg):
        _ = MetaKernel(KERNELS + '/metakernel.tm', baz='qux')

    # Update only one symbol among many
    mock_mk_content(
        'empty',
        {
            'PATH_VALUES': ['foo', 'bar'],
            'PATH_SYMBOLS': ['FOO', 'BAR'],
            'KERNELS_TO_LOAD': ['$BAR/dummy.ti', '$FOO/dummy.tpc'],
        },
    )

    mk = MetaKernel(KERNELS + '/metakernel.tm', bar=KERNELS, foo=KERNELS)

    assert len(mk.kernels) == 2
    assert mk.kernels[0] == KERNELS + '/dummy.ti'
    assert mk.kernels[1] == KERNELS + '/dummy.tpc'
    assert mk.urls == [None, None]

    # No remote url in the content
    mock_mk_content('empty', {'KERNELS_TO_LOAD': []})
    mk = MetaKernel(KERNELS + '/metakernel.tm')

    assert mk.remote is None
    assert mk.urls == []

    # Multiple remote url in the content
    mock_mk_content('https://foo\nftp://bar', {'KERNELS_TO_LOAD': []})

    mk = MetaKernel(KERNELS + '/metakernel.tm')

    assert mk.remote == 'https://foo'

    mk = MetaKernel(KERNELS + '/metakernel.tm', remote=1)

    assert mk.remote == 'ftp://bar'

    with raises(IndexError, match='list index out of range'):
        _ = MetaKernel(KERNELS + '/metakernel.tm', remote=2)

    # Override remote in content
    mk = MetaKernel(KERNELS + '/metakernel.tm', remote='https://example.org')

    assert mk.remote == 'https://example.org'

    # Missing kernel with known remote
    mock_mk_content(
        'https://example.org',
        {
            'PATH_VALUES': ['.'],
            'PATH_SYMBOLS': ['FOO'],
            'KERNELS_TO_LOAD': ['$FOO/dummy.tf'],
        },
    )

    err_msg = re.escape(
        '`./dummy.tf` was not found locally. '
        'You can add `download=True` to download it automatically '
        'from `https://example.org` '
        'or/and you can change the kernel path value(s) by adding '
        "`foo='path/to/my/kernels'`."
    )

    with raises(MissingKernelError, match=err_msg):
        _ = MetaKernel(KERNELS + '/metakernel.tm')

    # Missing kernel with unknown remote (and multiple symbols)
    mock_mk_content(
        'empty',
        {
            'PATH_VALUES': ['.', '..'],
            'PATH_SYMBOLS': ['FOO', 'BAR'],
            'KERNELS_TO_LOAD': ['$BAR/dummy.tpc'],
        },
    )

    err_msg = re.escape(
        '`../dummy.tpc` was not found locally. '
        'You can add `download=True` to download it automatically '
        'or/and you can change the kernel path value(s) by adding '
        "`foo|bar='path/to/my/kernels'`."
    )

    with raises(MissingKernelError, match=err_msg):
        _ = MetaKernel(KERNELS + '/metakernel.tm')

    # Download kernel without remote
    err_msg = re.escape(
        'Impossible to download the missing kernel `../dummy.tpc`. '
        "You need to provide a `remote='[https|http|ftp]://kernels.domain.tld/'`."
    )

    with raises(MissingKernelsRemoteError, match=err_msg):
        _ = MetaKernel(KERNELS + '/metakernel.tm', download=True)


def test_metakernel_single_download(monkeypatch, mock_mk_content):
    """Test Metakernel single download."""

    class MockPoolStarmapWget:
        """Mock multiprocessing Pool starmap function to by-pass wget."""

        def __init__(self, **kwargs):
            pass

        def __enter__(self):
            return self

        def __exit__(self, *args):
            pass

        @staticmethod
        def starmap(_, iterable):
            """Mock pool starmap function to create dummy files."""
            for _, fname in iterable:
                f = Path(fname)
                f.parent.mkdir(parents=True, exist_ok=True)  # Create parents
                f.write_text('foo c-kernel', encoding='utf-8')  # Add dummy values

    monkeypatch.setattr(_metakernel, 'Pool', MockPoolStarmapWget)

    # Download missing kernel
    mock_mk_content(
        'https://example.org',
        {
            'PATH_VALUES': ['.'],
            'PATH_SYMBOLS': ['KERNELS'],
            'KERNELS_TO_LOAD': ['$KERNELS/ck/dummy.bc'],
        },
    )

    ck = Path(KERNELS) / 'ck' / 'dummy.bc'

    if ck.exists():
        ck.unlink()

    if ck.parent.exists():
        ck.parent.rmdir()

    mk = MetaKernel(KERNELS + '/metakernel.tm', kernels=KERNELS, download=True)

    assert mk.kernels[0] == KERNELS + '/ck/dummy.bc'
    assert ck.exists()
    assert ck.read_text(encoding='utf-8') == 'foo c-kernel'

    ck.unlink()
    ck.parent.rmdir()

    # Kernel path too long (> 255 characters)
    with raises(BufferError):
        _ = MetaKernel(KERNELS + '/metakernel.tm', kernels=255 * 'a')


def test_metakernel_duplicate_download(monkeypatch, mock_mk_content):
    """Test MetaKernel duplicate download."""
    urls = []

    class MockPoolStarmapURL:
        """Mock multiprocessing Pool starmap function to store requested urls."""

        def __init__(self, **kwargs):
            pass

        def __enter__(self):
            return self

        def __exit__(self, *args):
            pass

        @staticmethod
        def starmap(_, iterable):
            """Mock pool starmap function to store the requested urls."""
            for url, _ in iterable:
                urls.append(url)

    monkeypatch.setattr(_metakernel, 'Pool', MockPoolStarmapURL)

    # Try to download missing kernels with duplicates
    mock_mk_content(
        'https://example.org',
        {
            'PATH_VALUES': ['.'],
            'PATH_SYMBOLS': ['KERNELS'],
            'KERNELS_TO_LOAD': [
                '$KERNELS/foo.bc',
                '$KERNELS/foo.bc',
                '$KERNELS/foo.bc',
                '$KERNELS/bar.bc',
                '$KERNELS/baz.bc',
            ],
        },
    )

    assert not urls

    MetaKernel(KERNELS + '/metakernel.tm', kernels=KERNELS, download=True)

    assert len(urls) == 3
    assert sorted(urls) == [
        'https://example.org/bar.bc',
        'https://example.org/baz.bc',
        'https://example.org/foo.bc',
    ]


def test_metakernel_disable_multi_thread_download(monkeypatch, mock_mk_content):
    """Test MetaKernel disabled multi thread download."""
    urls = []

    def mock_wget(url, _):
        """Mock wget download function."""
        urls.append(url)

    monkeypatch.setattr(_metakernel, 'wget', mock_wget)

    mock_mk_content(
        'https://example.org',
        {
            'PATH_VALUES': ['.'],
            'PATH_SYMBOLS': ['KERNELS'],
            'KERNELS_TO_LOAD': [
                '$KERNELS/foo.bc',
                '$KERNELS/bar.bc',
                '$KERNELS/baz.bc',
            ],
        },
    )

    assert not urls

    MetaKernel(KERNELS + '/metakernel.tm', kernels=KERNELS, download=True, n_threads=1)

    assert len(urls) == 3
    assert sorted(urls) == [
        'https://example.org/bar.bc',
        'https://example.org/baz.bc',
        'https://example.org/foo.bc',
    ]


def test_metakernel_load_kernels(caplog):
    """Test MetaKernel kernels loader."""
    debug_spice_pool(True)
    SpicePool.purge()
    caplog.clear()

    mk = MetaKernel(KERNELS + '/metakernel.tm', kernels=KERNELS, load_kernels=True)

    assert len(SpicePool) == 3
    assert mk in SpicePool
    assert KERNELS + '/naif0012.tls' in SpicePool
    assert KERNELS + '/dummy.tf' in SpicePool

    # Check metakernel hash
    assert hash(mk) in SpicePool.MK_HASH.values()

    # Check debugger message
    reload_msg = [
        'The content of the pool changed -> the metakernel will be reloaded.',
        'Purge the pool',
        'Add `metakernel.tm` in the SPICE pool',
        'Cache metakernel original hash.',
    ]

    assert caplog.messages == reload_msg

    # Re-load the pool with the same metakernel (no reload)
    caplog.clear()

    mk.load_kernels()
    assert not caplog.messages

    # Remove metakernel
    SpicePool.remove(mk)

    assert len(SpicePool) == 0


def test_metakernel_content_details(mk):
    """Test MetaKernel kernel content details."""
    assert mk.title == 'Dummy kernel'
    assert mk.description == 'Example of single value.'

    assert display_html(mk) == (
        '<h3>📚 Dummy kernel</h3>\n'
        '<p>Example of single value.</p>\n'
        '<hr/>\n'
        '<dl>\n'
        f'<dt><b>Location:</b></dt><dd>{KERNELS}/metakernel.tm</dd>\n'
        '<dt><b>SKD version:</b></dt><dd>v999_20210531_001</dd>\n'
        '<dt><b>MK identifier:</b></dt><dd>metakernel_v999_20210531_001</dd>\n'
        f'<dt><b>$KERNELS</b></dt><dd>{KERNELS}</dd>\n'
        '</dl>\n'
        '<hr/>\n'
        '<table>'
        '<thead><tr><th>Kernels</th><th>Type</th><th>Size</th></tr></thead>'
        '<tbody>'
        '<tr><td>$KERNELS/naif0012.tls</td><td>⏱️ LSK</td><td>5 kB</td></tr>'
        '<tr><td>$KERNELS/dummy.tf</td><td>🖼️ FK</td><td>2 kB</td></tr>'
        '</tbody></table>'
    )

    assert display_html(mk.summary) == (
        '<table>'
        '<thead><tr><th>Types</th><th>Count</th><th>Size</th></tr></thead>'
        '<tbody>'
        '<tr><td>⏱️ LSK</td><td>1</td><td>5 kB</td></tr>'
        '<tr><td>🖼️ FK</td><td>1</td><td>2 kB</td></tr>'
        '<tr><td><b>Total</b></td><td>2</td><td>7 kB</td></tr>'
        '</tbody></table>'
    )
