"""Test SPICE times module."""

from pathlib import Path

import numpy as np
from numpy.testing import assert_array_equal

import spiceypy as sp

from pytest import raises

from planetary_coverage.debug import debug_spice_pool
from planetary_coverage.html import display_html
from planetary_coverage.spice import MetaKernel, SpicePool, check_kernels


KERNELS = str((Path(__file__).parent / '..' / 'data').resolve())
tls = KERNELS + '/naif0012.tls'

mk = MetaKernel(KERNELS + '/metakernel.tm', kernels=KERNELS)
mk_kernels_hash = hash((
    hash(mk),
    hash(KERNELS + '/naif0012.tls'),
    hash(KERNELS + '/dummy.tf'),
))


def test_spice_pool():
    """Test SPICE pool."""
    sp.kclear()

    assert int(SpicePool) == len(SpicePool) == 0

    # __add__
    _ = SpicePool + tls

    assert int(SpicePool) == 1
    assert hash(SpicePool) == hash((hash(tls),))

    # Try to add a kernel already present in the pool
    with raises(ValueError):
        SpicePool.add(tls)

    # Purge before adding a list of kernels
    SpicePool.add([None, tls], purge=True)

    assert int(SpicePool) == 1  # Only the valid kernels are loaded.

    # __eq__
    assert SpicePool == (tls,)  # Tuple
    assert SpicePool == [tls]  # List
    assert SpicePool == tls  # Str
    assert SpicePool == hash((hash(tls),))  # Hash of hashes
    assert SpicePool == (None, tls)  # Discard `None` values
    assert SpicePool != 12345

    # __contains__
    assert tls in SpicePool
    assert 'foo' not in SpicePool

    # __iter__
    for k in SpicePool:
        assert k == tls

    # __sub__
    _ = SpicePool - tls

    with raises(ValueError):
        SpicePool.remove(tls)  # Already removed

    assert not SpicePool
    assert len(SpicePool) == 0

    # Add an array of kernels
    _ = SpicePool + [tls]
    assert len(SpicePool) == 1

    # __getitem__
    assert SpicePool['DELTET/DELTA_T_A'] == 32.184

    # Purge
    SpicePool.purge()
    assert not SpicePool


def test_spice_pool_add(monkeypatch):
    """Test SPICE pool add new kernel."""
    sp.kclear()

    spice_pool = []

    def mock_furnsh(kernel):
        """Mock spiceypy furnsh function."""
        spice_pool.append(kernel)

    monkeypatch.setattr(sp, 'furnsh', mock_furnsh)

    assert not spice_pool

    # Relative path
    SpicePool.add('foo.ti')
    assert spice_pool == ['foo.ti']

    # Absolute path
    SpicePool.add(['/bar.bsp'])
    assert spice_pool == ['foo.ti', '/bar.bsp']

    # Expanded usr path
    SpicePool.add(['~/baz.bc'])
    assert spice_pool == ['foo.ti', '/bar.bsp', f'{Path.home()}/baz.bc']


def test_spice_logger(caplog):
    """Test SPICE Pool logger."""
    sp.kclear()

    # Enable debugger
    debug_spice_pool(True)

    assert caplog.messages == ['Change logger level to DEBUG']
    caplog.clear()

    SpicePool.add(tls)
    assert caplog.messages == [f'Add `{tls}` in the SPICE pool']
    caplog.clear()

    SpicePool.remove(tls)
    assert caplog.messages == [f'Remove {tls}']
    caplog.clear()

    SpicePool.purge()
    assert caplog.messages == ['Purge the pool']
    caplog.clear()

    # Enable debugger with INFO level
    debug_spice_pool('info')
    assert caplog.messages == ['Change logger level to INFO']
    caplog.clear()

    SpicePool.add(tls, purge=True)
    SpicePool.remove(tls)

    assert caplog.messages == ['Purge the pool']
    caplog.clear()

    # Disable debugger
    debug_spice_pool(False)

    SpicePool.add(tls, purge=True)
    SpicePool.remove(tls)

    assert caplog.messages == []


def test_spice_pool_content_details():
    """Test spice pool kernel content details."""
    SpicePool.purge()
    assert display_html(SpicePool) == (
        '<p><span style="color: #d62728">⌀</span> <em>No kernel in the pool</em></p>'
    )

    SpicePool.add([
        KERNELS + '/naif0012.tls',
        KERNELS + '/dummy.ti',
    ])

    assert display_html(SpicePool) == (
        '<table>'
        '<thead><tr><th>Kernels</th><th>Type</th><th>Size</th></tr></thead>'
        '<tbody>'
        f'<tr><td>{KERNELS}/naif0012.tls</td><td>⏱️ LSK</td><td>5 kB</td></tr>'
        f'<tr><td>{KERNELS}/dummy.ti</td><td>🔬 IK</td><td>7 kB</td></tr>'
        '</tbody></table>'
    )

    assert display_html(SpicePool.summary) == (
        '<table>'
        '<thead><tr><th>Types</th><th>Count</th><th>Size</th></tr></thead>'
        '<tbody>'
        '<tr><td>⏱️ LSK</td><td>1</td><td>5 kB</td></tr>'
        '<tr><td>🔬 IK</td><td>1</td><td>7 kB</td></tr>'
        '<tr><td><b>Total</b></td><td>2</td><td>13 kB</td></tr>'
        '</tbody></table>'
    )


def test_spice_pool_windows(monkeypatch):
    """Test spice pool windows."""
    SpicePool.add(tls, purge=True)  # Dummy load (masked by monkeypatch)

    kernels = [
        ('lsk/naif0012.tls', 'TEXT', None, None),
        ('ck/dummy.bc', 'CK', None, None),
        ('pck/dummy.bpc', 'PCK', None, None),
        ('pck/dummy.tpc', 'TEXT', None, None),
        ('spk/dummy.bsp', 'SPK', None, None),
    ]

    body_names = {
        'JUICE': -28,
        'JUICE_SPACECRAFT': -28_000,
        'GANYMEDE': 503,
    }

    body_codes = {
        -28: 'JUICE',
        -28_000: 'JUICE_SPACECRAFT',
        503: 'GANYMEDE',
    }

    monkeypatch.setattr(sp, 'ktotal', lambda _: len(kernels))
    monkeypatch.setattr(sp, 'kdata', lambda i, _: kernels[i])
    monkeypatch.setattr(sp, 'bods2c', lambda name: body_names[name])
    monkeypatch.setattr(sp, 'bodc2n', lambda code: body_codes[code])
    monkeypatch.setattr(sp, 'ckobj', lambda *args: [-28_000])
    monkeypatch.setattr(
        sp,
        'ckcov',
        lambda *args: [
            783500680,
            783870237,
            880011923,
            880901281,
            942003659,
            942745795,
            975964826,
            976605399,
        ],
    )
    monkeypatch.setattr(sp, 'wncard', lambda cover: len(cover) // 2)
    monkeypatch.setattr(sp, 'cell_int', lambda *args: [503])
    monkeypatch.setattr(sp, 'cell_double', lambda *args: [94305664.183, 757080065.183])
    monkeypatch.setattr(sp, 'pckfrm', lambda *args: None)
    monkeypatch.setattr(sp, 'pckcov', lambda *args: None)
    monkeypatch.setattr(sp, 'spkobj', lambda *args: [503])
    monkeypatch.setattr(sp, 'spkcov', lambda *args: [707313600, 1577880000])

    assert len(SpicePool) == 5

    # Single values: JUICE_SPACECRAFT (CK only)
    ets_windows = SpicePool.windows('JUICE_SPACECRAFT', fmt='ET')

    assert isinstance(ets_windows, dict)
    assert len(ets_windows) == 1
    assert 'JUICE_SPACECRAFT' in ets_windows
    assert 'GANYMEDE' not in ets_windows

    ets_kernels = ets_windows['JUICE_SPACECRAFT']

    assert isinstance(ets_kernels, dict)
    assert len(ets_kernels) == 1
    assert list(ets_kernels.keys()) == [
        'ck/dummy.bc',
    ]

    assert list(ets_kernels.values()) == [
        [
            # ck/dummy.bc
            [783500680, 783870237],
            [880011923, 880901281],
            [942003659, 942745795],
            [975964826, 976605399],
        ]
    ]

    # Single values: GANYMEDE (PCK and SPK)
    ets_windows = SpicePool.windows('GANYMEDE', fmt='ET')

    assert isinstance(ets_windows, dict)
    assert len(ets_windows) == 1
    assert 'GANYMEDE' in ets_windows
    assert 'JUICE_SPACECRAFT' not in ets_windows

    ets_kernels = ets_windows['GANYMEDE']

    assert isinstance(ets_kernels, dict)
    assert len(ets_kernels) == 2
    assert list(ets_kernels.keys()) == [
        'pck/dummy.bpc',
        'spk/dummy.bsp',
    ]

    assert list(ets_kernels.values()) == [
        [
            # pck/dummy.bpc
            [94305664.183, 757080065.183],
        ],
        [
            # spk/dummy.bsp
            [707313600, 1577880000],
        ],
    ]

    # Multiple values
    ets_windows = SpicePool.windows('JUICE_SPACECRAFT', 'GANYMEDE', fmt='ET')

    assert isinstance(ets_windows, dict)
    assert len(ets_windows) == 2
    assert 'JUICE_SPACECRAFT' in ets_windows
    assert 'GANYMEDE' in ets_windows

    juice_sc = ets_windows['JUICE_SPACECRAFT']
    ganymede = ets_windows['GANYMEDE']

    assert len(juice_sc) == 1
    assert len(ganymede) == 2

    assert list(juice_sc.keys()) == [
        'ck/dummy.bc',
    ]
    assert list(ganymede.keys()) == [
        'pck/dummy.bpc',
        'spk/dummy.bsp',
    ]

    assert list(juice_sc.values()) == [
        [
            # ck/dummy.bc
            [783500680, 783870237],
            [880011923, 880901281],
            [942003659, 942745795],
            [975964826, 976605399],
        ]
    ]

    assert list(ganymede.values()) == [
        [
            # pck/dummy.bpc
            [94305664.183, 757080065.183],
        ],
        [
            # spk/dummy.bsp
            [707313600, 1577880000],
        ],
    ]

    # UTC output format (default)
    windows = SpicePool.windows('JUICE_SPACECRAFT')
    juice_sc = list(windows['JUICE_SPACECRAFT'].values())

    assert len(juice_sc) == 1

    # ck/dummy.bc
    assert_array_equal(
        juice_sc[0].astype(str),
        [
            ['2024-10-29T19:03:30.818', '2024-11-03T01:42:47.817'],
            ['2027-11-20T19:44:13.817', '2027-12-01T02:46:51.817'],
            ['2029-11-07T07:39:49.817', '2029-11-15T21:48:45.817'],
            ['2030-12-05T09:19:16.817', '2030-12-12T19:15:29.817'],
        ],
    )

    # TDB output format
    windows = SpicePool.windows('GANYMEDE', fmt='TDB')
    ganymede = list(windows['GANYMEDE'].values())

    assert len(ganymede) == 2

    # pck/dummy.bpc
    assert_array_equal(
        ganymede[0].astype(str),
        [
            ['2002-12-28 00:01:04.182 TDB', '2023-12-29 00:01:05.182 TDB'],
        ],
    )

    # spk/dummy.bsp
    assert_array_equal(
        ganymede[1].astype(str),
        [
            ['2022-06-01 00:00:00.000 TDB', '2050-01-01 00:00:00.000 TDB'],
        ],
    )

    # Invalid output format
    with raises(TypeError):
        _ = SpicePool.windows('JUICE_SPACECRAFT', fmt='INVALID_FMT')

    # No windows for a unknown reference
    err = 'The windows for JUICE was not found.'
    with raises(KeyError, match=err):
        _ = SpicePool.windows('JUICE')


def test_spice_pool_coverage(monkeypatch):
    """Test spice pool references coverage."""
    SpicePool.add(tls, purge=True)

    def mock_windows(*args, **kwargs):
        """Mock SpicePool window method."""
        return {
            -28_000: {
                'pck/dummy.bpc': [
                    [94305664.183, 757080065.183],
                ],
                'spk/dummy.bsp': [
                    [707313600, 1577880000],
                ],
            }
        }

    monkeypatch.setattr(SpicePool, 'windows', mock_windows)

    # Take only the coverage in ET
    ets = SpicePool.coverage('JUICE_SPACECRAFT', fmt='ET')

    assert len(ets) == 2
    assert_array_equal(ets, [94305664.183, 1577880000])

    # Take only the coverage in UTC
    utc = SpicePool.coverage('JUICE_SPACECRAFT', fmt='UTC')

    assert len(utc) == 2
    assert_array_equal(
        utc,
        (
            np.datetime64('2002-12-27T23:59:59.999'),
            np.datetime64('2049-12-31T23:58:50.816'),
        ),
    )

    # Take only the coverage in TDB
    tdb = SpicePool.coverage('JUICE_SPACECRAFT', fmt='TDB')

    assert len(tdb) == 2
    assert_array_equal(
        tdb, ['2002-12-28 00:01:04.182 TDB', '2050-01-01 00:00:00.000 TDB']
    )

    # Invalid output format
    with raises(TypeError):
        _ = SpicePool.coverage('JUICE_SPACECRAFT', fmt='INVALID_FMT')

    # Coverage with multiple resources and overlap
    def mock_windows_overlap(*args, **kwargs):
        """Mock SpicePool window method with two references with overlap."""
        return {
            -28_000: {
                'pck/dummy.bpc': [
                    [94305664.183, 757080065.183],
                ],
            },
            503: {
                'spk/dummy.bsp': [
                    [707313600, 1577880000],
                ]
            },
        }

    monkeypatch.setattr(SpicePool, 'windows', mock_windows_overlap)

    ets = SpicePool.coverage('JUICE_SPACECRAFT', 'GANYMEDE', fmt='ET')

    assert len(ets) == 2
    assert_array_equal(ets, [707313600, 757080065.183])

    # Coverage with multiple resources and overlap
    def mock_windows_no_overlap(*args, **kwargs):
        """Mock SpicePool window method with two references with no overlap."""
        return {
            -28_000: {
                'ck/dummy.bc': [
                    [783500680, 783870237],
                    [880011923, 880901281],
                    [942003659, 942745795],
                    [975964826, 976605399],
                ]
            },
            503: {
                'pck/dummy.bpc': [
                    [94305664.183, 757080065.183],
                ],
            },
        }

    monkeypatch.setattr(SpicePool, 'windows', mock_windows_no_overlap)

    err = (
        r'MAX start time \(2024-10-29 19:04:40.000 TDB\) '
        r'is after MIN stop time \(2023-12-29 00:01:05.182 TDB\).'
    )

    with raises(ValueError, match=err):
        _ = SpicePool.coverage('JUICE_SPACECRAFT', 'GANYMEDE')


def test_spice_pool_gaps(monkeypatch):
    """Test spice pool references windows gaps."""
    SpicePool.add(tls, purge=True)

    def kdata(i, _):
        """SPICE kdata mock."""
        return [
            ('lsk/naif0012.tls', 'TEXT', None, None),
            ('ck/dummy.bc', 'CK', None, None),
        ][i]

    def bods2c(name):
        """SPICE bods2c mock."""
        return {
            'JUICE': -28,
            'JUICE_SPACECRAFT': -28_000,
            'JUICE_SPACECRAFT_PLAN': -28_001,
        }[name]

    def bodc2n(code):
        """SPICE bodc2n mock."""
        return {
            -28: 'JUICE',
            -28_000: 'JUICE_SPACECRAFT',
            -28_001: 'JUICE_SPACECRAFT_PLAN',
        }[code]

    def mock_ckobj(*args):
        """SPICE ckobj mock."""
        return [-28, -28_000, -28_001]

    def mock_ckcov(_, ref, *args):
        """SPICE ckcov mock."""
        return (
            [
                # JUICE_SPACECRAFT_PLAN
                734878907,
                748682868,
                880011923,
                880901281,
                942003659,
                942745795,
                975964826,
                976605399,
            ]
            if int(ref) == -28_001
            else [
                # JUICE_SPACECRAFT
                707313600,
                783500680,
                783870237,
                1577880000,
            ]
            if int(ref) == -28_000
            else [
                # JUICE
                734748206.842,
                1128434338.683,
            ]
        )

    monkeypatch.setattr(sp, 'ktotal', lambda _: 2)
    monkeypatch.setattr(sp, 'kdata', kdata)
    monkeypatch.setattr(sp, 'bods2c', bods2c)
    monkeypatch.setattr(sp, 'bodc2n', bodc2n)
    monkeypatch.setattr(sp, 'ckobj', mock_ckobj)
    monkeypatch.setattr(sp, 'ckcov', mock_ckcov)

    # No gaps in the temporal coverage window
    assert SpicePool.gaps('JUICE').size == 0

    # Single gap
    gaps = SpicePool.gaps('JUICE_SPACECRAFT', fmt='TDB')

    assert len(gaps) == 1
    assert_array_equal(
        gaps, (['2024-10-29 19:04:40.000 TDB', '2024-11-03 01:43:57.000 TDB'],)
    )

    # Multiple gaps
    gaps = SpicePool.gaps('JUICE_SPACECRAFT_PLAN', fmt='TDB')

    assert len(gaps) == 3
    assert_array_equal(
        gaps,
        [
            ['2023-09-22 19:27:48.000 TDB', '2027-11-20 19:45:23.000 TDB'],
            ['2027-12-01 02:48:01.000 TDB', '2029-11-07 07:40:59.000 TDB'],
            ['2029-11-15 21:49:55.000 TDB', '2030-12-05 09:20:26.000 TDB'],
        ],
    )

    # Multiple references (sorted gaps by start time)
    gaps = SpicePool.gaps('JUICE_SPACECRAFT', 'JUICE_SPACECRAFT_PLAN', fmt='TDB')

    assert len(gaps) == 4
    assert_array_equal(
        gaps,
        (
            ['2023-09-22 19:27:48.000 TDB', '2027-11-20 19:45:23.000 TDB'],
            ['2024-10-29 19:04:40.000 TDB', '2024-11-03 01:43:57.000 TDB'],
            ['2027-12-01 02:48:01.000 TDB', '2029-11-07 07:40:59.000 TDB'],
            ['2029-11-15 21:49:55.000 TDB', '2030-12-05 09:20:26.000 TDB'],
        ),
    )


def test_spice_pool_brief(monkeypatch):
    """Test spice pool brief coverage."""
    SpicePool.add(tls, purge=True)

    def kdata(i, _):
        """SPICE kdata mock."""
        return [
            ('lsk/naif0012.tls', 'TEXT', None, None),
            ('pck/dummy.bpc', 'PCK', None, None),
            ('spk/dummy.bsp', 'SPK', None, None),
        ][i]

    def bodc2n(code):
        """SPICE bodc2n mock."""
        if code == -28_001:
            raise ValueError

        return {
            -28: 'JUICE',
            -28_000: 'JUICE_SPACECRAFT',
            503: 'GANYMEDE',
            504: 'CALLISTO',
        }[code]

    monkeypatch.setattr(sp, 'ktotal', lambda _: 3)
    monkeypatch.setattr(sp, 'kdata', kdata)
    monkeypatch.setattr(sp, 'bods2c', int)
    monkeypatch.setattr(sp, 'bodc2n', bodc2n)

    def mock_pck_cov(*args):
        """Mock PCK coverage method"""
        return [[94305664.183, 757080065.183]]

    def mock_spk_cov(_, ref):
        """Mock SPK coverage method"""
        return (
            [
                [783500680, 783870237],
                [880011923, 880901281],
                [942003659, 942745795],
                [975964826, 976605399],
            ]
            if ref < 0
            else [[707313600, 1577880000]]
        )

    def mock_spice_pool_cov(_, ext):
        """Mock SPICE pool coverage method."""
        if ext == 'PCK':
            return [503, 504], mock_pck_cov

        if ext == 'SPK':
            return [-28, -28_000, -28_001, 503], mock_spk_cov

        return set(), None

    monkeypatch.setattr(SpicePool, '_cov', mock_spice_pool_cov)

    brief = SpicePool.brief(fmt='TDB')

    assert isinstance(brief, dict)
    assert list(brief.keys()) == [
        # Sorted by decreasing negatives
        -28,
        -28_000,
        # Skip unknown key: -28,001
        # then increasing positives
        503,
        504,
    ]

    assert list(brief.values()) == [
        ('2024-10-29 19:04:40.000 TDB', '2030-12-12 19:16:39.000 TDB'),  # -28
        ('2024-10-29 19:04:40.000 TDB', '2030-12-12 19:16:39.000 TDB'),  # -28,000
        ('2002-12-28 00:01:04.182 TDB', '2050-01-01 00:00:00.000 TDB'),  # 503
        ('2002-12-28 00:01:04.182 TDB', '2023-12-29 00:01:05.182 TDB'),  # 504
    ]


def test_spice_load_mk():
    """Test metakernel loader."""
    SpicePool.add(mk, purge=True)

    assert len(SpicePool) == 3

    # Check metakernel cache hash
    assert hash(mk) in SpicePool.MK_HASH.values()

    # Check kernel contains in the pool
    assert mk in SpicePool
    assert KERNELS + '/naif0012.tls' in SpicePool
    assert KERNELS + '/dummy.tf' in SpicePool

    # Check that all the kernels in the mk where loaded
    kernels = SpicePool.kernels

    assert kernels[0].endswith('.tm')
    assert 'metakernel' in kernels[0]
    assert 'naif0012.tls' in kernels[1]
    assert 'dummy.tf' in kernels[2]

    # Check pool content
    assert hash(SpicePool) == mk_kernels_hash

    # Empty the pool
    SpicePool.purge()
    assert hash(SpicePool) != mk_kernels_hash

    assert len(SpicePool) == 0
    assert not SpicePool.MK_HASH

    # Get metakernel hash (independent of the pool content)
    assert SpicePool.hash(mk) == mk_kernels_hash


def test_spice_pool_check_kernels(caplog):
    """Test Spice Pool kernel checker decorator."""
    debug_spice_pool(True)
    SpicePool.purge()
    caplog.clear()

    class A:
        """Dummy class with kernels and hash."""

        @property
        def kernels(self):
            """Object required kernels."""
            return (tls,)

        def __hash__(self):
            return SpicePool.hash(self.kernels)

        @check_kernels
        def get_first_kernel(self):
            """Decorated dummy function that require the kernels to be loaded."""
            return self.kernels[0]

        @property
        @check_kernels
        def last_kernel(self):
            """Decorated dummy property that require the kernels to be loaded."""
            return self.kernels[-1]

    a = A()

    # Check the attributes
    assert a.kernels == (tls,)
    assert hash(a) == hash((hash(tls),))

    # Check the pool content (the kernels are not loaded)
    assert len(SpicePool) == 0
    assert not caplog.messages

    # When the decorated function is call, the pool is checked and kernels/mk are loaded.
    assert a.get_first_kernel() == tls

    reload_msg = [
        'The content of the pool changed -> the kernels will be reloaded.',
        'Purge the pool',
        f'Add `{tls}` in the SPICE pool',
    ]

    assert caplog.messages == reload_msg

    # Check the pool content
    assert len(SpicePool) == 1
    assert SpicePool.kernels == (tls,)

    # Re-run the same function but this time the pool match the kernels
    caplog.clear()

    assert a.get_first_kernel() == tls
    assert not caplog.messages

    # Checker on property (no need to reload the pool)
    assert a.last_kernel == tls
    assert not caplog.messages

    # Class without kernels attribute
    class B:
        """Dummy class without explicit hash."""

        @check_kernels
        def dummy(self):
            """Decorated dummy function."""

    b = B()

    with raises(AttributeError):
        b.dummy()
