"""Test SPICE reference module."""

from pathlib import Path

import spiceypy as sp

from pytest import approx, raises

from planetary_coverage.spice import SpiceBody, SpiceFrame, SpiceRef


KERNELS = str((Path(__file__).parent / '..' / 'data').resolve())
tpc = KERNELS + '/dummy.tpc'


def test_ref_body_code_validity():
    """Test SPICE body code validity."""
    # Sun
    assert SpiceBody.is_valid(10)

    assert not SpiceBody.is_valid(0)
    assert not SpiceBody.is_valid(11)

    # Planets
    assert SpiceBody.is_valid(199)  # Mercury
    assert SpiceBody.is_valid(299)  # Venus
    assert SpiceBody.is_valid(399)  # Earth
    assert SpiceBody.is_valid(499)  # Mars
    assert SpiceBody.is_valid(599)  # Jupiter

    # Satellites
    assert SpiceBody.is_valid(301)  # Moon
    assert SpiceBody.is_valid(401)  # Phobos
    assert SpiceBody.is_valid(402)  # Deimos
    assert SpiceBody.is_valid(501)  # Io

    assert not SpiceBody.is_valid(101)  # No Mercury satellite
    assert not SpiceBody.is_valid(201)  # No Venus satellite
    assert not SpiceBody.is_valid(302)  # No additional Earth satellite
    assert not SpiceBody.is_valid(403)  # No additional Mars satellite

    assert not SpiceBody.is_valid(100)
    assert not SpiceBody.is_valid(1_000)

    # Comets
    assert SpiceBody.is_valid(1_000_001)  # Arend comet
    assert SpiceBody.is_valid(1_999_999)  # Not discover yet

    assert not SpiceBody.is_valid(1_000_000)
    assert not SpiceBody.is_valid(2_000_000)

    # Comet fragments
    assert SpiceBody.is_valid(50_000_001)  # Shoemaker Levy 9-W  (fragment  1)
    assert SpiceBody.is_valid(50_000_023)  # Shoemaker Levy 9-P2 (fragment 8B)

    # Permanently numbered asteroids (original schema / 7 digits)
    assert SpiceBody.is_valid(2_000_001)  # Ceres
    assert SpiceBody.is_valid(2_999_999)

    assert not SpiceBody.is_valid(3_000_000)

    # Provisional numbered asteroids (original schema / 7 digits)
    assert SpiceBody.is_valid(3_000_001)
    assert SpiceBody.is_valid(3_999_999)

    assert not SpiceBody.is_valid(4_000_000)

    # Permanently numbered asteroids (extended schema / 8 digits)
    assert SpiceBody.is_valid(20_000_001)
    assert SpiceBody.is_valid(49_999_999)

    assert not SpiceBody.is_valid(20_000_000)
    assert not SpiceBody.is_valid(50_000_000)

    # Provisional numbered asteroids (extended schema / 8 digits)
    assert SpiceBody.is_valid(50_000_024)
    assert SpiceBody.is_valid(99_999_999)

    assert not SpiceBody.is_valid(100_000_000)

    # Permanently numbered asteroids in multi-body system (9-digits)
    assert SpiceBody.is_valid(920_065_803)  # Didymos
    assert SpiceBody.is_valid(120_065_803)  # Dimorphos

    assert SpiceBody.is_valid(999_999_999)
    assert SpiceBody.is_valid(120_000_001)

    assert not SpiceBody.is_valid(120_000_000)  # 1 + 8-digit not valid
    assert not SpiceBody.is_valid(950_000_000)  # 9 + 8-digit not valid
    assert not SpiceBody.is_valid(1_000_000_000)

    # Exceptions
    assert SpiceBody.is_valid(9_511_010)  # Gaspra


def test_ref_body_sun_attrs():
    """Test SPICE body sun attributes."""
    sp.kclear()

    # Sun body
    body = SpiceBody('SUN')

    assert body == 'SUN'
    assert body == 10

    assert not body.is_planet
    assert not body.is_satellite
    assert not body.is_comet
    assert not body.is_asteroid

    assert body.frame == 'IAU_SUN'
    assert isinstance(body.frame, SpiceFrame)

    assert isinstance(body.parent, SpiceBody)
    assert body.parent == 'SUN'

    assert body.barycenter == 'SOLAR SYSTEM BARYCENTER'


def test_ref_body_planet_attrs():
    """Test SPICE body planet attributes."""
    sp.kclear()
    sp.furnsh(tpc)

    # Planet body
    body = SpiceBody('JUPITER')

    assert body == 'JUPITER'
    assert body == 599

    assert body.is_planet
    assert not body.is_satellite
    assert not body.is_comet
    assert not body.is_asteroid

    assert body.frame == 'IAU_JUPITER'
    assert isinstance(body.frame, SpiceFrame)

    assert body.mu == approx(126_686_534.92, abs=1e-2)

    assert isinstance(body.parent, SpiceBody)
    assert body.parent == 'SUN'

    assert body.barycenter == 'JUPITER BARYCENTER'

    assert body['GM'] == approx(1.267e8, abs=1e5)  # BODY599_GM

    with raises(KeyError):
        _ = body['RADII']  # BODY599_RADII is not present in `dummy.tpc`


def test_ref_body_satellite_attrs():
    """Test SPICE body satellite attributes."""
    sp.kclear()
    sp.furnsh(tpc)

    # Satellite body
    body = SpiceBody('Ganymede')

    assert str(body) == body.name == 'GANYMEDE'
    assert int(body) == body.id == 503
    assert body.code == '503'

    assert not body.is_planet
    assert body.is_satellite
    assert not body.is_comet
    assert not body.is_asteroid

    assert body.frame == 'IAU_GANYMEDE'
    assert isinstance(body.frame, SpiceFrame)

    assert tuple(body.radii) == (2631.2, 2631.2, 2631.2)
    assert body.radius == body.r == approx(2631.2, abs=1e-2)
    assert body.re == body.rp == 2631.2
    assert body.f == 0

    assert isinstance(body.parent, SpiceBody)
    assert body.parent == 'JUPITER'
    assert body.parent.mu == approx(126_686_534.92, abs=1e-2)

    assert body.barycenter == 'JUPITER BARYCENTER'

    assert tuple(body['RADII']) == (2631.2, 2631.2, 2631.2)  # BODY503_RADII

    with raises(KeyError):
        _ = body['GM']  # BODY503_GM is not present in `dummy.tpc`


def test_ref_body_comet_attrs():
    """Test SPICE body comet attributes."""
    sp.kclear()

    # Comet body
    body = SpiceBody('TEMPEL_1')
    assert SpiceBody.is_valid(1_000_093)

    assert str(body) == body.name == 'TEMPEL 1'
    assert int(body) == body.id == 1_000_093
    assert body.code == '1000093'

    assert not body.is_planet
    assert not body.is_satellite
    assert body.is_comet
    assert not body.is_asteroid

    assert body.frame == 'IAU_TEMPEL_1'
    assert isinstance(body.frame, SpiceFrame)

    assert isinstance(body.parent, SpiceBody)
    assert body.parent == 'SUN'
    assert body.barycenter == 'SOLAR SYSTEM BARYCENTER'

    with raises(KeyError):
        _ = body.radii


def test_ref_body_asteroid_original_schema_attrs():
    """Test SPICE body asteroid original schema attributes."""
    sp.kclear()

    # Permanently numbered asteroids (original schema / 7 digits)
    body = SpiceBody('CERES')

    assert str(body) == body.name == 'CERES'
    assert int(body) == body.id == 2_000_001
    assert body.code == '2000001'

    assert not body.is_planet
    assert not body.is_satellite
    assert not body.is_comet
    assert body.is_asteroid

    assert body.frame == 'IAU_CERES'
    assert isinstance(body.frame, SpiceFrame)

    assert isinstance(body.parent, SpiceBody)
    assert body.parent == 'SUN'
    assert body.barycenter == 'SOLAR SYSTEM BARYCENTER'

    with raises(KeyError):
        _ = body.radii


def test_ref_body_asteroid_extended_schema_attrs():
    """Test SPICE body asteroid extended schema attributes."""
    sp.kclear()

    # Permanently numbered asteroids (extended schema / 8 digits)
    body = SpiceBody('LEUCUS')  # Single body

    assert str(body) == body.name == 'LEUCUS'
    assert int(body) == body.id == 20_011_351
    assert body.code == '20011351'

    assert not body.is_planet
    assert not body.is_satellite
    assert not body.is_comet
    assert body.is_asteroid

    assert body.frame == 'IAU_LEUCUS'
    assert isinstance(body.frame, SpiceFrame)

    assert isinstance(body.parent, SpiceBody)
    assert body.parent == 'SUN'
    assert body.barycenter == 'SOLAR SYSTEM BARYCENTER'

    # Permanently numbered asteroids in multi-body system (9-digits)
    body = SpiceBody('Didymos')  # Primary body

    assert str(body) == body.name == 'DIDYMOS'
    assert int(body) == body.id == 920_065_803
    assert body.code == '920065803'

    assert not body.is_planet
    assert not body.is_satellite
    assert not body.is_comet
    assert body.is_asteroid

    assert body.frame == 'IAU_DIDYMOS'
    assert isinstance(body.frame, SpiceFrame)

    assert isinstance(body.parent, SpiceBody)
    assert body.parent == 'SUN'
    assert body.barycenter == 'DIDYMOS BARYCENTER'

    # Permanently numbered asteroids in multi-body system (9-digits)
    body = SpiceBody('DIMORPHOS')  # Secondary body

    assert str(body) == body.name == 'DIMORPHOS'
    assert int(body) == body.id == 120_065_803
    assert body.code == '120065803'

    assert not body.is_planet
    assert body.is_satellite
    assert not body.is_comet
    assert body.is_asteroid

    assert body.frame == 'IAU_DIMORPHOS'
    assert isinstance(body.frame, SpiceFrame)

    # Parent body (prefixed with 9)
    assert isinstance(body.parent, SpiceBody)
    assert body.parent == 'DIDYMOS'
    assert int(body.parent) == 920_065_803
    assert body.parent.code == '920065803'

    # Barycenter (without prefixed)
    assert isinstance(body.barycenter, SpiceRef)
    assert body.barycenter == 'DIDYMOS BARYCENTER'
    assert int(body.barycenter) == 20_065_803
    assert body.barycenter.code == '20065803'


def test_ref_body_asteroid_exceptions_attrs():
    """Test SPICE body exceptions attributes."""
    sp.kclear()

    # Gaspra exception (old rule)
    body = SpiceBody('Gaspra')

    assert str(body) == body.name == 'GASPRA'
    assert int(body) == body.id == 9_511_010
    assert body.code == '9511010'

    assert not body.is_planet
    assert not body.is_satellite
    assert not body.is_comet
    assert body.is_asteroid

    # Comet Shoemaker Levy-9 fragments
    body = SpiceBody('SHOEMAKER-LEVY 9-W')

    assert str(body) == body.name == 'SHOEMAKER-LEVY 9-W'
    assert int(body) == body.id == 50_000_001
    assert body.code == '50000001'

    assert not body.is_planet
    assert body.is_satellite
    assert body.is_comet
    assert not body.is_asteroid

    assert isinstance(body.parent, SpiceBody)
    assert body.parent == 'JUPITER'
    assert body.barycenter == 'JUPITER BARYCENTER'


def test_ref_body_err():
    """Test SPICE body errors."""
    with raises(KeyError):
        _ = SpiceBody('JUICE')

    # Asteroids barycenter is not a valid SpiceBody
    with raises(KeyError):
        _ = SpiceBody('DIDYMOS BARYCENTER')
