"""Test SPICE reference module."""

from pathlib import Path

import spiceypy as sp

from pytest import raises

from planetary_coverage.spice import SpiceFrame, SpiceInstrument, SpiceRef


KERNELS = str((Path(__file__).parent / '..' / 'data').resolve())
tf = KERNELS + '/dummy.tf'
ti = KERNELS + '/dummy.ti'
tpc = KERNELS + '/dummy.tpc'


def test_ref_frame_code_validity():
    """Test SPICE frame code validity."""
    sp.kclear()
    sp.furnsh([tpc, tf])

    assert SpiceFrame.is_valid('J2000')
    assert SpiceFrame.is_valid('IAU_GANYMEDE')
    assert SpiceFrame.is_valid('JUICE_SPACECRAFT')
    assert SpiceFrame.is_valid('JUICE_JANUS')
    assert SpiceFrame.is_valid('DSS-14_TOPO')
    assert SpiceFrame.is_valid('DUMMY_REF')
    assert not SpiceFrame.is_valid('FOO')

    # Int values are not accepted
    assert not SpiceFrame.is_valid(1)
    assert not SpiceFrame.is_valid(-28_000)


def test_ref_frame_attrs():
    """Test SPICE frame attributes."""
    sp.kclear()
    sp.furnsh([tpc, tf, ti])

    # Inertial frame
    frame = SpiceFrame('J2000')

    assert frame == 'J2000'
    assert frame == 1

    assert frame.class_type == 'Inertial frame'

    # Errors
    err = '`FRAME_1_CENTER` was not found in the kernel pool.'
    with raises(KeyError, match=err):
        _ = frame.center

    with raises(NotImplementedError):
        _ = frame.frame

    with raises(KeyError):
        _ = frame['FOO']

    # PCK body-fixed frame
    frame = SpiceFrame('IAU_GANYMEDE')

    assert frame == 'IAU_GANYMEDE'
    assert frame == 10_025

    assert frame.class_type == 'PCK body-fixed frame'

    # CK frame
    frame = SpiceFrame('JUICE_SPACECRAFT')

    assert frame == 'JUICE_SPACECRAFT'
    assert frame == -28_000

    assert frame.class_type == 'CK frame'

    assert frame.center == 'JUICE_SPACECRAFT'
    assert isinstance(frame.center, SpiceFrame)

    assert frame.sclk == 'JUICE'
    assert frame.spk == 'JUICE'

    # Fixed offset frame
    frame = SpiceFrame('JUICE_JANUS')

    assert frame == 'JUICE_JANUS'
    assert frame == -28_200

    assert frame.class_type == 'Fixed offset frame'

    assert frame.center == 'JUICE_JANUS'
    assert isinstance(frame.center, SpiceInstrument)

    # Station frame
    frame = SpiceFrame('DSS-14_TOPO')

    assert frame == 'DSS-14_TOPO'
    assert frame == 1_399_014

    assert frame.class_type == 'Fixed offset frame'

    assert frame.center == 'DSS-14'
    assert isinstance(frame.center, SpiceRef)

    # Dummy frame
    frame = SpiceFrame('DUMMY_REF')

    assert frame == 'DUMMY_REF'
    assert frame == -28_999

    with raises(sp.stypes.NotFoundError):
        _ = frame.class_type


def test_ref_frame_err():
    """Test SPICE frame errors."""
    with raises(KeyError):
        _ = SpiceFrame('JUICE')
