"""Test SPICE reference module."""

from pathlib import Path

import spiceypy as sp

from pytest import raises

from planetary_coverage.spice import SpiceFrame, SpiceInstrument, SpiceSpacecraft


KERNELS = str((Path(__file__).parent / '..' / 'data').resolve())
tf = KERNELS + '/dummy.tf'
ti = KERNELS + '/dummy.ti'
tls = KERNELS + '/naif0012.tls'
tsc = KERNELS + '/dummy.tsc'


def test_ref_sc_code_validity():
    """Test SPICE spacecraft code validity."""
    assert SpiceSpacecraft.is_valid(-1)
    assert SpiceSpacecraft.is_valid(-999)
    assert SpiceSpacecraft.is_valid(-100_001)
    assert SpiceSpacecraft.is_valid(-119_999)

    assert not SpiceSpacecraft.is_valid(0)
    assert not SpiceSpacecraft.is_valid(10)
    assert not SpiceSpacecraft.is_valid(-1_000)
    assert not SpiceSpacecraft.is_valid(-100_000)
    assert not SpiceSpacecraft.is_valid(-120_000)

    assert not SpiceSpacecraft.is_valid('-0')
    assert not SpiceSpacecraft.is_valid('-00')
    assert not SpiceSpacecraft.is_valid('-000')


def test_ref_sc_attrs():
    """Test SPICE spacecraft attributes."""
    sp.kclear()
    sp.furnsh([tf, ti, tsc])

    sc = SpiceSpacecraft('JUICE')

    assert sc == 'JUICE'
    assert sc == -28

    # Spacecraft attribute
    assert isinstance(sc.spacecraft, SpiceSpacecraft)
    assert sc.spacecraft == sc

    # Frame attribute
    assert sc.frame == 'JUICE_SPACECRAFT'
    assert isinstance(sc.frame, SpiceFrame)

    # Boresight attribute
    assert tuple(sc.boresight) == (0, 0, 1)

    # Instruments list
    assert len(sc.instruments) == 2

    inst = sc.instruments[0]

    assert isinstance(inst, SpiceInstrument)
    assert inst == 'JUICE_JANUS'
    assert inst == -28200

    assert inst.frame == 'JUICE_JANUS'
    assert isinstance(inst.frame, SpiceFrame)
    assert inst.shape == 'RECTANGLE'

    # 2nd instrument in the list
    assert sc.instruments[1] == 'JUICE_UVS'

    # Instrument by suffix reference
    inst = sc.instr('JANUS')

    assert isinstance(inst, SpiceInstrument)
    assert inst == 'JUICE_JANUS'
    assert inst == -28200

    assert inst.frame == 'JUICE_JANUS'
    assert isinstance(inst.frame, SpiceFrame)
    assert inst.shape == 'RECTANGLE'

    # Instrument by full reference
    inst = sc.instr('JUICE_JANUS')

    assert isinstance(inst, SpiceInstrument)
    assert inst == 'JUICE_JANUS'

    # Invalid instrument name
    err = 'Unknown reference: `FOO`'
    with raises(ValueError, match=err):
        _ = sc.instr('FOO')

    # Change spacecraft default boresight
    assert tuple(sc.BORESIGHT) == (0, 0, 1)

    sc.BORESIGHT = (-1, 0, 0)
    assert tuple(sc.BORESIGHT) == (-1, 0, 0)
    assert tuple(sc.boresight) == (-1, 0, 0)

    # Direct change on the property is not allowed
    with raises(AttributeError):
        sc.boresight = (0, 1, 0)


def test_ref_sc_sclk():
    """Test SPICE spacecraft clock."""
    sp.kclear()
    sp.furnsh([tls, tsc])

    sc = SpiceSpacecraft('JUICE')

    assert sc.sclk('2022-MAY-31-23:59:59.999 TDB') == 0.0
    assert sc.sclk('2033-01-01') == 21_896_158_850_903.266


def test_ref_sc_err():
    """Test SPICE spacecraft errors."""
    with raises(KeyError):
        _ = SpiceSpacecraft('Ganymede')
