"""Test SPICE reference module."""

from pathlib import Path

import spiceypy as sp

from pytest import raises

from planetary_coverage.spice import (
    SpiceBody,
    SpiceFrame,
    SpiceInstrument,
    SpiceObserver,
    SpiceRef,
    SpiceSpacecraft,
)
from planetary_coverage.spice.references import SpiceRefRegex


KERNELS = str((Path(__file__).parent / '..' / 'data').resolve())
tf = KERNELS + '/dummy.tf'
ti = KERNELS + '/dummy.ti'


def test_spice_ref_code_validity():
    """Test SPICE reference code validity."""
    assert SpiceRef.is_valid(-1)
    assert SpiceRef.is_valid(1)

    with raises(ValueError):
        _ = SpiceRef.is_valid('FOO')


def test_spice_ref_attrs():
    """Test SPICE reference attributes."""
    sp.kclear()

    # Reference with string
    ref = SpiceRef('Jupiter Barycenter')

    assert str(ref) == ref.name == 'JUPITER BARYCENTER'
    assert int(ref) == ref.id == 5
    assert ref.code == '5'
    assert ref.frame == 'IAU_JUPITER_BARYCENTER'
    assert isinstance(ref.frame, SpiceFrame)
    assert ref.encode(encoding='utf-8') == b'JUPITER BARYCENTER'

    # Reference with positive integer
    ref = SpiceRef(5)

    assert str(ref) == 'JUPITER BARYCENTER'
    assert int(ref) == 5
    assert ref.code == '5'

    # Reference with negative integer
    ref = SpiceRef(-28)

    assert str(ref) == 'JUICE'
    assert int(ref) == -28
    assert ref.code == '-28'

    # __eq__
    assert ref == 'JUICE'
    assert ref == -28

    assert ref != 'Juice'  # Case-sensitive
    assert ref != 28

    # __getitem__
    with raises(KeyError):
        _ = ref['DUMMY']

    # Nested references
    assert SpiceRef(SpiceRef('JUICE')) == -28
    assert SpiceRef(SpiceRef(-28)) == 'JUICE'

    # Not implemented errors
    ref = SpiceRef('Jupiter Barycenter')

    with raises(NotImplementedError):
        _ = ref.spacecraft

    with raises(NotImplementedError):
        _ = ref.sclk('2035-01-01')

    # Invalid SPICE reference
    with raises(ValueError):
        _ = SpiceRef('FOO')


def test_spice_ref_promotion():
    """Test SPICE reference promotion."""
    sp.kclear()
    sp.furnsh([tf, ti])

    assert isinstance(SpiceRef('Jupiter Barycenter'), SpiceRef)
    assert isinstance(SpiceRef('Ganymede'), SpiceBody)
    assert isinstance(SpiceRef('JUICE'), SpiceSpacecraft)
    assert isinstance(SpiceRef('JUICE_JANUS'), SpiceInstrument)
    assert isinstance(SpiceRef('JUICE_SPACECRAFT'), SpiceFrame)

    # Spice observer parent class
    assert isinstance(SpiceRef('JUICE'), SpiceObserver)
    assert isinstance(SpiceRef('JUICE_JANUS'), SpiceObserver)

    # Spice observer promotion
    assert isinstance(SpiceObserver('JUICE'), SpiceSpacecraft)
    assert isinstance(SpiceObserver('JUICE_JANUS'), SpiceInstrument)

    with raises(KeyError):
        _ = SpiceObserver('Ganymede')


def test_spice_ref_hash():
    """Test SPICE reference hash."""
    sp.kclear()

    mercury_barycenter = SpiceRef('Mercury Barycenter')
    assert mercury_barycenter == 1
    assert hash(mercury_barycenter) == hash('MERCURY BARYCENTER')

    j2000 = SpiceRef('J2000')
    assert j2000 == 1
    assert hash(j2000) == hash('J2000')

    # Code SPICE ID collision
    assert mercury_barycenter == j2000

    # No collision in dict definition
    data = {
        mercury_barycenter: 'Mercury barycenter reference',
        j2000: 'J2000 reference',
    }

    assert len(data) == 2
    assert 'MERCURY BARYCENTER' in data
    assert 'J2000' in data
    assert 1 not in data

    assert data['MERCURY BARYCENTER'] == 'Mercury barycenter reference'
    assert data['J2000'] == 'J2000 reference'

    with raises(KeyError):
        _ = data[1]


def test_spice_ref_regex():
    """Test SPICE reference regular expressions."""
    # Simple test
    re_1 = SpiceRefRegex('10')

    assert 10 in re_1
    assert '10' in re_1
    assert 0 not in re_1
    assert '-1' not in re_1

    # Optional with fix pattern enforced
    re_2 = SpiceRefRegex(r'^[1-9]99$')

    assert 199 in re_2
    assert 999 in re_2
    assert 99 not in re_2
    assert 123 not in re_2
    assert 1_000 not in re_2
    assert -199 not in re_2
    assert -999 not in re_2

    # Multiple values
    re_3 = SpiceRefRegex(r'301|401|402')

    assert 301 in re_3
    assert 402 in re_3
    assert 403 not in re_3
    assert 342 not in re_3

    # Excluded pattern
    re_4 = SpiceRefRegex(r'[5-9](?!00|99)\d{2}')

    assert 301 not in re_4
    assert 500 not in re_4
    assert 501 in re_4
    assert 598 in re_4
    assert 599 not in re_4
    assert 900 not in re_4
    assert 901 in re_4
    assert 998 in re_4
    assert 999 not in re_4
    assert 1_000 not in re_4

    # Combined patterns
    assert 399 in re_2 | re_3 | re_4
    assert 301 in re_2 | re_3 | re_4
    assert 300 not in re_2 | re_3 | re_4
    assert 302 not in re_2 | re_3 | re_4
    assert 502 in re_2 | re_3 | re_4
    assert 599 in re_2 | re_3 | re_4
    assert 100 not in re_2 | re_3 | re_4
    assert 1_000 not in re_2 | re_3 | re_4

    # String representation
    assert str(re_2) == r'^[1-9]99$'
    assert str(re_3 | re_4) == r'^301|401|402$|^[5-9](?!00|99)\d{2}$'

    # Raise not implemented error
    with raises(NotImplementedError):
        re_1 | 'foo'
