"""Test SPICE toolbox helper module."""

from pathlib import Path

import numpy as np
from numpy.testing import assert_array_almost_equal as assert_array

import spiceypy as sp

from pytest import approx, raises

from planetary_coverage.spice import (
    SpiceBody,
    SpiceInstrument,
    SpiceObserver,
    SpicePool,
    attitude,
    ocentric2ographic,
    quaternions,
    toolbox,
)
from planetary_coverage.spice.toolbox import (
    angular_size,
    azel,
    boresight_pt,
    deg,
    et,
    fov_pts,
    groundtrack_velocity,
    illum_angles,
    intersect_pt,
    local_time,
    pixel_scale,
    planetographic,
    radec,
    rlonlat,
    sc_state,
    solar_longitude,
    station_azel,
    sub_obs_pt,
    sun_pos,
    target_position,
    target_separation,
    true_anomaly,
    type_check,
)


KERNELS = str((Path(__file__).parent / '..' / 'data').resolve())

tpc = KERNELS + '/dummy.tpc'
tf = KERNELS + '/dummy.tf'
ti = KERNELS + '/dummy.ti'


def test_toolbox_type_checker(monkeypatch):
    """Test SPICE toolbox type checker."""
    SpicePool.add([tf, ti], purge=True)

    monkeypatch.setattr(sp, 'str2et', lambda *_: 1234.5)

    # Generic
    assert type_check(1.2, float) == 1.2
    assert type_check(12, float) == 12.0
    assert type_check('12.3', float) == 12.3
    assert type_check('12', float, int) == 12

    # ET time input
    assert type_check(123.4, float, et) == 123.4
    assert type_check('2000-01-01', float, et) == 1234.5
    assert_array(type_check(['2000-01-01'], float, et), [1234.5])
    assert_array(type_check([123.4], float, et), [123.4])
    assert_array(type_check(np.array([123.4], dtype=np.float64), float, et), [123.4])

    # Spice Body
    assert type_check('Ganymede', SpiceBody) == 503
    assert type_check(503, SpiceBody) == 'GANYMEDE'

    # Spice Observer
    assert type_check('JUICE', SpiceObserver) == -28
    assert type_check(-28, SpiceObserver) == 'JUICE'

    # Spice Instrument
    assert type_check('JUICE_JANUS', SpiceInstrument) == -28_200
    assert type_check(-28_200, SpiceInstrument) == 'JUICE_JANUS'

    # Errors
    with raises(KeyError):
        _ = type_check(-28_200, SpiceBody)

    with raises(KeyError):
        _ = type_check(503, SpiceObserver)

    with raises(KeyError):
        _ = type_check(-28, SpiceInstrument)


def test_toolbox_dpr():
    """Test SPICE toolbox radians to degree convertor."""
    assert deg(0) == 0
    assert deg(np.pi / 2) == approx(90, abs=1e-6)
    assert deg(np.pi) == approx(180, abs=1e-6)
    assert deg(3 * np.pi / 2) == approx(270, abs=1e-6)

    # With a 1D list
    assert_array(deg([0, np.pi / 2, np.pi]), [0, 90, 180], decimal=6)


def test_toolbox_rlonlat():
    """Test SPICE toolbox planetocentric coordinates convertor."""
    assert_array(rlonlat([1, 0, 0]), (1, 0, 0))  # r, lon_e, lat
    assert_array(rlonlat([1, 1, 0]), (np.sqrt(2), 45, 0))
    assert_array(rlonlat([0, 1, 0]), (1, 90, 0))
    assert_array(rlonlat([0, -1, 0]), (1, 270, 0))
    assert_array(rlonlat([1, 0, 1]), (np.sqrt(2), 0, 45))
    assert_array(rlonlat([0, 0, 1]), (1, 0, 90))

    # SPICE examples
    assert_array(rlonlat([0, 0, 0]), [0, 0, 0], decimal=3)
    assert_array(rlonlat([1, 0, 0]), [1, 0, 0], decimal=3)
    assert_array(rlonlat([0, 1, 0]), [1, 90, 0], decimal=3)
    assert_array(rlonlat([0, 0, 1]), [1, 0, 90], decimal=3)
    assert_array(rlonlat([-1, 0, 0]), [1, 180, 0], decimal=3)
    assert_array(rlonlat([0, -1, 0]), [1, 270, 0], decimal=3)  # <- Changed for 270°E
    assert_array(rlonlat([0, 0, -1]), [1, 0, -90], decimal=3)
    assert_array(rlonlat([1, 1, 0]), [1.4142, 45, 0], decimal=3)
    assert_array(rlonlat([1, 0, 1]), [1.4142, 0, 45], decimal=3)
    assert_array(rlonlat([0, 1, 1]), [1.4142, 90, 45], decimal=3)
    assert_array(rlonlat([1, 1, 1]), [1.7320, 45, 35.2643], decimal=3)

    # Test on array
    r, lon_e, lat = rlonlat([[0, 1], [1, 1], [1, 1]])  # Shape: (3, 2)
    assert_array(r, [1.4142, 1.7320], decimal=3)
    assert_array(lon_e, [90, 45], decimal=3)
    assert_array(lat, [45, 35.2643], decimal=3)

    # NaN support
    assert_array(rlonlat([0, 0, np.nan]), (np.nan, np.nan, np.nan))
    assert_array(rlonlat([0, np.nan, 0]), (np.nan, np.nan, np.nan))
    assert_array(rlonlat([np.nan, 0, 0]), (np.nan, np.nan, np.nan))
    assert_array(rlonlat([np.nan, np.nan, np.nan]), (np.nan, np.nan, np.nan))

    r, lon_e, lat = rlonlat([[0, np.nan, 1, 1], [1, 1, np.nan, 1], [1, 1, 1, np.nan]])
    assert_array(r, [1.4142, np.nan, np.nan, np.nan], decimal=3)
    assert_array(lon_e, [90, np.nan, np.nan, np.nan], decimal=3)
    assert_array(lat, [45, np.nan, np.nan, np.nan], decimal=3)


def test_toolbox_planetographic():
    """Test SPICE toolbox planetographic coordinates convertor."""
    SpicePool.add(tpc, purge=True)

    assert_array(
        planetographic('MARS', [0, -2620.678914818178, 2592.4089088569672]),
        (300.0, 90.0, 45.0),  # alt, lon_w, lat
        decimal=3,
    )

    # Array input
    alt, lon_w, lat = planetographic(
        'MARS',
        [
            (3396.19, 0, 0),
            (-3396.19, 0, 0),
            (-3406.19, 0, 0),
            (-3386.19, 0, 0),
            (0, -3396.19, 0),
            (0, 3396.19, 0),
            (0, 0, 3376.2),
            (0, 0, -3376.2),
            (0, 0, 0),
        ],
    )

    assert_array(alt, [0, 0, 10, -10, 0, 0, 0, 0, -3_376.2], decimal=1)
    assert_array(lon_w, [0, 180, 180, 180, 90, 270, 0, 0, 0], decimal=1)
    assert_array(lat, [0, 0, 0, 0, 0, 0, 90, -90, 90], decimal=1)

    # Array input - Exception Earth: East longitudes
    alt, lon_e, lat = planetographic(
        'EARTH',
        [
            (6378.14, 0, 0),
            (-6378.14, 0, 0),
            (-6388.14, 0, 0),
            (-6368.14, 0, 0),
            (0, -6378.14, 0),
            (0, 6378.14, 0),
            (0, 0, 6356.75),
            (0, 0, -6356.75),
            (0, 0, 0),
        ],
    )

    assert_array(alt, [0, 0, 10, -10, 0, 0, 0, 0, -6_356.75], decimal=1)
    assert_array(lon_e, [0, 180, 180, 180, 270, 90, 0, 0, 0], decimal=1)  # East long
    assert_array(lat, [0, 0, 0, 0, 0, 0, 90, -90, 90], decimal=1)

    # Input dimension error
    err = r'Input dimension must be `\(3,\)` or `\(N, 3\)` not `\(0,\)`\.'
    with raises(ValueError, match=err):
        planetographic('MARS', [])

    err = r'Input dimension must be `\(3,\)` or `\(N, 3\)` not `\(4,\)`\.'
    with raises(ValueError, match=err):
        planetographic('MARS', [1, 2, 3, 4])

    err = r'Input dimension must be `\(3,\)` or `\(N, 3\)` not `\(3, 2\)`\.'
    with raises(ValueError, match=err):
        planetographic('MARS', np.zeros((3, 2)))


def test_toolbox_ocentric2ographic():
    """Test SPICE toolbox planetocentric to planetographic convertor."""
    SpicePool.add(tpc, purge=True)

    lon_w, lat = ocentric2ographic('MARS', 90, 45)

    assert lon_w == approx(270.0, abs=1e-2)
    assert lat == approx(45.34, abs=1e-2)

    # With array input
    lons_w, lats = ocentric2ographic('MARS', [90, 0, 45], [0, 90, 45])

    assert_array(lons_w, [270, 0, 315], decimal=2)
    assert_array(lats, [0, 90, 45.34], decimal=2)

    # Input dimension errors
    err = (
        r'East longitude and latitude inputs must have the same dimension: '
        r'\(\) vs\. \(2,\)'
    )
    with raises(ValueError, match=err):
        ocentric2ographic('MARS', 1, [2, 3])


def test_toolbox_radec():
    """Test SPICE toolbox RA/DEC vector in J2000 coordinates."""
    assert_array(radec([1, 0, 0]), (0, 0))
    assert_array(radec([1, 1, 0]), (45, 0))
    assert_array(radec([0, 1, 0]), (90, 0))
    assert_array(radec([0, -1, 0]), (270, 0))
    assert_array(radec([1, 0, 1]), (0, 45))
    assert_array(radec([0, 0, 1]), (0, 90))

    # SPICE examples
    assert_array(radec([0, 0, 0]), [0, 0], decimal=3)
    assert_array(radec([1, 0, 0]), [0, 0], decimal=3)
    assert_array(radec([0, 1, 0]), [90, 0], decimal=3)
    assert_array(radec([0, 0, 1]), [0, 90], decimal=3)
    assert_array(radec([-1, 0, 0]), [180, 0], decimal=3)
    assert_array(radec([0, -1, 0]), [270, 0], decimal=3)
    assert_array(radec([0, 0, -1]), [0, -90], decimal=3)
    assert_array(radec([1, 1, 0]), [45, 0], decimal=3)
    assert_array(radec([1, 0, 1]), [0, 45], decimal=3)
    assert_array(radec([0, 1, 1]), [90, 45], decimal=3)
    assert_array(radec([1, 1, 1]), [45, 35.2643], decimal=3)

    # Test on array
    ra, dec = radec([[0, 1], [1, 1], [1, 1]])  # Shape: (3, 2)
    assert_array(ra, [90, 45], decimal=3)
    assert_array(dec, [45, 35.2643], decimal=3)

    # NaN support
    assert_array(radec([0, 0, np.nan]), (np.nan, np.nan))
    assert_array(radec([0, np.nan, 0]), (np.nan, np.nan))
    assert_array(radec([np.nan, 0, 0]), (np.nan, np.nan))
    assert_array(radec([np.nan, np.nan, np.nan]), (np.nan, np.nan))

    ra, dec = radec([[0, np.nan, 1, 1], [1, 1, np.nan, 1], [1, 1, 1, np.nan]])
    assert_array(ra, [90, np.nan, np.nan, np.nan], decimal=3)
    assert_array(dec, [45, np.nan, np.nan, np.nan], decimal=3)


def test_toolbox_azel():
    """Test SPICE toolbox azimuth and elevation."""
    assert_array(azel([1, 0, 0]), (0, 0))
    assert_array(azel([1, 1, 0]), (315, 0))  # clockwise by default
    assert_array(azel([0, 1, 0]), (270, 0))
    assert_array(azel([0, -1, 0]), (90, 0))
    assert_array(azel([1, 0, 1]), (0, 45))
    assert_array(azel([0, 0, 1]), (0, 90))
    assert_array(azel([0, 0, 0]), [0, 0], decimal=3)
    assert_array(azel([0, -1, 1]), (90, 45))
    assert_array(radec([1, 1, 1]), [45, 35.2643], decimal=3)

    # Conventions
    assert_array(azel([1, 1, 0], az_spice=True), (45, 0))  # counterclockwise
    assert_array(azel([1, 0, 1], el_spice=False), (0, -45))  # Neg upper plane

    # Test on array
    az, el = azel([[0, 1], [-1, 1], [1, 1]])  # Shape: (3, 2)
    assert_array(az, [90, 315], decimal=3)
    assert_array(el, [45, 35.2643], decimal=3)

    # NaN support
    assert_array(azel([0, 0, np.nan]), (np.nan, np.nan))
    assert_array(azel([0, np.nan, 0]), (np.nan, np.nan))
    assert_array(azel([np.nan, 0, 0]), (np.nan, np.nan))
    assert_array(azel([np.nan, np.nan, np.nan]), (np.nan, np.nan))

    az, el = azel([[0, np.nan, 1, 1], [-1, 1, np.nan, 1], [1, 1, 1, np.nan]])
    assert_array(az, [90, np.nan, np.nan, np.nan], decimal=3)
    assert_array(el, [45, np.nan, np.nan, np.nan], decimal=3)


def test_toolbox_sub_obs_pt(monkeypatch):
    """Test SPICE toolbox sub-observer point calculation."""
    monkeypatch.setattr(sp, 'str2et', lambda *_: 1041422469.184)
    monkeypatch.setattr(
        sp,
        'subpnt',
        lambda *_: ([504.15405255, -649.50320473, -2499.43748038], None, None),
    )

    xyz = sub_obs_pt('2033-01-01', 'JUICE', 'GANYMEDE')

    assert_array(xyz, (504.154, -649.503, -2499.437), decimal=3)

    r, lon_e, lat = rlonlat(xyz)

    assert r == approx(2631.2, abs=1e-2)
    assert lon_e == approx(307.82, abs=1e-2)
    assert lat == approx(-71.79, abs=1e-2)

    # Array of times
    xyz = sub_obs_pt(['2033-01-01', '2033-01-01'], 'JUICE', 'GANYMEDE')

    assert xyz.shape == (3, 2)
    assert_array(xyz[:, 0], (504.154, -649.503, -2499.437), decimal=3)
    assert_array(xyz[:, 0], xyz[:, 1], decimal=3)

    r, lon_e, lat = rlonlat(xyz)

    assert r[0] == approx(2631.2, abs=1e-2)
    assert lon_e[0] == approx(307.82, abs=1e-2)
    assert lat[0] == approx(-71.79, abs=1e-2)


def test_toolbox_target_position(monkeypatch):
    """Test SPICE toolbox observer -> target position."""
    SpicePool.add(tf, purge=True)

    monkeypatch.setattr(sp, 'str2et', lambda *_: 1041422469.184)

    def spkpos(*args):
        """Mock Juice position seen from Ganymede frame.

        spkpos('JUICE', 1041422468.1839253, 'IAU_GANYMEDE', 'NONE', 'GANYMEDE')

        """
        return (
            [604.7307270168204, -779.0764414332297, -2998.0650497737743],
            0.010527661953128513,
        )

    monkeypatch.setattr(sp, 'spkpos', spkpos)

    xyz = target_position('2033-01-01', 'JUICE', 'GANYMEDE')

    assert_array(xyz, (604.730, -779.076, -2998.065), decimal=3)

    r, lon_e, lat = rlonlat(xyz)

    assert r == approx(3156.11, abs=1e-2)
    assert lon_e == approx(307.82, abs=1e-2)
    assert lat == approx(-71.79, abs=1e-2)

    # Array of times
    xyz = target_position(['2033-01-01', '2033-01-01'], 'JUICE', 'GANYMEDE')

    assert xyz.shape == (3, 2)
    assert_array(xyz[:, 0], (604.730, -779.076, -2998.065), decimal=3)
    assert_array(xyz[:, 0], xyz[:, 1], decimal=3)


def test_toolbox_sun_pos(monkeypatch):
    """Test SPICE toolbox target sun position."""
    monkeypatch.setattr(sp, 'str2et', lambda *_: 1041422469.184)

    def spkpos(*args):
        """Mock Sun position seen from Ganymede.

        spkpos('SUN', 1041422469.184, 'IAU_GANYMEDE', 'NONE', 'GANYMEDE')

        """
        return (
            [-7.55786120e08, -6.99653122e07, -5.01253665e06],
            2531.8655883470833,
        )

    monkeypatch.setattr(sp, 'spkpos', spkpos)

    xyz = sun_pos('2033-01-01', 'GANYMEDE')

    assert_array(xyz, (-755_786_120, -69_965_312, -5_012_537), decimal=0)

    r, lon_e, lat = rlonlat(xyz)

    assert r == approx(759_034_208, abs=1)
    assert lon_e == approx(185.29, abs=1e-2)
    assert lat == approx(-0.38, abs=1e-2)

    # Array of times
    xyz = sun_pos(['2033-01-01', '2033-01-01'], 'GANYMEDE')

    assert xyz.shape == (3, 2)
    assert_array(xyz[:, 0], (-755_786_120, -69_965_312, -5_012_537), decimal=0)
    assert_array(xyz[:, 0], xyz[:, 1], decimal=0)


def test_toolbox_angular_size(monkeypatch):
    """Test SPICE toolbox target angular size."""
    SpicePool.add([tpc, tf], purge=True)

    monkeypatch.setattr(sp, 'str2et', lambda *_: 1041422469.184)

    def spkpos(*args):
        """Mock Ganymede position seen from Juice frame.

        spkpos('GANYMEDE', 1041422468.1839253, 'JUICE_SPACECRAFT', 'NONE', 'JUICE')

        """
        return ([-479227.8970898, -44723.19534804, 3025912.73981675], 10.220246810160733)

    monkeypatch.setattr(sp, 'spkpos', spkpos)

    ang = angular_size('2033-01-01', 'JUICE', 'GANYMEDE')

    assert ang == approx(0.098407, abs=1e-6)

    # Array of times
    ang = angular_size(['2033-01-01', '2033-01-01'], 'JUICE', 'GANYMEDE')

    assert_array(ang, [0.098407, 0.098407], decimal=6)


def test_toolbox_station_azel(monkeypatch):
    """Test SPICE toolbox station azimuth and elevation."""
    SpicePool.add(tf, purge=True)
    monkeypatch.setattr(sp, 'str2et', lambda *_: 94651200)

    def mock_spkpos(*args):
        """Mock Venus position seen from Earth DSS-14 tracking station.

        sp.spkpos('VENUS', 94651200, 'DSS-14_TOPO', 'CN+S', 'DSS-14')

        """
        return (
            [-1342845.74601778, 80542035.83408652, -38648040.3275655],
            298.02218315575544,
        )

    monkeypatch.setattr(sp, 'spkpos', mock_spkpos)

    az, el = station_azel('2003 Jan 01 00:00:00 TDB', 'DSS-14', 'VENUS')

    assert az == approx(269.045, abs=1e-3)
    assert el == approx(-25.630, abs=1e-3)

    # Array of times
    az, el = station_azel(
        ['2003 Jan 01 00:00:00 TDB', '2003 Jan 01 00:00:00 TDB'], 'DSS-14', 'VENUS'
    )

    assert_array(az, [269.045, 269.045], decimal=3)
    assert_array(el, [-25.630, -25.630], decimal=3)


def test_toolbox_sc_state(monkeypatch):
    """Test SPICE toolbox observer -> position state."""
    monkeypatch.setattr(sp, 'str2et', lambda *_: 1041422469.184)

    def spkezr(*args):
        """Mock SPKEZR SPICE function.

        spkezr('GANYMEDE', 1041422468.1839253, 'IAU_GANYMEDE', 'NONE', 'JUICE')

        """
        return (
            [
                -604.7307270168204,
                779.0764414332297,
                2998.0650497737743,
                0.8940939952044199,
                -1.4141796637306154,
                0.5527125241664237,
            ],
            -0.010527661953128513,
        )

    monkeypatch.setattr(sp, 'spkezr', spkezr)

    state = sc_state('2033-01-01', 'JUICE', 'GANYMEDE')

    assert_array(state[:3], (604.730, -779.076, -2998.065), decimal=3)
    assert_array(state[3:], (-0.89, 1.41, -0.55), decimal=2)

    r, lon_e, lat = rlonlat(state[:3])

    assert r == approx(3156.11, abs=1e-2)
    assert lon_e == approx(307.82, abs=1e-2)
    assert lat == approx(-71.79, abs=1e-2)

    # Array of times
    state = sc_state(['2033-01-01', '2033-01-01'], 'JUICE', 'GANYMEDE')

    assert state.shape == (6, 2)
    assert_array(state[:3, 0], (604.730, -779.076, -2998.065), decimal=3)
    assert_array(state[3:, 0], (-0.89, 1.41, -0.55), decimal=2)
    assert_array(state[:, 0], state[:, 1], decimal=3)


def test_toolbox_attitude(monkeypatch):
    """Test SPICE toolbox c-matrix attitude."""
    SpicePool.add(tf, purge=True)

    monkeypatch.setattr(sp, 'str2et', lambda *_: 1041422469.184)

    def pxform(*args):
        """Mock PXFORM SPICE function.

        pxform('J2000', 'JUICE_SPACECRAFT', et)

        """
        return np.array([
            [0.97297918, 0.1840944, -0.1393584],
            [-0.18397789, 0.25345236, -0.94969155],
            [-0.13951218, 0.94966897, 0.28047317],
        ])

    monkeypatch.setattr(sp, 'pxform', pxform)

    # With SPICE Spacecraft attribute
    cmat = attitude('2033-01-01', 'JUICE')

    assert_array(
        cmat,
        [[0.973, 0.184, -0.139], [-0.184, 0.253, -0.95], [-0.14, 0.95, 0.28]],
        decimal=3,
    )

    # Projection of the z-axis on the sky
    vec = cmat.T @ [0, 0, 1]

    assert_array(vec, [-0.14, 0.95, 0.28], decimal=3)

    ra, dec = radec(vec)

    assert ra == approx(98.36, abs=1e-2)
    assert dec == approx(16.29, abs=1e-2)

    # Array of times
    cmat = attitude(['2033-01-01', '2033-01-01'], 'JUICE')

    assert cmat.shape == (3, 3, 2)
    assert_array(
        cmat[..., 0],
        [[0.973, 0.184, -0.139], [-0.184, 0.253, -0.95], [-0.14, 0.95, 0.28]],
        decimal=3,
    )
    assert_array(cmat[..., 0], cmat[..., 1], decimal=3)

    # With an invalid SPICE reference
    with raises(KeyError):
        _ = attitude('2033-01-01', 'JUPITER_BARYCENTER')


def test_toolbox_quaternions():
    """Test SPICE toolbox c-matrix quaternions."""
    m = [
        [0.97297918, 0.1840944, -0.1393584],
        [-0.18397789, 0.25345236, -0.94969155],
        [-0.13951218, 0.94966897, 0.28047317],
    ]

    # with a list
    q = quaternions(m)

    assert_array(q, [0.792, 0.600, 0.000, -0.116], decimal=3)

    # with a numpy array
    q = quaternions(np.array(m))

    assert_array(q, [0.792, 0.600, 0.000, -0.116], decimal=3)

    # with multiple numpy array
    mat = np.broadcast_to(np.reshape(m, (3, 3, 1)), (3, 3, 2))

    assert mat.shape == (3, 3, 2)
    quat = quaternions(mat)

    assert quat.shape == (4, 2)

    q0, q1, q2, q3 = quat

    assert_array(q0, [0.792, 0.792], decimal=3)
    assert_array(q1, [0.600, 0.600], decimal=3)
    assert_array(q2, [0.000, 0.000], decimal=3)
    assert_array(q3, [-0.116, -0.116], decimal=3)

    # Invalid shapes
    with raises(ValueError):
        _ = quaternions([])  # empty

    with raises(ValueError):
        _ = quaternions([1, 2, 3])  # 1D array

    with raises(ValueError):
        _ = quaternions([[1, 2, 3]])  # (1, 3) array

    with raises(ValueError):
        _ = quaternions([[1], [2], [3]])  # (3, 1) array

    with raises(ValueError):
        _ = quaternions([[[1, 2, 3]]])  # (1, 1, 3) array

    with raises(ValueError):
        _ = quaternions([[[[1, 2, 3]]]])  # 4D array


def test_toolbox_intersect_pt(monkeypatch):
    """Test SPICE toolbox intersect point."""
    monkeypatch.setattr(sp, 'str2et', lambda *_: 1041422469.184)

    def sincpt(*args):
        """Mock SINCPT SPICE function.

        sp.sincpt('ELLIPSOID', 'GANYMEDE', 1041422469.184, 'IAU_GANYMEDE',
                  'NONE', 'JUICE', 'JUICE_JANUS', (0, 0, 1))

        """
        return (
            np.array([503.58955867, -648.19888815, -2499.88983713]),
            1041422469.184,
            np.array([-100.24692452, 129.46313315, 498.72749321]),
        )

    monkeypatch.setattr(sp, 'sincpt', sincpt)

    # Single time
    xyz = intersect_pt('2033-01-01', 'JUICE', 'GANYMEDE', 'JUICE_JANUS', (0, 0, 1))

    assert_array(xyz, [503.6, -648.2, -2499.9, True], decimal=1)

    # Array of times
    xyz = intersect_pt(
        ['2033-01-01', '2033-01-01'], 'JUICE', 'GANYMEDE', 'JUICE_JANUS', (0, 0, 1)
    )

    assert xyz.shape == (4, 2)
    assert_array(xyz[:, 0], (503.6, -648.2, -2499.9, True), decimal=1)
    assert_array(xyz[:, 0], xyz[:, 1], decimal=3)

    # Array of rays
    xyz = intersect_pt(
        '2033-01-01', 'JUICE', 'GANYMEDE', 'JUICE_JANUS', [(0, 0, 1), (0, 0, 1)]
    )

    assert xyz.shape == (4, 2)
    assert_array(xyz[:, 0], (503.6, -648.2, -2499.9, True), decimal=1)
    assert_array(xyz[:, 0], xyz[:, 1], decimal=3)

    # Array of times and rays
    xyz = intersect_pt(
        ['2033-01-01', '2033-01-01'],
        'JUICE',
        'GANYMEDE',
        'JUICE_JANUS',
        [(0, 0, 1), (0, 0, 1)],
    )

    assert xyz.shape == (4, 2)
    assert_array(xyz[:, 0], (503.6, -648.2, -2499.9, True), decimal=1)
    assert_array(xyz[:, 0], xyz[:, 1], decimal=3)

    # Invalid dimension between the times and rays
    with raises(ValueError):
        _ = intersect_pt(
            ['2033-01-01', '2033-01-01'],
            'JUICE',
            'GANYMEDE',
            'JUICE_JANUS',
            [(0, 0, 1), (0, 0, 1), (0, 0, 1)],
        )

    # No intersection
    def sincpt_err(*args):
        """Mock SINCPT SPICE error function."""
        raise sp.stypes.NotFoundError

    monkeypatch.setattr(sp, 'sincpt', sincpt_err)

    xyz = intersect_pt('2033-01-01', 'JUICE', 'GANYMEDE', 'JUICE_JANUS', (0, 0, 1))

    assert_array(xyz, (np.nan, np.nan, np.nan, False))

    # Limb intersection
    def tangpt(*args):
        """Mock TANGPT SPICE function.

        sp.tangpt('ELLIPSOID', 'GANYMEDE', 1041422469.184, 'IAU_GANYMEDE',
                  'NONE', 'TANGENT POINT', 'JUICE', 'JUICE_JANUS', (0, 0, 1))

        """
        return (
            np.array([228462.81229751, 389573.63509928, -150640.01005072]),
            473452.0767441353,
            3026738.9831437827,
            np.array([1262.66008718, 2153.08161144, -832.55181143]),
            1041422469.184,
            np.array([-2882235.38329951, 986640.11060504, -323367.01819632]),
        )

    monkeypatch.setattr(sp, 'tangpt', tangpt)

    xyz = intersect_pt(
        '2033-01-01', 'JUICE', 'GANYMEDE', 'JUICE_JANUS', (0, 0, 1), limb=True
    )

    assert_array(xyz, (228462.8, 389573.6, -150640.0, False), decimal=1)

    # With an invalid SPICE reference
    with raises(KeyError):
        _ = intersect_pt(
            '2033-01-01',
            'JUPITER_BARYCENTER',
            'GANYMEDE',
            'JUICE_JANUS',
            (0, 0, 1),
            limb=True,
        )


def test_toolbox_boresight_pt(monkeypatch):
    """Test SPICE toolbox boresight point."""
    SpicePool.add([tf, ti], purge=True)

    monkeypatch.setattr(sp, 'str2et', lambda *_: 1041422469.184)

    def sincpt(*args):
        """Mock SINCPT SPICE function.

        sp.sincpt('ELLIPSOID', 'GANYMEDE', et, 'IAU_GANYMEDE',
                  'NONE','JUICE', 'JUICE_JANUS', (0, 0, 1))

        """
        return (
            np.array([503.58955867, -648.19888815, -2499.88983713]),
            1041422469.184,
            np.array([-100.24692452, 129.46313315, 498.72749321]),
        )

    monkeypatch.setattr(sp, 'sincpt', sincpt)

    # Instrument with single time
    xyz = boresight_pt('2033-01-01', 'JUICE_JANUS', 'GANYMEDE')

    assert_array(xyz, [503.6, -648.2, -2499.9, True], decimal=1)

    # Spacecraft with single time
    xyz = boresight_pt('2033-01-01', 'JUICE', 'GANYMEDE')

    assert_array(xyz, [503.6, -648.2, -2499.9, True], decimal=1)

    # Array of times
    xyz = boresight_pt(['2033-01-01', '2033-01-01'], 'JUICE_JANUS', 'GANYMEDE')

    assert xyz.shape == (4, 2)
    assert_array(xyz[:, 0], (503.6, -648.2, -2499.9, True), decimal=1)
    assert_array(xyz[:, 0], xyz[:, 1], decimal=3)

    # With an invalid SPICE reference
    with raises(KeyError):
        _ = boresight_pt('2033-01-01', 'JUPITER_BARYCENTER', 'GANYMEDE')


def test_toolbox_fov_pts(monkeypatch):
    """Test SPICE toolbox FOV pts."""
    SpicePool.add([tf, ti], purge=True)

    monkeypatch.setattr(sp, 'str2et', lambda *_: 1041422469.184)

    def sincpt(*args):
        """Mock SINCPT SPICE function.

        sp.sincpt('ELLIPSOID', 'GANYMEDE', et, 'IAU_GANYMEDE',
                  'NONE','JUICE', 'JUICE_JANUS', (0, 0, 1))

        """
        return (
            np.array([503.58955867, -648.19888815, -2499.88983713]),
            1041422469.184,
            np.array([-100.24692452, 129.46313315, 498.72749321]),
        )

    monkeypatch.setattr(sp, 'sincpt', sincpt)

    # Single time
    xyz = fov_pts('2033-01-01', 'JUICE_JANUS', 'GANYMEDE')

    assert xyz.shape == (4, 24)
    assert_array(xyz[:, 0], (503.6, -648.2, -2499.9, True), decimal=1)
    assert_array(xyz[:, 0], xyz[:, 1], decimal=3)

    # Array of times
    xyz = fov_pts(['2033-01-01', '2033-01-01'], 'JUICE_JANUS', 'GANYMEDE')

    assert xyz.shape == (4, 2, 24)
    assert_array(xyz[:, 0, 0], (503.6, -648.2, -2499.9, True), decimal=1)
    assert_array(xyz[:, 0, 0], xyz[:, 0, 1], decimal=3)
    assert_array(xyz[:, 0, 0], xyz[:, 1, 0], decimal=3)

    # Tuple of npts on the contour
    xyz = fov_pts('2033-01-01', 'JUICE_JANUS', 'GANYMEDE', npt=(0, 1))

    assert xyz.shape == (4, 6)  # 4 corners + 1 point on 2 edges = 6 pts

    # With an invalid SPICE reference (even a SpiceSpacecraft)
    with raises(KeyError):
        _ = fov_pts('2033-01-01', 'JUICE', 'GANYMEDE')


def test_toolbox_target_separation(monkeypatch):
    """Test SPICE toolbox angular target separation."""
    monkeypatch.setattr(sp, 'str2et', lambda *_: 1041422469.184)
    monkeypatch.setattr(sp, 'trgsep', lambda *_: np.pi / 6)

    assert target_separation('2033-01-01', 'JUICE', 'GANYMEDE', 'CALLISTO') == approx(
        30, abs=1e-6
    )

    assert_array(
        target_separation(['2033-01-01', '2033-01-02'], 'JUICE', 'GANYMEDE', 'CALLISTO'),
        [30, 30],
        decimal=6,
    )


def test_toolbox_local_time(monkeypatch):
    """Test SPICE toolbox local solar time."""
    monkeypatch.setattr(sp, 'str2et', lambda *_: 1041422469.184)

    def et2lst(et, body, lon, typein):
        """Mock ET2LST SPICE function.

        [
            sp.et2lst(1041422469.184, 503, lon, 'PLANETOCENTRIC')
            for lon in [
                0,
                np.pi / 12,
                np.pi / 2,
                np.pi / 12 + np.pi / 2,
                -np.pi / 2,
                3 * np.pi / 2,
            ]
        ]

        """
        return {
            np.radians(0): (23, 38, 50, '23:38:50', '11:38:50 P.M.'),
            np.radians(15): (0, 38, 50, '00:38:50', '12:38:50 A.M.'),
            np.radians(90): (5, 38, 50, '05:38:50', '05:38:50 A.M.'),
            np.radians(105): (6, 38, 50, '06:38:50', '06:38:50 A.M.'),
            np.radians(-90): (17, 38, 50, '17:38:50', '05:38:50 P.M.'),
            np.radians(270): (17, 38, 50, '17:38:50', '05:38:50 P.M.'),
        }[lon]

    # sp.et2lst(1041422469.184, 503, 0, 'PLANETOCENTRIC')
    monkeypatch.setattr(sp, 'et2lst', et2lst)

    h = local_time('2033-01-01', 0, 'GANYMEDE')
    assert h == approx(23.64722, abs=1e-5)  # 23 + 38 / 60 + 50 / 3600

    h = local_time('2033-01-01', 15, 'GANYMEDE')  # + 1 hour (π / 12)
    assert h == approx(0.64722, abs=1e-5)

    h = local_time('2033-01-01', 90, 'GANYMEDE')  # + 6 hours (π / 2)
    assert h == approx(5.64722, abs=1e-5)

    h = local_time('2033-01-01', 105, 'GANYMEDE')  # + 7 hours (π / 2 + π / 12)
    assert h == approx(6.64722, abs=1e-5)

    h = local_time('2033-01-01', -90, 'GANYMEDE')  # - 6 hours (-π / 2)
    assert h == approx(17.64722, abs=1e-5)

    h = local_time('2033-01-01', 270, 'GANYMEDE')  # - 6 hours (-π / 2)
    assert h == approx(17.64722, abs=1e-5)

    # 1D arrays
    h = local_time(['2033-01-01', '2033-01-01'], 0, 'GANYMEDE')
    assert_array(h, [23.64722, 23.64722], decimal=5)

    h = local_time('2033-01-01', [0, 15, 90, 105, -90, 270], 'GANYMEDE')
    assert_array(
        h,
        [
            23.64722,
            0.64722,
            5.64722,
            6.64722,
            17.64722,
            17.64722,
        ],
        decimal=5,
    )

    h = local_time(['2033-01-01', '2033-01-01'], [0, 270], 'GANYMEDE')
    assert_array(h, [23.64722, 17.64722], decimal=5)

    # Different size arrays
    with raises(ValueError):
        _ = local_time(['2033-01-01'], [0, 270], 'GANYMEDE')

    # NaN support
    h = local_time(['2033-01-01', '2033-01-01'], [0, np.nan], 'GANYMEDE')
    assert_array(h, [23.64722, np.nan], decimal=5)


def test_toolbox_illum_angle(monkeypatch):
    """Test SPICE toolbox illumination angles calculation."""
    monkeypatch.setattr(sp, 'str2et', lambda *_: 1041422469.184)

    def ilumin(*args):
        """SPICE illumination mock.

        ilumin('ELLIPSOID', 'GANYMEDE', 1041422469.184,
               'IAU_GANYMEDE', 'NONE', 'JUICE',
               [504.15405255, -649.50320473, -2499.43748038])

        """
        return (
            1041422469.184,
            [-100.57660764, 129.573131, 498.62761071],
            1.7332730463714863,
            1.7332731566600097,
            2.5091746534257484e-07,
        )

    monkeypatch.setattr(sp, 'ilumin', ilumin)

    # Sub-observer position (XYZ) from subpnt
    pt = [504.15405255, -649.50320473, -2499.43748038]

    i, e, p = illum_angles('2033-01-01', 'JUICE', 'GANYMEDE', pt)

    assert i == approx(99.31, abs=1e-2)
    assert e == approx(0, abs=1e-2)
    assert p == approx(99.31, abs=1e-2)

    # 1D arrays
    i, e, p = illum_angles(['2033-01-01', '2033-01-01'], 'JUICE', 'GANYMEDE', pt)

    assert len(i) == len(e) == len(p) == 2
    assert i[0] == i[1] == approx(99.31, abs=1e-2)
    assert e[0] == e[1] == approx(0, abs=1e-2)
    assert p[0] == p[1] == approx(99.31, abs=1e-2)

    pts = np.transpose([pt, pt])
    i, e, p = illum_angles('2033-01-01', 'JUICE', 'GANYMEDE', pts)

    assert len(i) == len(e) == len(p) == 2
    assert i[0] == i[1] == approx(99.31, abs=1e-2)
    assert e[0] == e[1] == approx(0, abs=1e-2)
    assert p[0] == p[1] == approx(99.31, abs=1e-2)

    i, e, p = illum_angles(['2033-01-01', '2033-01-01'], 'JUICE', 'GANYMEDE', pts)

    assert len(i) == len(e) == len(p) == 2
    assert i[0] == i[1] == approx(99.31, abs=1e-2)
    assert e[0] == e[1] == approx(0, abs=1e-2)
    assert p[0] == p[1] == approx(99.31, abs=1e-2)

    # Different size arrays
    with raises(ValueError):
        _ = illum_angles(['2033-01-01'], 'JUICE', 'GANYMEDE', pts)

    # NaN support
    pts = np.transpose([
        [504.15405255, -649.50320473, -2499.43748038],
        [504.15405255, -649.50320473, np.nan],
        [504.15405255, np.nan, -2499.43748038],
        [np.nan, -649.50320473, -2499.43748038],
        [np.nan, np.nan, np.nan],
    ])

    i, e, p = illum_angles('2033-01-01', 'JUICE', 'GANYMEDE', pts)

    assert len(i) == len(e) == len(p) == 5
    assert_array(i, [99.31, np.nan, np.nan, np.nan, np.nan], decimal=2)
    assert_array(e, [0, np.nan, np.nan, np.nan, np.nan], decimal=2)
    assert_array(p, [99.31, np.nan, np.nan, np.nan, np.nan], decimal=2)


def test_toolbox_solar_longitude(monkeypatch):
    """Test SPICE toolbox target solar longitude."""
    monkeypatch.setattr(sp, 'str2et', lambda *_: 1041422469.184)

    def lspcn(*args):
        """SPICE lspcn mock.

        lspcn('JUPITER', 1041422469.184, 'NONE')

        """
        return 6.175111888528254  # radians

    monkeypatch.setattr(sp, 'lspcn', lspcn)

    ls = solar_longitude('2033-01-01', 'GANYMEDE')

    assert ls == approx(353.80, abs=1e-2)

    # Array of times
    ls = solar_longitude(['2033-01-01', '2033-01-01'], 'GANYMEDE')

    assert len(ls) == 2
    assert ls[0] == ls[1] == approx(353.80, abs=1e-2)


def test_toolbox_true_anomaly(monkeypatch):
    """Test SPICE toolbox target true anomaly."""
    monkeypatch.setattr(sp, 'str2et', lambda *_: 1041422469.184)

    def spkezr(*args):
        """SPICE spkezr mock.

        spkezr('JUPITER', 1041422469.184, 'JUICE_JUPITER_IF_J2000', 'NONE', 'GANYMEDE')

        """
        return (
            [
                -1_040_889.80,
                253_373.917,
                -1143.85103,
                -2.58534596,
                -10.5600172,
                2.34602842e-02,
            ],
            3.573421604290011,
        )

    def gdpool(*args):
        """SPICE gdpool mock.

        gdpool('BODY599_GM', 0, 1)

        """
        return [126_686_534.922]

    def oscltx(*args):
        """SPICE oscltx mock.

        oscltx([1_040_889.80, -253_373.917, 1143.85103,
                2.58534596, 10.5600172, -2.34602842e-02],
                1041422469.184,
                126_686_534.922)

        """
        return [
            1_069_248.3426561414,
            0.0014106088241957206,
            0.002408863408621217,
            3.3620404490671434,
            0.7544890865790931,
            1.9252337120999292,
            1_041_422_468.184,
            126_686_534.922,
            1.9278779469246907,
            1_070_758.7644178143,
            6_18_517.6488648178,
        ]

    monkeypatch.setattr(sp, 'spkezr', spkezr)
    monkeypatch.setattr(sp, 'gdpool', gdpool)
    monkeypatch.setattr(sp, 'oscltx', oscltx)

    theta = true_anomaly('2033-01-01', 'GANYMEDE', frame='JUICE_JUPITER_IF_J2000')

    assert theta == approx(110.459, abs=1e-3)

    # Array of times
    theta = true_anomaly(
        ['2033-01-01', '2033-01-01'], 'GANYMEDE', frame='JUICE_JUPITER_IF_J2000'
    )

    assert len(theta) == 2
    assert theta[0] == theta[1] == approx(110.459, abs=1e-3)


def test_toolbox_groundtrack_velocity(monkeypatch):
    """Test SPICE toolbox target groundtrack velocity."""
    radii = [1562.6, 1560.3, 1559.5]  # Mock Europa radii

    monkeypatch.setattr(toolbox.SpiceBody, 'radii', radii)

    # 7E1 at CA - 6h
    # spkezr('EUROPA', 1025648613.184, 'IAU_EUROPA', 'NONE', 'JUICE')
    state_0 = [
        24_730.9408,
        -73_366.9913,
        -821.515833,
        -2.67960556,
        2.65983984,
        -0.0407036100,
    ]

    # 7E1 at CA
    # sp.spkezr('EUROPA', 1025670214.184, 'IAU_EUROPA', 'NONE', 'JUICE')
    state_1 = [
        1_337.52175,
        5.03647822,
        -1_433.79130,
        0.309068987,
        4.02873071,
        0.293659468,
    ]

    gt_0 = groundtrack_velocity('GANYMEDE', state_0)
    gt_1 = groundtrack_velocity('GANYMEDE', state_1)

    assert gt_0 == approx(0.034, abs=1e-3)
    assert gt_1 == approx(3.219, abs=1e-3)

    # Array of times
    gt = groundtrack_velocity('GANYMEDE', np.transpose([state_0, state_1]))

    assert len(gt) == 2
    assert gt[0] == approx(0.034, abs=1e-3)
    assert gt[1] == approx(3.219, abs=1e-3)

    # Invalid shape
    with raises(ValueError):
        _ = groundtrack_velocity('GANYMEDE', [state_0, state_1])


def test_toolbox_pixel_scale(monkeypatch):
    """Test SPICE toolbox instrument pixel scale."""
    SpicePool.add([tf, ti], purge=True)

    radius = 2631.2  # Mock Ganymede radius (km)

    monkeypatch.setattr(toolbox.SpiceBody, 'radius', radius)

    assert pixel_scale('JUICE_JANUS', 'GANYMEDE', 0, 5_000 + radius) == approx(
        0.075, abs=1e-3
    )

    assert np.isnan(pixel_scale('JUICE_JANUS', 'GANYMEDE', 90, 5_000 + radius))

    # Array of times
    assert_array(
        pixel_scale('JUICE_JANUS', 'GANYMEDE', 0, [500 + radius, 5_000 + radius]),
        [0.0075, 0.075],
        decimal=4,
    )

    # Invalid shape
    with raises(ValueError):
        _ = pixel_scale('JUICE_JANUS', 'GANYMEDE', [0, 1, 2], [3, 4])
